/*
** Copyright (c) 2003, Scott Lystig Fritchie.  All rights reserved.
** Some contributions Copyright (c) Chris Newcombe, 2006.  All rights reserved.
** See the file "../LICENSE" for license details.
*/

#include <stdio.h>
#include <sys/time.h> /* because logging macros call gettimeofday() */
#include <unistd.h>

#include <erl_driver.h>


#ifdef DRIVER_USING_PTHREADS
#include <pthread.h>
extern pthread_mutex_t logging_output_global_mutex;
#endif


/* These functions expose the standard malloc/realloc/free interface,
 * but underneath call driver_alloc/realloc/free_binary.
 *
 * The functions return a pointer to useable memory of the correct size.
 * But they also allocate the proper driver-binary header (immediately
 * before the usable memory).  The function edtk_alloced_ptr2ErlDrvBinary()
 * will convert the pointer returned from edtk_driver_alloc/realloc_wrapper
 * to an Erlang driver binary (i.e. will subtract the size of the
 * driver-binary header).
 *
 * This means that the resulting memory can be sent to Erlang
 * via driver_send_term() / driver_output_term() and use of the 
 * following macro: e.g.
 *
 *    msgcount = LOAD_BINARY(msg, msgcount, edtk_alloced_ptr2ErlDrvBinary(ptr), offset, length);
 *
 * Note: from Ericsson:
 *
 *  They are all thread safe in the emulator with smp support, but they will 
 *  probably never be thread safe in the emulator without smp support.
 *   
 *  The thread unsafe part is the reference counting of binaries. Making 
 *  them thread safe would cause overhead for all reference counted 
 *  binaries, not just driver binaries. We don't want to slow down all 
 *  binary management as it is now, but we might make them thread safe in 
 *  the future if threaded drivers should become very common (but don't 
 *  count on it).
 *
 *  Rickard Green, Erlang/OTP
 *
 */

void * edtk_driver_binary_malloc(size_t);
void * edtk_driver_binary_realloc(void *, size_t);
void   edtk_driver_binary_free(void *);

ErlDrvBinary * edtk_alloced_ptr2ErlDrvBinary(void *);


/*
 * Debug logging 
 */


/* To avoid mangled log lines in multi-threaded drivers we need a mutex.
 * Unfortunately we don't have any way of mutexing with the Erlang VM's output to stdout.
 * Maybe we could find a way to send logging as messages to the IO group-leader process
 * -- but then it would be subject to delays from queuing, which might be confusing.
 */
#ifdef DRIVER_USING_PTHREADS

#define EDTK_LOGGING_FPRINTF(stream, str, ...)                  \
    pthread_mutex_lock(& logging_output_global_mutex);          \
    fprintf(stream, str, ##__VA_ARGS__);                        \
    pthread_mutex_unlock(& logging_output_global_mutex);

#else

#define EDTK_LOGGING_FPRINTF(stream, str, ...)  \
    fprintf(stream, str, ##__VA_ARGS__);
    
#endif


/* All driver-level logging goes to edtk_log_file, which defaults to stderr.
 *
 * All port-level logging goes to desc->port_log_file, which defaults to stderr
 * but can be changed with direct_port_logging_to_file().
 *
 * IMPORTANT: The port logging macros have a theoretically-unsafe-but-practically-safe policy of
 * NOT acquiring desc->mutex even though they read desc->port_log_file and desc->port_label.
 * It's too onerous to require the caller to call a different macro if they happen
 * to be logging while holding the mutex.
 * This is 'practically safe' as those things don't change apart from during port-coordinator
 * initialization, when nothing else is happening in the driver.
 */
 
extern FILE * edtk_log_file;


/* Some helper functions to hard-abort if an invariant is violated */
#define EDTK_PORT_LOG_AND_ABORT(desc, str, ...)                         \
    {                                                                   \
        struct timeval curr_time;                                       \
        gettimeofday(&curr_time, NULL);                                 \
        EDTK_LOGGING_FPRINTF(desc->port_log_file, "\r\n\r\n%s: %s: %d:%lx: %lu:%lu " str ". ABORTING!\r\n\r\n", desc->port_label, __FUNCTION__, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, ##__VA_ARGS__); \
        fflush(desc->port_log_file);                                    \
        exit(EXIT_FAILURE);                                             \
    }

/* this variant does not require desc */
#define EDTK_LOG_AND_ABORT(str, ...) \
    {                                                                   \
        struct timeval curr_time;                                       \
        gettimeofday(&curr_time, NULL);                                 \
        EDTK_LOGGING_FPRINTF(edtk_log_file, "\r\n\r\nEDTK %s: %d:%lx: %lu:%lu " str ". ABORTING!\r\n\r\n", __FUNCTION__, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, ##__VA_ARGS__); \
        fflush(edtk_log_file);                                          \
        exit(EXIT_FAILURE);                                             \
    }


/* There are 2 types of logging; port-specific, which requires a port descriptor_t,
 * and generic 'driver' logging, which is done by code that does not have access to
 * a port descriptor.
 */

/* These constants also appear in erl_template.gsl.
 * Code in ?DRV:adjust_port_logging() converts a list of atoms to these constants.
 */
#define EDTK_PORT_LOG_COMMAND_OUTPUTV              (0x00000001)
#define EDTK_PORT_LOG_COMMAND_READY_ASYNC          (0x00000002)
#define EDTK_PORT_LOG_COMMAND_INVOKE_START         (0x00000004)
#define EDTK_PORT_LOG_COMMAND_INVOKE_END           (0x00000008)
#define EDTK_PORT_LOG_VALMAP_START                 (0x00000010)
#define EDTK_PORT_LOG_VALMAP_STOP                  (0x00000020)
#define EDTK_PORT_LOG_GROUP_VALMAP_CLEANUP_OUTPUTV (0x00000040)
#define EDTK_PORT_LOG_GROUP_VALMAP_CLEANUP_WORK    (0x00000080)
#define EDTK_PORT_LOG_SHUTDOWN                     (0x00000100) /* if this is set then the driver's stop() function will also enable GROUP_VALMAP_CLEANUP_WORK, VALMAP_START, and VALMAP_STOP */
#define EDTK_PORT_LOG_ALLOC_FREE                   (0x00000200)
#define EDTK_PORT_LOG_CALLSTATE_DRIVER_BINARIES    (0x00000400)
#define EDTK_PORT_LOG_REPLY_XTRA_DRIVER_BINARIES   (0x00000800)
#define EDTK_PORT_LOG_THREADPOOL_MANAGEMENT        (0x00001000)
#define EDTK_PORT_LOG_THREADPOOL_ENQUEUE_Q_LEN     (0x00002000)
#define EDTK_PORT_LOG_THREADPOOL_DEQUEUE_Q_LEN     (0x00004000)
#define EDTK_PORT_LOG_ASYNC_DONE_LIST_BATCH_SIZE   (0x00008000)
#define EDTK_PORT_LOG_SEPARATE_COMMANDS            (0x00010000) /* purely cosmetic -- logs a blank line in outputv */
/* these are modifiers*/
#define EDTK_PORT_LOG_VERBOSE                      (0x40000000) /* TODO: treat as a modifier */
#define EDTK_PORT_LOG_WITH_TIMESTAMPS              (0x80000000)
/* convenience */
#define EDTK_PORT_LOG_ALL                          (0xFFFFFFFF)

/* TODO: pthread_self() is only unique within a process, so if an Erlang node is running
 * multiple BDB environments in pipe-mode then should need to log the process id too.
 * (Note: pipe_driver_p will be set if that is the case.)
 * For now we rely on the port label to descriminate, just to minimize code size
 *(this is used a lot and is already too big) */

#define EDTK_PORT_LOG_FORCE(desc, str, ...)                             \
    if ( desc->port_log_mask & EDTK_PORT_LOG_WITH_TIMESTAMPS ) {   \
        struct timeval curr_time;                                       \
        gettimeofday(&curr_time, NULL);                                 \
        EDTK_LOGGING_FPRINTF(desc->port_log_file, "%s: %s: %d:%lx: %lu:%lu " str "\r\n", desc->port_label, __FUNCTION__, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, ##__VA_ARGS__); \
    } else {                                                            \
        EDTK_LOGGING_FPRINTF(desc->port_log_file, "%s: %s: %d:%lx: " str "\r\n", desc->port_label, __FUNCTION__, getpid(), pthread_self(), ##__VA_ARGS__); \
    }                                                                   \

#define EDTK_PORT_LOG(mask, desc, str, ...)             \
    if ( desc->port_log_mask & (mask) ) {          \
        EDTK_PORT_LOG_FORCE(desc, str, ##__VA_ARGS__);  \
    }

/* In c_c/h_template.gsl there are wrappers for malloc/realloc that will do port-specific logging if
 * necessary.  We cannot declare them here as they use descriptor_t which is not declared yet.
 * (it is declared in c_h_template.gsl).
 * ** Do not confuse these with edtk_driver_binary_malloc/realloc/free()
 * ** These are just for optional logging.
 *
 * void * EDTK_PORT_MALLOC(descriptor_t, size_t);
 * void * EDTK_PORT_REALLOC(descriptor_t, void *, size_t);
 * void   EDTK_PORT_FREE(descriptor_t, void *);
 */


/* 'driver-wide' logging facility, for when we don't have a particular port descriptor.
 * These are controlled by the ?DRV:adjust_driver_logging() API
 * Code for that function in erl_template.gsl converts a list of atoms to these constants.
 */
#define EDTK_LOG_INIT             (0x00000001)
#define EDTK_LOG_SHUTDOWN         (0x00000002)
#define EDTK_LOG_ALLOC_FREE       (0x00000004)
#define EDTK_LOG_PIPE_VERBOSE     (0x20000000)
#define EDTK_LOG_VERBOSE          (0x40000000)
#define EDTK_LOG_WITH_TIMESTAMPS  (0x80000000)
#define EDTK_LOG_ALL              (0xFFFFFFFF)

extern unsigned int edtk_log_mask;  /* defined in erl_driver_tk.c, initialized in c_c_template.gsl:driver_init() and pipe-main.c */

#define EDTK_LOG_FORCE(str, ...)                                        \
    if ( edtk_log_mask & EDTK_LOG_WITH_TIMESTAMPS ) {                   \
        struct timeval curr_time;                                       \
        gettimeofday(&curr_time, NULL);                                 \
        EDTK_LOGGING_FPRINTF(edtk_log_file, "EDTK: %s: %d:%lx: %lu:%lu " str "\r\n", __FUNCTION__, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, ##__VA_ARGS__); \
    } else {                                                            \
        EDTK_LOGGING_FPRINTF(edtk_log_file, "EDTK: %s: %d:%lx: " str "\r\n", __FUNCTION__, getpid(), pthread_self(), ##__VA_ARGS__); \
    }                                                                   \

#define EDTK_LOG(mask, str, ...)                \
    if ( edtk_log_mask & (mask) ) {             \
        EDTK_LOG_FORCE(str, ##__VA_ARGS__);     \
    }

/* Wrappers for malloc/free, so we can do logging if necessary.
 * ** Do not confuse these with edtk_driver_binary_malloc/realloc/free()
 * ** These are just for optional logging */
void * EDTK_MALLOC(size_t);
void * EDTK_REALLOC(void *, size_t);
/* this one can be a macro... */
#define EDTK_FREE(mem) \
    EDTK_LOG(EDTK_LOG_ALLOC_FREE, "freeing malloced memory 0x%p", mem); \
    free(mem);



/*
 * ErlIOVec manipulation functions, used for parsing the command sent to outputv 
 */

/* This is our own  is used for parsing the command sent to outputv */
int edtk_ev_forward_N(ErlIOVec *ev, int n, int *pp, int *qp, int);

/* The rest were stolen from efile_drv.c */

/* char *EV_CHAR_P(ErlIOVec *ev, int p, int q) */
#define EV_CHAR_P(ev, p, q) \
    (((char *)(ev)->iov[(q)].iov_base) + (p))

/* char *EV_UCHAR_P(ErlIOVec *ev, int p, int q) */
#define EV_UCHAR_P(ev, p, q) \
    (((unsigned char *)(ev)->iov[(q)].iov_base) + (p))

/* int EV_GET_CHAR(ErlIOVec *ev, char *p, int *pp, int *qp) */
#define EV_GET_CHAR(ev, p, pp, qp) \
    (*(pp)+1 <= (ev)->iov[*(qp)].iov_len \
     ? (*(p) = *EV_CHAR_P(ev, *(pp), *(qp)), \
        *(pp) = (*(pp)+1 < (ev)->iov[*(qp)].iov_len \
                 ? *(pp)+1 \
                 : ((*(qp))++, 0)), \
        !0) \
     : 0)

#define EV_GET_UINT8(ev, p, pp, qp)	EV_GET_CHAR(ev, p, pp, qp)

/* int EV_GET_UINT16(ErlIOVec *ev, Uint32 *p, int *pp, int *qp) */
#define EV_GET_UINT16(ev, p, pp, qp) \
    (*(pp)+2 <= (ev)->iov[*(qp)].iov_len \
     ? (*((unsigned short *) p) = (*EV_UCHAR_P(ev, *(pp), *(qp)) << 8)\
                | *EV_UCHAR_P(ev, *(pp)+1, *(qp)), \
        *(pp) = (*(pp)+2 < (ev)->iov[*(qp)].iov_len \
                 ? *(pp)+2 \
                 : ((*(qp))++, 0)), \
        !0) \
     : 0)

/* int EV_GET_UINT32(ErlIOVec *ev, Uint32 *p, int *pp, int *qp) */
#define EV_GET_UINT32(ev, p, pp, qp) \
    (*(pp)+4 <= (ev)->iov[*(qp)].iov_len \
     ? (*((unsigned long *) p) = (*EV_UCHAR_P(ev, *(pp), *(qp)) << 24) \
                | (*EV_UCHAR_P(ev, *(pp)+1, *(qp)) << 16) \
                | (*EV_UCHAR_P(ev, *(pp)+2, *(qp)) << 8)\
                | *EV_UCHAR_P(ev, *(pp)+3, *(qp)), \
        *(pp) = (*(pp)+4 < (ev)->iov[*(qp)].iov_len \
                 ? *(pp)+4 \
                 : ((*(qp))++, 0)), \
        !0) \
     : 0)

/* char * EV_GETPOS(ErlIOVec *ev, int p, int q) */
#define EV_GETPOS(ev, p, q) \
    ((q) < (ev)->vsize \
    ? ((ev)->iov[(q)].iov_base + p) \
    : NULL)


/*
 * Term manipulation functions, stolen from inet_drv.c
 * These are used to build the reply message (in ready_async
 * or the reply_???() functions, which is then sent via
 * driver_send_term() or driver_output_term().
 */

#define LOAD_NIL(vec, i) \
  (((vec)[(i)] = ERL_DRV_NIL), \
  (i+1))

#define LOAD_ATOM(vec, i, atom) \
  (((vec)[(i)] = ERL_DRV_ATOM), \
  ((vec)[(i)+1] = (atom)), \
  (i+2))

#define LOAD_INT(vec, i, val) \
  (((vec)[(i)] = ERL_DRV_INT), \
  ((vec)[(i)+1] = (ErlDrvTermData)(val)), \
  (i+2))

#define LOAD_PORT(vec, i, port) \
  (((vec)[(i)] = ERL_DRV_PORT), \
  ((vec)[(i)+1] = (port)), \
  (i+2))

#define LOAD_PID(vec, i, pid) \
  (((vec)[(i)] = ERL_DRV_PID), \
  ((vec)[(i)+1] = (pid)), \
  (i+2))

#define LOAD_BINARY(vec, i, bin, offs, len) \
  (((vec)[(i)] = ERL_DRV_BINARY), \
  ((vec)[(i)+1] = (ErlDrvTermData)(bin)), \
  ((vec)[(i)+2] = (len)), \
  ((vec)[(i)+3] = (offs)), \
  (i+4))

#define LOAD_STRING(vec, i, str, len) \
  (((vec)[(i)] = ERL_DRV_STRING), \
  ((vec)[(i)+1] = (ErlDrvTermData)(str)), \
  ((vec)[(i)+2] = (len)), \
  (i+3))

#define LOAD_STRING_CONS(vec, i, str, len) \
  (((vec)[(i)] = ERL_DRV_STRING_CONS), \
  ((vec)[(i)+1] = (ErlDrvTermData)(str)), \
  ((vec)[(i)+2] = (len)), \
  (i+3))

#define LOAD_TUPLE(vec, i, size) \
  (((vec)[(i)] = ERL_DRV_TUPLE), \
  ((vec)[(i)+1] = (size)), \
  (i+2))

#define LOAD_LIST(vec, i, size) \
  (((vec)[(i)] = ERL_DRV_LIST), \
  ((vec)[(i)+1] = (size)), \
  (i+2))

#define get_int32(s) ((((unsigned char*) (s))[0] << 24) | \
                      (((unsigned char*) (s))[1] << 16)  | \
                      (((unsigned char*) (s))[2] << 8) | \
                      (((unsigned char*) (s))[3]))


/* These are type tages for our protocol in pipe-mode (over stdout)
 * that emulates driver_send_term() / driver_output_term().
 */
#define	PIPE_DRIVER_TERM_NIL	0
#define	PIPE_DRIVER_TERM_ATOM	1
#define	PIPE_DRIVER_TERM_PORT	2
#define	PIPE_DRIVER_TERM_INT	3
#define	PIPE_DRIVER_TERM_TUPLE	4
#define	PIPE_DRIVER_TERM_BINARY	5
#define	PIPE_DRIVER_TERM_STRING	6
#define	PIPE_DRIVER_TERM_LIST	7

