
%%% File    : berkeley_db_helpers.erl
%%% Author  : Chris Newcombe <>
%%% Description : 
%%% Created :  7 Sep 2006 by Chris Newcombe <>

-module(berkeley_db_helpers).


-export([do_txn/2,             % throws on error
         do_txn/3,
         do_txn_no_throw/3,    % mainly for use by servers

         %% The following are convenience functions for use from an Erlang shell on a running system.
         %% e.g.
         %%   berkeley_db_helpers:do_txn(PortCoordPid, {berkeley_db_helpers, txn_fun_write_coredump, "core.requested1"}, 0).
         %%
         txn_fun_write_coredump/2,
         txn_fun_adjust_port_logging/2,
         txn_fun_env_set_verbose/2,
         txn_fun_keys_values_foldl/2,
         txn_fun_keys_values_bulk_foldl/2,
         keys_values_foldl/4,
         keys_values_foldl_from_cursor/4,
         keys_values_foldl_from_cursor_without_close/4,
         keys_values_bulk_foldl/5,
         keys_values_bulk_foldl_from_cursor/5,
         keys_values_bulk_foldl_from_cursor_without_close/5,
         txn_fun_list_all_open_databases/2,
         list_all_on_disk_databases/1,
         txn_fun_md5_all_keys_values_in_all_on_disk_databases/2,
         txn_fun_md5_all_keys_values_in_multiple_databases/2,
         txn_fun_do_fun/2
        ]).


% do_txn calls group_valmap_cleanup, which needs #bdb_port_handle.group_valmap_id
-include("berkeley_db_drv.hrl").

-define(BDB,      berkeley_db_drv).
-define(BDB_U,    berkeley_db_utilities).
-define(BDB_PC,   berkeley_db_port_coordinator).


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


%%% Function: do_txn()
%%%
%%% Description: 
%%%
%%%   This is a convenience wrapper.
%%%   It Executes a TxnFun, and optionally retry some number of times if a BDB deadlock occurs.
%%%   The TxnFun is executed by the caller process -- i.e. no new process is spawned.
%%%
%%%   NOTE: The target port-coordinator must be on the same node as the caller
%%%         (due to the fact that Erlang/OTP ports cannot be used across hosts).
%%%
%%%   TxnFun must be  
%%%
%%%      {Module, Function, ClientDataTerm}
%%%
%%%   where the following exported function must exist
%%%
%%%      Module:Function(BdbPortHandle, ClientDataTerm) -> ResultTerm
%%%
%%%   Note that TxnFun is (unfortunately) NOT a normal fun (closure) or even a new-style 'fun M:F/A'.
%%%   This is done to keep a uniform interface with berkeley_db_replication_group:do_txn_on_master 
%%%   (which in fact is implemented in terms of this function).
%%%
%%%   [Sidebar: do_txn_on_master cannot use normal funs as TxnFun might be shipped to another host
%%%   (to execute on the master), and the code for TxnFun on that other host might be older or newer than 
%%%   the code that creates TxnFun on the client.  The rules for fun() code upgrades are complicated,
%%%   and relatively undocumented.  It appears that the MD5 signatures of the modules must match, which 
%%%   would prevent staggered upgrades. Using the Module:Function form to specify a TxnFun means that the 
%%%   client code can be a different version than the version on the server.]
%%%
%%%   IMPORTANT: TxnFun must **NOT** call ?BDB_PC:release_bdb_port_handle() on BdbPortHandle
%%%   This should be obvious as the TxnFun is passed the handle, and does not obtain it itself.
%%%
%%% Returns    
%%%
%%%   ResultOfTxnFun
%%%
%%% or throws a 2-tuple.
%%%
%%% If TxnFun itself throws then that exception is captured and rethrown as one of...
%%%
%%%   {txn_fun_error, ThrownException}                   (e.g. if TxnFun called throw)
%%%   {txn_fun_error, {{exit, Reason}, StackTrace}}      (e.g. if TxnFun called exit(Reason))
%%%   {txn_fun_error, {{error, Reason}, StackTrace}}     (e.g. if TxnFun hit a badmatch/badarg/function_clause/case_clause etc)
%%%   {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}} (if TxnFun hit BDB deadlocks and exhausted NumDeadlockRetries)
%%%   {txn_fun_error, {return_with_open_resources, 
%%%                    NumResources, 
%%%                    TxnFunResult}}                    (if TxnFun returned with an open cursor/txn)
%%%
%%%    [NOTE that we only include stack-traces with exits and errors,
%%%    not with throws.  This is optimization, as stack traces take
%%%    time to construct and thrown exceptions might be quite common -
%%%    e.g. {bdb_error, bdb_DB_NOT_FOUND}.  The contents of Erlang
%%%    exits and errors (particularly the latter) are often very
%%%    terse, and a stack trace helps hugely in understanding them.
%%%    But the contents of 'normal' (thrown) exceptions are under the
%%%    control of the application so can be as descriptive as
%%%    necessary. (Unfortunately there currently doesn't seem to be
%%%    any way to get a currently stacktrace before calling throw; the
%%%    closest is erlang:process_display(self(), backtrace) but that
%%%    writes a string to stderror and the format is not fully
%%%    defined.)]
%%%
%%% or an internal error (these are examples, *not* an exhaustive list)...
%%%
%%%   {bdb_helper_error, Reason}
%%%   {bdb_helper_error, {acquire_bdb_port_handle, Reason}}
%%%   {bdb_helper_error, {group_valmap_cleanup, Reason}}
%%%   {bdb_helper_error, {release_bdb_port_handle, Reason, TxnFunResult}}
%%%
%%% Examples:
%%%
%%% If the specified port_coordinator is down when do_txn() is called then it will throw:
%%%
%%%   {acquire_bdb_port_handle, {{exit, {noproc, {gen_server, call, [PortCoordinator, acquire_bdb_port_handle]}}}, _StackTrace}},
%%%
%%% If the BDB port closes while the TxnFun is running then the underlying
%%% 'badarg' (from erlang:port_command) is caught and the following is thrown instead:
%%% 
%%%   {bdb_helper_error, {release_bdb_port_handle, {port_is_unreachable, ErlangPort}}}
%%%
%%% (Note: a txn_fun_error is not thrown, as release_bdb_port_handle is always 
%%% called internally, and that calls group_valmap_cleanup internally, which 
%%% attempts to use the port.)
%%%
%%%
%%% Example of a simple TxnFun
%%%
%%%         my_txn(BdbPortHandle, {MyDbName, MyKey, MyData}) -> 
%%%
%%%                 %% we almost always need at least one DB handle (which we must NOT close - they are shared)
%%%                 DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, MyDbName),
%%%          
%%%                 %% now do some work and auto-commit it...
%%%                 ?BDB:db_put(BdbPort, DB, void, MyKey, MyData, []),
%%%          
%%%                 %% and/or do a multi-part transaction and explicitly commit it...
%%%                 Txn = ?BDB:txn_begin(BdbPort, void, []),
%%%                 ?BDB:db_put(BdbPort, DB, Txn, MyKey, MyData, []),
%%%                 ?BDB:db_put(BdbPort, DB, Txn, <<"some other key">>, MyData, []),
%%%                 ?BDB:txn_commit(BdbPort, Txn, []),
%%%         end
%%% 
%%% 


%% This hits the default port-coordinator.
do_txn(TxnFun, NumDeadlockRetries) ->
    do_txn(berkeley_db_port_coordinator, TxnFun, NumDeadlockRetries).


%% A variant to hit a specific port_coordinator.
%% PortCoord must be an atom (name of a locally-registered process) or a *local* pid 
%% (as we call acquire_bdb_port_handle and the resulting handle can only be used locally)
do_txn(PortCoord, TxnFun, NumDeadlockRetries) ->
    case do_txn_no_throw(PortCoord, TxnFun, NumDeadlockRetries) of
        {ok, Result}    -> Result;
        {error, Reason} -> throw(Reason)
    end.


%% We want a 'no throw' variant as it may be called by a server that wishes to return 
%% a term to be rethrown on the client side.
%%
%% This returns one of
%%
%%   {ok,    Result}
%%   {error, {txn_fun_error,    Reason}}
%%   {error, {bdb_helper_error, Reason}}
%%
%% This catches and convert **ALL** TxnFun throws, exits, and 'erlang errors' (e.g. badmatch)
%% and converts them to appropriate txn_fun_error tuples (including stack traces in the case of 
%% Erlang errors and exits).
%%
%% This also catches and converts throws/exits/errors from 'internal' code to error tuples (tagged accordingly).
%% It also catches and coverts any 'erlang errors' that might be considered dynamic errrors 
%% - e.g. if the port closes unexpectedly then attempts to use it (e.g. for group_valmap_cleanup)
%% will crash with error:badarg and a stack-trace that comes from erlang:port_command.  Those 
%% errors are converted to appropriately-tagged error terms.

do_txn_no_throw(PortCoord, {Module, Function, ClientData}, NumDeadlockRetries) when is_atom(PortCoord) ->
    %% This variant is a minor optimization, to avoid repeated local lookups of the process name.
    case whereis(PortCoord) of 
        undefined -> {error, {bdb_helper_error, {noproc, PortCoord}}};
        Pid       -> do_txn_no_throw(Pid, {Module, Function, ClientData}, NumDeadlockRetries)
    end;

do_txn_no_throw(PortCoord, {Module, Function, ClientData}, NumDeadlockRetries) ->

    try ?BDB_PC:acquire_bdb_port_handle(PortCoord) of

        #bdb_port_handle{erlang_port = ErlangPort} = BdbPortHandle ->

            %% We need a sub-function in order to retry if there is a BDB deadlock
            %% This must return either {ok, TxnFunResult} or {error, SomeError}.
            %% It should not throw.
            RetryWrapperFun = 

                fun(_NumRetriesLeft, _) when _NumRetriesLeft < 0 ->
                        {error, {txn_fun_error, {bdb_error, bdb_DB_LOCK_DEADLOCK}}};

                   (NumRetriesLeft, ThisFun) ->
                        try 
                            {ok, Module:Function(BdbPortHandle, ClientData)}
                        catch 
                            %% If we got a BDB deadlock then we may wish to retry.
                            %% Otherwise convert the error to something that will be
                            %% rethrown by the client.

                            {bdb_error, bdb_DB_LOCK_DEADLOCK} ->

                                %% Before we retry we need to close any cursors, abort any txns

                                try ?BDB:group_valmap_cleanup(BdbPortHandle, 
                                                              BdbPortHandle#bdb_port_handle.group_valmap_id) of

                                    NumResourcesCleaned ->
                                        ?INFO_LOGGER("TxnFun got deadlock. Cleaned ~w resources and retrying ~w more times~n", [NumResourcesCleaned, NumRetriesLeft]),
                                        ThisFun(NumRetriesLeft - 1, ThisFun)

                                catch
                                    %% The port_coordinator might fail in between TxnFun throwing the deadlock and our call 
                                    %% to group_valmap_cleanup. We would notice this by finding that the BDB port is suddenly 
                                    %% closed (a badarg error from erlang:port_command).
                                    %% That is a dynamic error, and we want to distinguish it clearly from other badarg errors 
                                    %% (which are probably bugs).  However, we don't need to do so here, as we are guaranteed to 
                                    %% call group_valmap_cleanup *again* below, as part of release_bdb_port_handle, and that will 
                                    %% presumably get the same port error.
                                    %% Any such error from the second group_valmap_cleanup will mask TxnFunResult anyway (as TxnFun might 
                                    %% have returned with open txns/cursors -- so it's not safe to expose TxnFunResult).
                                    %% It is such a rare case (for a port to be unexpectedly closed) that it is not worth attempting 
                                    %% to avoid the call to release_bdb_port_handle/group_valmap cleanup (e.g. by throwing here, and 
                                    %% catching around the call to RetryWrapperFum), even though we know that release_bdb_port_handle
                                    %% will fail (and why).  Instead, we leverage the error-handling around the call to 
                                    %% release_bdb_port_handle, to convert the 'badarg' port error to something more meaningful.
                                    %%
                                    %% It could be argued that other exceptions of type 'error' be allowed to pass through 
                                    %% (with their stack-traces intact) as they are probably bugs in group_valmap_cleanup, or the
                                    %% BDB driver.  (We do allow that in other places in this function.)
                                    %% But in this case that would require more code for a vanishingly rare case, so we just 
                                    %% capture and comvert all errors here.  
                                    %%
                                    %% Note that the value we return from here is very likely irrelevant -- it will almost certainly 
                                    %% be lost when group_valmap_cleanup fails again (inside release_bdb_port_handle).
                                    %% But it is worth capturing just in case that second call to group_valmap_cleanup 
                                    %% does magically succeed for some reason (to debug that level of insanity we need as
                                    %% much information as possible).

                                    GroupCleanupClass:GroupCleanupReason -> 
                                        {error, {bdb_helper_error, {group_valmap_cleanup_on_deadlock, {{GroupCleanupClass, GroupCleanupReason}, erlang:get_stacktrace()}}}}
                                end;

                            %% Normal exceptions (e.g. other bdb errors) are the most common failure.
                            %% We don't include 'throw' with TxnFunReason, as it makes it harder for the caller to pattern-match on.
                            %% And we don't include the stack-trace because they are apparenly fairly expensive to construct, 
                            %% and this is the most common form of exception.
                            throw:TxnFunReason       -> {error, {txn_fun_error, TxnFunReason}};

                            %% The caller could cause errors; e.g. 
                            %% - Providing bad/undefined Module or Function (or mismatched parameters, giving a function_clause error)
                            %% - TxnFun might call exit() or erlang:error/fault().
                            %% - If the port-coordinator terminates while the TxnFun is executing then calls to APIs like 
                            %%   ?BDB_PC:get_db_handle() etc will fail (typically an exit() from within gen_server:call).
                            %% - But a TxnFun is most likely to notice a port_coordinator failure by finding that the BDB port 
                            %%   has been unexpectedly closed (symptom: a badarg error from erlang:port_command).
                            %%   See comments immediately above (in catch-clause for group_valmap_cleanup) for how we 
                            %%   handle that (basically leave it to re-occur in release_bdb_port_handle).
                            TxnFunClass:TxnFunReason -> {error, {txn_fun_error, {{TxnFunClass, TxnFunReason}, erlang:get_stacktrace()}}}
                        end
                end,

            %% Actually run TxnFun.
            %% Note that this is not in a try/catch -- it is expected to be no-throw 
            %% (apart from maybe 'erlang errors' that might be bugs).

            RetryWrapperFunResult = RetryWrapperFun(NumDeadlockRetries, RetryWrapperFun),

            %% We must always cleanup, even if TxnFun has returned an error
            try ?BDB_PC:release_bdb_port_handle(BdbPortHandle) of

                ok -> RetryWrapperFunResult   % final success (this is returned to caller)

            catch
                {bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResources}} ->

                    case RetryWrapperFunResult of 

                        {ok, TxnFunResult} ->
                            %% TxnFun returned normally, but accidentally left resources open
                            %% We obviously consider this to be an error (so will throw on the client),
                            %% but we do return TxnFun's return value to help with diagnostics.
                            %% We include the stack-trace to make the txn_fun_error a 3-tuple (like all the others)
                            %% and to aid debugging by giving the name of the TxnFun (in the args to do_txn_no_throw).
                            {error, {txn_fun_error, {return_with_open_resources, NumResources, TxnFunResult}}};

                        _ ->
                            %% Here TxnFun failed for some reason, so it's not suprising that 
                            %% it left resources open.  We ignore the latter, and return the real TxnFun error.
                            RetryWrapperFunResult
                    end;

                %% Other exceptions:
                %%
                %% If release_bdb_port_handle() failed to contact the port_coordinator (to tell it to unlink the caller)
                %% (e.g. if the port_coordiantor happened to crash after TxnFun returned but before we call 
                %% release_bdb_port_handle) then gen_server:call() call exit() such that we get:
                %%
                %%   exit:{noproc, {gen_server, call, [Server, Args]}}
                %%   exit:{killed, {gen_server, call, [Server, Args]}}
                %%
                %% In that case we *could* in theory simply ignore the exit, and just return the result of TxnFun, 
                %% as the caller doesn't care (at the moment) that the port coordinator died after TxnFun completed.
                %%
                %% But note that it is NOT safe to ignore All exits/errors from release_bdb_port_handle, because 
                %% release_bdb_port_handle also calls group_valmap_cleanup in the context of the caller process.
                %% If group_valmap_cleanup finds that the port has failed then it crashes with
                %%
                %%   error:badarg   and erlang:get_stacktrace returns: [{erlang, port_command, [ErlangPort, _]}, {berkeley_db_drv, group_valmap_cleanup,2}, {berkeley_db_port_coordinator, release_bdb_port_handle,1}, {berkeley_db_helpers,do_txn_no_throw,3} | _]
                %%
                %% In that case it is NOT safe to return TxnFunResult, as if TxnFun left resources open (e.g. did not commit the txn) 
                %% then it's result is invalid e.g. the transaction will be rolled back, either in the C port stop() function, 
                %% or in BDB recovery).
                %%
                %% Currently we don't distinguish between the cases -- we just return an error and do *not* include TxnFunResult.
                %%
                %% Final note: if the port-coordinator is on a different node to the caller of do_txn() 
                %% (which is illegal) and the target node is unreachable then gen_server:call exits with:
                %%
                %%   exit:{nodedown, {gen_server, call, _}}
                %%
                %% If the node *does* exist then the TxnFun will probably crash with 
                %%
                %%   error:badarg   and erlang:get_stacktrace returns: [{erlang, port_command, [ErlangPort, _]} | _]
                %%  
                %% and even if the TxnFun does not crash like that, release_bdb_port_handle() certainly will
                %% (when it calls group_valmap_cleanup).

                %% If the BDB port is closed then that is a dynamic error, and we want to distinguish it clearly 
                %% from other badarg errors.
                error:badarg ->
                    case erlang:get_stacktrace() of 

                        [{erlang, port_command, [ErlangPort, _]} | _] = ReleaseStackTrace ->
                            {error, {bdb_helper_error, {release_bdb_port_handle, {{port_is_unreachable, ErlangPort}, ReleaseStackTrace}}}};

                        ReleaseStackTrace ->
                            %% For any other error we just convert it like the default clause does.
                            {error, {bdb_helper_error, {release_bdb_port_handle, {{error, badarg}, ReleaseStackTrace}}}}
                    end;

                %% This handles all of the gen_server:call exits mentioned earlier (the most common type of failure)
                %% We don't expect any other thrown errors, so we do include the atom 'throw' if one occurs.
                ReleaseClass:ReleaseReason -> 
                    {error, {bdb_helper_error, {release_bdb_port_handle, {{ReleaseClass, ReleaseReason}, erlang:get_stacktrace()}}}}
            end

    catch 
        %% acquire_bdb_port_handle failed.
        %% See comments in the catch-clause for release_bdb_port_handle, above.
        %% However, unlike release_bdb_port_handle, acquire_bdb_port_handle doesn't do anything significant 
        %% on the client side; it's just a gen_server:call.
        %% We don't expect any thrown errors, so we do include the atom 'throw' if one occurs.
        AcquireClass : AcquireReason -> 
            {error, {bdb_helper_error, {acquire_bdb_port_handle, {{AcquireClass, AcquireReason}, erlang:get_stacktrace()}}}}
    end.


%%
%% Helpers for use from an Erlang shell
%%

txn_fun_write_coredump(BdbPort, Filename) ->
    ?BDB:write_coredump(BdbPort, Filename).


txn_fun_adjust_port_logging(BdbPort, {Operation, Flags}) ->
    ?BDB:adjust_port_logging(BdbPort, Operation, Flags).


txn_fun_env_set_verbose(BdbPort, {EnableDisable, VerboseTypeList}) ->

    lists:foreach(fun(VerboseType) -> ?BDB:env_set_verbose(BdbPort, VerboseType, EnableDisable) end,
                  VerboseTypeList).


txn_fun_keys_values_foldl(BdbPort, {DbName, Fun, InitAcc}) ->

    keys_values_foldl(BdbPort,
                      ?BDB_PC:get_db_handle(BdbPort, bdb_DB_UNKNOWN, DbName),
                      Fun,
                      InitAcc).


txn_fun_keys_values_bulk_foldl(BdbPort, {DbName, SizeOfChunk, Fun, InitAcc}) ->

    keys_values_bulk_foldl(BdbPort,
                           ?BDB_PC:get_db_handle(BdbPort, bdb_DB_UNKNOWN, DbName),
                           SizeOfChunk,
                           Fun,
                           InitAcc).


%% Currently this is always the default isolation level                                                                                                                         
%% -- i.e. serializable, not read-committed (degree 2) or read-uncommitted (degree 1)                                                                                           
keys_values_foldl(BdbPort, DB, Fun, InitAcc) -> 

    keys_values_foldl_from_cursor(BdbPort,
                                  ?BDB:db_cursor(BdbPort, DB, void, []),
                                  Fun,
                                  InitAcc).

keys_values_foldl_from_cursor(BdbPort, Cursor, Fun, Acc) ->

    try ?BDB:c_get(BdbPort, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
        {Key, Value} -> 
            keys_values_foldl_from_cursor(BdbPort, Cursor, Fun, Fun({Key, Value}, Acc))
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            ?BDB:c_close(BdbPort, Cursor),
            Acc
    end.

keys_values_foldl_from_cursor_without_close(BdbPort, Cursor, Fun, Acc) ->

    try ?BDB:c_get(BdbPort, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
        {Key, Value} -> 
            keys_values_foldl_from_cursor_without_close(BdbPort, Cursor, Fun, Fun({Key, Value}, Acc))
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            Acc
    end.


%% This variant uses c_get_multiple, which can greatly reduce the number 
%% of calls made to BDB.

keys_values_bulk_foldl(BdbPort, DB, SizeOfChunk, Fun, InitAcc) -> 

    keys_values_bulk_foldl_from_cursor(BdbPort,
                                       ?BDB:db_cursor(BdbPort, DB, void, []),
                                       SizeOfChunk,
                                       Fun,
                                       InitAcc).

keys_values_bulk_foldl_from_cursor(BdbPort, Cursor, SizeOfChunk, Fun, Acc) ->

    try ?BDB:c_get_multiple(BdbPort, Cursor, <<>>, <<>>, SizeOfChunk, [bdb_DB_NEXT]) of
        KeysValuesInChunk -> 
            NewAcc = lists:foldl(Fun, Acc, KeysValuesInChunk),
            keys_values_bulk_foldl_from_cursor(BdbPort, Cursor, SizeOfChunk, Fun, NewAcc)
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            ?BDB:c_close(BdbPort, Cursor),
            Acc
    end.

keys_values_bulk_foldl_from_cursor_without_close(BdbPort, Cursor, SizeOfChunk, Fun, Acc) ->

    try ?BDB:c_get_multiple(BdbPort, Cursor, <<>>, <<>>, SizeOfChunk, [bdb_DB_NEXT]) of
        KeysValuesInChunk -> 
            NewAcc = lists:foldl(Fun, Acc, KeysValuesInChunk),
            keys_values_bulk_foldl_from_cursor_without_close(BdbPort, Cursor, SizeOfChunk, Fun, NewAcc)
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            Acc
    end.


txn_fun_list_all_open_databases(BdbPort, _) ->
    ?BDB_PC:list_all_open_databases(BdbPort).


%% This one is very useful for verifying replication correctness 
%% - run it on all sites to check that they have bit-identical contents.

txn_fun_md5_all_keys_values_in_all_on_disk_databases(BdbPort, SizeOfChunk) ->

    %% We sort the DBIds to ensure that we traverse them in the same order on all sites.
    SortedDBIds = lists:sort( list_all_on_disk_databases(BdbPort) ),

    txn_fun_md5_all_keys_values_in_multiple_databases(BdbPort, {SizeOfChunk, SortedDBIds}).


list_all_on_disk_databases(BdbPort) ->

    InternalDatabasesToIgnore = ["__db.rep.db"],

    DbFilenames = filelib:wildcard("*.db", get_value_for_key(env_dir, ?BDB_PC:get_config(BdbPort))) 
        -- InternalDatabasesToIgnore,

    %% Because we are accessing *.db files on disk, we don't know if they 
    %% contain multiple internal databases.  
    %% So we have to open each one to see, and generate the necessary {Filename, InternalName}
    %% tuples to pass to txn_fun_md5_all_keys_values_in_multiple_databases.

    lists:foldl(
      fun(Filename, Acc) ->
              %%io:format("Examining ~p~n", [Filename]),
              DB = ?BDB:db_create(BdbPort, []),
              ?BDB:db_open(BdbPort, DB, void, Filename, "", bdb_DB_UNKNOWN, [bdb_DB_RDONLY], 0),

              %% TODO: Before BDB v4.6 there is NO WAY to tell if we have opened the meta-data 'directory'
              %% for a multi-database (the keys are are database names), or whether we have opened a 
              %% single-database-in-a-single-file (the keys are real application data).
              %% This is fixed in BDB v4.6 with the addition of the DB->get_multiple test.
              %% For now we fake it by trying to open an internal database using the first key
              Cursor = ?BDB:db_cursor(BdbPort, DB, void, []),
              try ?BDB:c_get(BdbPort, Cursor, <<>>, <<>>, [bdb_DB_FIRST]) of

                  {FirstKey, _} ->

                      FirstKeyStr = binary_to_list(FirstKey),
                      %%io:format("Trying to open ~p~n", [{Filename, FirstKeyStr}]),
                      FirstInternalDB = ?BDB:db_create(BdbPort, []),
                      try ?BDB:db_open(BdbPort, FirstInternalDB, void, Filename, FirstKeyStr, bdb_DB_UNKNOWN, [bdb_DB_RDONLY, bdb_DB_NOMMAP], 0) of

                          ok ->
                              %% List all of the internal databases, in sorted order -- 
                              %% and cons them onto the front of Acc.
                              %% Don't close the cursor, so we can close it in 'after' for simpler code.
                              lists:sort(
                                keys_values_foldl_from_cursor_without_close(
                                  BdbPort, 
                                  Cursor, 
                                  fun({KeyDBName, _}, InnerAcc) -> 
                                          [{Filename, binary_to_list(KeyDBName)} | InnerAcc] 
                                  end, 
                                  [{Filename, FirstKeyStr} | Acc]))
                      catch 
                          {bdb_error, einval} ->
                              %% first key is not the name of an internal database, so just add the filenamea
                              %%io:format("~p is NOT the name of an internal database~n", [{Filename, FirstKeyStr}]),
                              [Filename | Acc]

                      after
                          ?BDB:db_close(BdbPort, FirstInternalDB, [])
                      end

              catch
                  {bdb_error, bdb_DB_NOTFOUND} ->
                      %% empty database, so just add the filename
                      %%io:format("~p is an empty physical database~n", [Filename]),
                      [Filename | Acc]

              after
                  ?BDB:c_close(BdbPort, Cursor),
                ?BDB:db_close(BdbPort, DB, [])
              end
      end,
      [],
      DbFilenames).


get_value_for_key(Key, AssocList) ->

    case lists:keysearch(Key, 1, AssocList) of
        {value, {Key, Value}} -> Value;
        false                 -> throw({missing_required_key, Key, AssocList})
    end.


%% ListDBIds is a list of the form returned by ?BDB_PC:list_all_open_databases
txn_fun_md5_all_keys_values_in_multiple_databases(BdbPort, {SizeOfChunk, ListDBIds}) ->

    erlang:md5_final(
      lists:foldl(
        %% We MD5 the name of the database before we do its key values, so we can't confuse the contents of  
        %% one database with another.
        fun({DBFilename, DBName} = DBId, MD5Context) ->
                %% md5_update() takes an iolist, so we can do it in one call, 
                %% and we use a separator that is illegal in stand-alone DBFilenames
                md5_all_keys_values(BdbPort, SizeOfChunk, DBId, erlang:md5_update(MD5Context, [DBFilename, 0, DBName]));
           (DBFilename, MD5Context) ->
                md5_all_keys_values(BdbPort, SizeOfChunk, DBFilename, erlang:md5_update(MD5Context, DBFilename))
        end,
        erlang:md5_init(),
        lists:sort(ListDBIds))).  % sort the list, to avoid false negatives

md5_all_keys_values(BdbPort, SizeOfChunk, DBId, MD5Context) ->

    io:format("Running MD5 on all key/values in ~p~n", [DBId]),
    keys_values_bulk_foldl_from_cursor(
      BdbPort,
      ?BDB:db_cursor(BdbPort, 
                     ?BDB_PC:get_db_handle_with_db_config(BdbPort, 
                                                          bdb_DB_UNKNOWN, 
                                                          DBId, 
                                                          %% remove DB_CREATE as it is illegal with DB_UNKNOWN
                                                          [{db_open_flags, [bdb_DB_AUTO_COMMIT,
                                                                            bdb_DB_READ_UNCOMMITTED]}
                                                          ]),
                     void, 
                     []),
      SizeOfChunk,
      fun({Key, Value}, InnerMD5Context) ->
              erlang:md5_update(InnerMD5Context, [Key, Value]) % takes an iolist, so we can do it in one call
      end,
      MD5Context).


 
%% This is useful if we want to write a new function from the shell.
%% AFAIK, functions defined in the shell cannot be referenced via Module:Function.
%% But you _can_ define funs, which are compiled on demand in the shell
%% (and the code is even sent across nodes if necessar)
txn_fun_do_fun(BdbPort, Fun) ->
    Fun(BdbPort).

