
%%
%% berkeley_db_msg_queue.hrl
%%


%%
%% Distributed deadlocks are theoretically possible with this module (although extremeley unlikely).
%%
%% This is because the sendpump process holds locks on the sender's databases while it 
%% sends the message to the receiverpump, while the receiverpump appends and commits to it's own 
%% database, and while the receiverpump sends an ack back.
%% The DB_QUEUE has record level locking so this is not an issue for that database.  However, if the current 
%% record does not entirely fit into the DB_QUEUE'S fixed-length record, then a lock will be held in 
%% the DB_BTREE too.  
%% (This is one reason why it is strongly advisable that most records fit in the DB_QUEUE record length)
%%
%% The lock on the DB_BTREE may block enqueing of new messages by workers on the sender side.
%%
%% The workers are likely to be attempting to enqueue the message in the context of a transaction that 
%% is holding locks on arbitray other sender-side databases.
%%
%% The receivepump may block trying to append and commit the message locally due to a sink worker 
%% attempting to consume a message and holding a similar lock on the receiver's queue DB_BTREE.
%%
%% If the receiver sink worker happens to also be attempting to enqueue a message in an entirely separate
%% queue that goes back to the sender's node, then the receiver sink worker may block on a DB_BTREE lock
%% on that outgoing queueing, because it's receiver (now on the original sender node) is likewise blocked.
%% The second receiver could be blocked because a sink worker on it's node (the original sender node)
%% may be holding a DB_BTREE lock (consuming a message) while attempting to do work on arbitrary databases 
%% in that environment.  And THAT may be blocked on locks held by the original sender workers.
%%
%% In summary, if two 2 nodes have 2 queues between them (one in each direction)
%% with endpoints in the same BDB environment on each node, then lock cycles can occur due to the nature 
%% of enqueuer and sink processes wanting to do other work as part of the same transaction that 
%% enqueues or consumes a message.  The msg queue DB_QUEUE and DB_BTREE databases at each end of each 
%% queue are entirely independent and so cannot deadlock, but the transactions that modify them can 
%% (and almost certainly will, for that is the entent) share common databases (and therefore common locks).
%%
%% This situation is likely to be exacerbated if we ever have multiple concurrent sendpumps/receivepumps to 
%% improve throughput (at the expense of in-order delivery).  In this case, the DB_BTREEs at each end 
%% may cause lock cycles too.
%%
%% We use lock/transaction timeouts to detect and handle such distributed deadlocks.
%%
%% Berkeley DB only enforces timeouts when the deadlock detector is called.  Therefore the the caller should 
%% ensure that all port coordinators used in applications that use this module run a periodic deadlock detector 
%% -- ideally at a frequency (e.g. this value plus a bit / 2 ) that will detect timeouts of this duration quickly.
%% There is a port coordinator config option to enable a periodic deadlock detector, so this is trivial to do.

-define(BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS, (10 * 1000)).


%%
%% bdb_msg_q represents a combined DB_QUEUE and DB_BTREE (the later is used
%% to hold any overflow for messages that won't fit in the DB_QUEUE's fixed-length record).
%%
%% The object does not change if it is used for enqueuing or de-queuing. 
%%

-record(bdb_msg_q, {qname, 
                    bdb_port_handle, 
                    db_q, 
                    q_reclen,               % kept so we know how to split large messages across db_q and db_btree
                    db_btree,
                    sequence_server_name}). % cached to avoid re-creating it on every enqueue call.


%% For convenience and performance (to avoid unecessary conversions at both ends)
%% the enqueue API accepts any of the following as messages
%%   - arbitrary Erlang terms 
%%   - arbitrary Erlang terms that are *already* encoded via term_to_binary
%%   - or raw binaries (which are not
%%
%% Erlang terms are encoded with term_to_binary (and flagged as sucn).
%% Raw binaries are sent untouched.
%%
%% The consume API does NOT automatically call binary_to_term on 
%% messages that are flagged as external-term format, as the consumer 
%% may want to keep them in that format, without paying an potentially expensive 
%% and entirely redundant conversion.
%%
%% So the consume API returns a flag to say if the message is an external 
%% term or not, and leaves it to the application to decide what to do.
%%
%% The flag is a bit mask, so we can add more flags later.

-define(BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM, 16#1).


