
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/file.h>

#include <erl_driver.h>
#include <erl_driver_tk.h>

#include "my-berkeley_db.h"

#include <time.h>


static int make_dbt(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount, DBT *dbt, char *orig_buf);
static int enqueue_untagged_event_notification(descriptor_t *desc, int event_command_code, u_int32_t payload_uint32, long payload_long);


/* Support for writing coredumps from a live application (and continuing afterwards).
 * We use this from the my_errcall callback (for BDB's logging of serious internal conditions)
 * and also expose a driver command to write coredumps at will (useful from remote-shells).
 *
 * We use Google's coredumper library, which requires us to hold a mutex around the call to WriteCoredump().
 * Therefore this support means EDTK and examples/berkeley_db both require DRIVER_USING_PTHREADS is set.
 * Neither will currently compile without that.
 * (Obviously the EDTK support for private threadpools also requires it.)
 */
#ifdef DRIVER_USING_PTHREADS

/* This mutex projects access to all of the following statics */
static pthread_mutex_t write_coredump_global_mutex = PTHREAD_MUTEX_INITIALIZER;

/* For writing core-dumps using http://goog-coredumper.sourceforge.net/ */
#include <google/coredumper.h>
#include <dlfcn.h>
static int    coredumperLibraryOpened   = 0;
static void * coredumperLibraryHandle   = NULL;
static int (*pfnWriteCoreDump)(const char *);

/* In the errcall callback handler we will write a coredump if the msg string from BDB matches (or optionally fails to match)
 * a regular expression, so we need this regex library.
 * Note: I don't know if this is thread-safe (it's README does not mention thread-safety at all), so
 * all calls are made while holding write_coredump_global_mutex. */
static int    pcreLibraryOpened   = 0;
static void * pcreLibraryHandle   = NULL;
static pcre * (*pfnPCRECompile)(const char *, int, const char **, int *, const unsigned char *) = NULL;
static int    (*pfnPCREExec)(const pcre *, const pcre_extra *, PCRE_SPTR, int, int, int, int *, int) = NULL;
static void   (**pfnPCREFree)(void *) = NULL;
/* (Other associated variables -- e.g. for throttling the rate of
 *  coredumps -- are added to descriptor_t by berkeley_db.xml).
 * **NOTE: those desc members are protected write_coredump_global_mutex, NOT desd->mutex,
 * for simplicity of code / mutex ordering). */

#endif /* DRIVER_USING_PTHREADS */


/* Support for passing a piece of data back from event_notify callback
 * via Thread Local Storage.
 * This is used to cause txn_commit, txn_abort, db_put(void-txn) to
 * throw:
 *         {bdb_error, not_durably_replicated}
 *
 * NOTE: this appears to be a global variable -- i.e. is not part of descriptor_t,
 * and therefore appears to be shared across all open BDB ports/environments in this Erlang node.
 * But that is not the case, as this is *only* called from private threadpool threads
 * (the BDB driver *requires* that private threadpools be enabled, and the Erlang VM's
 * threadpool not be used, to avoid several kinds of deadlock).
 * As each threadpool is private to a BDB port/environment, this variable is
 * not shared across ports/environments (indeed it is not even shared
 * with a single port/environment -- there is a separate instance per thread,
 * which is the entire point.
 * IMPORTANT: if EDTK is ever changed to support multiple ports sharing global threadpools
 * then this code will break badly.  It must then be changed to use
 * the POSIX functions to allocate a thread-local key (which is stored in descriptor_t).
 * See pthread_key_create, pthread_(set|get)specific, pthread_key_delete
 */
static __thread int tls_not_durably_replicated_error;


/*
** We flip around the return value of db_env_create so that we normally
** return the new DB_ENV *, which makes EDTK's simple valmap mechanism
** happy.
*/

DB_ENV *
my_db_env_create(u_int32_t flags, int *my_error_value)
{
    DB_ENV	*dbenv;
    int		err;

    if ((err = db_env_create(&dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    return dbenv;
}


/* We modify env_open to do the locking necessary to meet BDB's documented preconditions
 * See detailed comments at the top of berkeley_db_port_coordinator.erl for details. */
int
my_db_env_open(DB_ENV *dbenv, const char *home_dir, u_int32_t flags, int mode)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int            fd = -1;
    int            res, err;

    /* Only an assert on an int, so we don't bother with desc->mutex */
    assert( -1 == desc->env_ownership_advisory_lock_fd );

    /* home_dir CAN be NULL, e.g. for a purely in-memory environment using DB_PRIVATE.
     * We only do locking if it is not NULL */
    if ( NULL == home_dir )
    {
        if ( ! (flags & DB_PRIVATE) )
        {
            EDTK_PORT_LOG_FORCE(desc, "home_dir may only be NULL if DB_PRIVATE is set");
            return EINVAL;
        }

        /* If DB_PRIVATE is set then we fall through to call env-open
         * and if that is successful we will store fd (-1) in desc->env_ownership_advisory_lock_fd.
         * (If it fails then we don't attempt to unlock or close fd as it is -1) */
    }
    else
    {
        char lock_filename[1024];

        res = snprintf(lock_filename, sizeof(lock_filename), "%s/__erlang_edtk.lockfile", home_dir);

       /* Upon successful return, these functions return the number of characters
        * printed  (not  including  the  trailing \0 used  to  end  output to
        * strings).  The functions snprintf and vsnprintf do not write more  than
        * size  bytes (including the trailing \0).  If the output was truncated
        * due to this limit then the return value is  the  number  of  characters
        * (not  including the trailing \0) which would have been written to the
        * final string if enough space had been available. Thus, a  return  value
        * of  size  or  more means that the output was truncated.
        * Earlier implementations return -1 if the output was truncated. */

        if ( (res < 0) || (res >= sizeof(lock_filename)) ) {
            EDTK_PORT_LOG_FORCE(desc, "home_dir path is too long (to construct lockfile name): %s", home_dir);
            return ENAMETOOLONG;
        }
        
        if ( -1 == (fd = open(lock_filename, O_CREAT | O_RDWR, (S_IWUSR|S_IRUSR)|(S_IRGRP|S_IWGRP)|(S_IROTH|S_IWOTH))) ) {
            err = errno;
            EDTK_PORT_LOG_FORCE(desc, "Failed (%d) to open file descriptor for advisory lock on file %s", err, lock_filename);
            return err;
        }

        if ( 0 != flock(fd, LOCK_EX | LOCK_NB) ) {
            /* Possible errors are:
             * EWOULDBLOCK    The file is locked and the LOCK_NB flag was selected.
             * EBADF          fd is not a not an open file descriptor.
             * EINTR          While  waiting  to  acquire  a lock, the call was interrupted by
             *                delivery of a signal caught by a handler.
             * EINVAL         operation is invalid.
             * ENOLCK         The kernel ran out of memory for allocating lock records. */
            err = errno;
            EDTK_PORT_LOG_FORCE(desc, "flock failed (%d) on fd %d for lockfile %s", err, fd, lock_filename);
            close(fd);        /* must not leak the open fd */       
            return err;
        }
        EDTK_PORT_LOG_FORCE(desc, "flock succeeded on fd %d for lockfile %s", fd, lock_filename);
    }
    
    if ( 0 != (res = dbenv->open(dbenv, home_dir, flags, mode)) )
    {
        if ( fd != -1 ) {
            EDTK_PORT_LOG_FORCE(desc, "env_open failed so unlocking and closing lockfile");
            flock(fd, LOCK_UN);
            close(fd);
        }
        return res;
    }

    /* both flock and env_open succeeded */
    pthread_mutex_lock(&desc->mutex);
    desc->env_ownership_advisory_lock_fd = fd;  /* might be -1 (if home_dir is NULL and DB_PRIVATE was specified */
    pthread_mutex_unlock(&desc->mutex);

    return res;
}

/* We modify env_close to cooperate with my_db_env_open on the locking necessary to meet BDB's documented preconditions. */
static int do_env_close_and_release_advisory_lock(DB_ENV *dbenv, u_int32_t flags)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int res_env_close, fd;

    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "dbenv = 0x%p", (void*)dbenv);

    res_env_close = dbenv->close(dbenv, flags);

    /* We _can_ reach here without having acquired the advisory lock, because the
     * lock is acquired before env_open but the DB_ENV handle is created
     * earlier in env_create */
    
    pthread_mutex_lock(&desc->mutex);
    fd = desc->env_ownership_advisory_lock_fd;
    pthread_mutex_unlock(&desc->mutex);

    if ( fd != -1 )
    {
        EDTK_PORT_LOG_FORCE(desc, "releasing flock on fd %d", fd);

        /* Even if dbenv->close failed, we release the advisory lock. */
        flock(fd, LOCK_UN);
        close(fd);

        pthread_mutex_lock(&desc->mutex);
        desc->env_ownership_advisory_lock_fd = -1;
        pthread_mutex_unlock(&desc->mutex);
    }

    return res_env_close;
}

int
my_db_env_close(DB_ENV *dbenv, u_int32_t flags)
{
    return do_env_close_and_release_advisory_lock(dbenv, flags);
}

/* This variant is used for valmap cleanup, as the function signature must just be the valmap value. */
int
my_db_env_close_valmap_cleanup(DB_ENV *dbenv)
{
    return do_env_close_and_release_advisory_lock(dbenv, 0);
}


void 
my_msgcall(const DB_ENV *dbenv, const char *msg)
{
    /* IMPORTANT: we CANNOT acquire desc->mutex in this callback, as a threadpool thread may have
     * it locked while doing BDB operations and any BDB operation might call this function
     * (e.g. bdb_stop during driver shutdown does that).
     * It is reasonable to require threads to not hold desc->mutex while calling BDB,
     * but is too risky (easy to forget/miss, and then one day a BDB operation tries to log a message...)
     * We already have a theoretically-unsafe-but-practically-safe policy of
     * allowing use of the PORT_LOGGING macros without holding desc->mutex, even
     * though they read from desc->port_log_file and desc->port_label.
     * This is 'practically safe' as those things don't change apart from during port-coordinator
     * initialization, when nothing else is happening in the driver.
     * So it is similarly 'practically safe' to read those fields without acquring the mutex here. */
       
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    struct timeval curr_time;
    gettimeofday(&curr_time, NULL);
    
    EDTK_LOGGING_FPRINTF(desc->port_log_file, "%s: %d:%lx: %lu:%lu: MSG: %s\r\n", desc->port_label, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, msg);
    /* as this is only an informational message, we don't flush the log file */
}


/*
 * Support for writing coredumps - see comments at top of file.
 */

#ifdef DRIVER_USING_PTHREADS

/* For some reason, I can't get automatic loading of Google's libcoredumper.so
 * to work.  So to avoid "undefined symbol: WriteCoreDump" errors, I load it
 * and find the symbol myself.  It works, at least.
 * This ASSUMES that the caller holds write_coredump_global_mutex */
void open_coredumper_library_while_holding_mutex(descriptor_t * desc)
{
    if ( ! coredumperLibraryOpened )
    {
        coredumperLibraryOpened = 1;
        if ( NULL == (coredumperLibraryHandle = dlopen("libcoredumper.so", RTLD_LOCAL | RTLD_LAZY)) )
        {
            EDTK_PORT_LOG_FORCE(desc, "Failed to dlopen libcoredumper.so : %s", dlerror());
        }
        else
        {
            if ( NULL == (pfnWriteCoreDump = dlsym(coredumperLibraryHandle, "WriteCoreDump")) ) {
                EDTK_PORT_LOG_FORCE(desc, "Failed to find symbol WriteCoreDump in libcoredumper.so : %s", dlerror());
            }
        }
    }
}


/* We ASSUME that the caller has already acquired write_coredump_mutex,
 * as it also protects our own static variables like errcall_coredump_interval_secs.
 * Returns 0 on success, non-zero (errno) on failure */
static int write_coredump_while_holding_mutex(descriptor_t * desc, const char * filename)
{
    int               res;
    time_t            now;
    char              time_buf[30];       /* docs say 'at least 26 bytes' */

    time(&now);
    ctime_r(&now, time_buf);
    time_buf[sizeof(time_buf) - 1] = 0;   /* ensure null-termination */
    time_buf[strlen(time_buf) - 1] = 0;   /* the asctime functions append a \n for no good reason */

    open_coredumper_library_while_holding_mutex(desc);
    
    if ( NULL == pfnWriteCoreDump )
    {
        res = ELIBACC;
    }
    else
    {
        EDTK_PORT_LOG_FORCE(desc, "%s: Writing coredump to %s", time_buf, filename);
        res = (*pfnWriteCoreDump)(filename);

        /* Get the end-time */
        time(&now);
        ctime_r(&now, time_buf);
        time_buf[sizeof(time_buf) - 1] = 0;   /* ensure null-termination */
        time_buf[strlen(time_buf) - 1] = 0;   /* the asctime functions append a \n for no good reason */

        if ( 0 == res )
        {
            EDTK_PORT_LOG_FORCE(desc, "%s: Sucessfully written coredump to %s", time_buf, filename);
        }
        else
        {
            int idx;
            
            res = errno;
            idx = res - 1;
            if (  0 <= idx  &&  idx < NUM_ARY_ELEMS(errno_atoms)
               && errno_atoms[idx].code != 0 )
            {
                EDTK_PORT_LOG_FORCE(desc, "%s: Failed to write coredump %s : %s", time_buf, filename, errno_atoms[idx].atom_text);
            }
            else
            {
                EDTK_PORT_LOG_FORCE(desc, "%s: Failed to write coredump %s : unknown errno %d", time_buf, filename, res);
            }
        }
    }
    return res;
}


/*
 * Support for conditional writing of coredumps in the BDB errcall callback
 * (based on regex matching of the error string from BDB)
 */
   
/* This ASSUMES that the caller holds write_coredump_mutex */
void open_pcre_library_while_holding_mutex(descriptor_t * desc)
{
    if ( ! pcreLibraryOpened )
    {
        pcreLibraryOpened = 1;
        if ( NULL == (pcreLibraryHandle = dlopen("libpcre.so", RTLD_LOCAL | RTLD_LAZY)) )
        {
            EDTK_PORT_LOG_FORCE(desc, "Failed to dlopen libpcre.so : %s",  dlerror());
        }
        else
        {
            if ( NULL == (pfnPCRECompile = dlsym(pcreLibraryHandle, "pcre_compile")) ) {
                EDTK_PORT_LOG_FORCE(desc, "Failed to find symbol pcre_compile in libpcre.so : %s", dlerror());
            }
            if ( NULL == (pfnPCREExec = dlsym(pcreLibraryHandle, "pcre_exec")) ) {
                EDTK_PORT_LOG_FORCE(desc, "Failed to find symbol pcre_exec in libpcre.so : %s", dlerror());
            }
            if ( NULL == (pfnPCREFree = dlsym(pcreLibraryHandle, "pcre_free")) ) {
                EDTK_PORT_LOG_FORCE(desc, "Failed to find symbol pcre_free in libpcre.so : %s", dlerror());
            }
        }
    }
}


int my_set_write_coredump_on_errcall(const DB_ENV *dbenv, int coredump_interval_secs, int regex_mode, const char * path_prefix, const char *regex)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int            res;

    pthread_mutex_lock(& write_coredump_global_mutex);

    open_pcre_library_while_holding_mutex(desc);

    /* set_write_coredump_on_errcall() is typically called early -- in port-coordinator initialization.
     * So we take advantage of that, and load the coredumper library too, so the first coredump
     * won't be slower than it needs to be -- *and* we can visibly fail if we can't find a library. */
    open_coredumper_library_while_holding_mutex(desc);

    if ( (NULL == pfnPCRECompile)  ||  (NULL == pfnWriteCoreDump) )
    {
        res = ELIBACC;
    }
    else
    {
        pcre  *        compiled_regex;
        const char *   error;
        int            erroffset;
        
        compiled_regex = (*pfnPCRECompile)(regex,                /* the pattern */
                                           0,                    /* default options */
                                           &error,               /* for error message */
                                           &erroffset,           /* for error offset */
                                           NULL);                /* use default character tables */
        if ( NULL == compiled_regex )
        {
            EDTK_PORT_LOG_FORCE(desc, "PCRE compilation failed at offset %d: %s", erroffset, error);
            res = EINVAL;        
        }
        else
        {
            /* These members of desc are protected by write_coredump_global_mutex, NOT desc->mutex */
            desc->errcall_coredump_interval_secs  = coredump_interval_secs;
            desc->errcall_coredump_regex_mode     = regex_mode;

            snprintf(desc->errcall_coredump_path_prefix, sizeof(desc->errcall_coredump_path_prefix), path_prefix);
            desc->errcall_coredump_path_prefix[sizeof(desc->errcall_coredump_path_prefix) - 1] = 0;
            
            free_coredump_compiled_regex(desc);
            desc->errcall_coredump_compiled_regex = compiled_regex;
            desc->errcall_last_coredump_time      = 0;  /* reset, so the change takes effect immediately */
            res = 0;
        }
    }

    pthread_mutex_unlock(& write_coredump_global_mutex);
    return res;
}

void free_coredump_compiled_regex(descriptor_t *desc)
{
    if (  desc->errcall_coredump_compiled_regex != NULL
       && pfnPCREFree != NULL )
    {
        (**pfnPCREFree)(desc->errcall_coredump_compiled_regex);
        desc->errcall_coredump_compiled_regex = NULL;
    }
}

int
my_write_coredump(DB_ENV *dbenv, const char *filename)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int            res;

    pthread_mutex_lock(& write_coredump_global_mutex);

    if ( strlen(filename) )
    {
        res = write_coredump_while_holding_mutex(desc, filename);
    }
    else
    {
        time_t now;
        char default_filename[128];

        time(&now);
        snprintf(default_filename, sizeof(default_filename), "core_requested.%ld", now);
        default_filename[sizeof(default_filename)-1] = 0;

        res = write_coredump_while_holding_mutex(desc, default_filename);
    }

    pthread_mutex_unlock(& write_coredump_global_mutex);
    return res;
}

#endif /* DRIVER_USING_PTHREADS */


void 
my_errcall(const DB_ENV *dbenv, const char *errpfx, const char *msg)
{
    /* IMPORTANT: we CANNOT acquire desc->mutex in this callback.
     *            -- see my_msgcall() earlier, for rationale */
    
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    struct timeval curr_time;
    time_t now;                /* TODO: use timeval_subtract() instead of difftime() to avoid asking for the system time twice */
    double time_difference;
    char   time_buf[30];       /* docs say 'at least 26 bytes' */
    
    gettimeofday(&curr_time, NULL);

    EDTK_LOGGING_FPRINTF(desc->port_log_file, "%s: %d:%lx: %lu:%lu: ERROR: %s: %s\r\n", desc->port_label, getpid(), pthread_self(), curr_time.tv_sec, curr_time.tv_usec, errpfx ? errpfx : "", msg);

    /* As this is potentially a serious error, we flush the log file
     * (although the default of stderr is already auto-flushing). */
    if ( stderr != desc->port_log_file ) {
        fflush(desc->port_log_file);
    }

    /* Optionally write coredump, based on regex matching of msg string from BDB */
#ifdef DRIVER_USING_PTHREADS   

    /* TODO: use timeval_subtract() instead of difftime() to avoid asking for the system time twice */
    time(&now);

    /* just for logging a human-readable timestamp to associate with coredump (EDTK logging may give raw gettimeofday() values as well) */
    ctime_r(&now, time_buf);
    time_buf[sizeof(time_buf) - 1] = 0;   /* ensure null-termination */
    time_buf[strlen(time_buf) - 1] = 0;   /* the asctime functions append a \n for no good reason */

    pthread_mutex_lock(& write_coredump_global_mutex);
    
    /* These members of desc are protected by write_coredump_global_mutex, NOT desc->mutex */
    if ( 0.0L == desc->errcall_coredump_interval_secs )
    {
        EDTK_PORT_LOG_FORCE(desc, "%s: Not writing coredump as they are disabled with desc->errcall_coredump_interval_secs = 0", time_buf);
    }
    else
    {
        if ( (time_difference = difftime(now, desc->errcall_last_coredump_time)) < desc->errcall_coredump_interval_secs )
        {
            EDTK_PORT_LOG_FORCE(desc, "%s: Not writing coredump as they are limited to one every %f seconds but only %f seconds have passed", time_buf, desc->errcall_coredump_interval_secs, time_difference);
        }
        else
        {
            /* Do a regular-expression match against msg, to see if we want to write a core-dump */
            open_pcre_library_while_holding_mutex(desc);

            if ( NULL == *pfnPCREExec )
            {
                EDTK_PORT_LOG_FORCE(desc, "%s: Not writing coredump as *pfnPCREExec", time_buf);
            }
            else
            {
                int res = (*pfnPCREExec)(desc->errcall_coredump_compiled_regex, /* the compiled pattern */
                                         NULL,                                   /* no extra data - we didn't study the pattern */
                                         msg,                                    /* the subject string */
                                         strlen(msg),                            /* the length of the subject */
                                         0,                                      /* start at offset 0 in the subject */
                                         0,                                      /* default options */
                                         NULL,                                   /* output vector for substring information */
                                         0);                                     /* number of elements in the output vector */

                /* if ( 0 == res ) {
                 *    EDTK_PORT_LOG_FORCE(desc,"MATCHES!");
                 *} else {
                 *    EDTK_PORT_LOG_FORCE(desc,"NOT matches!");
                 *}
                 */

                if (  ((PCRE_ERROR_NOMATCH == res)  &&  (desc->errcall_coredump_regex_mode == REGEX_IS_EXCLUSIVE))    /* no match, and we want to exclude all that match the pattern (and write core-dumps for the remainder) */
                   || ((0 == res)                   &&  (desc->errcall_coredump_regex_mode == REGEX_IS_INCLUSIVE)) )  /* match, and we want to write core-dumps for the specific pattern */
                {
                    char filename[200];
                    snprintf(filename, sizeof(filename),
                             "%score_bdb_errcall.%ld",
                             desc->errcall_coredump_path_prefix,
                             now);
                    filename[sizeof(filename)-1] = 0;

                    write_coredump_while_holding_mutex(desc, filename);

                    desc->errcall_last_coredump_time = now;
                }
                else
                {
                    /* EDTK_PORT_LOG_FORCE(desc, "%s: Not writing coredump as regex-match result = %d and desc->errcall_coredump_regex_mode = %d", time_buf, res, desc->errcall_coredump_regex_mode); */
                }
            }
        }
    }

    pthread_mutex_unlock(& write_coredump_global_mutex);
#endif /* DRIVER_USING_PTHREADS */
}


/* May need this later */
#if 0
/* From the GNU C libray manual:
 * Subtract the `struct timeval' values X and Y,
 * storing the result in RESULT.
 * Return 1 if the difference is negative, otherwise 0.  */
int
timeval_subtract(struct timeval * result, struct timeval *x, struct timeval *y)
{
  /* Perform the carry for the later subtraction by updating y. */
  if (x->tv_usec < y->tv_usec) {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }

  /* Compute the time remaining to wait.
     tv_usec is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;

  /* Return 1 if result is negative. */
  return x->tv_sec < y->tv_sec;
}
#endif


int
my_env_set_errfile(DB_ENV *dbenv, char *file, char *mode)
{
    FILE	*fp;

    if ( NULL == (fp = fopen(file, mode)) ) {
	return errno;
    }
    dbenv->set_errfile(dbenv, fp);
    return 0;
}

int
my_db_set_errfile(DB *db, char *file, char *mode)
{
    FILE	*fp;

    if ( NULL == (fp = fopen(file, mode)) ) {
	return errno;
    }
    db->set_errfile(db, fp);
    return 0;
}

DB *
my_db_create(DB_ENV *dbenv, u_int32_t flags, int *my_error_value)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    DB	*db;
    int	err;

    if ((err = db_create(&db, dbenv, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    db->app_private = desc;
    return db;
}

int
my_db_close_valmap_cleanup(DB *db)
{
    descriptor_t * desc = (descriptor_t *)(db->app_private);

    /* See detailed comments in port-coordinator's close_environment function
     * for why we pass the DB_NOSYNC flag if the DB handle is transactional. */
    u_int32_t      flags = db->get_transactional(db) ? DB_NOSYNC : 0;
    
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "db = 0x%p, flags = %d", (void*)db, flags);
    return db->close(db, flags);
}

int
my_db_close(DB *db, u_int32_t flags)
{
    descriptor_t * desc = (descriptor_t *)(db->app_private);
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "db = 0x%p", (void*)db);
    return db->close(db, flags);
}

DBC *
my_cursor_create(DB *db, DB_TXN *db_txn, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = db->cursor(db, db_txn, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    /* TODO: if/when DBC gets app_private field, set it here from db->app_private */
    return cursorp;
}

DBC *
my_cursor_dup(DBC *cursor, u_int32_t flags, int *my_error_value)
{
    DBC	*cursorp;
    int	err;

    if ((err = cursor->c_dup(cursor, &cursorp, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }
    /* TODO: if/when DBC gets app_private field, copy it here */
    return cursorp;
}

int
my_cursor_close_valmap_cleanup(DBC *cursor)
{
    /* TODO: as DBC does not currently have an app_private field we do not have access to desc here */
    EDTK_LOG(EDTK_LOG_VERBOSE, "new cursor = 0x%p", (void*)cursor);
    return cursor->c_close(cursor);
}


/* transaction wrappers */

DB_TXN *
my_txn_begin(DB_ENV *dbenv, DB_TXN *parent_db_txn, u_int32_t flags,
             int *my_error_value)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    DB_TXN * new_txn;
    int	err;

    if ((err = dbenv->txn_begin(dbenv, parent_db_txn, & new_txn, flags)) != 0) {
	*my_error_value = err;
	return NULL;
    }

    /* TODO: if/when DB_TXN gets app_private field, set it here from db->app_private */
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "db_txn = 0x%p", (void*)new_txn);
    return new_txn;
}

/* This is only used as a valmap cleanup function */
int
my_txn_abort_valmap_cleanup(DB_TXN *db_txn)
{
    /* TODO: as DB_TXN does not currently have an app_private field we do not have access to desc here */
    EDTK_LOG(EDTK_LOG_VERBOSE, "new db_txn = 0x%p", (void*)db_txn);
    return db_txn->abort(db_txn);
}


int
my_db_verify(DB *db, char *file, char *database, char *outfilename,
	     u_int32_t flags)
{
    FILE * outfile = NULL;

    if ( NULL == outfilename ) {
	return EINVAL;
    }
    if ( NULL == (outfile = fopen(outfilename, "w")) ) {
	return errno;
    }
    return db->verify(db, file, database, outfile, flags);
}

/* These can ONLY be called from reply_xtra_XXX (xreturn functions), from inside ready_async.
 * They must increment *members by 1.
 * They currently ALWAYS return *members (if 1 this indicates success, but it's a number, not 'true').
 * Failure is handled by **hard-aborting** the OS process
 *
 * They also adjust *members and *msgcount as appropriate.
 * IMPORTANT: The caller will ASSERT that *members == 1 (but *msgcount can be more than that) */

int
make_dbt_key(descriptor_t *desc, callstate_t *c,
             ErlDrvTermData *msg, int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.key_in, c->i.ekey_in);
}

int
make_dbt_data(descriptor_t *desc, callstate_t *c,
              ErlDrvTermData *msg, int *members, int *msgcount)
{

    return make_dbt(desc, c, msg, members, msgcount, &c->i.data_in, c->i.edata_in);
}

static int
make_dbt(descriptor_t *     desc,
         callstate_t *      c,
         ErlDrvTermData *   msg,
	 int *              members,
         int *              msgcount,
         DBT *              dbt,
         char *             orig_buf)
{
    char *         bin_mem;
    ErlDrvBinary * bin;

    if ( sys_info.smp_support )
    {
        /* We ARE running with SMP support (or in pipe-mode)
         * so driver_alloc/realloc/free_binary ARE thread safe,
         * so (in env_create) we told BDB to use edtk_driver_binary_alloc/realloc/free
         * instead of standard malloc/realloc/free.
         * So if BDB allocated the data, then we don't need to copy it.
         */
        
        if (dbt->data != orig_buf)
        {
            /* The DBT is nolonger pointing to the original input data.
             * This means that BDB must have allocated new data.
             * As we have told BDB to use edtk_driver_binary_malloc instead of malloc,
             * we already have a driver binary here, so we don't need to copy the data.
             * Note: the BDB allocated data is  freed in free_callstate() by a call to EDTK_PORT_FREE_BDB_MEMORY()
             * as ready_async is not guaranteed to be called (e.g. if driver is closing)
             * However, if ready_async is called, then this binary will have already been sent
             * to the Erlang VM by ready_async/reply_XXX/driver_send_term, so the 
             * the Erlang VM will have incremented its reference count, and it won't actually
             * be freed until the Erlang VM gc's all references to it.
             */
            assert( DB_DBT_MALLOC == dbt->flags );
            bin = edtk_alloced_ptr2ErlDrvBinary(dbt->data);
        }
        else
        {
            /* The DBT is still pointing to the original input data.
             * As we don't know who owns that data, we must allocate a copy here.
             *
             * TODO: The input data is usually (always?) a driver binary (part of the input).
             * If this function had a slightly richer interface (an ErlDrvBinary* binary, char *start, size_t len)
             * we could simply use that binary (increment its refcount), and avoid the alloc/copy.
             * e.g. Without that optimization the Key argument to db_get/c_get is usually copied here
             * (but not always; e.g. if c_get has been told to fetch the key with the data).
             */
            if ( NULL == (bin_mem = edtk_driver_binary_malloc(dbt->size)) ) {
                EDTK_PORT_LOG_AND_ABORT(desc, "edtk_driver_binary_alloc failed %d", dbt->size);
            }
            memcpy(bin_mem, dbt->data, dbt->size);
            bin = edtk_alloced_ptr2ErlDrvBinary(bin_mem);

            /* We need to ensure that the binary is freed properly.
             * We could either set dbt->data = bin_mem and let the
             * call to EDTK_PORT_FREE_BDB_MEMORY() in free_callstate()
             * free it, or we can append it to desc->driver_binaries_to_free_at_end_of_reply_xtra so that 
             * reply_xtra_XXX frees it when it is done. We do the latter, as it is clearer.
             * (Note that in either case, the binary will already have been sent to the Erlang VM,
             * so it's refcount will have been incremented inside driver_send_term,
             * so it won't actually be freed).
             *
             * Appending to desc->driver_binaries_to_free_at_end_of_reply_xtra[] would superficially appear
             * to require that we acquire desc->mutex.
             * However, so long as we ONLY modify this in reply_xtra_XXX (code that
             * is directly called by ready_async), then we don't need the mutex.
             * The Erlang VM guarantees that ready_async will only be called
             * by one thread at a time for a given port instance  (this is true whether
             * port-level or driver-level locking is being done by the Erlang VM)
             */
            assert(desc->num_driver_binaries_to_free_at_end_of_reply_xtra < MAX_RETURN_TERMS);
            desc->driver_binaries_to_free_at_end_of_reply_xtra[desc->num_driver_binaries_to_free_at_end_of_reply_xtra++]
                = bin;
            EDTK_PORT_LOG(EDTK_PORT_LOG_REPLY_XTRA_DRIVER_BINARIES, desc, "desc now has %d reply_xtra binaries; added a copied-BDB-input-DBT binary 0x%p", desc->num_driver_binaries_to_free_at_end_of_reply_xtra, bin);
        }
    }
    else
    {
        /* We are NOT running with SMP support (and are not in pipe-mode)
         * so driver_alloc/realloc/free_binary are NOT thread safe,
         * so BDB is still using standard malloc/realloc/free.
         *
         * So if BDB did allocate the data, we need to allocate a new driver
         * binary to hold it (safe here, as we are only called from ready_async
         * by the single Erlang VM thread), copy the data, and free the original
         * that was allocated by BDB.
         *
         * TODO
         * Actually, for simplicity we currently ALWAYS copy the data, even if BDB did not
         * allocate it.  But in many cases (e.g. the key argument to db_get)
         * the data we are returning is usually (always) a driver binary (part of the input).
         * So we _could_ just use the same binary (increment it's refcount before free_callstate calls
         * driver_free_binary on it).
         * e.g. Without that optimization the Key argument is usually copied here
         * (but not always; e.g. if c_get has been told to fetch the key with the data).
         * HOWEVER, edtk_alloced_ptr2ErlDrvBinary() will NOT work on input data
         * -- e.g. in pipe mode, the input data is buried at some offset inside the
         * single large 'fake' ErlDrvBinary created by pipe-main.c
         * So this function (make_dbt) would have to 
         */
    
        if ( NULL == (bin_mem = edtk_driver_binary_malloc(dbt->size)) ) {
            EDTK_PORT_LOG_AND_ABORT(desc, "edtk_driver_binary_malloc failed");
        }
        memcpy(bin_mem, dbt->data, dbt->size);
        bin = edtk_alloced_ptr2ErlDrvBinary(bin_mem);

        /* If BDB did malloc data for this object then we must free
         * it, now that we have copied it to a driver binary.
         * However, we free it in free_callstate not here, because
         * this function (make_dbt) might not be called.  If driver
         * shutdown has begun after invoke_XX has called BDB but
         * before ready_async is called then ready_async won't
         * actually be called (and therefore cannot call us). Only
         * free_callstate will be called. */

        /* To free the new binary we HAVE to append it to desc->driver_binaries_to_free_at_end_of_reply_xtra
         * (see earlier comments) as dbt->data might already be pointing at data allocated by BDB
         * with malloc (and which will be freed in free_callstate).
         * The earlier comments about not needing to acquire desc->mutex
         * also apply here. */
        assert(desc->num_driver_binaries_to_free_at_end_of_reply_xtra < MAX_RETURN_TERMS);
        desc->driver_binaries_to_free_at_end_of_reply_xtra[desc->num_driver_binaries_to_free_at_end_of_reply_xtra++] = bin;
        EDTK_PORT_LOG(EDTK_PORT_LOG_REPLY_XTRA_DRIVER_BINARIES, desc, "desc now has %d reply_xtra binaries; added a copied-BDB-DBT binary 0x%p", desc->num_driver_binaries_to_free_at_end_of_reply_xtra, bin);
    }

    *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, dbt->size);
    *members = 1;
    return 1;
}


/* This is called when txn_prepare is successful (by a hack in berkeley_db.xml).
 * It is called for each immediate child of the prepared txn, NOT the prepared txn itself
 * (as it clears the slots of each txn it is called on, *after* recursing on it's children. */
void discard_descendents_of_prepared_txn(descriptor_t * desc, int idx)
{
    if ( desc->valmap_db_txn[idx].child_count != 0 )
    {
        int child_idx;
        EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "prepared parent=%d has child_count %d", idx, desc->valmap_db_txn[idx].child_count);
        for (child_idx = 0; child_idx < desc->num_valmap_db_txn_slots; child_idx++)
        {
            if ( desc->valmap_db_txn[child_idx].parent_index == idx )
            {
                EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "found child %d (of parent %d)", child_idx, idx);

                /* Recurse to the leaf nodes and then cleanup on the way back out */
                discard_descendents_of_prepared_txn(desc, child_idx);

                /* Most parents will have few children, and they will be found early.  
                 * So it's probably worth stopping the scan of the array when we've done them all.*/
                if ( desc->valmap_db_txn[idx].child_count == 0 ) {
                   break;
                }
            }
        }
    }
    
    desc->valmap_db_txn[idx].value = NULL;

    /* IMPORTANT: in cleanup_$(.name)_valmap_indes there is a hack section here for "cleanup_additional_valmap_struct_members".
     * berkeley_db.xml has a hack that frees desc->valmap_db_txn[i].distributed_txn_guid.
     * We KNOW that cannot be set for children of a prepared txn.
     * But any other such hacks for db_txn must be included here!!
     * (Maintenance headache of course; berkeley_db.xml does have documentation for this. */         
    assert( NULL == desc->valmap_db_txn[idx].distributed_txn_guid );

    desc->valmap_db_txn[idx].flags = 0;

    ++(desc->valmap_db_txn[idx].generation_id); /* to prevent reuse of expired valmap tuples that the Erlang application might still have */
    /* EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "i=%d, made generation-id=%d", idx, desc->valmap_db_txn[idx].generation_id); */

    if ( desc->valmap_db_txn[idx].parent_index >= 0 ) {
        /* This was a child (the only time it's not a child is the top-level prepared txn).
         * Decrement parent's child count. */
        --(desc->valmap_db_txn[desc->valmap_db_txn[idx].parent_index].child_count);
        EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "i=%d is a child with parent=%d. Have reduced parent's child count to %d", idx, desc->valmap_db_txn[idx].parent_index, desc->valmap_db_txn[desc->valmap_db_txn[idx].parent_index].child_count);
    }

    desc->valmap_db_txn[idx].parent_index = -1;

    /* desc->valmap_db_txn[idx].group_id = -1; */    /* *not* a special value, but handy for debugging */
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "finished discarding children of prepared parent=%d", idx);
}


/* This is only called after a successful call to txn_recover.
 * c->o.num_unresolved_txns is the length of the list.
 * Any unresolved transactions have been installed in
 * i.e. valmap_db_txn[*].value              is the DB_TXN *,
 * and  valmap_db_txn[*].distributed_txn_id is the GUID
 *
 * Note that c->o.num_unresolved_txns has already been capped to
 * avoid overflowing the valmap_db_txn array
 * AND avoid overflowing MAX_RETURN_TERMS when constructing this list.
 * See code in berkeley_db.xml (hack section for txn_recover).
 * If this ever creates more than (7*N) + 1 stack entries then that code needs to change.
 */
int
my_make_list_unresolved_txns(descriptor_t *desc,
                             callstate_t *c,
                             ErlDrvTermData *msg,
                             int *members,
                             int *msgcount)
{
    int txn_idx;

    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "num unresolved txns = %d", (int)(c->o.num_unresolved_txns));
    for (txn_idx = 0; txn_idx < c->o.num_unresolved_txns; ++txn_idx)
    {
        /* Append <<DistTxnGUID>>, #bdb_valmap_db_txn{}.
         * This allocates memory, so may fail. */
        int inner_members = 0;
        /* This consumes 6 stack slots, and creates 2 members. */
        /* EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "before my_LOAD_unresolved_DistTxnId_and_valmap_record *msgcount=%d", *msgcount); */
        if ( 0 == my_LOAD_unresolved_DistTxnId_and_valmap_record(desc, txn_idx, msg, &inner_members, msgcount) )
        {
            EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "after (failed) my_LOAD_unresolved_DistTxnId_and_valmap_record() *msgcount=%d inner_members=%d. Calling txn_discard on remaining %d unresolved txns and returning list of %d txns", *msgcount, inner_members, (int)(c->o.num_unresolved_txns - txn_idx), txn_idx);
            /* Failed.  We explicitly txn_discard all of the remaing transactions
             * and return what we have.
             * This function is only called by read_async, so doing this calls
             * txn_discard from the Erlang VM thread. That's OK as it does not acquire
             * locks, and this should never happen anyway.
             *
             * Note: if we fail to allocate on the first txn then we CANNOT simplj 
             * return an empty list, as that would tell the port_coordinator that there are NO
             * unresolved txns remaining.  That is not safe -- BDB will NOT currently reject attempts
             * to create new transactions if there are unresolved txns, it will just CORRUPT the log.
             * So if we would return an empty list we append an atom to it, which will 
             * cause a badmatch in the port_coordinator */
            
            for ( ; txn_idx < c->o.num_unresolved_txns; ++txn_idx )
            {
                EDTK_PORT_LOG_FORCE(desc, "************ discarding txn #%d", txn_idx);
                desc->valmap_db_txn[txn_idx].value->discard(desc->valmap_db_txn[txn_idx].value, 0);

                EDTK_PORT_LOG(EDTK_PORT_LOG_ALLOC_FREE, desc, "Freeing distributed txn guid for txn #%d", txn_idx);
                EDTK_PORT_FREE(desc, desc->valmap_db_txn[txn_idx].distributed_txn_guid);
                desc->valmap_db_txn[txn_idx].distributed_txn_guid = NULL;
                desc->valmap_db_txn[txn_idx].value                = NULL;
            }
            if ( txn_idx == 0 ) {
                *msgcount = LOAD_ATOM(msg, *msgcount, am_enomem);
                ++(*members);
            }
            break;
        }
        else {
            /* EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "after (success) my_LOAD_unresolved_DistTxnId_and_valmap_record *msgcount=%d inner_members=%d", *msgcount, inner_members); */
        }
        /* make it into a 2-tuple */
        *msgcount = LOAD_TUPLE(msg, *msgcount, inner_members);
        ++(*members);
    }

    /* The list has to end with NIL */
    *msgcount = LOAD_NIL(msg, *msgcount);
    ++(*members);
    
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "added %d out of %ld (*msgcount=%d, *members=%d)", txn_idx, c->o.num_unresolved_txns, *msgcount, *members);

    /* In berkeley_db.xml, this xtra-return type is declared as a 'list'.
     * So reply_xtra will do the final "*msgcount = LOAD_LIST(msg, *msgcount, members);" */

    return *members;
}


/* Helper which adds these two parts to a message.
 * Note, these are NOT wrapped in an enclosing tuple.
 *
 *    DistTxnGUIDBinary, #bdb_valmap_db_txn{}
 *
 * This consumes 6 stack slots, and creates 2 members.
 *
 * Returns 0 if fails, N (number of added members) if successfull.
 * Can only fail due to out-of-memory, and will not change anything if it does fail.
 *
 * If this ever adds more than 6 slots then code in berkeley_db.xml
 * (hack section for txn_recover) must change.
 */
int my_LOAD_unresolved_DistTxnId_and_valmap_record(descriptor_t *desc,
                                                   int txn_idx,
                                                   ErlDrvTermData *msg,
                                                   int *members,
                                                   int *msgcount)
{
    char *         memory_in_binary;
    ErlDrvBinary * binary;

    /* First the binary */
    if ( NULL == (memory_in_binary = edtk_driver_binary_malloc(DB_XIDDATASIZE)) ) {
        EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "edtk_driver_binary_malloc failed");
        *members = 0;
        return 0;
    }
    memcpy(memory_in_binary, desc->valmap_db_txn[txn_idx].distributed_txn_guid, DB_XIDDATASIZE);
    binary = edtk_alloced_ptr2ErlDrvBinary(memory_in_binary);

    /* Ensure that we decrement it's refcount at the end of reply_xtra_XXX
     * (*AFTER* being send to Erlang -- i.e. Erlang will have incremented
     * it's refcount, so it won't really be freed) */
    assert(desc->num_driver_binaries_to_free_at_end_of_reply_xtra < MAX_RETURN_TERMS);
    desc->driver_binaries_to_free_at_end_of_reply_xtra[desc->num_driver_binaries_to_free_at_end_of_reply_xtra++] = binary;
    EDTK_PORT_LOG(EDTK_PORT_LOG_REPLY_XTRA_DRIVER_BINARIES, desc, "desc now has %d reply_xtra binaries; added an unresolved-distributed-txn-guid binary 0x%p", desc->num_driver_binaries_to_free_at_end_of_reply_xtra, binary);

    *msgcount = LOAD_BINARY(msg, *msgcount, binary, 0, DB_XIDDATASIZE);

    /* Now the valmap_db_txn.
     * This MUST match the record definition generated by hrl_template.gsl.
     * Similar code is in reply_ok_valmap(). */
    *msgcount = LOAD_ATOM(msg, *msgcount, am_valmap_db_txn);
    *msgcount = LOAD_PORT(msg, *msgcount, driver_mk_port(desc->port));
    *msgcount = LOAD_INT(msg, *msgcount, txn_idx);
    *msgcount = LOAD_INT(msg, *msgcount, desc->valmap_db_txn[txn_idx].generation_id);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 4);
    /* If this ever adds more than 6 slots then code in berkeley_db.xml
     * (hack section for txn_recover) must change. */

    *members = 2;
    return 2;  /* two new Erlang terms added */
}


/*
** This is called when there is a prepared transaction that must be resolved incrementally.
** It enqueues a command that will send an the event to the ports connected process (the port_coordinator)
** as an untagged_event message.
** The command includes the txn_idx, AND the generation_id from the txn_id slot.
** The generation_id allows the 'send event' code (in berkeley_db.xml, injected into ready_async)
** to determine if the transaction has been commited or aborted while the 'send event' command
** was waiting to be processed. (If that happens the event is not sent.)
**
** This ASSUMES that the caller owns desc->mutex.
**
** As the transaction is holding locks, this event is critical to availability.
** So we HARD-ABORT the OS process if we fail to allocage memory for the command.
*/
void my_send_need_to_resolve_prepared_txn_notification(descriptor_t * desc, int txn_idx)
{
    if ( 0 == enqueue_untagged_event_notification(desc,
                                                  BDB_NEED_TO_RESOLVE_PREPARED_TXN_NOTIFICATION,
                                                  (u_int32_t)txn_idx,
                                                  desc->valmap_db_txn[txn_idx].generation_id) ) {
        EDTK_PORT_LOG_AND_ABORT(desc, "enqueue_untagged_event_notification failed");
    }
}


/*
** This is a callback, installed with DbEnv->set_event_notify().
** It sends the event to the ports connected process (the port_coordinator)
** as an 'untagged' message.
**
** As some of the events are critical to availability/data-integrity
** (DUPMASTER, NEWMASTER, MASTER, DOWNGRADED_TO_CLIENT etc)
** we **hard-abort** the OS process if we fail to allocate memory for the event.
**
** TODO
** For some types of events, order of delivery is important.
** e.g. NEWMASTER is often followed very quickly by STARTUPDONE.
** If they are delivered in reverse order, then berkeley_db_replication_group.erl
** will change the local site's status to {replica, not_in_sync) and
** it will stay that way.  That will affect the load-balancing done
** by ?BDB_RG_H:do_txn_on_any_site() etc.
**
** One mechanism to help guarantee preservation of order is implemented below
** -- enqueuing events on the far end (tail) of desc->async_done_list rather
** than the 0(1) head (because async_done_list is normally treated as a LIFO
** stack, not a FIFO queue).
**
** However, as this function may be called by multiple threads inside BDB code,
** there are still thread race conditions.  To defeat those, we need for 
** Sleepycat must guarantee that a dependent event is not triggered until after
** the notification callback for the first has returned.  i.e. STARTUPDONE
** is not triggered until after NEWMASTER has been processed.
** That is NOT true in v4.6, but is fixed in v4.6
**
** (Sidebar: for a while I was worried that we might need a stronger ordering mechanism
** if the Erlang VM is running in SMP mode and multiple schedulers may call ready_input
** to process (send) the current contents of desc->async_done_list.
** Reminder: ready_input captures the entire current contents of desc->async_done_list
** and then releases its mutex and begins to process the captured contents.
** So new entries are added to (the empty) desc->async_done_list in parallel with
** old contents being processed (sent).  So if another VM scheduler thread is
** allowed to call ready_input then (due to vagaries of thread scheduling) some
** contents of the *later (second)* batch may be processed before the first
** -- i.e. order inversion is possible.
** However, I believe that both of the VM's locking schemes for drivers
** (driver-level and port-level locking) make this scenario impossible.)
*/
void my_handle_sleepycat_event_notification(DB_ENV *dbenv, u_int32_t event, void *event_info)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);

    /* TODO Temporary logging to show that STARTUPDONE sometimes arrives before the associated NEWMASTER which confuses the application.
     *      Sleepycat [#15265] */
    if ( (event < 1)  ||  (event > NUM_ARY_ELEMS(bdb_public_event_code_atoms) )) {
        EDTK_PORT_LOG_FORCE(desc, "BAD sleepycat event code %d", event);
    } else {
        EDTK_PORT_LOG_FORCE(desc, "Got sleepycat event %s", bdb_public_event_code_atoms[event - 1].atom_text);
    }
    /* TODO: end of temporary logging */

    switch ( event )
    {
    case DB_EVENT_REP_NEWMASTER:
    {
        /* NEWMASTER has a payload; the integer environment-id of the site that is now master.
         * (This is the eid returned by repmgr_add_remote_site() */
        int * master_eid_p = (int *)event_info;
        if ( NULL == master_eid_p ) {
            EDTK_PORT_LOG_AND_ABORT(desc, "No event_info with DB_EVENT_REP_NEWMASTER");
        }
        if ( 0 == enqueue_untagged_event_notification(desc, BDB_SLEEPYCAT_EVENT_NOTIFICATION, event, *master_eid_p) ) {
            EDTK_PORT_LOG_AND_ABORT(desc, "enqueue_untagged_event_notification failed");
        }
        break;
    }

    case DB_EVENT_REP_PERM_FAILED:
    {
        /* Tell my_txn_commit, my_txn_prepare, my_db_put, my_db_open that
         * this transaction is not guaranteed durable under failover */
        
        /* EDTK_PORT_LOG_FORCE(desc, "got PERM_FAILED, so setting tls_not_durably_replicated_error"); */
        tls_not_durably_replicated_error = 1;

        /* now FALL THROUGH to send the event to the port-coordinator too
         * (we might want to shutdown the master if we get too many PERM_FAILED events */
    }
    
    default:
        /* The other events don't have a payload */
        if ( 0 == enqueue_untagged_event_notification(desc, BDB_SLEEPYCAT_EVENT_NOTIFICATION, event, 0) ) {
            EDTK_PORT_LOG_AND_ABORT(desc, "enqueue_untagged_event_notification failed");
        }
        break;
    }
}


/* The following wrappers only exist to test tls_not_durably_replicated_error */
int my_txn_commit(DB_TXN * db_txn, u_int32_t flags)
{
    int ret;

    /* assume that this operation will be durably replicated */
    tls_not_durably_replicated_error = 0;

    if ( 0 == (ret = db_txn->commit(db_txn, flags)) )
    {
        /* BDB says the operation succeeded -- but did it cause a PERM_FAILED event? */
        if ( 0 != tls_not_durably_replicated_error )
        {
            ret = DB_REP_NOTPERM;
        }
    }
    
    return ret;
}

int my_txn_prepare(DB_TXN *db_txn, u_int8_t distributed_txn_guid[DB_XIDDATASIZE])
{
    int ret;

    tls_not_durably_replicated_error = 0;

    if ( 0 == (ret = db_txn->prepare(db_txn, distributed_txn_guid)) )
    {
        if ( 0 != tls_not_durably_replicated_error )
        {
            ret = DB_REP_NOTPERM;
        }
    }
    
    return ret;
}

int my_db_put(DB *db, DB_TXN *db_txn, DBT *key, DBT *data, u_int32_t flags)
{
    int ret;

    if ( NULL != db_txn )
    {
        /* Many db_put operations have an explicit txn, so
         * it's probably worth avoiding the overhead of the thread-local-storage
         * clear/test
         */
        ret = db->put(db, db_txn, key, data, flags);
    }
    else
    {
        /* db_txn is 'void'
         * Now, if
         *  - transactions are enabled (usually are; we could call db->get_transactional(), but that would mostly be waster)
         *  - and replication is enabled (we have no fast way to tell)
         * then this is AUTOCOMMIT is set (or this is an illegal operation).
         * (Note that we cannot simply test if AUTOMCOMMIT is set in flags, as there is a global override).
         * So it might hit a PERM_FAILED event.
         */
        tls_not_durably_replicated_error = 0;

        if ( 0 == (ret = db->put(db, db_txn, key, data, flags)) )
        {
            if ( 0 != tls_not_durably_replicated_error )
            {
                ret = DB_REP_NOTPERM;
            }
        }
    }    
    return ret;
}

/* see my_db_put for comments */
int my_db_open(DB *db, DB_TXN *db_txn, const char *file,
               const char *database, DBTYPE type, u_int32_t flags, int mode)
{
    int ret;

    if ( NULL != db_txn )
    {
        ret = db->open(db, db_txn, file, database, type, flags, mode);
    }
    else
    {
        tls_not_durably_replicated_error = 0;

        if ( 0 == (ret = db->open(db, db_txn, file, database, type, flags, mode)) )
        {
            if ( 0 != tls_not_durably_replicated_error )
            {
                ret = DB_REP_NOTPERM;
            }
        }
    }    
    return ret;
}


 /* Helper for sending untagged events to ports connected process.  See code in berkeley_db.xml.
 *
 * Returns 0 if FAILED, 1 if successful.
 *
 * This does NOT attempt to acquire desc->mutex.
 * It DOES acquire and release desc->async_done_list_mutex.
 */
static int enqueue_untagged_event_notification(descriptor_t *desc,
                                               int           event_command_code,
                                               u_int32_t     payload_uint32,
                                               long          payload_long)
{
    async_queue_entry_t * entry;
    callstate_t *         c;
    int                   res;
    
    if ( NULL == (entry = EDTK_PORT_MALLOC(desc, sizeof(async_queue_entry_t))) ) {
        EDTK_PORT_LOG(EDTK_PORT_LOG_ALLOC_FREE, desc, "failed to allocate memory for async_queue_entry");
        return 0;
    }

    if ( NULL == (c = bdb_alloc_callstate(event_command_code, desc)) ) {
        EDTK_PORT_LOG(EDTK_PORT_LOG_ALLOC_FREE, desc, "failed to allocate memory for callstate");
        return 0;
    }

    c->payload_uint32   = payload_uint32;
    c->payload_long     = payload_long;

    entry->next         = NULL;
    entry->prev         = NULL;
    entry->async_data   = c;
    entry->async_invoke = (void(*)(void *))-1;    /* won't be used (but not NULL as NULL is poison pill -- we want accidental use to cause a crash) */
    entry->async_free   = bdb_free_callstate;

    /* We don't want to invert the order of delivery of events that happen to be triggered very close together,
     * ready_input sends messages starting from desc->async_done_list and working backwards.
     * So we need to find the end of that list.
     * Note: ready_async does enqueue normal responses onto the O(1) end (i.e. desc->async_done_list)
     * so those responses ARE delivered in inverted order.  But that's fine as 'normal' responses
     * are almost always sent to *different* processes.  ;And even when they do go to the same
     * process (e.g. group_valmap_cleanup_non_block responses going to the port-coordinator
     * when it is shutting down) then we handle that with selective receives on the tags.     
     * So the order-preseving scheme *only* applies to event notifications (which are all sent to the port's
     * connected process -- i.e. the port_coordinator).
     */
    pthread_mutex_lock(&desc->async_done_list_mutex);
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "adding bdb_untagged_event operation %p to done_list", (void*)(entry->async_data));
    if ( ! desc->async_done_list )
    {
        desc->async_done_list = entry;
    }
    else
    {
        async_queue_entry_t * end = desc->async_done_list;
        while ( end->next ) {
            end = end->next;
        }
        end->next = entry;
        /* We don't need to set
         *   entry->prev = end;
         * as async_done_list is treated as a singly-linked list (only the next ptrs are used)
         */
    }
    pthread_mutex_unlock(&desc->async_done_list_mutex);

    /* Tell the Erlang VM to call bdb_ready_input (which will call bdb_ready_async to send the message).
     *
     * We DON'T test the return value of write() because I can't think of anything reasonable to do if it fails.
     * Aborting the process seems too aggressive, as the next normal command will cause the message to be sent
     * i.e. the message won't be lost, just delayed, unless no more messages can possibly be sent at all,
     * in which case the Erlang stubs for normal commands will start to get timeouts.
     * It coild be argued that a delay in receiving an event like DUPMASTER is dangerous, but if the entire driver's
     * reply channel is broken, then no other commands (e.g. txn_commit) will be returning anyway. */
     
    res = write(desc->async_done_signal_pipe_fds[1], "0", 1);
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "write() to done_signal fd returned %d", res);

    return 1;  /* success */
}


/* See berkeley_db_port_coordinator:close_environment for detailed comments about this.
 * Return 0 on success (safe to proceed)
 *        non-zero (posix code) on FAILURE */
int check_is_safe_for_port_coordinator_to_close_db_handles(DB_ENV *dbenv)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int i, ret;
    assert( dbenv );
    ret = EINVAL;
    
    pthread_mutex_lock(&desc->mutex);

    /* Check there are NO open cursors */
    for (i = 0; i < desc->num_valmap_cursor_slots; i++)
    {
        if (desc->valmap_cursor[i].value != NULL) {
            EDTK_PORT_LOG_FORCE(desc, "cursor #%d is allocated", i);
            goto done;
        }
        if ( desc->valmap_cursor[i].flags & VALMAP_RESERVED ) {
            EDTK_PORT_LOG_FORCE(desc, "cursor #%d is flagged as reserved", i);
            goto done;
        }
    }

    /* Check there are NO open txns apart from prepared (unresolved) txns.
     * If there are any prepared txns they must NOT be in-use */
    for (i = 0; i < desc->num_valmap_db_txn_slots; i++)
    {
        if (desc->valmap_db_txn[i].value != NULL)
        {
            if ( desc->valmap_db_txn[i].flags & VALMAP_CANNOT_AUTO_CLEANUP )
            {
                if ( desc->valmap_db_txn[i].flags & VALMAP_EXCL_LOCK ) {
                    EDTK_PORT_LOG_FORCE(desc, "txn #%d is allocated and flagged as VALMAP_CANNOT_AUTO_CLEANUP (i.e. is prepared), but is also flagged as in-use", i);
                    goto done;
                }
            }
            else
            {
                EDTK_PORT_LOG_FORCE(desc, "txn #%d is allocated but is not flagged as VALMAP_CANNOT_AUTO_CLEANUP (i.e. not prepared)", i);
                goto done;
            }
        }
    }

    /* Check that ALL db handles belong to the port-coordinator's group id (0)
     * and none of them are in-use. */
    for (i = 0; i < desc->num_valmap_db_slots; i++)
    {
        if (desc->valmap_db[i].value != NULL)
        {
            if ( desc->valmap_db[i].group_id != 0 ) {
                EDTK_PORT_LOG_FORCE(desc, "db #%d is allocated but is not owned by the port-coordinator (group id %d)", i, desc->valmap_db[i].group_id);
                goto done;
            }
            if ( desc->valmap_db[i].flags & VALMAP_EXCL_LOCK ) {
                EDTK_PORT_LOG_FORCE(desc, "db #%d is allocated and owned by port-coordinator, but is flagged as EXCL_LOCK (in-use)", i);
                goto done;
            }
            if ( desc->valmap_db[i].shared_lock_refcount != 0 ) {
                EDTK_PORT_LOG_FORCE(desc, "db #%d is allocated and owned by port-coordinator, but has non-zero shared_lock_refcount %d", i, desc->valmap_db[i].shared_lock_refcount);
                goto done;
            }
        }
    }

    /* Check that there is only a single dbenv handle, that it is allocated and belong to the port-coordinator's group id (0)
     * and that it is only in-use by this command (shared_lock_refcount == 1) */
    if ( 1 != desc->num_valmap_dbenv_slots ) {
        EDTK_PORT_LOG_FORCE(desc, "more than one (%d) dbenv is allocated", desc->num_valmap_dbenv_slots);
        goto done;
    }
    if (NULL == desc->valmap_dbenv[0].value)
    {
        EDTK_PORT_LOG_FORCE(desc, "dbenv[0] is NOT allocated");
        goto done;
    }
    if ( desc->valmap_dbenv[0].group_id != 0 ) {
        EDTK_PORT_LOG_FORCE(desc, "dbenv #0 is allocated but is not owned by the port-coordinator (group id %d)", desc->valmap_dbenv[0].group_id);
        goto done;
    }
    if ( desc->valmap_dbenv[0].flags & VALMAP_EXCL_LOCK ) {
        EDTK_PORT_LOG_FORCE(desc, "dbenv #0 is allocated and owned by port-coordinator, but is flagged as EXCL_LOCK (in-use)");
        goto done;
    }
    if ( desc->valmap_dbenv[0].shared_lock_refcount != 1 ) {
        EDTK_PORT_LOG_FORCE(desc, "dbenv #0 is allocated and owned by port-coordinator, but shared_lock_refcount (%d) != 1", desc->valmap_dbenv[0].shared_lock_refcount);
        goto done;
    }

    ret = 0; /* success */
    
  done:
    pthread_mutex_unlock(&desc->mutex);
    return ret;
}


/* Returns 1 on FAILURE (any txn still in use), 0 on success (and *pout is set to NumUnresolvedTxns) */
int my_make_db_close_safe_if_there_are_unresolved_transactions(DB_ENV *dbenv, long *pout)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int i;
    assert( dbenv );
    *pout = 0;
    pthread_mutex_lock(&desc->mutex);
    for (i = 0; i < desc->num_valmap_db_txn_slots; i++) {
        if ( desc->valmap_db_txn[i].flags & VALMAP_EXCL_LOCK ) {
            EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "txn #%d is in-use", i);
            return 1; /* failure */
        }
        if ( desc->valmap_db_txn[i].flags & VALMAP_CANNOT_AUTO_CLEANUP ) {
            assert(desc->valmap_db_txn[i].value != NULL);
            assert(desc->valmap_db_txn[i].parent_index == -1);
            EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "found prepared txn #%d", i);
            ++(*pout);
        }
    }
    
    if (*pout == 0) {
        EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "no unresolved txns found");
    }
    else {
        EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "there are %ld unresolved txns, so setting dbenv_shutdown (to prevent logging in DB::close)", *pout);
        dbenv->dbenv_shutdown = 1;
    }
    pthread_mutex_unlock(&desc->mutex);
    return 0; /* success */
}


/* The caller is responsible for calling erl_free_compound() on the returned value.
 * Also this function does not  free  c->o.rep_stat_p  -- free_callstate will do that later. */
ETERM * make_rep_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    return erl_format(
        "[{st_status,~i},"                    /* u_int32_t  Current replication status. */
        "{st_next_lsn,{~i,~i}},"              /* DB_LSN  Next LSN to use or expect. */
        "{st_waiting_lsn,{~i,~i}},"           /* DB_LSN  LSN we're awaiting, if any. */
        "{st_next_pg,~i},"                    /* db_pgno_t  Next pg we expect. */
        "{st_waiting_pg,~i},"                 /* db_pgno_t  pg we're awaiting, if any. */
        "{st_dupmasters,~i},"                 /* u_int32_t  # of times a duplicate master condition was detected.+ */
        "{st_env_id,~i},"                     /* int  Current environment ID. */
        "{st_env_priority,~i},"               /* int  Current environment priority. */
        "{st_bulk_fills,~i},"                 /* u_int32_t  Bulk buffer fills. */
        "{st_bulk_overflows,~i},"             /* u_int32_t  Bulk buffer overflows. */
        "{st_bulk_records,~i},"               /* u_int32_t  Bulk records stored. */
        "{st_bulk_transfers,~i},"             /* u_int32_t  Transfers of bulk buffers. */
        "{st_client_rerequests,~i},"          /* u_int32_t  Number of forced rerequests. */
        "{st_client_svc_req,~i},"             /* u_int32_t  Number of client service requests received by this client. */
        "{st_client_svc_miss,~i},"            /* u_int32_t  Number of client service requests missing on this client. */
        "{st_gen,~i},"                        /* u_int32_t  Current generation number. */
        "{st_egen,~i},"                       /* u_int32_t  Current election gen number. */
        "{st_log_duplicated,~i},"             /* u_int32_t  Log records received multiply.+ */
        "{st_log_queued,~i},"                 /* u_int32_t  Log records currently queued.+ */
        "{st_log_queued_max,~i},"             /* u_int32_t  Max. log records queued at once.+ */
        "{st_log_queued_total,~i},"           /* u_int32_t  Total # of log recs. ever queued.+ */
        "{st_log_records,~i},"                /* u_int32_t  Log records received and put.+ */
        "{st_log_requested,~i},"              /* u_int32_t  Log recs. missed and requested.+ */
        "{st_master,~i},"                     /* int  Env. ID of the current master. */
        "{st_master_changes,~i},"             /* u_int32_t  # of times we've switched masters. */
        "{st_msgs_badgen,~i},"                /* u_int32_t  Messages with a bad generation #.+ */
        "{st_msgs_processed,~i},"             /* u_int32_t  Messages received and processed.+ */
        "{st_msgs_recover,~i},"               /* u_int32_t  Messages ignored because this site was a client in recovery.+ */
        "{st_msgs_send_failures,~i},"         /* u_int32_t  # of failed message sends.+ */
        "{st_msgs_sent,~i},"                  /* u_int32_t  # of successful message sends.+ */
        "{st_newsites,~i},"                   /* u_int32_t  # of NEWSITE msgs. received.+ */
        "{st_nsites,~i},"                     /* int  Current number of sites we will assume during elections. */
        "{st_nthrottles,~i},"                 /* u_int32_t  # of times we were throttled. */
        "{st_outdated,~i},"                   /* u_int32_t  # of times we detected and returned an OUTDATED condition.+ */
        "{st_pg_duplicated,~i},"              /* u_int32_t  Pages received multiply.+ */
        "{st_pg_records,~i},"                 /* u_int32_t  Pages received and stored.+ */
        "{st_pg_requested,~i},"               /* u_int32_t  Pages missed and requested.+ */
        "{st_startup_complete,~i},"           /* u_int32_t  Site completed client sync-up. */
        "{st_txns_applied,~i},"               /* u_int32_t  # of transactions applied.+ */
        "{st_elections,~i},"                  /* u_int32_t  # of elections held.+ */
        "{st_elections_won,~i},"              /* u_int32_t  # of elections won by this site.+ */
        "{st_election_cur_winner,~i},"        /* int  Current front-runner. */
        "{st_election_gen,~i},"               /* u_int32_t  Election generation number. */
        "{st_election_lsn,{~i,~i}},"          /* DB_LSN  Max. LSN of current winner. */
        "{st_election_nsites,~i},"            /* int  # of "registered voters". */
        "{st_election_nvotes,~i},"            /* int  # of "registered voters" needed. */
        "{st_election_priority,~i},"          /* int  Current election priority. */
        "{st_election_status,~i},"            /* int  Current election status. */
        "{st_election_tiebreaker,~i},"        /* u_int32_t  Election tiebreaker value. */
        "{st_election_votes,~i},"             /* int  Votes received in this round. */
        "{st_election_sec,~i},"               /* u_int32_t  Last election time seconds. */
        "{st_election_usec,~i}]",             /* u_int32_t  Last election time useconds. */
        
        (int)(c->o.rep_stat_p->st_status),
        (int)(c->o.rep_stat_p->st_next_lsn.file), (int)(c->o.rep_stat_p->st_next_lsn.offset),
        (int)(c->o.rep_stat_p->st_waiting_lsn.file), (int)(c->o.rep_stat_p->st_waiting_lsn.offset),
        (int)(c->o.rep_stat_p->st_next_pg),
        (int)(c->o.rep_stat_p->st_waiting_pg),
        (int)(c->o.rep_stat_p->st_dupmasters),
        (int)(c->o.rep_stat_p->st_env_id),
        (int)(c->o.rep_stat_p->st_env_priority),
        (int)(c->o.rep_stat_p->st_bulk_fills),
        (int)(c->o.rep_stat_p->st_bulk_overflows),
        (int)(c->o.rep_stat_p->st_bulk_records),
        (int)(c->o.rep_stat_p->st_bulk_transfers),
        (int)(c->o.rep_stat_p->st_client_rerequests),
        (int)(c->o.rep_stat_p->st_client_svc_req),
        (int)(c->o.rep_stat_p->st_client_svc_miss),
        (int)(c->o.rep_stat_p->st_gen),
        (int)(c->o.rep_stat_p->st_egen),
        (int)(c->o.rep_stat_p->st_log_duplicated),
        (int)(c->o.rep_stat_p->st_log_queued),
        (int)(c->o.rep_stat_p->st_log_queued_max),
        (int)(c->o.rep_stat_p->st_log_queued_total),
        (int)(c->o.rep_stat_p->st_log_records),
        (int)(c->o.rep_stat_p->st_log_requested),
        (int)(c->o.rep_stat_p->st_master),
        (int)(c->o.rep_stat_p->st_master_changes),
        (int)(c->o.rep_stat_p->st_msgs_badgen),
        (int)(c->o.rep_stat_p->st_msgs_processed),
        (int)(c->o.rep_stat_p->st_msgs_recover),
        (int)(c->o.rep_stat_p->st_msgs_send_failures),
        (int)(c->o.rep_stat_p->st_msgs_sent),
        (int)(c->o.rep_stat_p->st_newsites),
        (int)(c->o.rep_stat_p->st_nsites),
        (int)(c->o.rep_stat_p->st_nthrottles),
        (int)(c->o.rep_stat_p->st_outdated),
        (int)(c->o.rep_stat_p->st_pg_duplicated),
        (int)(c->o.rep_stat_p->st_pg_records),
        (int)(c->o.rep_stat_p->st_pg_requested),
        (int)(c->o.rep_stat_p->st_startup_complete),
        (int)(c->o.rep_stat_p->st_txns_applied),
        (int)(c->o.rep_stat_p->st_elections),
        (int)(c->o.rep_stat_p->st_elections_won),
        (int)(c->o.rep_stat_p->st_election_cur_winner),
        (int)(c->o.rep_stat_p->st_election_gen),
        (int)(c->o.rep_stat_p->st_election_lsn.file), (int)(c->o.rep_stat_p->st_election_lsn.offset),
        (int)(c->o.rep_stat_p->st_election_nsites),
        (int)(c->o.rep_stat_p->st_election_nvotes),
        (int)(c->o.rep_stat_p->st_election_priority),
        (int)(c->o.rep_stat_p->st_election_status),
        (int)(c->o.rep_stat_p->st_election_tiebreaker),
        (int)(c->o.rep_stat_p->st_election_votes),
        (int)(c->o.rep_stat_p->st_election_sec),
        (int)(c->o.rep_stat_p->st_election_usec)
        );
}


/* The caller is responsible for calling erl_free_compound() on the returned value.
 * Also this function does not  freec ->o.lock_stat_p  -- free_callstate will do that later. */
ETERM * make_lock_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    return erl_format(
        "[{st_id,~i},"            /* u_int32_t  Last allocated locker ID. */
        "{st_cur_maxid,~i},"      /* u_int32_t  Current maximum unused ID. */
        "{st_maxlocks,~i},"       /* u_int32_t  Maximum number of locks in table. */
        "{st_maxlockers,~i},"     /* u_int32_t  Maximum num of lockers in table. */
        "{st_maxobjects,~i},"     /* u_int32_t  Maximum num of objects in table. */
        "{st_nmodes,~i},"         /* int  Number of lock modes. */
        "{st_nlocks,~i},"         /* u_int32_t  Current number of locks. */
        "{st_maxnlocks,~i},"      /* u_int32_t  Maximum number of locks so far. */
        "{st_nlockers,~i},"       /* u_int32_t  Current number of lockers. */
        "{st_maxnlockers,~i},"    /* u_int32_t  Maximum number of lockers so far. */
        "{st_nobjects,~i},"       /* u_int32_t  Current number of objects. */
        "{st_maxnobjects,~i},"    /* u_int32_t  Maximum number of objects so far. */
        "{st_nrequests,~i},"      /* u_int32_t  Number of lock gets. */
        "{st_nreleases,~i},"      /* u_int32_t  Number of lock puts. */
        "{st_nupgrade,~i},"       /* u_int32_t  Number of lock upgrades. */
        "{st_ndowngrade,~i},"     /* u_int32_t  Number of lock downgrades. */
        "{st_lock_wait,~i},"      /* u_int32_t  Lock conflicts w/ subsequent wait */
        "{st_lock_nowait,~i},"    /* u_int32_t  Lock conflicts w/o subsequent wait */
        "{st_ndeadlocks,~i},"     /* u_int32_t  Number of lock deadlocks. */
        "{st_locktimeout,~i},"    /* db_timeout_t  Lock timeout. */
        "{st_nlocktimeouts,~i},"  /* u_int32_t  Number of lock timeouts. */
        "{st_txntimeout,~i},"     /* db_timeout_t  Transaction timeout. */
        "{st_ntxntimeouts,~i},"   /* u_int32_t  Number of transaction timeouts. */
        "{st_region_wait,~i},"    /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"  /* u_int32_t  Region lock granted without wait. */
        "{st_regsize,{~i, ~i}}]",  /* roff_t  Region size. {GB, B} as it won't fit in an int on a 64-bit machine */

        (int)(c->o.lock_stat_p->st_id),
        (int)(c->o.lock_stat_p->st_cur_maxid),
        (int)(c->o.lock_stat_p->st_maxlocks),
        (int)(c->o.lock_stat_p->st_maxlockers),
        (int)(c->o.lock_stat_p->st_maxobjects),
        (int)(c->o.lock_stat_p->st_nmodes),
        (int)(c->o.lock_stat_p->st_nlocks),
        (int)(c->o.lock_stat_p->st_maxnlocks),
        (int)(c->o.lock_stat_p->st_nlockers),
        (int)(c->o.lock_stat_p->st_maxnlockers),
        (int)(c->o.lock_stat_p->st_nobjects),
        (int)(c->o.lock_stat_p->st_maxnobjects),
        (int)(c->o.lock_stat_p->st_nrequests),
        (int)(c->o.lock_stat_p->st_nreleases),
        (int)(c->o.lock_stat_p->st_nupgrade),
        (int)(c->o.lock_stat_p->st_ndowngrade),
        (int)(c->o.lock_stat_p->st_lock_wait),
        (int)(c->o.lock_stat_p->st_lock_nowait),
        (int)(c->o.lock_stat_p->st_ndeadlocks),
        (int)(c->o.lock_stat_p->st_locktimeout),
        (int)(c->o.lock_stat_p->st_nlocktimeouts),
        (int)(c->o.lock_stat_p->st_txntimeout),
        (int)(c->o.lock_stat_p->st_ntxntimeouts),
        (int)(c->o.lock_stat_p->st_region_wait),
        (int)(c->o.lock_stat_p->st_region_nowait),
        (int)(c->o.lock_stat_p->st_regsize >> 30), (int)(c->o.lock_stat_p->st_regsize & ((1<<30)-1))); /* 2^30 = 1GB */
}

ETERM * make_mutex_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    return erl_format(
        "[{st_mutex_align,~i},"            /* u_int32_t  Mutex alignment */
        "{st_mutex_tas_spins,~i},"        /* u_int32_t  Mutex test-and-set spins */
        "{st_mutex_cnt,~i},"              /* u_int32_t  Mutex count */
        "{st_mutex_free,~i},"             /* u_int32_t  Available mutexes */
        "{st_mutex_inuse,~i},"            /* u_int32_t  Mutexes in use */
        "{st_mutex_inuse_max,~i},"        /* u_int32_t  Maximum mutexes ever in use */
        "{st_region_wait,~i},"            /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"          /* u_int32_t  Region lock granted without wait. */
        "{st_regsize,{~i,~i}}]",          /*roff_t  Region size. */

        (int)(c->o.mutex_stat_p->st_mutex_align),
        (int)(c->o.mutex_stat_p->st_mutex_tas_spins),
        (int)(c->o.mutex_stat_p->st_mutex_cnt),
        (int)(c->o.mutex_stat_p->st_mutex_free),
        (int)(c->o.mutex_stat_p->st_mutex_inuse),
        (int)(c->o.mutex_stat_p->st_mutex_inuse_max),
        (int)(c->o.mutex_stat_p->st_region_wait),
        (int)(c->o.mutex_stat_p->st_region_nowait),
        (int)(c->o.mutex_stat_p->st_regsize >> 30), (int)(c->o.mutex_stat_p->st_regsize & ((1<<30)-1))); /* 2^30 = 1GB */
}

/* private helper for make_txn_stat_external_term() */
ETERM * make_active_txn_term(DB_TXN_ACTIVE * active_txn_p)
{
    ETERM * active_txn_term;
    ETERM * xid_term = (active_txn_p->status == TXN_PREPARED)
                       ? erl_mk_binary((char *)(active_txn_p->xid), DB_XIDDATASIZE)
                       : erl_mk_atom("not_prepared");
    if ( NULL == xid_term ) {
        return NULL;
    }

    if ( NULL == (active_txn_term = erl_format(
                      "{~i,"                            /* u_int32_t  Transaction ID (as returned by DB_TXN->id()) */
                      "[{parent_txn_id,~i},"            /* u_int32_t  Transaction ID of parent */
                      "{proces_id,~i},"                 /* pid_t  Process owning txn ID */
                      "{thread_id,~i},"                 /* db_threadid_t Thread owning txn ID */
                      "{lsn,{~i,~i}},"                  /* DB_LSN   LSN when transaction began */
                      "{read_lsn,{~i,~i}},"             /* DB_LSN   Read LSN for MVCC */
                      "{mvcc_ref,~i},"                  /* u_int32_t  MVCC reference count */
                      "{status,~a},"                    /* u_int32_t  status */
                      "{xid,~w},"                       /* u_int8_t[DB_XIDDATASIZE]  global transaction ID (if prepared) */
                      "{name,~s}]}",                    /* char[51]  50 bytes of name, nul termination */

                      (int)(active_txn_p->txnid),
                      (int)(active_txn_p->parentid),
                      (int)(active_txn_p->pid),
                      (int)(active_txn_p->tid),
                      (int)(active_txn_p->lsn.file),      (int)(active_txn_p->lsn.offset),
                      (int)(active_txn_p->read_lsn.file), (int)(active_txn_p->read_lsn.offset),
                      (int)(active_txn_p->mvcc_ref),
                      bdb_TXN_status_code_atoms[active_txn_p->status - 1].atom_text,
                      xid_term,
                      (int)(active_txn_p->name))) )
    {
        erl_free_compound(xid_term);
        return NULL;
    }
    return active_txn_term;
}

ETERM * make_txn_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    int               i;
    ETERM *           active_txn_term_list;
    ETERM *           txn_stat_term;
    
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");
    
    if ( NULL == (active_txn_term_list = erl_mk_empty_list()) ) {  /* we will cons onto the front of this */
        return NULL;
    }

    for (i = 0; i < c->o.txn_stat_p->st_nactive; ++i)
    {
        ETERM * active_txn_term;
        ETERM * new_cons;

        if ( NULL == (active_txn_term = make_active_txn_term(& c->o.txn_stat_p->st_txnarray[i])) ) {
            erl_free_compound(active_txn_term_list);
            return NULL;
        }
        if ( NULL == (new_cons = erl_cons(active_txn_term, active_txn_term_list)) ) {
            erl_free_compound(active_txn_term);
            erl_free_compound(active_txn_term_list);
            return NULL;
        }
        active_txn_term_list = new_cons;
    }

    if ( NULL == (txn_stat_term = erl_format(
                      "[{active_txns,~w},"            /* list of active transactions */
                      "{st_last_ckp,{~i,~i}},"          /* DB_LSN   lsn of the last checkpoint */
                      "{st_time_ckp,{~i,~i,~i}},"       /* time_t  time of last checkpoint   (a tuple compatible with erlang:now()) */
                      "{st_last_txnid,~i},"             /* u_int32_t  last transaction id given out */
                      "{st_maxtxns,~i},"                /* u_int32_t  maximum txns possible */
                      "{st_naborts,~i},"                /* u_int32_t  number of aborted transactions */
                      "{st_nbegins,~i},"                /* u_int32_t  number of begun transactions */
                      "{st_ncommits,~i},"               /* u_int32_t  number of committed transactions */
                      "{st_nactive,~i},"                /* u_int32_t  number of active transactions */
                      "{st_nsnapshot,~i},"              /* u_int32_t  number of snapshot transactions */
                      "{st_nrestores,~i},"              /* u_int32_t  number of restored transactions after recovery. */
                      "{st_maxnactive,~i},"             /* u_int32_t  maximum active transactions */
                      "{st_maxnsnapshot,~i},"           /* u_int32_t  maximum snapshot transactions */
                      "{st_region_wait,~i},"            /* u_int32_t  Region lock granted after wait. */
                      "{st_region_nowait,~i}"           /* u_int32_t  Region lock granted without wait. */
                      "{st_regsize,{~i,~i}}]",          /*roff_t  Region size. */

                      active_txn_term_list,
                      (int)(c->o.txn_stat_p->st_last_ckp.file), (int)(c->o.txn_stat_p->st_last_ckp.offset),
                      (int)(c->o.txn_stat_p->st_time_ckp / 1000000), (int)(c->o.txn_stat_p->st_time_ckp % 1000000), 0, /* time_t is seconds since 00:00:00 on January 1, 1970 UTC.  erlang:now is same but microseconds. */
                      (int)(c->o.txn_stat_p->st_last_txnid),
                      (int)(c->o.txn_stat_p->st_maxtxns),
                      (int)(c->o.txn_stat_p->st_naborts),
                      (int)(c->o.txn_stat_p->st_nbegins),
                      (int)(c->o.txn_stat_p->st_ncommits),
                      (int)(c->o.txn_stat_p->st_nactive),
                      (int)(c->o.txn_stat_p->st_nsnapshot),
                      (int)(c->o.txn_stat_p->st_nrestores),
                      (int)(c->o.txn_stat_p->st_maxnactive),
                      (int)(c->o.txn_stat_p->st_maxnsnapshot),
                      (int)(c->o.txn_stat_p->st_region_wait),
                      (int)(c->o.txn_stat_p->st_region_nowait),
                      (int)(c->o.txn_stat_p->st_regsize >> 30), (int)(c->o.txn_stat_p->st_regsize & ((1<<30)-1)))) ) /* 2^30 = 1GB */
    {
        erl_free_compound(active_txn_term_list);
        return NULL;
    }
    return txn_stat_term;
}

/* private helper for make_memp_stat_external_term */
ETERM * make_memp_fstat_term(DB_MPOOL_FSTAT * mpool_fstat_p)
{
    return erl_format(
        "{~s,"                            /* char[51] File name. */
        "[{st_pagesize,~i},"              /* u_int32_t  Page size. */
        "{st_map,~i},"                    /* u_int32_t  Pages from mapped files. */
        "{st_cache_hit,~i},"              /* u_int32_t  Pages found in the cache. */
        "{st_cache_miss,~i},"             /* u_int32_t  Pages not found in the cache. */
        "{st_page_create,~i},"            /* u_int32_t  Pages created in the cache. */
        "{st_page_in,~i},"                /* u_int32_t  Pages read in. */
        "{st_page_out,~i}]}",             /* u_int32_t  Pages written out. */

        mpool_fstat_p->file_name,
        (int)(mpool_fstat_p->st_pagesize),
        (int)(mpool_fstat_p->st_map),
        (int)(mpool_fstat_p->st_cache_hit),
        (int)(mpool_fstat_p->st_cache_miss),
        (int)(mpool_fstat_p->st_page_create),
        (int)(mpool_fstat_p->st_page_in),
        (int)(mpool_fstat_p->st_page_out));
}

ETERM * make_memp_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    ETERM *           fstat_term_list;
    DB_MPOOL_FSTAT ** fstat_pp;
    ETERM *           memp_stat_term;
    
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    if ( NULL == (fstat_term_list = erl_mk_empty_list()) ) {  /* we will cons onto the front of this */
        return NULL;
    }

    /* EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "fstat array is at %p\n", (void*)(c->o.mpool_fstat_ary_p))); */
    if ( (fstat_pp = c->o.mpool_fstat_ary_p) != NULL )   /* might be no array at all if there are no files */
    {
        while ( *fstat_pp != NULL )
        {
            ETERM * fstat_term;
            ETERM * new_cons;

            /* EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "at %p (idx %d), adding %s\n", (void*)(fstat_pp), fstat_pp - c->o.mpool_fstat_ary_p, (*fstat_pp)->file_name); */
            if ( NULL == (fstat_term = make_memp_fstat_term(*fstat_pp)) ) {
                erl_free_compound(fstat_term_list);
                return NULL;
            }
            if ( NULL == (new_cons = erl_cons(fstat_term, fstat_term_list)) ) {
                erl_free_compound(fstat_term);
                erl_free_compound(fstat_term_list);
                return NULL;
            }
            fstat_term_list = new_cons;
            ++fstat_pp;
        }
    }
    /* EDTK_DBT_INFO(desc, "end of fstat array is at %p (array idx %d)\n", (void*)(fstat_pp), fstat_pp - c->o.mpool_fstat_ary_p); */

    if ( NULL == (memp_stat_term = erl_format(
                      "[{file_stats,~w},"          /* fstat array */
                      "{st_cache_size,{~i,~i}},"   /* 2*u_int32_t  Total cache size: {GB, B} */
                      "{st_ncache,~i},"            /* u_int32_t  Number of caches. */
                      "{st_regsize,{~i,~i}},"      /* roff_t Region size.           {GB, B} */
                      "{st_mmapsize,{~i, ~i}},"    /* size_t Maximum file size for mmap. {GB, B} */
                      "{st_maxopenfd,~i},"         /* int Maximum number of open fd's. */
                      "{st_maxwrite,~i},"          /* int Maximum buffers to write. */
                      "{st_maxwrite_sleep,~i},"    /* int Sleep after writing max buffers. */
                      "{st_map,~i},"               /* u_int32_t  Pages from mapped files. */
                      "{st_cache_hit,~i},"         /* u_int32_t  Pages found in the cache. */
                      "{st_cache_miss,~i},"        /* u_int32_t  Pages not found in the cache. */
                      "{st_page_create,~i},"       /* u_int32_t  Pages created in the cache. */
                      "{st_page_in,~i},"           /* u_int32_t  Pages read in. */
                      "{st_page_out,~i},"          /* u_int32_t  Pages written out. */
                      "{st_ro_evict,~i},"          /* u_int32_t  Clean pages forced from the cache. */
                      "{st_rw_evict,~i},"          /* u_int32_t  Dirty pages forced from the cache. */
                      "{st_page_trickle,~i},"      /* u_int32_t  Pages written by memp_trickle. */
                      "{st_pages,~i},"             /* u_int32_t  Total number of pages. */
                      "{st_page_clean,~i},"        /* u_int32_t  Clean pages. */
                      "{st_page_dirty,~i},"        /* u_int32_t  Dirty pages. */
                      "{st_hash_buckets,~i},"      /* u_int32_t  Number of hash buckets. */
                      "{st_hash_searches,~i},"     /* u_int32_t  Total hash chain searches. */
                      "{st_hash_longest,~i},"      /* u_int32_t  Longest hash chain searched. */
                      "{st_hash_examined,~i},"     /* u_int32_t  Total hash entries searched. */
                      "{st_hash_nowait,~i},"       /* u_int32_t  Hash lock granted with nowait. */
                      "{st_hash_wait,~i},"         /* u_int32_t  Hash lock granted after wait. */
                      "{st_hash_max_wait,~i},"     /* u_int32_t  Max hash lock granted after wait. */
                      "{st_region_nowait,~i},"     /* u_int32_t  Region lock granted with nowait. */
                      "{st_region_wait,~i},"       /* u_int32_t  Region lock granted after wait. */
                      "{st_mvcc_frozen,~i},"       /* u_int32_t  Buffers frozen. */
                      "{st_mvcc_thawed,~i},"       /* u_int32_t  Buffers thawed. */
                      "{st_mvcc_freed,~i},"        /* u_int32_t  Frozen buffers freed. */
                      "{st_alloc,~i},"             /* u_int32_t  Number of page allocations. */
                      "{st_alloc_buckets,~i},"     /* u_int32_t  Buckets checked during allocation. */
                      "{st_alloc_max_buckets,~i}," /* u_int32_t  Max checked during allocation. */
                      "{st_alloc_pages,~i},"       /* u_int32_t  Pages checked during allocation. */
                      "{st_alloc_max_pages,~i},"   /* u_int32_t  Max checked during allocation. */
                      "{st_io_wait,~i}]",          /* u_int32_t  Thread waited on buffer I/O. */

                      fstat_term_list,
                      (int)(c->o.mpool_stat_p->st_gbytes), (int)(c->o.mpool_stat_p->st_bytes),
                      (int)(c->o.mpool_stat_p->st_ncache),
                      (int)(c->o.mpool_stat_p->st_regsize >> 30),  (int)(c->o.mpool_stat_p->st_regsize  & ((1<<30)-1)), /* 2^30 = 1GB */
                      (int)(c->o.mpool_stat_p->st_mmapsize >> 30), (int)(c->o.mpool_stat_p->st_mmapsize & ((1<<30)-1)),
                      (int)(c->o.mpool_stat_p->st_maxopenfd),
                      (int)(c->o.mpool_stat_p->st_maxwrite),
                      (int)(c->o.mpool_stat_p->st_maxwrite_sleep),
                      (int)(c->o.mpool_stat_p->st_map),
                      (int)(c->o.mpool_stat_p->st_cache_hit),
                      (int)(c->o.mpool_stat_p->st_cache_miss),
                      (int)(c->o.mpool_stat_p->st_page_create),
                      (int)(c->o.mpool_stat_p->st_page_in),
                      (int)(c->o.mpool_stat_p->st_page_out),
                      (int)(c->o.mpool_stat_p->st_ro_evict),
                      (int)(c->o.mpool_stat_p->st_rw_evict),
                      (int)(c->o.mpool_stat_p->st_page_trickle),
                      (int)(c->o.mpool_stat_p->st_pages),
                      (int)(c->o.mpool_stat_p->st_page_clean),
                      (int)(c->o.mpool_stat_p->st_page_dirty),
                      (int)(c->o.mpool_stat_p->st_hash_buckets),
                      (int)(c->o.mpool_stat_p->st_hash_searches),
                      (int)(c->o.mpool_stat_p->st_hash_longest),
                      (int)(c->o.mpool_stat_p->st_hash_examined),
                      (int)(c->o.mpool_stat_p->st_hash_nowait),
                      (int)(c->o.mpool_stat_p->st_hash_wait),
                      (int)(c->o.mpool_stat_p->st_hash_max_wait),
                      (int)(c->o.mpool_stat_p->st_region_nowait),
                      (int)(c->o.mpool_stat_p->st_region_wait),
                      (int)(c->o.mpool_stat_p->st_mvcc_frozen),
                      (int)(c->o.mpool_stat_p->st_mvcc_thawed),
                      (int)(c->o.mpool_stat_p->st_mvcc_freed),
                      (int)(c->o.mpool_stat_p->st_alloc),
                      (int)(c->o.mpool_stat_p->st_alloc_buckets),
                      (int)(c->o.mpool_stat_p->st_alloc_max_buckets),
                      (int)(c->o.mpool_stat_p->st_alloc_pages),
                      (int)(c->o.mpool_stat_p->st_alloc_max_pages),
                      (int)(c->o.mpool_stat_p->st_io_wait))) )
    {
        erl_free_compound(fstat_term_list);
        return NULL;
    }
    return memp_stat_term;
}

ETERM * make_log_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    return erl_format(
        "[{st_magic,~i},"             /* u_int32_t  Log file magic number. */
        "{st_version,~i},"            /* u_int32_t  Log file version number. */
        "{st_mode,~i},"               /* int Log file permissions mode. */
        "{st_lg_bsize,~i},"           /* u_int32_t  Log buffer size. */
        "{st_lg_size,~i},"            /* u_int32_t  Log file size. */
        "{st_record,~i},"             /* u_int32_t  Records entered into the log. */
        "{st_w_bytes,~i},"            /* u_int32_t  Bytes to log. */
        "{st_w_mbytes,~i},"           /* u_int32_t  Megabytes to log. */
        "{st_wc_bytes,~i},"           /* u_int32_t  Bytes to log since checkpoint. */
        "{st_wc_mbytes,~i},"          /* u_int32_t  Megabytes to log since checkpoint. */
        "{st_wcount,~i},"             /* u_int32_t  Total I/O writes to the log. */
        "{st_wcount_fill,~i},"        /* u_int32_t  Overflow writes to the log. */
        "{st_rcount,~i},"             /* u_int32_t  Total I/O reads from the log. */
        "{st_scount,~i},"             /* u_int32_t  Total syncs to the log. */
        "{st_region_wait,~i},"        /* u_int32_t  Region lock granted after wait. */
        "{st_region_nowait,~i},"      /* u_int32_t  Region lock granted without wait. */
        "{st_cur_file,~i},"           /* u_int32_t  Current log file number. */
        "{st_cur_offset,~i},"         /* u_int32_t  Current log file offset. */
        "{st_disk_file,~i},"          /* u_int32_t  Known on disk log file number. */
        "{st_disk_offset,~i},"        /* u_int32_t  Known on disk log file offset. */
        "{st_regsize,{~i,~i}},"       /* roff_t     Region size {GB, B} */
        "{st_maxcommitperflush,~i},"  /* u_int32_t  Max number of commits in a flush. */
        "{st_mincommitperflush,~i}]", /* u_int32_t  Min number of commits in a flush. */

        (int)(c->o.log_stat_p->st_magic),
        (int)(c->o.log_stat_p->st_version),
        (int)(c->o.log_stat_p->st_mode),
        (int)(c->o.log_stat_p->st_lg_bsize),
        (int)(c->o.log_stat_p->st_lg_size),
        (int)(c->o.log_stat_p->st_record),
        (int)(c->o.log_stat_p->st_w_bytes),
        (int)(c->o.log_stat_p->st_w_mbytes),
        (int)(c->o.log_stat_p->st_wc_bytes),
        (int)(c->o.log_stat_p->st_wc_mbytes),
        (int)(c->o.log_stat_p->st_wcount),
        (int)(c->o.log_stat_p->st_wcount_fill),
        (int)(c->o.log_stat_p->st_rcount),
        (int)(c->o.log_stat_p->st_scount),
        (int)(c->o.log_stat_p->st_region_wait),
        (int)(c->o.log_stat_p->st_region_nowait),
        (int)(c->o.log_stat_p->st_cur_file),
        (int)(c->o.log_stat_p->st_cur_offset),
        (int)(c->o.log_stat_p->st_disk_file),
        (int)(c->o.log_stat_p->st_disk_offset),
        (int)(c->o.log_stat_p->st_regsize >> 30), (int)(c->o.log_stat_p->st_regsize & ((1<<30)-1)), /* 2^30 = 1GB */
        (int)(c->o.log_stat_p->st_maxcommitperflush),
        (int)(c->o.log_stat_p->st_mincommitperflush));
}

/* We only need this helper because we need to call DB->get_type in order to decode the output */
int my_db_stat(DB *db, DB_TXN *db_txn, DBTYPE * db_type_p, void ** db_stat_p, u_int32_t flags)
{
    int res;
    if ( (res = db->get_type(db, db_type_p)) != 0 )
        return res;
    
    return db->stat(db, db_txn, db_stat_p, flags);
}

ETERM * make_db_stat_external_term(descriptor_t *desc, callstate_t *c)
{
    int idx = c->o.db_type - 1;

    EDTK_PORT_LOG(EDTK_PORT_LOG_VERBOSE, desc, "calling erl_format");

    assert(idx >= 0 && idx < NUM_ARY_ELEMS(bdb_DBTYPE_code_atoms) );
    
    switch ( c->o.db_type )
    {
    case DB_BTREE:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{bt_magic,~i},"                  /* u_int32_t  Magic number. */
            "{bt_version,~i},"                /* u_int32_t  Version number. */
            "{bt_metaflags,~i},"              /* u_int32_t  Metadata flags. */
            "{bt_nkeys,~i},"                  /* u_int32_t  Number of unique keys. */
            "{bt_ndata,~i},"                  /* u_int32_t  Number of data items. */
            "{bt_pagesize,~i},"               /* u_int32_t  Page size. */
            "{bt_minkey,~i},"                 /* u_int32_t  Minkey value. */
            "{bt_re_len,~i},"                 /* u_int32_t  Fixed-length record length. */
            "{bt_re_pad,~i},"                 /* u_int32_t  Fixed-length record pad. */
            "{bt_levels,~i},"                 /* u_int32_t  Tree levels. */
            "{bt_int_pg,~i},"                 /* u_int32_t  Internal pages. */
            "{bt_leaf_pg,~i},"                /* u_int32_t  Leaf pages. */
            "{bt_dup_pg,~i},"                 /* u_int32_t  Duplicate pages. */
            "{bt_over_pg,~i},"                /* u_int32_t  Overflow pages. */
            "{bt_empty_pg,~i},"               /* u_int32_t  Empty pages. */
            "{bt_free,~i},"                   /* u_int32_t  Pages on the free list. */
            "{bt_int_pgfree,~i},"             /* u_int32_t  Bytes free in internal pages. */
            "{bt_leaf_pgfree,~i},"            /* u_int32_t  Bytes free in leaf pages. */
            "{bt_dup_pgfree,~i},"             /* u_int32_t  Bytes free in duplicate pages. */
            "{bt_over_pgfree,~i}]",           /* u_int32_t  Bytes free in overflow pages. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_magic),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_version),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_metaflags),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_nkeys),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_ndata),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_pagesize),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_minkey),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_re_len),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_re_pad),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_levels),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_int_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_leaf_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_dup_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_over_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_empty_pg),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_free),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_int_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_leaf_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_dup_pgfree),
            (int)(((DB_BTREE_STAT*)(c->o.db_stat_p))->bt_over_pgfree));

    case DB_HASH:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{hash_magic,~i},"                /* u_int32_t  Magic number. */
            "{hash_version,~i},"              /* u_int32_t  Version number. */
            "{hash_metaflags,~i},"            /* u_int32_t  Metadata flags. */
            "{hash_nkeys,~i},"                /* u_int32_t  Number of unique keys. */
            "{hash_ndata,~i},"                /* u_int32_t  Number of data items. */
            "{hash_pagesize,~i},"             /* u_int32_t  Page size. */
            "{hash_ffactor,~i},"              /* u_int32_t  Fill factor specified at create. */
            "{hash_buckets,~i},"              /* u_int32_t  Number of hash buckets. */
            "{hash_free,~i},"                 /* u_int32_t  Pages on the free list. */
            "{hash_bfree,~i},"                /* u_int32_t  Bytes free on bucket pages. */
            "{hash_bigpages,~i},"             /* u_int32_t  Number of big key/data pages. */
            "{hash_big_bfree,~i},"            /* u_int32_t  Bytes free on big item pages. */
            "{hash_overflows,~i},"            /* u_int32_t  Number of overflow pages. */
            "{hash_ovfl_free,~i},"            /* u_int32_t  Bytes free on ovfl pages. */
            "{hash_dup,~i},"                  /* u_int32_t  Number of dup pages. */
            "{hash_dup_free,~i}]",            /* u_int32_t  Bytes free on duplicate pages. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_magic),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_version),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_metaflags),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_nkeys),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ndata),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_pagesize),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ffactor),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_buckets),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_free),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_bfree),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_bigpages),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_big_bfree),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_overflows),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_ovfl_free),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_dup),
            (int)(((DB_HASH_STAT*)(c->o.db_stat_p))->hash_dup_free));
        
    case DB_QUEUE:
        return erl_format(
            "[{db_type,~a},"                  /* we add this so caller doesn't have to call db_get_type */
            "{qs_magic,~i},"                  /* u_int32_t  Magic number. */
            "{qs_version,~i},"                /* u_int32_t  Version number. */
            "{qs_metaflags,~i},"              /* u_int32_t  Metadata flags. */
            "{qs_nkeys,~i},"                  /* u_int32_t  Number of unique keys. */
            "{qs_ndata,~i},"                  /* u_int32_t  Number of data items. */
            "{qs_pagesize,~i},"               /* u_int32_t  Page size. */
            "{qs_extentsize,~i},"             /* u_int32_t  Pages per extent. */
            "{qs_pages,~i},"                  /* u_int32_t  Data pages. */
            "{qs_re_len,~i},"                 /* u_int32_t  Fixed-length record length. */
            "{qs_re_pad,~i},"                 /* u_int32_t  Fixed-length record pad. */
            "{qs_pgfree,~i},"                 /* u_int32_t  Bytes free in data pages. */
            "{qs_first_recno,~i},"            /* u_int32_t  First not deleted record. */
            "{qs_cur_recno,~i}]",             /* u_int32_t  Next available record number. */

            bdb_DBTYPE_code_atoms[idx].atom_text,    /* text to convert to atom -- i.e. NOT .atom !! */
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_magic),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_version),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_metaflags),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_nkeys),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_ndata),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pagesize),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_extentsize),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pages),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_re_len),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_re_pad),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_pgfree),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_first_recno),
            (int)(((DB_QUEUE_STAT*)(c->o.db_stat_p))->qs_cur_recno));
        
    default:
        return NULL;
    }

}


int make_db_compact_xreturn(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount)
{
    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_free);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_free);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_examine);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_examine);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_levels);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_levels);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_deadlock);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_deadlock);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_pages_truncated);
    *msgcount = LOAD_INT(msg, *msgcount, c->i.compact_data.compact_pages_truncated);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);

    *msgcount = LOAD_ATOM(msg, *msgcount, am_compact_end_key);
    /* ... handle any necessarh copying of the data that was allocated by BDB */
    make_dbt(desc, c, msg, members, msgcount, &c->o.end_key, NULL);    
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
   
    /* The list has to end with NIL */
    *msgcount = LOAD_NIL(msg, *msgcount);

    *members = 7; /* including NIL l */
    return 7;
}


/* This must advance *members by 1.
 *
 * This is used to convert
 *    0 to 'ok'
 *  +ve to posix (e.g. 'enomem', 'einval')
 *  -ve to BDB   (e.g. 'bdb_DB_NOTFOUND', 'bdb_DB_LOCK_DEADLOCK')
 *
 * Unrecognized codes are converted to
 *         {unknown_error_code, int()}
 */
void convert_bdb_return_code_to_driver_term(descriptor_t *desc,
                                            int error_code,
                                            ErlDrvTermData *msg,
                                            int *members,
                                            int *msgcount)
{
    if ( error_code == 0 )
    {
        *msgcount = LOAD_ATOM(msg, *msgcount, am_ok);
        *members += 1;
        return;
    }
    else if ( error_code > 0 )
    {
        /* POSIX (error_codeno) */
        int idx = error_code - 1;
        if (  idx < NUM_ARY_ELEMS(errno_atoms)
           && errno_atoms[idx].code != 0 )
        {
            *msgcount = LOAD_ATOM(msg, *msgcount, errno_atoms[idx].atom);
            *members += 1;
            return;
        }
        else {
            /* unknown POSIX (positive) error_codeor */
        }
    }
    else if ( error_code >= -30999  &&  error_code <= -30800 )
    {
        /* In documented range of a BDB public error code */
        int idx = 0 - (-30999 - error_code);
        if (  idx < NUM_ARY_ELEMS(bdb_public_error_code_atoms)
           && errno_atoms[idx].code != 0 )
        {
            *msgcount = LOAD_ATOM(msg, *msgcount, bdb_public_error_code_atoms[idx].atom);
            *members += 1;
            return;
        }
        else {
            /* We don't have a definition for this code */
        }
    }
    else {
        /* negative but not in BDB public range or Erlang/BDB range */
    }

    *msgcount = LOAD_ATOM(msg, *msgcount, am_unknown_error_code);
    *msgcount = LOAD_INT(msg, *msgcount, error_code);
    *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
    *members += 1;
}

int make_dbtype(descriptor_t *desc, callstate_t *c, ErlDrvTermData *msg, int *members, int *msgcount)
{
    int idx = c->o.db_type - 1;
    assert(idx >= 0 && idx < NUM_ARY_ELEMS(bdb_DBTYPE_code_atoms) );
    *msgcount = LOAD_ATOM(msg, *msgcount, bdb_DBTYPE_code_atoms[idx].atom);
    *members += 1;
    return 1;
}


ETERM * make_log_filenames_external_term(descriptor_t *desc, callstate_t *c)
{
    ETERM * list_so_far = erl_mk_empty_list();
    
    if (c->o.log_filenames_p != NULL)
    {
        /* we want to preserve the original order, so add them to the list in reverse order */
        char ** filename_p = c->o.log_filenames_p;
        while ( *filename_p != NULL ) { /* find NULL */
            ++filename_p;
        }
        ptrdiff_t num_filenames = filename_p - c->o.log_filenames_p;
        while ( num_filenames-- )
        {
            --filename_p;
            list_so_far = erl_cons(erl_mk_estring(*filename_p, strlen(*filename_p)),
                                   list_so_far);
        }
        /* c->o.log_filenames_p will be freed in free_callstate() */
    }    
    return list_so_far;
}


/* This is a temporary hack; it becomes part of rep_set_timeout in BDB v4.6 */
int my_rep_set_master_checkpoint_sleep(DB_ENV * dbenv, u_int32_t period_seconds)
{
    dbenv->dbenv_rep_checkpoint_delay = period_seconds;
    return 0;
}

/* This is a hack to interupt BDB calls that are doing long os_sleeps
 * (e.g. txn_checkpoint if my_rep_set_master_checkpoint_sleep has been set
 * to tens or hundreds of seconds.
 * After this function is called once, ALL current and future os_sleep
 * calls for this BDB environment handle will return immediately.
 * The effect of this call is only reset by closing and re-opening the environment.
 */
int my_env_interrupt_os_sleep(DB_ENV * dbenv)
{
    descriptor_t * desc = (descriptor_t *)(dbenv->app_private);
    int res;

    if ( (res = write(dbenv->dbenv_interrupt_os_sleep_pipe_fds[1], "0", 1)) != 1 ) {
        /* We ignore any error from write, and always return success */
        int err = errno;
        EDTK_PORT_LOG_FORCE(desc, "warning: write() returned %d (errno = %d)", res, err);
    }
    return 0;
}


int make_c_get_multiple_xreturn(descriptor_t *desc, callstate_t *c,  ErlDrvTermData *msg, int *members, int *msgcount)
{
    ErlDrvBinary * bin;
    DBTYPE dbtype;
    
    /* c_get_multiple has (in outputv) allocated a buffer with edtk_driver_binary_malloc
     * and stored the pointer in c->i.data_in.data.
     * All we need to do is to return that driver binary.
     *
     * Note: we don't need to worry about the buffer that was allocated (c->i.data_in->ulen) being
     * bigger than the size of the returned data.  BDB appears to always set c->i.data_in->size
     * to the value of c->i.data_in->ulen (not so good in pipe mode, but fine in linked-in mode).
     * So we *don't* need to driver_realloc_binary to 'shrink' the binary before sending it
     * to the Erlang VM.
     */
    
    assert( DB_DBT_USERMEM == c->i.data_in.flags );
    bin = edtk_alloced_ptr2ErlDrvBinary(c->i.data_in.data);

    /* Note: The format of the binary data depends on the type of the underlying
     * database. The public cursor API does not give a way to access that type,
     * so we use this hack into the private data members. */
    dbtype = c->i.dbc->dbtype;
     
    /* debug_print_c_get_multiple(desc, &c->i.data_in, c->i.flags, dbtype); */

     if ( c->i.flags & DB_MULTIPLE_KEY )
     {
         /*    keys/records (no duplicates) from btree/hash
          * or keys/records from queue/recno */
         if ( (DB_BTREE == dbtype) || (DB_HASH == dbtype) )
         {
             /* The normal case. post_process_c_get_multiple_result knows
              * to expect this as we return a single binary
              * and Flags implies DB_MULTIPLE_KEY (i.e. does NOT contain DB_MULTIPLE).
              * Even though BDB sets c->i.key_in, we don't send it
              * as it is redundant (keys are embedded in the data) */

             *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, c->i.data_in.size);
             *members = 1;
             return 1;
         }
         else
         {
             assert( (DB_QUEUE == dbtype) || (DB_RECNO == dbtype) );
             
             /* DB_MULTIPLE_KEY was used on a queue or recno database,
              * requiring DB_MULTIPLE_RECNO_NEXT to decode the result.
              * So we send an atom to tell post_process_c_get_multiple_result
              * how to interpret the binary.
              * Even though BDB sets c->i.key_in, we don't send it
              * as it is redundant (keys (recnos) are embedded in the data) */

             *msgcount = LOAD_ATOM(msg, *msgcount, am_recno_keys);
             *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, c->i.data_in.size);
             *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
             *members = 1;
             return 1;
         }
     }
     else
     {
         assert( c->i.flags & DB_MULTIPLE );

         /*    *duplicate* records from btree/hash, 
          * or consecutive records from queue/recno, and first record number is returned in c->i.key_in.
          *
          * if DB_BTREE or DB_HASH:
          *   The normal case (for duplicates). post_process_c_get_multiple_result knows
          *   to expect this as we return a single binary and Flags contains DB_MULTIPLE.
          *
          * if DB_QUEUE or DB_RECNO:
          *
          *   DB_MULTIPLE_KEY was used on a queue or recno database,
          *   giving a list of records, and a single recno initial-key.
          *   recno is an unsigned int, but LOAD_INT is signed,
          *   so we send it as a binary DBT and let the Erlang code convert it.
          *
          * Either way, we return the key and the encoded binary.
          */
         {
             int members = 0;
             make_dbt_key(desc, c, msg, &members, msgcount);
         }
         *msgcount = LOAD_BINARY(msg, *msgcount, bin, 0, c->i.data_in.size);
         *msgcount = LOAD_TUPLE(msg, *msgcount, 2);
         *members = 1;
         return 1;
     }
}

#if 0
/* emulate post_process_c_get_multiple_result() from berkeley_db.xml */
void debug_print_c_get_multiple(descriptor_t * desc, DBT * dbt, u_int32_t flags)
{
#define MIN(a,b) ((a)>(b)?(b):(a))

    void * p;

    DB_MULTIPLE_INIT(p, dbt);
        
    if ( flags & DB_MULTIPLE )
    {
        void * retdata;
        size_t retdlen;
        do
        {
            DB_MULTIPLE_NEXT(p, dbt, retdata, retdlen);
            if ( p != NULL )
            {
                if ( NULL == retdata )
                {
                    EDTK_PORT_LOG_FORCE(desc, "deleted record");
                }
                else
                {
                    char dbuff[1024];
                    memcpy(dbuff, retdata, MIN(sizeof(dbuff), retdlen));
                    dbuff[NUM_ARY_ELEMS(dbuff) - 1] = 0;
                    EDTK_PORT_LOG_FORCE(desc, "record %s", dbuff);
                }
            }
        } while (p != NULL);
    }
    else if ( flags & DB_MULTIPLE_KEY )
    {
        void * retkey;
        size_t retklen;
        void * retdata;
        size_t retdlen;
        do
        {
            DB_MULTIPLE_KEY_NEXT(p, dbt, retkey, retklen, retdata, retdlen);
            if ( p != NULL )
            {
                char kbuff[1024];
                char dbuff[1024];
                memcpy(kbuff, retkey, MIN(sizeof(kbuff), retklen));
                kbuff[NUM_ARY_ELEMS(kbuff) - 1] = 0;
                memcpy(dbuff, retdata, MIN(sizeof(dbuff), retdlen));
                dbuff[NUM_ARY_ELEMS(dbuff) - 1] = 0;
                EDTK_PORT_LOG_FORCE(desc, "key %s, record %s", kbuff, dbuff);
            }
        } while (p != NULL);
    }
    else
    {
        EDTK_PORT_LOG_FORCE(desc, "no DB_MULTIPLE_KEY/DB_MULTIPLE set");            
    }
}
#endif
