%%%-------------------------------------------------------------------
%%% File    : berkeley_db_port_coordinator.erl
%%% Author  : Chris Newcombe
%%% Description : 
%%%
%%%   A gen_server that owns a single instance of an EDTK-generated
%%%   berkeley_db port.
%%%
%%%   This server does *not* abstract the BDB API.
%%%   Instead, it allows multiple Erlang processes to share 
%%%   a single BDB port in a safe way.
%%%   
%%%   Pros of this approach:
%%%        - For linked-in drivers, commands and replies go 
%%%          directly to/from the port, without any intermediate server.
%%%          (In pipe-mode, the server forwards the replies to the 
%%%           correct client processes).
%%%
%%%        - Less code, as no need to abstract the BDB API.
%%%
%%%   Cons:
%%%        - Clients must obey some (fairly simple) restrictions.
%%%
%%%        - The full BDB API is exposed, which can cause havoc. e.g. Closing a 
%%%          transaction that has a cursor open causes an environment panic,
%%%          which in linked-in mode will abort the entire Erlang VM; 
%%%
%%%          *** You MUST read the BDB documentation before using this code ***.
%%%
%%%        - driver_caller and driver_send_term are not supported in pipe-mode.
%%%          So this server recieves the messages, and we have to arrange 
%%%          for the client's pid to be included in the message, so the server
%%%          can forward the message to the client.
%%%
%%%   The server also offers some important conveniences
%%%        - multiple registered instances are supported (one per BDB environment)
%%%        - simple configuration with good defaults for linked & pipe modes
%%%        - convenient replication, checkpointing, deadlock-detection, memp_trickle, etc.
%%%
%%%   --- API ---
%%%
%%%   init()
%%%
%%%        - Opens a single port instance, configures the BDB environment (DB_ENV)
%%%          for this port, and opens the DB_ENV (e.g. running recovery if requested).
%%%
%%%        - As part of opening the DB_ENV, performs locking as required to 
%%%          satisfy BDB's pre-conditions (to avoid crashes and/or data-corruption).
%%%          See below for details.
%%%
%%%        - Traps exits.  If an exit is received from a client, 
%%%          handle_info will call ?BDB:group_valmap_cleanup(DeadClientGroupValmapId).
%%% 
%%%         If replication is enabled then a gen_event process is created with the same name as the port coordinator
%%%         process name, plus a suffix.
%%%         Clients can retrieve that name with get_replication_gen_event_proc_name/0,1.
%%%         (Internally, the server must call dynamic_replication_gen_event_proc_name/0 to get the same information.)
%%%     
%%%
%%%   acquire_bdb_port_handle()
%%%
%%%        returns   BdbPortHandle   (includes the DBENV handle as member 'dbenv')
%%%      or throws   {bdb_port_coordinator_error, Reason}
%%%
%%%      The BdbPortHandle contains a unique GroupValmapId that will be used to 
%%%      track resources (e.g. transactions, cursors) opened by the caller, and clean them 
%%%      up if the caller dies.
%%%
%%%      The BdbPortHandle is bound to the pid of the caller, and should not be passed to 
%%%      other processes.
%%%
%%%      Side-effects: server starts monitoring the calling process (if call is successful)
%%%
%%%      Client may then:
%%%        - Directly hit the returned port handle via functions in 
%%%          berkeley_db_drv.erl (the reply will be sent to the client).
%%%        - Open its own transactions and cursors (**but see below**)
%%%
%%%      Client MUST
%%%        - When opening transactions and cursors, client **MUST** 
%%%          specify the UniqueGroupValmapId returned from acquire_bdb_port_handle().
%%%          This allows the server can clean up if the client exits.
%%%
%%%      Client must NOT
%%%        - Pass the BdbPortHandle to any other processes (it won't work anyway).
%%%
%%%        - Open any Db handles itself.  It must use get_db_handle().
%%%
%%%          Note that get_db_handle does not support opening (creation) of databases within a 
%%%          transaction (because there is a race condition -- other clients may attempt to use 
%%%          that DB handle before the original client commits their transaction -- which BDB 
%%%          prevents, and worse, the original client may abort the transaction and leave the 
%%%          server with a cached but invalid DB handle, which is then handed out to other clients).
%%%          
%%%          This means that if a client *must* create a DB handle, it must do it itself.  
%%%          However, this is potentially dangerous:
%%%
%%%          - DB handles cost about 600 bytes of heap, plus a BDB mutex and a locker (and a valmap slot).
%%%            (Fortunately, they don't each require an OS file descriptor.)
%%%            So if many Erlang processes do this, there may be resource problems.
%%%
%%%          - If a client dies while holding a DB handle that it created/opened (i.e. group_valmap_cleanup sees a DB handle),
%%%            and if there happen to be any open txn handles (not even in-use at that instant, just open)
%%%            then the OS process (i.e. probably the entire Erlang VM) will be hard aborted.
%%%            This cannot happen if clients stick to using get_db_handle (as the port_coordinator owns 
%%%            all of the handles).
%%%
%%%        - Close any Db handles, or the DBENV handle.
%%%
%%%        - Remove or truncate databases, or remove the environment.
%%%
%%%        - Pass it's GroupValmapId to any other process.
%%%          In general, processes and GroupValmapIds should have a 
%%%          one-to-one relationship).
%%%
%%%        - Allow ANY concurrent operations to use the same 
%%%          transaction or cursor handle.  This means not specifying 
%%%          the same handle to multiple concurrent non_block operations,
%%%          and NOT passing transaction or cursor handles to other 
%%%          processes.
%%%
%%%
%%%   get_db_handle(DBType, DBName)   DBType is one of bdb_DB_BTREE, bdb_DB_HASH, DB_QUEUE etc.
%%%                                   DBName is a Filename or {Filename, DatabaseNameWithinFile}
%%%
%%%          returns   DB
%%%         or throws  {bdb_port_coordinator_error, Reason}
%%%
%%%        This returns a copy of a shared Db handle for a database.
%%%        If the database is not open it is opened, and if it does 
%%%        not exist it is created.  The gen_server caches these and
%%%        reuses them.  
%%%
%%%        Note: when creating the database it is NOT possible for the caller to provide 
%%%        a Txn handle to be passed to db_open, because that txn may later be rolled back
%%%        by the caller (and we will have cached an invalid handle).  Also there would be a 
%%%        condition: other clients may request that handle before the original caller 
%%%        commits the txn.  BDB returns an error if a DB handle is used from another txn 
%%%        when the txn that created that DB has not yet been committed.
%%%        If the caller really needs to create a database as part of a transaction then 
%%%        it should create the database handle itself.  
%%%        (Each DB handle requires about 600 bytes of heap plus a locker and a mutex
%%%        so this approach should be applied with caution -- e.g. after the transaction 
%%%        commits, consider closing the private handle and asking the port_coordinator 
%%%        for a cached/cacheable handle.)
%%%
%%%        If the database is already open then the type must match
%%%        and the config list must be 'compatible' (not contradict
%%%        the existing config list)
%%%
%%%      Client must NOT close Db handles, or truncate or delete databases.
%%%
%%%
%%%   clone_bdb_port_handle()
%%%           
%%%          returns BdbPort handle useable by the calling process but which
%%%          SHARES the same group_valmap_id as the port handle that was given as an argument.
%%%
%%%          The returned clone handle SHOULD NOT CURRENTLY be passed to release_bdb_port_handle().
%%%          Caveat Emptor: THIS MAY CHANGE IN FUTURE
%%%
%%%        This API allows a set of related processes to operate on a shared pool of BDB resources.
%%%
%%%        This is an *unusual* use-case, and support is limited (and may change in some ways 
%%%        in future).  So read the following carefully.
%%%
%%%        The motivating use-case is when a parent process wants to open a parent txn,
%%%        and then spawn several child processes each of which will open a child txn.
%%%
%%%        [Sidebar: BDB requires, and EDTK enforces, that only a single thread is using (calling BDB)
%%%         with ANY txn handle that belongs to a single hierarchy. So ports using parent/child 
%%%         txns must coordinate their actions or retry if they get a valmap_in_use error.] 
%%%
%%%        When all of the child txns have completed, the parent will commit the parent process.
%%%        The child processes cannot use acquire_bdb_port_handle() as they will each get a new 
%%%        group_valmap_id, and currently parent/child valmaps must all belong to the same 
%%%        group id. It would require a huge change to EDTK to support cross-group valmap relationships
%%%        because that would imply a dependency graph across group_valmap_ids too.  e.g. 
%%%        Consider the case where the child process opens a cursor in its child txn, and 
%%%        then the parent process crashes.  To cleanup the parent txn requires cleaning-up the 
%%%        cursors in the child processes/groups, as well as the child txns of the parent txn.
%%%
%%%        By allowing a set of processes to share a group-id that complexity can be avoided.
%%%        However, there are other important implementation decisions:
%%%
%%%        Should the process that calls clone_bdb_port_handle be registered with the port coordinator,
%%%        and if so, how should the port_coordinator manage them? (so it monitors to them and know if they exit)?
%%%        If we do register them, there are several options:
%%%         
%%%        a. If one of the processes with a 'clone handle' exits,
%%%        should the port coordinator run group-valmap-cleanup on the
%%%        shared group?  (this would invalidate handles owned by the
%%%        other processes)
%%%         
%%%        b. If one of the processes with a 'clone handle' exits,
%%%        should the port-coordinator kill all of the processes using
%%%        that are using that group, and then run group-cleanup for
%%%        that group.
%%%         
%%%        c. Or should the port-coordinator keep a ref-count on the
%%%        group-id, and only run group-cleanup after the last process
%%%        using that group exits?
%%%         
%%%        This is complicated by the fact that the 'standard' helpers; do_txn, do_txn_on_master etc 
%%%        *always* run group-cleanup before returning, so those functions implement choice (a), 
%%%        and it would be too much of a performance hit for 'normal' code to change this.
%%%         
%%%        Current implementation:
%%%
%%%        Processes that call clone_bdb_port_handle are ** NOT ** registered with the port-coordinator,  
%%%        so group_valmap_cleanup is ONLY called if the 'root parent' process (the one that originally 
%%%        called acquire_bdb_port_handle) finally exits.
%%%
%%%        This means that if child processes open cursors, other txns etc, those resources will be 
%%%        accumulated in the valmap group, but the parent process won't (by iteself) have handles 
%%%        to them or know anything about them. This gives rise to the following:
%%%        
%%%        CRITICAL SAFETY PRE-CONDITION:
%%%
%%%           The 'parent' process MUST ALWAYS obtain explicit consent from the children that it is safe 
%%%           to commit (*or* to explicitly abort) the parent txn. (Otherwise, the child processes may 
%%%           still have cursors open in a child txn, so closing the parent txn will panic the entire BDB environment.)
%%%     
%%%           If the parent cannot contact any child process then the parent should either exit 
%%%           (and let the port-coordinator clean-up), or explicitly call group-valmap-cleanup itself, 
%%%           and start again.   
%%%           Either way, group_valmap_cleanup is called, which will close any cursors before aborting the parent txn.
%%%           The parent *cannot* itself call txn_abort in this situation (because it won't close cursors first).
%%%
%%%        Probably the ideal future model (if full dependency-analysis across valmap groups is not implemented)
%%%        is option (c) -- register the processes that have cloned handles, and do reference-counting to decide 
%%%        when to clean-up. But note that the above safety-precondition will STILL apply even with that model.
%%%
%%%
%%%   list_all_open_databases()
%%%
%%%          returns   list of DBId   where DBId = Filename or {Filename, InternalName}
%%%         or throws  {bdb_port_coordinator_error, Reason}
%%%
%%%
%%%   close_unused_database(DBName, UserTxn)   DBName is a Filename or {Filename, DatabaseNameWithinFile}
%%%                                             UserTxn is required (cannot be 'void')
%%%
%%%          returns   ok
%%%         or throws  {bdb_port_coordinator_error, Reason}
%%%
%%%        This API allows a database to closed in preparation for calling env_dbremove() or env_dbrename() 
%%%        (i.e. before physically deleting or renaming the database).
%%%
%%%        IMPORTANT: This is an **EXTREMELY DANGEROUS** API, with stringent preconditions for safe use.
%%%        Incorrect use will both
%%%              - crash Berkeley DB very badly (undefined behaviour)
%%%              - **almost certainly corrupt the transaction log and make the entire BDB environment UNRECOVERABLE**.
%%%
%%%        Also, you should think strongly before using this API if you are using replication as:
%%%
%%%              - DB handles closed on the master are NOT automatically closed on replicas.
%%%
%%%              - At the time of writing there is a Berkeley DB bug (#15076 or TAR 5815135.992)
%%%                that will cause replication to hang if a database is physically removed on the 
%%%                master but there are DB handles open for that database on replicas.
%%%
%%%        The pre-conditions for this API can ONLY be achieved by conventions implemented and enforced by
%%%        the application (for a suggested convention, see below).
%%%
%%%        The port coordinator does NOT enforce these preconditions, or even detect if they have been
%%%        violated.  The port_coordinator does NOT track which clients have requested a given DB 
%%%        handle, as this would be both expensive (it is a very common operation) and pointless (clients 
%%%        can open their own private handles to the database without knowledge or monitoring by the 
%%%        port coordinator).
%%%
%%%        The pre-conditions are:
%%%
%%%        1. DBName must have been opened by at least one client using get_db_handle().
%%%           (Note that this is done as a side-effect of the coordination scheme described below.)
%%%
%%%        2a. No client may be using (or later begin to use) the shared DB handle for the passed database
%%%           (i.e. the handle returned by get_db_handle() & co).
%%%
%%%        2b. Subtle point: 2a implies that this DB handle CANNOT EVER be used in distributed transaction. 
%%%            This is because the port_coordinator may be told to do incremental cleanup of distributed txns
%%%            - so a 2PC Txn handle may outlive the other valmap slot entries owned by that group
%%%            (basically, the port_coordinator takes ownership of the txn).
%%%            So, if you do need to combine close_unused_database() with (hopefully few) distributed transactions, 
%%%            then processes performing the distributed transactions should open their own DB handles for those transactions.
%%%            HOWEVER, doing so is playing **Russian Roulette**, as if the client crashes with an unresolved prepared txn
%%%            the OS process will be HARD-ABORTED, as it is not safe to close a DB handle that is involved 
%%%            in an unresolved txn.
%%%            Note that most applications do not need to call close_unused_database, in which case, 
%%%            using get_db_handle() with txn_prepare is quite safe (as the port_coordinator owns 
%%%            all of the DB handles and guarantees that the DB handles will stay open longer than an unresolved txns,
%%%            (or, if the port_coordinator cannot resolve all prepared txns before shutting down, then it (the port_coordinator)
%%%            will make the final round of db_close calls safe -- i.e. prevent them from logging via a Sleepycat hack 
%%%            - but it cannot do that for the db_close inside close_unused_database, as the Sleepycat hack manipulates a global flag).
%%%
%%%        3. [Related, but not a true precondition of close_unused_database]
%%%
%%%           As the goal of calling close_unused_database() is in preparation for 
%%%           immediately calling env_dbremove() or env_dbrename() it must be noted that 
%%%           those BDB APIs alsl require that there are **ZERO** open database handles for the 
%%%           database in question.
%%%
%%%           This function (close_unused_database) only closes the handle owned by the port_coordinator.
%%%           So to safely use those BDB APIs the application is also responsible for:
%%%
%%%                - preventing ANY application process from holding or opening a private DB handle 
%%%                  for DBName, until after it has been renamed/removed.
%%%
%%%                - preventing ANY application process from calling get_db_handle() & co. 
%%%                  on DBName, until after it has been renamed/removed.
%%%
%%%        Applications can enforce this preconditions above by ensuring that all operations that use 
%%%        a database which might ever be passed to close_unused_database() first obtain and hold a shared (read) lock 
%%%        on some other object outside of this database.
%%%
%%%        IMPORTANT: the operations should obtain the lock **before** they call get_db_handle(), or db_open(),
%%%        or before they use a privately-cached DB handle obtained from those calls.
%%%        The operat8on should only release that lock after it is certain that it will not refer to the DB handle again 
%%%        (without re-acquiring the read lock first).
%%%
%%%        An example of the 'other object' that must be locked might a 'meta-data' record in a different database 
%%%        -- note that it must be in an entirely different physical file if you want to remove a file-level database.
%%%
%%%        Then the application code that calls close_unused_database() should obtain an exclusive lock on the 
%%%        'meta-data' object (thus waiting for any existing operations that use the database have finished, 
%%%        and also preventing any new operations from starting to using the database)
%%%
%%%
%%%   release_bdb_port_handle()
%%%
%%%      returns   ok
%%%
%%%      or throw  {bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResourcesCleanedUp}}
%%%
%%%      Side-effects: server stops monitoring the client process.
%%%
%%%      (If the server is shutdown or learns of a serious error 
%%%       -- e.g. DB_RUNRECOVERY -- then it will kill all of it's active 
%%%       clients -- i.e. any that have called acquire_bdb_port_handle() but not yet 
%%%       called release_bdb_port_handle).
%%%
%%%      Client should
%%%        - Have closed all of it's cursors and transactions
%%%          before calling this function.
%%%          To err on the side of caution, the server calls 
%%%          group_valmap_cleanup(), just in case the client did not close all 
%%%          of it's txn and cursor handles (hence the error return value).
%%%          But it is very bad practice to rely on this. 
%%%
%%%      Client must NOT
%%%        - Use the port instance again after calling this function.
%%%
%%%
%%%   get_replication_gen_event_proc_name()
%%%
%%%      Returns the locally-registered name of a gen_event process that 
%%%      will publish replication events (if replication is enabled).
%%%     
%%%
%%%   Additional server responsibilities:
%%%
%%%   - If the BDB port is opened in 'pipe' mode then the port driver 
%%%     cannot use driver_caller/driver_send_term to send replies
%%%     directly to the Erlang process that sent each command.
%%%     Instead, the replies come to this server as it is the 
%%%     port's connected process.  Therefore we must forward the 
%%%     replies to the correct clients. (The client's pid is included 
%%%     in the reply, so we don't have to do any book-keeping).
%%%
%%%     Important: It is critical that we garbage-collect binaries 
%%%     aggresively.  For this reason, if pipe-mode is used 
%%%     the port_coordinator is spawned using {fullsweep_after, 0}
%%%     so the collector will consider reducing reference counts of 
%%%     binaries whenever the process is garbage collected.
%%%     Still there is the concern that the port_coordinator won't 
%%%     be garbage-collected very often, as it's State does not 
%%%     change much (only when clients register/unregister, and 
%%%     when DB handles are first opened).  In pipe-mode we might 
%%%     want to add explicit calls to the collector on a timer.
%%%     However, the last paragraph of the following mailining-list 
%%%     post suggests that we shouldn't have to do this.
%%%         
%%%        From: Bjorn Gustavsson  
%%%        Date: 20 Oct 2003 14:02:08 +0200 
%%%
%%%        Garbage collections of binaries is a tricky issue.
%%%         
%%%        To clarify how binaries are removed: It is done by a combination of
%%%        garbage collection and reference counters. Binaries are represented in
%%%        each process as small objects on the heap. Those objects are also part
%%%        of a linked list (one list per process). After each garbage collection,
%%%        the linked list is scanned to which objects were not moved (thus, no longer
%%%        referenced). For each unreferenced object, the reference count in the binary
%%%        itself is decreased.
%%%         
%%%        As a consequence of the above, if a binary is referenced from the
%%%        old generation heap, it cannot be removed until a new major collection
%%%        (a.k.a. fullsweep collection) is done. How often a major collection occurs
%%%        depends on the application.
%%%         
%%%        If a process is known to have large binaries but small amounts of live
%%%        heap data, the process can be started using the spawn_opt/4, passing in the
%%%        '{fullsweep_after,0}' option. (See the documentation for spawn_opt/4 for
%%%        further details.) That will make every GC a major collection (fullsweep),
%%%        cleaning away any binaries that are no longer referenced.
%%%         
%%%        Note that a major collection copies all data that is is still alive; if you
%%%        have large amounts of live data on the heap doing a major collection every time
%%%        will typically be slower than the default generational GC (which only does a
%%%        major collection when there is no more place on the old generation heap).
%%%         
%%%        BTW, when a process aquires a new binary, it gets "penalty points" so that it
%%%        will be garbage collected sooner. But a binary that has survived one GC will
%%%        not generate further "penalty points".
%%%         
%%% 
%%% Evironment Locking
%%%
%%% [That section is perhaps an absurd amount of design analysis for a few lines of code, but 
%%%  the consequences of getting this wrong are catastrophic.]
%%%
%%% BDB requires that the application enforce various pre-conditions and invariants.
%%% If these are violated, crashes and/or irrevocable data-corruption may occur.
%%%
%%% Some of these pre-conditions are satisfied by EDTK (e.g. destruction order 
%%% of cursors, txns).
%%%
%%% Other preconditions are satisfied by Erlang's management of ports (e.g. the driver's 
%%% 'stop' function is guaranteed to be called if it's connected process exits).
%%%
%%% Other preconditions are satisified by code in the port-coordinator; e.g. 
%%% - the port-coordiantor detects if a registered client crashes, and calls 
%%%   group valmap cleanup for that client.
%%% - the port-coordinator's close_environment fynction mirrors as much as possible the b
%%%   ehaviour of bdb_stop, and if it finds it cannot handle the current state of things,
%%%   it defers to bdb_stop. (Tthe advantage of using close_environment instead of bdb_stop 
%%%   whenever possible is that aborting txns and closing databases might take significant time, 
%%%   and bdb_stop will block an entire Erlang VM scheduler thread until it completes).
%%%
%%% (The above 3 mechanisms combine to guarantee that if some (Erlang) process 
%%% crashes while using BDB then it's BDB resources (cursors, txns etc) are 
%%% cleaned up in the appropriate order, without having to kill any other Erlang 
%%% processes  Note: neither the port-coordinator nor EDTK provide any mechanism for 
%%% cleanup if an EDTK worker-thread fails while holding BDB resources; it is assumed 
%%% that the OS process running EDTK will crash and BDB recovery will be run when 
%%% the port is re-opened.)
%%% 
%%% One of the most important BDB preconditions is that when recovery is run on an 
%%% environment (which we do by default when starting a port-coordinator), no 
%%% other thread or process may be using the environment, or begin using the environment 
%%% until after recovery has finished.
%%% In many applications this requires holding a public mutex around the call to 
%%% env_open (if DB_ENV_RECOVER is set), and some secondary public indication (e.g. 
%%% just the existance of the public mutex object) that a thread/process is using the 
%%% environment and therefore any other thread/process should not run recovery.  
%%% Obviously both of those mechanisms must be somehow reset if the thread/process that 
%%% is running recovery or holding the environment fails.
%%%
%%% The EDTK/port-coordinator model for using BDB is 
%%%
%%%  p1. Only a single Erlang VM/node may access any given BDB environment at a time.
%%%
%%%      Sidebar: this means that only a single OS process will be accessing 
%%%      the environment (in either linked-in mode or pipe-mode), which means 
%%%      that DB_PRIVATE may safely be used if desired. (This may give a performance increase.)
%%%
%%%  p2. Within that VM, only a single port-instance may access the BDB environment at a time.
%%%
%%% The second precondition is tricky because a port-instance may live (slightly) longer than 
%%% the port-coordinator process that created the port, so we _cannot_ rely on local process  
%%% registration to act as an access-mutex (it's almost good enough, but not quite).
%%% e.g. If the port-coordinator is hard-killed with exit(Pid,kill), or crashes in close_environment,
%%% then bdb_stop is run (and likely has work to do), but the port-coordinator process nolonger 
%%% exist (and will send out EXIT signals). 
%%% i.e. the Erlang VM does *not* keep port-owning processes artificially alive until the 
%%% port's cleanup function has completed.  
%%% And in most applications it is very likely that a supervisor will restart the failed
%%% port-coordinator, and unless prevented, the restarted port-coordinator might open the e
%%% nvironment and run recovery while bdb_stop
%%% is still cleaning up the environment handle owned by the previous port instance.
%%% e.g.
%%%   - In pipe-mode the externally-spawned pipe-main program for the old port instance 
%%%     may still be running bdb_stop
%%%
%%%   - In linked-in mode with the SMP emulator and more than one scheduler thread,
%%%     one scheduler thread may still be running bdb_stop while the other runs 
%%%     the restarted port coordinator.  We *could* use driver-level locking to serialize  
%%%     these, but that would have serious performance implications if we want to open 
%%%     multiple environments within the same Erlang VM (and we certainly do want to do so).
%%%
%%% The other reason that we cannot rely on local process registration is that it does nothing 
%%% to prevent two port coordinators starting with different names but with the same environment 
%%% directory.  We could us a BDB-specific registry, perhaps an ETS table, for this, but it 
%%% would be ideal to have one solution that solves all of the problems; 
%%%
%%%  1. Prevent multiple nodes accessing the same environment simultaneously
%%%  2. Prevent different port-coordinators in the same node from accessing the same environment simultaneously
%%%  3. Prevent overlapping shutdown/restart of the same port-coordinator from accessing the same environment simultaneously
%%%  4. Allows safe use of DB_PRIVATE in both linked-in and pipe modes
%%%     (i.e. only a single OS process, with a single DB_ENV handle may access the environment at any time.)
%%%
%%% Numbers #1 and #3 appear to require a locking mechanism external to Erlang.
%%%
%%% The DB_REGISTER feature in v4.5 superficially appears to solve #1, #2 and #3.  
%%% However, DB_REGISTER requires that there be only a single DB_ENV handle per database 
%%% environment in each process. 
%%%   "As the DB_REGISTER locking is per-process, not per-thread, multiple DB_ENV handles in a 
%%%    single environment could race with each other, potentially causing data corruption."
%%% This is exactly the situation in #2 and #3, so DB_REGISTER won't provide the type of locking 
%%% that we need. 
%%% Note: DB_REGISTER is implicitly incompatible with DB_PRIVATE, as DB_REGISTER only 
%%% applies to multiple processes, and DB_PRIVATE restricts the application to single 
%%% process, so this excludes #4 as well.
%%% (Note that once our own locking solution is in place, DB_REGISTER is safe because 
%%% our locking solution guarantees that only 1 DB_ENV handle will be accessing 
%%% the environment at a time. Of course, as only a single process will be accessing the 
%%% environment, DB_REGISTER is essentially useless. The only benefit _might_ be 
%%% slightly faster startup as DB_REGISTER can apparently detect if recovery really needs 
%%% be run, and if it is not necessary it simply doesn't do it.
%%% However, I have no idea how reliable that decision is. (It would be hard to verify that 
%%% this feature is bullet-proof.)
%%%
%%% Which Locking Solution?
%%%
%%% Ideally the OS would provide an IPD mechanism that we could use for the lock.
%%% The properties that are required are
%%%   - nameable              (we would use the symlimk-resolved, absolute path of the env_dir)
%%%   - publicly visible
%%%   - automatically released if a process crashes
%%%
%%% e.g. I have used Windows public named mutexes for this in the past.
%%% Linux does have IPC semaphores.  However, we have to run on RHEL3 (2.4 kernel)
%%% and named public semaphores are not supported until 2.6 kernels.
%%% Even then, named semaphores are not automatically removed when the last process using them dies 
%%% -- they must be explicitly unlinked first (which removes the name immediately).  That's 
%%% not the right building-block.
%%%
%%% At first glance, normal files don't seem to be useful for this as tne linux/unix file APIs 
%%% have no facility to grant truly exclusive access to a file (even for writes!).
%%%
%%% Linux/unix/posix does provide the O_CREAT|O_EXCL flags to the posix open() call,
%%% but all that does is to atomically create a file if it does not exist, and return an error if it
%%% already exist.  But obviously the file is not automatically deleted if the creating process crasnes.
%%%
%%% Fortunately, linux does support the notion of 'advisory' locks on files.  These are locks which 
%%% *don't* actually affect the ability of a process to read or even write the specified file.
%%% Instead the file is simply used as an identity for the lock (a convenient namespace).
%%%
%%% The file specified for an advisory lock must actually exist, 
%%% so we will create a file in the BDB environment's 'home' directory.
%%%
%%% (It would be nice to use the home-directory itself as open() will open a directory 
%%% if O_RDONLY is used.  Or opendir() followed by dirfd() gives the file descriptor of 
%%% a directory.  However flock returns EBADF if we do this.)
%%%
%%% Q. Do we use the OS 'advisory lock' to mean "I own this BDB environment"?
%%%    Or do we write data to a file that says which process/port-coordinator 
%%%    owns the environment (i.e. the contents of that file are a semantic lock), 
%%%    and only use the OS-level advisory lock to control access to the file?
%%%
%%% With the former, we would hold the lock for the entire time that the environment is open,
%%% and only release it after we close the environment.  The OS will automatically 
%%% release the lock if the process owning it crashes.
%%% With the latter approache we would release the lock on the file after reading and modifying it, 
%%% (immediately before opening the BDB environment).
%%%
%%% Pros of "owning lock means I own this environment"
%%%   - [BIG] Just by acquiring the lock we *know* that we own the environment.
%%%   - [BIG] The will OS automatically release the lock if our process crashes.
%%%
%%% Cons of "own this environment"
%%%   - No-one else can safely read the file (even though unix will let them as the lock is only advisory)
%%%     as they don't know if the contents will be changed while they are reading 
%%%     -- i.e. there are zero consistency guarantees.
%%%     So if a process finds that the lock is held, it does NOT know who is holding it.
%%%     Possible workaround:
%%%     We could always put an MD5 of the (other) contents at the start or end of the file,
%%%     and re-read the entire file until the hash agreed with the contents.
%%%
%%% Pros of "owning lock means I am reading or modifying the contents of the lock file"
%%%   - If you own the lock then the contents of the lock file are stable (but not necessarily consistent/valid),
%%%     and the lock is only held for a very short while, so any interested process can read the file.
%%%
%%% Cons of "owning lock means I am reading or modifying the contents of the lock file"
%%%   - The contents of the file are tnot* guaranteed to be consistent, as an earlier process may have crashed 
%%%     while writing to the file. What do we do in this case? Act like the file is empty?
%%%
%%%   - [BIG] to decide if is safe to open the BDB environment a process must
%%%     - obtain the lock on the lock file
%%%     - read and parse its contents
%%%     - test if any OS pids mentioned in the file are still running or (if they are running) 
%%%       still have the same identity.
%%%     - make a decision, successfully modify the file acordingly, and release the lock.
%%%
%%% In the "file contents are a semantic lock" model the file must contain enough information to 
%%% identify the current or previous owner of the environment.
%%% As the lock-file is not automatically deleted or adjusted if any OS process crashes, 
%%% when attempting to acquire the lock (and finding the file not empty) we must 
%%% check whether the OS pids that it refers to is still running.  
%%% However, OS pids can be re-used quite quickly, so we need to store other identifying 
%%% information too -- e.g. the start_time and/or a hash of the /proc/cmdline of the pid, 
%%% to see if the pid's identity has changed since it was recorded in the lock file.
%%% 
%%% Possible design for lock-file contents and semantics:
%%%
%%% Normal contents are a tuple:
%%%
%%%   {OwnerErlangVM_OSPid, OwnerEDTK_OSPid, OwnerPortCoordinatorName}.
%%%
%%% OwnerErlangVM_OSPid      = os:getpid() 
%%%                            (this is passed to my_db_env_open as a parameter solely for this purpose)
%%t                           
%%% OwnerEDTK_OSPid          = getpid() in the my_db_env_open call.
%%%                            In linked-in mode this will be the same as OwnerErlangVM_OSPid.
%%%                            In pipe-mode this will be the pid of the 'pipe-main' instance that owns the environment.
%%%
%%% OwnerPortCoordinatorName = The full name (as an atom) of the port-coordinator that owns the environment.
%%%                            (this is passed to my_db_env_open as a parameter solely for this purpose)
%%%
%%% In practice, the 'pid' elements below might be tuples that record the actual OS pid and any other useful 
%%% information that may reveal if the pid has crashed and restarted (e.g. the process' start_time, and perhaps 
%%% a hash of it's /proc/<pid>/cmdline).
%%% The label 'Dead' really means not-running, or running but has materially changed identity (start-time/cmd-line are different).
%%% The label 'Live' really means is running, and identity is still the same.
%%%
%%% Lock-file contents                                 Semantic
%%%
%%% does not exist                                     Environment is not in use.
%%%                                                    ## Take the lock and do run recovery 
%%%                                                    (i.e. don't assume that the previous environment shutdown was clean)
%%%
%%% 'clean_shutdown'                                   Environment is not in use and was cleanly shutdown.
%%% (this state is optional, TBD)                      ## Take the lock and tell the port-coordinator NOT to run recovery.
%%%                                                    (This saves the cost of destroying and re-creating the environment 
%%%                                                    -- at the risk of serious problems if the environment is actually corrupt).
%%%
%%% {DeadVM_OSPid, _AnyDead_OSPid, _AnyPortCoordName}. The previous owner node has crashed and is not using the environment.
%%%                                                    (If _AnyDead_OSPid matches DeadVM_OSPid then it was running in linked in 
%%%                                                    mode, otherwise it was running it pipe-mode, but that doesn't matter here.)
%%%                                                    a restarting instance of the crashed node, but that doesn't matter.)
%%%                                                    ## Take the lock and do run recovery.
%%%                                                     
%%% {Dead_OSPid_1, Live_OSPid, _AnyPortCoordName}.     The previous owner node (which was running in pipe mode) has crashed, 
%%%                                                    and the 'pipe-main' process has not yet finished cleanup (or has hung).
%%%                                                    (If _AnyPortCoordName matches OurPortCoordName then we are _probably_ 
%%%                                                    a restarting instance of the crashed node, but that doesn't matter.)
%%%                                                    ## Return EAGAIN to tell the port-coordinator to wait and 
%%%                                                    ## try again later (up to a limited number of times).
%%%                                                    ## TODO: to resolve hung pipe-main processes we could return the 
%%%                                                    ## Live_OSPid of the pipe-main process, so if the port-coordinator is 
%%%                                                    ## close to giving up it could try killing that process.
%%%
%%% {LiveOtherVM_OSPid, _AnyPid, _AnyPortCoordName}.   Some *other* running Erlang node is currently using this environment.
%%%                                                    Even if _AnyPid was different (i.e. pipe-mode) and dead (showing 
%%%                                                    that the environment is at least for now unused), we specify that 
%%%                                                    applications are designed for only a single node to access an environment,
%%%                                                    and therefore the other node will probably restart its port-coordinator soon.
%%%                                                    ## Leave the lock alone, and fail the call with a non-retryable error.
%%%
%%% {ThisVM_OSPid, _AnyPid, OtherPortCoordName}.       Some *other* port-coordinator in this Erlang node is (or has been)
%%%                                                    currently using this environment.
%%%                                                    Even if _AnyPid was different (i.e. pipe-mode) and dead (showing 
%%%                                                    that the environment is at least for now unused), we specify that 
%%%                                                    environments should be used by a single port-coordinator, and this looks 
%%%                                                    like an attempt by 2 port-coordinators while one of the is temporarily down.
%%%                                                    We choose to not allow the second port-coordinator to step in.
%%%                                                    TODO: One possible legitimate case in which this policy would be deficient:
%%%                                                    if a live-code upgrade were done that changed the naming convention of port 
%%%                                                    coordinators, this policy would prevent the new coordinator from starting 
%%%                                                    until the entire node was restarted.
%%%                                                    ## Leave the lock alone, and fail the call.
%%%
%%% {ThisVM_OSPid, ThisVM_OSPid, OurPortCoordName}.    The current node (running in linked-in mode) owns the environment.
%%%                                                    We must be in SMP mode and some other VM thread is still running 
%%%                                                    bdb_stop on the old environment (or has hung).
%%%                                                    ## Return EAGAIN to tell the port-coordinator to wait and 
%%%                                                    ## try again later (up to a limited number of times).
%%%                                                    To fix a permanent hang here, we would have to shutdown our own VM.
%%%
%%% {ThisVM_OSPid, Live_OSPid, OurPortCoordName}.      The current node (running in pipe mode) owns the environment.
%%%                                                    and some other 'pipe-main' process has not yet finished cleanup (or has hung).
%%%                                                    ## Return EAGAIN to tell the port-coordinator to wait and 
%%%                                                    ## try again later (up to a limited number of times).
%%%                                                    TODO: to fix a hang here, we would have to shutdown this Erlang VM. 
%%%
%%% Clearly that is a LOT of complexity. I currently don't consider the extra complexity to be worth 
%%% the extra visibility (e.g. the lock-file contents could actually be a history of owners -- but that's 
%%% not incredibly useful either).
%%%
%%% So for now I will use the "owning the advisory lock means we own the BDB environment".
%%%
%%% Management of Advisory Lock
%%%
%%% We acquire the lock inside a wrapper around env_open.  It would be easier to do it from Erlang, 
%%% but the Erlang file libraries don't provided the necessary APIs.  And acquiring the lock in 
%%% env_open means that there is no way to avoid the lock by bypassing the port-coordinator; 
%%% the lock is embedded in the code generated/wrapped by EDTK, and there is no other way to open 
%%% an environment.
%%%
%%% We release the lock inside a wrapper around env_close -- and the same wrapper is used 
%%% as the valmap 'stop' function for dbenv by group_valmap_cleanup and bdb_stop, so we don't 
%%% rely on an explicit call to env_close to release the lock.
%%%
%%% Linux offers two advisory-locking APIs, the older flock(), and the posix-compliant (and more powerful) fcntl().
%%% We use flock() as it offers the semantics that we want:
%%%
%%% Pros of flock():
%%%    - locks apply to separate file descriptors in the same process (and therefore to separate threads in the same process) 
%%%      as well as across processes:
%%%       "If a process uses open(2) (or similar) to obtain more than one descriptor for the same file, 
%%%        these descriptors are treated independently by flock(). An attempt to lock the file using one 
%%%        of these file descriptors may be denied by a lock that the calling process has already placed 
%%%        via another descriptor. "
%%% Cons of flock()
%%%    - it doesn't work on NFS, but that's fine as BDB doesn't support NFS either.
%%%    - I found some reports of flock bugs in 2.4 series kernels (which I am using)
%%%      but the bugs only seem to apply if fork() is used.  Ours is a much simpler use-case.
%%%      http://groups.google.com/group/linux.kernel/browse_thread/thread/81769e274a7bff37/fedd8d26e105cea0?lnk=st&q=linux+2.4+flock&rnum=4&hl=en#fedd8d26e105cea0
%%%
%%% In comparison, fcntl() locks have these serious problems
%%%      - the locks only work between processes, NOT between threads within the same proces
%%%      - the locks are automatically released if the process closes ANY file
%%%        descriptor referring to a file on which locks are held.
%%% So to use fcntl we would first need to acquire a *driver-level mutex* (perhaps called 'manipulating_BDB_env_lock_mutex').
%%% In linked-in mode this would _partially_ serialize env_open/env_close of all BDB environments 
%%% running on a given node. This would be OK as we can release the mutex before actually opening the environment 
%%% (as opening the environment might take seconds or even minutes, if recovery is run and/or region-init is used).
%%% (In pipe-mode the mutex wouldn't affect anything because it would be local to a pipe-main process, so is never contended.)
%%%
%%% But for now, flock() seems simpler.
%%%
%%% All processes attempting to access a BDB environment do the following:
%%%
%%% 1. Attempt to open the lock file in the environments 'home' directory, 
%%%    using O_CREAT to create it if necessary.
%%%    (We don't use O_EXCL as it doesn't achieve anything we can use.)
%%%    Multiple processes and threads within the same process may do this concurrently.
%%%
%%%    Store the fd in a new descriptor_t member 'env_ownership_advisory_lock_fd'
%%%
%%% 2. Attempt to obtain an exclusive lock *without* blocking if the lock is already held.
%%%         flock(desc->env_ownership_advisory_lock_fd, LOCK_EX | LOCK_NB)
%%%    (**Remember: this is just an advisory lock; processes are free to ignore the lock 
%%%       and still read or even write the file).
%%%
%%%    If we fail to lock the file (i.e. EWOULDBLOCK), return EWOULDBLOCK to Erlang.
%%%    (the port-coordinator should retry some number of times with a small sleep between attempts).
%%%
%%%    If we successfully lock the file, proceed to open the environment.
%%%
%%% The release mechanism is symmetric:
%%%
%%% 1. Close the BDB environment
%%%
%%% 2. Release the advisory lock with flock(desc->env_ownership_advisory_lock_fd, LOCK_UN)
%%%
%%% 3. close desc->env_ownership_advisory_lock_fd, and set the member to -1.
%%%
%%%
%%% General TODO
%%%
%%% - in kill_all_support_procs_and_clients_and_close_environment,
%%%   recalculate the timout when waiting for group_valmap_cleanup_non_block 
%%%   and db_close_non_block to reply, so that we don't wait for 
%%%   a maximum of N * timeout.
%%%
%%%-------------------------------------------------------------------
-module(berkeley_db_port_coordinator).

-behaviour(gen_server).

%% Basic API (server is registered locally using this module's name)
-export([start_link/1, 
         start_link/2,
         stop/0, 
         acquire_bdb_port_handle/0,
         get_replication_gen_event_proc_name/0]).

%% Variants of the above that also take a process name (as their first argument)
%% which will be registered locally.
%% These are necessary if using multiple BDB environments in the same VM 
%% (each environment has its own own port_coordinator)
-export([start_link/3,
         stop/1, 
         acquire_bdb_port_handle/1,
         get_replication_gen_event_proc_name/1]).

%% APIs that take a bdb_port_handle (the port_coordinator process name is obtained from that handle)
-export([get_db_handle/3, 
         get_db_handle_with_db_config/4,
         clone_bdb_port_handle/1,
         list_all_open_databases/1,
         get_config/1,                     % also accepts a pid/procname (instead of a BdbPortHandle)
         close_unused_database/2,
         start_checkpointer_if_enabled/1,  % if do_parallel_trickle_and_checkpoint is enabled then 
         stop_checkpointer_if_running/1,   % these two functions also affect the cache trickler (because it's the same process)
         release_bdb_port_handle/1]).


%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, code_change/3]).

%% Internal functions exported for use with do_txn
-export([txn_fun_gtm_resolve_distributed_txn/2]).


%%
%% Internal declarations
%%

-define(REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX, "_replication_gen_event").

%% Support processes are checkpointer, cache_trickler, deadlock_detector, replication_gen_event, ...
%% Any that access the BDB environment are spawned in a 'logically suspended' state, and told to start
%% after the environment has been opened.
-record(support_proc, {name, pid, start_fun, stop_fun}).

-record(state, {bdb_port_handle,                          % includes dbenv
                clients_ets_table_id,                     % maps client pids to group_valmap_ids and monitor refs
                db_handle_cache,                          % a dict, mapping filenames to open DB handles 
                next_group_valmap_id,
                support_procs}).                          % list of support_proc records

-define(SERVER_GROUP_VALMAP_ID, 0).  % must be 0

%% In addition to the state record
%%
%% 1. Every port coordinator instance creates a private ETS table to store a hash table
%%    of registered clients:
%%
%%       {ClientPid, MonitorRef, GroupValmapId}
%%
%%    (This is an ETS table and not a dict in #state as it is modified every time a client 
%%    registers and unregisteres -- i.e. twice for every transaction unless client processes 
%%    are pooled and re-used, which they currently are not.  So this can generate a lot 
%%    of garbage at high transaction rates.)
%%
%% 2. we use the process dictionary to cache some **immutable** state 
%%    (created by init).  These are just parts of the initial server config that we need fast access to.
%%
%%    config_proplist
%%    file_permissions
%%    default_db_config_proplist


-include("berkeley_db_drv.hrl").
-define(BDB,      berkeley_db_drv).
-define(BDB_U,    berkeley_db_utilities).
-define(BDB_H,    berkeley_db_helpers).
-define(BDB_RG_H, berkeley_db_replication_group_helpers).
-define(BDB_GTM,  berkeley_db_global_transaction_manager).

-include("berkeley_db_replication_group_helpers.hrl").
-define(CS,       berkeley_db_replication_group_caller_state).

%% This is for inet:gethostname() and inet:gethostbyname() which are used in handle_replication_option_before_env_open/2.
%% -include("inet.hrl").


%%-define(INFO_LOGGER,              error_logger:info_msg).
%%-define(INFO_LOGGER_NO_TIMESTAMP, error_logger:info_msg).
%%-define(WARNING_LOGGER,           error_logger:warning_msg).
%%-define(ERROR_LOGGER,             error_logger:error_msg).
-define(INFO_LOGGER,              ?BDB_U:stdout_logger_with_timestamp).
-define(INFO_LOGGER_NO_TIMESTAMP, ?BDB_U:stdout_logger_no_timestamp).
-define(WARNING_LOGGER,           ?BDB_U:stdout_logger_with_timestamp).
-define(ERROR_LOGGER,             ?BDB_U:stdout_logger_with_timestamp).


%%
%% Config defaults
%%

%% env_dir is required, so there is no default.

%% If no explicit override of 'env_open_flags' is specified then we
%% use a set of 'base' flags (see below), 
%% plus, if replication is enabled, we add DB_INIT_REP.
%%
%% The caller might wish to override this to specify DB_PRIVATE.
%%
%% Note that we always specify DB_RECOVER (any user-override should very likely do the same).
%%
%% There is no need to specify DB_REGISTER as our locking scheme (described earlier)
%% makes this unnecessary and mostly useless (see earlier discussion).

-define(DEFAULT_PORT_START_FUNCTION,  start).  % start or start_pipe

-define(DEFAULT_ENV_OPEN_BASE_FLAGS,  [bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN]).

-define(DEFAULT_CONFIG_PROPLIST, 
        [{port_start_function,                                     ?DEFAULT_PORT_START_FUNCTION},

         %% A key 'spawn_options' is supported but NOT defined by default.
         %% The default is [{fullsweep_after, 0}] in 'pipe' mode (as the port_coordinator receives ALL binary data from BDB),
         %% and            []                     in 'linked-in' mode.
         %% Any specified value will entirely over-ride the default, so it is possible to make 'pipe' mode use [].

         {default_receive_reply_timeout_milliseconds,              ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},
         {group_valmap_cleanup_receive_reply_timeout_milliseconds, 2 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT}, % might have multiple things to do (although not usually)

         {env_cache_size_mb_and_fragments,                         {50, 1}},                

         % (see above for env_open_flags)
         {env_open_receive_reply_timeout_milliseconds,             20000},      % recovery time depends on checkpoint_period, write rate, env_cache_size (as we use REGION_INIT)
         {env_open_num_retries_and_retry_period_milliseconds,      {5, 200}},   % retry if BDB environment is in-use by another port (see discussion of locking earlier)

         {env_set_flags,                                           [bdb_DB_AUTO_COMMIT, 
                                                                    bdb_DB_LOG_AUTOREMOVE, % Note: AUTOREMOVE is NOT SAFE for use with replication -- see comments in berkeley_db_replication_group.erl
                                                                    bdb_DB_REGION_INIT, 
                                                                    bdb_DB_TIME_NOTGRANTED]},

         {cache_trickle_milliseconds_and_percent,                  {1000, 20}}, % or 'disabled' (see also checkpoint_receive_reply_timeout_milliseconds)
         {checkpoint_seconds_and_min_kb,                           {30, 100}},  % or 'disabled' (see also checkpoint_receive_reply_timeout_milliseconds)
         {do_parallel_trickle_and_checkpoint,                      enabled},    % (if enabled, need >= 2 threads in housekeeping)
         {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},   % disabled is same as {0, anything}
         {checkpoint_receive_reply_timeout_milliseconds,           4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},    % also used for memp_trickle

         {auto_deadlock_detect,                                    bdb_DB_LOCK_DEFAULT},                % or 'disabled'
         {deadlock_detector_milliseconds_and_policy,               {5000, 1000, bdb_DB_LOCK_EXPIRE}},   % or 'disabled' (**see edtk-1.2/TODO: currently required to prevent rare threadpool starvation**) period less than default_receive_reply_timeout_milliseconds
         {deadlock_detector_receive_reply_timeout_milliseconds,    ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},
         {default_global_txn_timeout_microseconds,                 3000 * 1000},                        % or 'disabled' (**see edtk-1.2/TODO: currently required to prevent rare threadpool starvation**) less than default_receive_reply_timeout_milliseconds
         {default_global_lock_timeout_microseconds,                disabled},

         {default_db_config_proplist,                              [{db_open_flags,  [bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT]}
                                                                    %% important: if you use {page_size_bytes, N} then also use {set_flags, [bdb_DB_CHKSUM]}
                                                                   ]}, 

         {max_num_concurrent_client_procs,                         2 * ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},  % adjust to prevent saturation of BDB (note: not all client procs start explicit txns)
         {min_limit_concurrent_transactions,                       ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},      % a BDB-enforced *lower* limit (real limit may be up to approx 3x this number)

         {{initial_num_valmap_slots, bdb_cursor},                  ?BDB_INITIAL_NUM_VALMAP_CURSOR_SLOTS},
         {{initial_num_valmap_slots, bdb_db_txn},                  ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS},
         {{initial_num_valmap_slots, bdb_db},                      ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS},

         {max_locks,                                               1000},
         {max_lockers,                                             1000},
         {max_locked_objects,                                      1000},
         %%{max_mutexes,                                           ????},       % OBTAIN FROM db_stat AFTER TESTING (default sizing is poor: SR: 14691) 
         {log_region_size_bytes,                                   65000},      % may need to increase if very many databases are created
         {log_file_size_bytes,                                     10485760},   % 10MB
         {log_buffer_size_bytes,                                   32768},      % increasing may improve performance of long-running or large txns, or highly-concurrent applications

         {replication,                                             disabled},   % see below for sub-options

         {distributed_transactions,                                disabled},   % see below for sub-options

         %% On stack size for Berkeley DB, Keith Bostic (VP Product Development, Sleepycat) says 
         %%
         %%  "We do not document a smallest, or even a "safe", stack size value for any release of Berkeley DB.
         %%   Berkeley DB puts relatively large objects on the stack, on the order of 1 or 2KB 
         %%   (for examples, a routine might have 2 1KB buffers to hold pathnames for a file system rename call).
         %%   Berkeley DB does not go deeply recursive.
         %%   If I had to pick a number out of the air, I'd guess that 8KB would be risky, 16KB would probably work, 
         %%   and 32KB would be safe.  I have no experimental data to prove that, however."
         %%
         %% IMPORTAHT: we now call pcre (regular expressions) from with the BDB errcall callback.
         %% There is also command that compiles the regex specified via coredump_on_errcall_interval_secs_and_regex.
         %%
         %% Also we call Google's coredumper library in the BDB errcall callback.
         %%
         %% The EDTK threadpool implementation uses neglible stack.
         %%
         %% IMPORTANT: currently the valmap_stop threadpool queue MUST have a size higher than will ever be used.
         %% If the limit is reached, the code that handles failure of the enqueue will NOT call the cleanup 
         %% function but WILL do the valmap-slot cleanup.  This will leak BDB handles, and locks.
         %% TODO: implement a "no limit on queue length" feature, and use for valmap_stop threadpool.
         %%%
         {{thread_count_stacksize_qlenlimit, workers},             {10, 32*1024, 1000}},
         {{thread_count_stacksize_qlenlimit, valmap_stop},         {10, 32*1024, (1 bsl 31)}}, % DO NOT CHANGE qlenlimit SEE ABOVE.  Separate note: If replication is enabled, these threads block for acks
         {{thread_count_stacksize_qlenlimit, consume_wait},         {0, 32*1024,    0}},
         {{thread_count_stacksize_qlenlimit, housekeeping},         {2, 32*1024,   20}}, % cache_trickle and txn checkpoint (see do_parallel_trickle_and_checkpoint, above); qlenlimit must be at least 3
         {{thread_count_stacksize_qlenlimit, lock_detect},          {1, 32*1024,   10}},

         {file_permissions,                                        8#644},      % user read/write, group read, world read

         {env_verbose_types,                                       [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER]},  % bdb_DB_VERB_REPLICATION

         {coredump_on_errcall_interval_secs_and_regex,             {300,                 % at most 1 core dump every 5 minutes (may be more if set_write_coredump_on_errcall() is called repeatedly as it resets the timer -- and port_coord init calls it)
                                                                    regex_is_exclusive,  % we *don't* want core-dumps for the following...
                                                                    "./",                % path prefix
                                                                    "(connecting to site)"                                       % repmgr noise
                                                                    "|(EOF on connection)"                                       % repmgr noise
                                                                    "|(socket writing failure)"                                  % repmgr noise
                                                                    "|(can't read from site)"                                    % repmgr noise
                                                                    "|(writing data)"                                            % repmgr noise
                                                                    "|(redundant incoming connection will be ignored)"           % repmgr noise
                                                                    "|(can't bind socket to listening address)"                  % repmgr noise
                                                                    "|(recovery of prepared but not yet committed transactions is incomplete)" % can happen if a read is done on a replica before a new master has resolve all prepared txns after failover
                                                                    "|(DB_UNKNOWN type specified with DB_CREATE)"                % get_db_handle noise
                                                                    "|(multiple databases specified but not supported by file)"  % get_db_handle_with_config noise
                                                                    "|(Prepare disallowed on child transactions)"                % txn_prepare noise
                                                                    "|(Operation locked out.  Waiting for replication lockout to complete)"
                                                                    "|(DB_ENV->get_open_flags: method not permitted before handle's open method)"
                                                                   }},

         {port_log_directory,                                      disabled},   % or a string, to log EDTK/BDB messages to <string>/<port-coordinator-name>.log
         {enable_port_logging_before_open_environment,             []},
         {enable_port_logging_before_close_environment,            [with_timestamps, shutdown, valmap_start, valmap_stop, group_valmap_cleanup_outputv, group_valmap_cleanup_work]},

         {run_txn_funs_on_BDB_event,                               []},         % list of {M, F, FixedArgs, NumDeadlockRetries}. To call: do_txn({M, F, {EventArgs,FixedArgs}}, NumDeadlockRetries)

         %% Some PANIC conditions could be dealt with by closing & re-opening the port, to run BEB recovery,
         %% but others may mean that memory is corrupted.
         %%
         %% In 'linked-in' mode (port_start_function = 'start') we want to restart the Erlang VM if BDB hits a PANIC.
         %% In 'spawned-process pipe' mode (port_start_function = 'start_pipe') we don't need to do this, 
         %% as BDB is running in a different OS process.
         %% Therefore we have 2 lists of funs, tagged with PortStartFunc.
         %%
         %% We use erlang:halt(Reason) instead of init:reboot(), as the latter attempts to cleanly shutdown all processes.
         {run_mode_specific_funs_on_BDB_PANIC, 

          [{start,      
            [fun(_State) -> 
                     Name = case process_info(self(), registered_name) of
                                {registered_name, PN} -> PN;
                                _                     -> unknown
                            end,
                     erlang:halt("BDB PANIC from " ++ atom_to_list(Name)) 
             end]},

           {start_pipe, 
            [fun(#state{clients_ets_table_id = ClientsEtsTable}) -> 
                     %% Hard-kill all of our clients (incase they might be trapping exits), 
                     %% and then return. The port-coordinator will then exit.
                     ets:foldl(fun({ClientPid, _, _}, _) ->
                                       ?ERROR_LOGGER("PANIC: killing client pid ~w~n", [ClientPid]),
                                       exit(ClientPid, kill);
                                  (_, _) -> % curr_num_clients or prev_group_valmap_id
                                       ok 
                               end,
                               [], 
                               ClientsEtsTable)
             end]}
          ]}
        ]).


%% env_set_flags:
%%  - DB handle operations for which no explicit transaction handle was specified, and which modify databases in the database environment, will be automatically enclosed within a transaction. 
%%    (Note: this does NOT apply to cursors; non-transactional cursors must NOT perform write operations in a transactional environment.)
%%  - Log files should be auto-removed asap (only needed for catastrophic recovery, which is too expensive).
%%  - Regions (most significantly, the page-cache) should be page-faulted into memory when created/joined.
%%  - Lock timeouts (e.g. used by berkeley_db_msg_queue.erl, and when using lock_vec to unblock a db_get(CONSUME_WAIT) 
%%    should cause DB_LOCK_NOTGRANTED, not DB_LOCK_DEADLOCK (so we can distinguish between the two)       
%%
%% min_limit_concurrent_transactions : 
%% We might want this to be more than the number of db_txn valmap slots because many operations use auto-commit transactions, 
%% which don't use a valmap slot (but do use an internal transaction).
%% However, BDB typically allows up to 3x this number of transactions; this only sets a lower limit, not a hard upper limit.
%% So the default (#valmap slots) might be fine.
%%
%% default_db_config_proplist:
%% A list containing any of the following (they are all optional):
%%   
%%    [{db_open_flags,                  [bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT, ...]},
%%     {page_size_bytes,                N},
%%     {set_flags,                      Flags},    (should include bdb_DB_CHKSUM if an explicit page-size is chosen)
%%     {set_btree_min_keys_per_page,    MinKey},
%%     {set_fixed_record_size_bytes,    NumBytes},
%%     {set_fixed_record_pad_byte,      ByteValue},
%%     {set_queue_extent_size_db_pages, NumDBPages}]
%%
%% replication :
%% This is disabled by default. 
%% If is enabled by the caller then it has a whole set of possible sub-options.
%% Multiple 'replication' entries may appear in the main config list. 
%% Each 'replication' entry must be a list of sub-options, or the atom 'disabled'.
%% If any replication entry is 'disabled' then any others must also be 'disabled' (i.e. you cannot provide conflicting config).
%% The sub-lists from all of the entries are concatenated (in order) and then treated as a normal proplist.

-define(REPLICATION_REQUIRED_KEYS, [this_hostname_port, 
                                    this_site_priority,
                                    peer_hostnames_ports
                                   ]).

%% If replication is explicitly enabled then these defaults are used for some keys (hence they are not listed in REPLICATION_REQUIRED_KEYS)
-define(DEFAULT_REPLICATION_CONFIG_IF_ENABLED, 
        [{initial_role_and_threadcount, {bdb_DB_REP_ELECTION, 2}},
         {ack_policy,                    bdb_DB_REPMGR_ACKS_QUORUM},
         {config_flags,                                   []},
         {allow_not_durably_replicated_db_open,     disabled},
         {limit_transfer_per_call_bytes,         1024 * 1024},
         {request_retransmit_after_min_max_records, {4, 128}},
         {{timeout, acks_microseconds},               500000}, % 0.5 seconds
         {{timeout, election_milliseconds},             3000},
         {{timeout, election_retry_milliseconds},       4000},
         {{timeout, connection_retry_milliseconds},     3000},
         {{internal_delay, master_checkpoint_seconds},     0}  % default is 30         
        ]).


%% Format for the other options:
%%
%%       {replication, [{this_hostname_port,   {"hostname", port},
%%                      {this_site_priority,   N},
%%                      {peer_hostnames_ports, [{"hostname", port}, ... ],
%%  optional:           {event_handlers,       [{ ModuleName|{ModuleName,Id}, Args}]}     % (see below)
%%                     ]}
%%
%% If replication is enabled then it spawns a gen_fsm (owned by the port coordinator) with a locally-registered 
%% public name, which available via the get_replication_gen_event_proc_name/0,1 API.
%% Any process may add/deleted handlers from this gen_event at any time.
%%
%% To avoid missing events during replication start up, handlers may be registered before repmgr_start is called 
%% by listing them in the optional event_handlers argument.  
%% This is only a convenience, not a requirement.
%% Note that event_handlers is currently processed from the server process, calling gen_event:add_handler, 
%% -- i.e. NOT calling add_sup_handler.  So the process that calls start_link will NOT currently
%% receive any {gen_event_EXIT,Handler,Reason} messages.  I think that's the behaviour we want.
%% If necessary, we could register these from the client process (in start_link) and use add_sup_handler, 
%% so the registering process would get those messages.
%%
%%
%% distributed_transactions :
%% 
%% This is disabled by default.  If enabled then the config works like that for 'replication'
%% -- i.e. multiple lists of sub-options are allowed and there are some defaults.

-define(DISTRIBUTED_TRANSACTIONS_REQUIRED_KEYS, [gtm_rg_caller_state]).

-define(DEFAULT_DISTRIBUTED_TRANSACTIONS_CONFIG_IF_ENABLED, [{gtm_timeout_milliseconds,          5000},
                                                             {resolve_max_num_attempts,            10},
                                                             {resolve_retry_period_milliseconds, 5000}]).

%% Format:
%%
%%       {distributed_transactions, [{gtm_rg_caller_state,      RgCallerState},  % must be added programmatically
%%                                   {gtm_timeout_milliseconds, 5000}
%%                                  ]}
%% NOTE: 
%%   The GTM is a 'weakly replicated application' -- see comments at the top of berkeley_db_global_transaction_manager.erl
%%   This means that a 'caller_state' record (#berkeley_db_replication_group_caller_state{}) is required to talk to it.
%%   Applications should call ?BDB_RG_H:make_initial_caller_state() to create this, and then include it in the 
%%   config proplist. (i.e. this cannot be read from config in a .app or .config file).
%%
%% If distributed_transactions is enabled *and replication is disabled* then init() runs 
%% txn_recover immediately after opening the environment, and if there are any prepared but unresovled 
%% transactions it attempts to ask the GTM whether they should be committed or aborted.
%%
%% If *both* distributed_transactions *and replication are enabled* then port_coordinator does NOT run 
%% txn_recover immediately.  Instead txn_recover is always run by a node when it receives A DB_REP_MASTER
%% event.
%%
%% Note: as of BDB v4.5, it is illegal to closee DB handles if there is a prepared but unresolved transaction
%% -- the log will be left in an unrecoverable state.  Calling txn_discard does not help (and is illegal in this context).
%% So if the port is closed while there are unresolved prepared txns, the driver uses a patch from Sleepycat
%% (patch.14785]) to prevent DB:close from logging the close.  
%% The port_coordinators own clean-shutdown code alos has to obey this 
%% rule - see the call to ?BDB:make_db_close_safe_if_there_are_unresolved_transactions(). 
%% Sleepycat plan to fix this problem in a future release (probably in a different way to the mechanism used 
%% by this patch, so the driver and port-coordinator code will have to change then).


-define(STRINGIZE(Arg), ??Arg).


% Debug helpers (poor-man's tracing)
-define(AT, io:format("~nAt line ~w~n", [?LINE])).   %% remove with regex    ^ *\?AT, *C-qC-j


%%====================================================================
%% API
%%====================================================================

-define(THROW_ERRORS(Expr),
           case Expr of
               {ok, Result}   -> Result;
               {error, Error} -> throw({bdb_port_coordinator_error, Error})
           end).

%%--------------------------------------------------------------------
%% Function: start_link() -> {ok,Pid} | ignore | {error,Error}
%% Description: Starts the server
%%--------------------------------------------------------------------

start_link(EnvDir) ->
    start_link(?MODULE, EnvDir, []).

start_link(EnvDir, ConfigProplist) ->
    start_link(?MODULE, EnvDir, ConfigProplist).

stop() ->
    gen_server:cast(?MODULE, stop).

acquire_bdb_port_handle() ->
    ?THROW_ERRORS( gen_server:call(?MODULE, acquire_bdb_port_handle) ).

%% TODO: An alternative would be to store dynamic_replication_gen_event_proc_name() in the process dictionary 
%% (during init) and fetch it with a gen_server:call()
get_replication_gen_event_proc_name() ->
    list_to_atom(?STRINGIZE(?MODULE) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).


%% These variants of the above are used when we want to use multiple port_coordinators 
%% (IMPORTANT: each port_coordiantor MUST use a DIFFERENT BDB ENV.
%% TODO: we could use a global ETS table to track which port coord is using 
%% which (absolute path) EnvDir.  However, most systems will only use one, so 
%% the extra complexity is not worth it (yet).

start_link(ServerRegisterName, EnvDir, ConfigProplist) ->
    SpawnOptions = 
        case lists:keysearch(spawn_options, 1, ConfigProplist) of
            {value, {spawn_options, Value}} -> 
                Value;
            false -> 
                %% No explicit spawn_options defined.
                %% In 'pipe' mode it is probably important that we at least use {fullsweep_after, 0}
                %% as the port_coordinator will receive ALL results from the BDB driver.
                case proplists:get_value(port_start_function, ConfigProplist, ?DEFAULT_PORT_START_FUNCTION) of
                    start      -> [];
                    start_pipe -> [{fullsweep_after, 0}]
                end
        end,
    gen_server:start_link({local, ServerRegisterName}, ?MODULE, [{env_dir, EnvDir} | ConfigProplist], [{spawn_opt, SpawnOptions}]).

stop(ServerRegisterName) ->
    gen_server:cast(ServerRegisterName, stop).

acquire_bdb_port_handle(ServerRegisterName) ->
    ?THROW_ERRORS( gen_server:call(ServerRegisterName, acquire_bdb_port_handle) ).

get_replication_gen_event_proc_name(ServerRegisterName) ->
    list_to_atom(atom_to_list(ServerRegisterName) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).


%%
%% APIs that require a BdbPortHandle
%%

get_db_handle(#bdb_port_handle{} = BdbPortHandle, DBType, DBName) ->
    get_db_handle_with_db_config(BdbPortHandle, DBType, DBName, []).

%% DBConfigProplist is for configuring the DB handle before db_open is called
%% (The default option list is simply empty.)
%%
%%    {set_fixed_record_size_bytes,      NumBytes}
%%    {set_fixed_record_pad_byte,       ByteValue}
%%    {set_queue_extent_size_db_pages,  NumDBPages}
%%

%% Convenience function for when caller only provides Filename (no DatabaseNameWithinFile)
get_db_handle_with_db_config(#bdb_port_handle{} = BdbPortHandle, DBType, DBFilename, DBConfigProplist) when is_list(DBFilename) ->
    get_db_handle_with_db_config(BdbPortHandle, DBType, {DBFilename, ""}, DBConfigProplist);

get_db_handle_with_db_config(#bdb_port_handle{port_coordinator_pid  = PortCoordPid,
                                              receive_reply_timeout = ReceiveReplyTimeout}, 
                             DBType, 
                             {DBFilename, DBName} = DBId, 
                             DBConfigProplist) 
  when is_list(DBFilename), is_list(DBName) ->

    %% As the server might need to call db_open, we use ReceiveReplyTimeout for the gen_server timeout.
    %% TODO: should probably add an optional override Timeout parameter
    ?THROW_ERRORS( gen_server:call(PortCoordPid, 
                                   {get_db_handle, {DBType, DBId, DBConfigProplist}},
                                   ReceiveReplyTimeout) ).


list_all_open_databases(#bdb_port_handle{port_coordinator_pid = PortCoordPid}) ->
    ?THROW_ERRORS( gen_server:call(PortCoordPid, list_all_open_databases) ).


clone_bdb_port_handle(#bdb_port_handle{} = BdbPort) ->
    BdbPort#bdb_port_handle{issued_to_pid = self()}.


%% **EXTREMELY DANGEROUS API** 
%% See documentation for stringent preconditions
close_unused_database(#bdb_port_handle{} = BdbPortHandle, DBFilename) when is_list(DBFilename) ->
    close_unused_database(BdbPortHandle, {DBFilename, ""});

close_unused_database(#bdb_port_handle{port_coordinator_pid = PortCoordPid}, 
                      {DBFilename, DBName} = DBId) 
  when is_list(DBFilename), is_list(DBName) ->

    ?THROW_ERRORS( gen_server:call(PortCoordPid, {close_unused_database, DBId}) ).


get_config(#bdb_port_handle{port_coordinator_pid = PortCoordPid}) ->
    ?THROW_ERRORS( gen_server:call(PortCoordPid, get_config) );

get_config(PortCoordinator) ->
    ?THROW_ERRORS( gen_server:call(PortCoordinator, get_config) ).


start_checkpointer_if_enabled(#bdb_port_handle{port_coordinator_pid = PortCoordPid}) ->
    ?THROW_ERRORS( gen_server:call(PortCoordPid, start_checkpointer_if_enabled) ).


stop_checkpointer_if_running(#bdb_port_handle{port_coordinator_pid = PortCoordPid}) ->
    ?THROW_ERRORS( gen_server:call(PortCoordPid, stop_checkpointer_if_running) ).


release_bdb_port_handle(
  #bdb_port_handle{group_valmap_id                            = GroupValmapId,
                   group_valmap_cleanup_receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup,
                   port_coordinator_pid                       = PortCoordPid} = BdbPortHandle) ->

    %% We need to call group_valmap_cleanup and we want it to be a
    %% serious error (throw) if the caller still owns any resources.
    %%
    %% Therefore the call to group_valmap_cleanup must be blocking, as
    %% it needs the return value.
    %%
    %% We don't want to block the port_coordinator server process for
    %% this call, as group_valmap_cleanup might take considerable
    %% time.
    %%
    %% Currently, rather than spawning a temporary process on the
    %% server to do it (and using gen_server:reply) it seems
    %% reasonable and fair to make the client call it.
    %%
    %% [Note As release_bdb_port_handle is often called if
    %% group_valmap_cleanup has real work to do (e.g. ?BDB_H:do_txn
    %% uses it that way), then we _do_ want to the possibly larger
    %% group_valmap_cleanup_receive_reply_timeout here.  As we are
    %% running on the client, we arrange for the server to put the
    %% value in BdbPortHandle, just for us to use.]
    %%
    %% Important detail: If this fails then the client will crash and
    %% the the port-coordinator will attempt to call
    %% group_valmap_cleanup again.  That second attempt will either
    %% discover that there is nothing to do (the first call completed)
    %% or that the first call is still in progress (so the failure on
    %% the client-side must have been a receive_reply_timeout) The
    %% server code handle_info('EXIT') and close_environment has code
    %% to deal with that -- see comments in those functions.
    %%
    %% IMPORTANT NOTE: if group_valmap_cleanup fails here (e.g. a receive_reply_timeout)
    %% then we still need to ensure that the port-coordinator de-registeres tne client.
    %% Which means that 
    %%  i. Nothing (not even ?BDB_H:do_txn et al) must eat the EXIT signal
    %% ii. The port_coordinator must ensure that group_valmap_cleanup 
    %%     completes before de-registering.
    %%

    BdbPortHandleForGroupValmapCleanup = BdbPortHandle#bdb_port_handle{receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup},

    NumResourcesCleanedUp = ?BDB:group_valmap_cleanup(BdbPortHandleForGroupValmapCleanup, GroupValmapId),

    %% We *always* tell the server to forget about this client (as cleanup has been done).
    %% We don't wrap this in a ?THROW_ERRORS() invocation because if NumResourcesCleanedUp is not 0 then 
    %% that is a more important error.

    %% This is a call as the client might want to terminate immediately after release_bdb_port_handle returns
    %%(so if we used a cast the EXIT signal would race the cast).
    ServerResult = gen_server:call(PortCoordPid, release_bdb_port_handle),

    %% Now throw if the client leaked any resources, or (if it didn't leak resources), if the server command returned an error.
    case NumResourcesCleanedUp of
        0 -> ?THROW_ERRORS(ServerResult);
        _ -> throw({bdb_port_coordinator_error, {release_bdb_port_handle, BdbPortHandle, still_owned_resources, NumResourcesCleanedUp}})            
    end.



%%====================================================================
%% gen_server callbacks
%%====================================================================

%%--------------------------------------------------------------------
%% Function: init(Args) -> {ok, State} |
%%                         {ok, State, Timeout} |
%%                         ignore               |
%%                         {stop, Reason}
%% Description: Initiates the server
%%
%% Args is a 'proplist' of tagged tuples
%%
%% Required
%%    {env_dir, "directory path"}                          
%%
%% Optional
%%    See defaults in comments at start.
%%
%%--------------------------------------------------------------------
init(SpecifiedConfigProplist) ->

    ?INFO_LOGGER("starting (~w)~n", [self()]),

    %% We only trap exits for clearer reporting if a support process fails.
    %% (We use erlang:monitor to track client processes.)
    process_flag(trap_exit, true),

    %% This is the only required arg (no default)
    EnvDir = case proplists:get_value(env_dir, SpecifiedConfigProplist) of 
                 undefined  -> throw({missing_required_config_option, env_dir});
                 Value      -> Value
             end,

    %% Create the ETS process to track registered clients and next_group_valmap_id

    %% ... we don't really need this to be a named table
    %%{registered_name, PortCoordinatorProcessName} = process_info(self(), registered_name),
    %%ClientsEtsTableName = list_to_atom(atom_to_list(PortCoordinatorProcessName) ++ "_clients"),

    ClientsEtsTableId   = ets:new(none, [set, protected, {keypos,1}]),  

    %% We explicitly trawck #clients, rather than use ets:info() to get the #keys, as the 
    %% latter returns a lot of junk and we don't want the garbage.
    true = ets:insert(ClientsEtsTableId, {curr_num_clients, 0}),

    %% We store the prev_group_valmap_id in ClientsETS table, as ETS provides the update_counter 
    %% API which does exactly what we need to adjust it.
    %% This means that neither acquire_bdb_port_handle and release_bdb_port_handle modify #state,
    %% and so garbage generation/collection is minimal for port-coordinator processes
    %% (**unless pipe mode is used, in which case every BDB reply message passes through 
    %% the port-coordinator as lists of bytes**).
    true = ets:insert(ClientsEtsTableId, {prev_group_valmap_id, 0}),

    %% Some options (replication and distributed_transactions) have lists of sub-options, and we want to treat
    %% multiple occurences of those parameters as cummulative (comcatenated) proplists in their own right.  
    %% So preprocess those first, otherwise get_dominant_proplist_values will discard all but the first occurance.
    %% Also, we cannot have a 'disabled' parameter and a sublist.  Either all occurences are 'disabled', or none of them are.
    PropList2 = ?BDB_U:normalize_sub_proplists_for_key(
                   replication, 
                   SpecifiedConfigProplist,
                   ?DEFAULT_REPLICATION_CONFIG_IF_ENABLED),

    PropList3 = ?BDB_U:normalize_sub_proplists_for_key(
                   distributed_transactions, 
                   PropList2,
                   ?DEFAULT_DISTRIBUTED_TRANSACTIONS_CONFIG_IF_ENABLED),

    %% Produce the final configuration 
    %% (append the default config to the list specified by the caller, and extract all of the first occurences (the final config).
    ConfigProplist = ?BDB_U:get_dominant_proplist_values(
                        [PropList3, % may contain an override of env_open_flags
                         [{env_open_flags, case proplists:get_value(replication, SpecifiedConfigProplist) of
                                               undefined -> ?DEFAULT_ENV_OPEN_BASE_FLAGS;
                                               disabled  -> ?DEFAULT_ENV_OPEN_BASE_FLAGS;
                                               _         -> [bdb_DB_INIT_REP] ++ ?DEFAULT_ENV_OPEN_BASE_FLAGS
                                           end}],
                         ?DEFAULT_CONFIG_PROPLIST]),

    %% Store the final config in the process dictionary, so that handle_init_option_before_env_open 
    %% can use it if necessary, and so that it will be logged if the server crashes.
    %% (Also the main code may check this if needed, e.g. to see if we are in linked-in or pipe-mode.
    %% If all commands start timing-out (deadlock) and we are in pipe-mode we can simply kill and 
    %% restart the server.  But in linked-in mode we will need to restart the entire Erlang VM.)
    put(config_proplist, ConfigProplist),

    %% Display the final config.
    lists:foreach(fun({Key, Value}) -> ?INFO_LOGGER_NO_TIMESTAMP("init: ~55w = ~200p~n", [Key, Value]) end, lists:keysort(1, ConfigProplist)),

    %% Actually open the BDB port instance
    PortStartFunc = get_value_for_key(port_start_function, ConfigProplist),

    DefaultBdbPortHandle 
        = #bdb_port_handle{erlang_port     = ErlangPort,
                           group_valmap_id = ?SERVER_GROUP_VALMAP_ID}
        = ?BDB:PortStartFunc(),

    %% Add the port_coordinator details, and possibly override receive_reply_timeout
    BdbPortHandle = DefaultBdbPortHandle#bdb_port_handle{
                      port_coordinator_name                      = get_port_coordinator_proc_name(),
                      port_coordinator_pid                       = self(),
                      receive_reply_timeout                      = get_value_for_key(default_receive_reply_timeout_milliseconds, 
                                                                                     ConfigProplist),
                      group_valmap_cleanup_receive_reply_timeout = get_value_for_key(group_valmap_cleanup_receive_reply_timeout_milliseconds, 
                                                                                     ConfigProplist)}, 
    
    ProcNameStr = atom_to_list(get_port_coordinator_proc_name()),
    ?BDB:set_port_label(BdbPortHandle, erlang:port_to_list(ErlangPort) ++ " " ++ ProcNameStr),

    {VersionMajor, VersionMinor, VersionPatch, VersionText} = ?BDB:db_version(BdbPortHandle),
    DriverSidePortNumber = ?BDB:get_port_number(BdbPortHandle),
    ?INFO_LOGGER("init: Successfully opened port ~w (on driver side: ~w). Berkeley DB version = ~w.~w.~w ~p~n", 
                 [ErlangPort, DriverSidePortNumber, VersionMajor, VersionMinor, VersionPatch, binary_to_list(VersionText)]),

    ?BDB:adjust_port_logging(BdbPortHandle, 
                             enable, 
                             get_value_for_key(enable_port_logging_before_open_environment, ConfigProplist)),

    case get_value_for_key(port_log_directory, ConfigProplist) of
        disabled -> 
            ok;  % default is to log to stderr
        Directory ->
            {A,B,C} = now(),
            LogFilename = Directory ++ "/" ++ ProcNameStr ++ "." ++ lists:flatten(io_lib:format("~6..0w~6..0w~6..0w", [A, B, C])),
            ?INFO_LOGGER("init: Directing EDTK/BDB logging to file ~p~n", [LogFilename]),
            ?BDB:direct_port_logging_to_file(BdbPortHandle, LogFilename),
            %% as the above code creates a unique file per port instance (and the filename specifies the port)
            %% we can make the file a lot shorter by not logging port-label on every line
            ?BDB:set_port_label(BdbPortHandle, "")
    end,

    %% Validate and apply the config options before openening the BDB environment.
    %% The config handler clauses have the option of adjusting the server State record.
    %% They can also move or add options to a list of AfterEnvOpenConfigOptions.

    {State1, AfterEnvOpenConfigOptions}
       = lists:foldl(fun(ConfigArg, State) -> handle_init_option_before_env_open(ConfigArg, State) end, 
                      {#state{bdb_port_handle         = BdbPortHandle, 
                              clients_ets_table_id    = ClientsEtsTableId,
                              db_handle_cache         = dict:new(), 
                              support_procs           = []}, 
                       []}, % AfterEnvOpenConfigOptions
                      ConfigProplist),

    ?INFO_LOGGER("init: Opening environment...~n"),
    {NumRetryAttempts, RetryPeriodMs} = get_value_for_key(env_open_num_retries_and_retry_period_milliseconds, ConfigProplist),

    env_open_with_retry(BdbPortHandle#bdb_port_handle{receive_reply_timeout 
                                                       = get_value_for_key(env_open_receive_reply_timeout_milliseconds, ConfigProplist)},
                        EnvDir, 
                        get_value_for_key(env_open_flags, ConfigProplist), 
                        get(file_permissions),
                        NumRetryAttempts + 1,  % MaxNumAttempts
                        RetryPeriodMs,
                        1),                    % AttemptNumber
    ?INFO_LOGGER("init: environment opened successfully.~n"),

    %% If distributed transactions are enabled then we may need to resolve any 
    %% previously prepared transactions before BDB will allow any new transactions.
    %% [ Arguably, we ought to call do_distributed_transaction_recovery EVERY time, 
    %% as a site may have been configured for distributed transactions before and left 
    %% unresolve txns, even though it is not configured for distributed txns now.
    %% However,
    %%  - this is both unlikely and a serious operator error (not a normal use-case)
    %%  - we cannot resolve any txns without config pointing to the global_transaction_manager
    %%  - it is safe, as BDB will reject txn_begin until all unresolved txns have been resolved.]
    %%
    %% We do this even if replication is enabled. (We cannot call repmgr_start if there 
    %% are unresolved transactions).

    case get_value_for_key(distributed_transactions, ConfigProplist) of

        disabled ->
            ok;

        DistributedTxnConfigList ->
            ?INFO_LOGGER("init: Calling do_distributed_transaction_recovery~n"),
            do_distributed_transaction_recovery(BdbPortHandle, bdb_DB_FIRST, DistributedTxnConfigList),
            ?INFO_LOGGER("init: do_distributed_transaction_recovery finished.~n")
    end,

    %% Now do all config that requires the environment to be open
    State2 = lists:foldl(fun(ConfigArg, State) -> handle_init_option_after_env_open(ConfigArg, State) end, 
                         State1,
                         AfterEnvOpenConfigOptions),

    %% Now start the optional cache_trickler/checkpointer support processes.
    %% (Other support_procs are started in handle_init_option_before/after_env_open
    %% but these two are a little entangled, so we do them together here
    %% for clarity.)
    %%
    %% Note: if replication is enabled then it is very likely that the caller 
    %% wants the checkpointer to only run on the master. To achieve this, 
    %% ?BDB_RG uses run_txn_funs_on_BDB_event to call start_checkpointer_if_enabled 
    %% and stop_checkpointer_if_running (which are only public APIs for that reason).
    %% However, we don't want to anticipate whether that will be done, so we still start all 
    %% of the processese here.  The first DB_REP_EVENT_CLIENT message will 
    %% presumably trigger a txn-fun that stops just the checkpointer.
    %% Also, stop_checkpointer_if_running might actually just tell the relevant support_proc 
    %% to suspend checkpoints (and e.g. continue doing memp_trickle, 
    %% if do_parallel_trickle_and_checkpoint is disabled so that a single process is doing both).
    StateOut 
        = #state{support_procs = SupportProcs} 
        = init_checkpointer_and_or_cache_trickler(State2),

    %% Now allow our support processes to begin work.
    %% (Some of these might have been created before the environment was opened, 
    %% so they are waiting for a message telling them to start).
    lists:foreach(fun(#support_proc{start_fun = StartFun} = SupportProc) ->
                     StartFun(SupportProc) 
                  end, 
                  SupportProcs),

    {ok, StateOut}.


env_open_with_retry(Port, Dir, Flags, Mode, MaxNumAttempts, RetryPeriodMs, AttemptNumber) ->
    try ?BDB:env_open(Port, Dir, Flags, Mode)
    catch
        {bdb_error, eagain} ->
            if AttemptNumber < MaxNumAttempts ->
                    io:format("init: env_open returned EAGAIN (environment locked) on attempt ~w of ~w: retrying in ~w milliseconds", [AttemptNumber, MaxNumAttempts, RetryPeriodMs]),
                    timer:sleep(RetryPeriodMs),
                    env_open_with_retry(Port, Dir, Flags, Mode, MaxNumAttempts, RetryPeriodMs, AttemptNumber + 1);
               true ->
                    exit(bdb_environment_is_locked)
            end
    end.


%%
%% Handlers for config options
%%

handle_init_option_before_env_open({port_start_function, _}, StateAndPostponedConfigOptions) ->
    %% Handled by init
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({spawn_options, _}, StateAndPostponedConfigOptions) ->
    %% Handled by init
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_dir, _}, StateAndPostponedConfigOptions) ->
    %% Handled by init
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({default_receive_reply_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({group_valmap_cleanup_receive_reply_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_cache_size_mb_and_fragments, {CacheSizeMB, Fragments}}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) ->
    ?BDB:env_set_cachesize(BdbPortHandle, 
                           CacheSizeMB div 1024,                 % gigabytes
                           (CacheSizeMB rem 1024) * 1024 * 1024, % bytes
                           Fragments),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_open_receive_reply_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_open_num_retries_and_retry_period_milliseconds, {NumRetries, PeriodMs}}, StateAndPostponedConfigOptions) 
  when is_integer(NumRetries), NumRetries >= 0, is_integer(PeriodMs), PeriodMs >= 0 ->
    %% Handled by init()
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_open_flags, _}, StateAndPostponedConfigOptions) ->
    %% Handled by env_open
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_set_flags, Flags}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_list(Flags) ->
    ?BDB:env_set_flags(BdbPortHandle, Flags, 1),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({cache_trickle_milliseconds_and_percent, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({cache_trickle_milliseconds_and_percent, {PeriodMilliseconds, CleanPercent}}, 
                                   StateAndPostponedConfigOptions) when is_integer(PeriodMilliseconds), PeriodMilliseconds >= 0, is_integer(CleanPercent), CleanPercent > 0, CleanPercent =< 100 ->
    %% We delay starting this as it may be entangled with the checkpointer.
    %% So we do them both (choosing to waiting until after opening the environment, 
    %% support_procs do have a mechanism to delay starting work).
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({checkpoint_seconds_and_min_kb, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({checkpoint_seconds_and_min_kb, {PeriodSeconds, MinKBytes}}, 
                                   StateAndPostponedConfigOptions) when is_integer(PeriodSeconds), PeriodSeconds >= 0, is_integer(MinKBytes), MinKBytes >= 0 ->
    %% See comments above for cache_trickle_milliseconds_and_percent
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({do_parallel_trickle_and_checkpoint, enabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({do_parallel_trickle_and_checkpoint, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({cache_throttle_io_write_ops_and_sleep_microseconds, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({cache_throttle_io_write_ops_and_sleep_microseconds, {SequentialWriteOps, SleepMicroseconds}} = Option, 
                                   {State, PostponedConfigOptions})
  when is_integer(SequentialWriteOps), SequentialWriteOps >= 0, is_integer(SleepMicroseconds), SleepMicroseconds >= 0 ->
    %% In BDB v4.5, mp_max_write affects the checkpoint done at the end of recovery inside env_open.
    %% So for (much) faster startup we postpone this until after env_open returns.
    {State, [Option | PostponedConfigOptions]};


handle_init_option_before_env_open({checkpoint_receive_reply_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    %% Handled when we start the checkpointer process
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({auto_deadlock_detect, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({auto_deadlock_detect, Policy}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_atom(Policy) ->
    ?BDB:set_lk_detect(BdbPortHandle, Policy),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({deadlock_detector_milliseconds_and_policy, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({deadlock_detector_milliseconds_and_policy, {InitialDelayMilliseconds, PeriodMilliseconds, Policy}}, 
                                   {#state{bdb_port_handle = BdbPortHandle,
                                           support_procs   = SupportProcs} = State, PostponedConfigOptions}) 
  when is_integer(InitialDelayMilliseconds), InitialDelayMilliseconds >= 0, is_integer(PeriodMilliseconds), PeriodMilliseconds >= 0, is_atom(Policy) ->

    ConfigProplist = get(config_proplist),
    DeadlockDetectorReceiveReplyTimeout = get_value_for_key(deadlock_detector_receive_reply_timeout_milliseconds, ConfigProplist),
    NewSupportProc =
        #support_proc{name      = deadlock_detector, 
                      pid       = proc_lib:spawn_link(
                                    fun() -> 
                                            deadlock_detector_process(
                                              BdbPortHandle,
                                              InitialDelayMilliseconds,
                                              PeriodMilliseconds, 
                                              Policy,
                                              DeadlockDetectorReceiveReplyTimeout) 
                                    end),
                      start_fun = fun(#support_proc{pid = Pid}) -> Pid ! go end,
                      stop_fun  = fun(#support_proc{pid = Pid}) -> 
                                          unlink(Pid),    % we *don't* want the EXIT signal
                                          exit(Pid, kill) % safe to kill this while lock_detect is still running (and we can close DBs etc in parallel)
                                  end},
    {State#state{support_procs = [NewSupportProc | SupportProcs]},
     PostponedConfigOptions};

handle_init_option_before_env_open({deadlock_detector_receive_reply_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    %% Handled when we start the deadlock-detector process
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_global_lock_timeout_microseconds, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({default_global_lock_timeout_microseconds, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    ?BDB:env_set_timeout(BdbPortHandle, N, bdb_DB_SET_LOCK_TIMEOUT),  
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({default_global_txn_timeout_microseconds, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({default_global_txn_timeout_microseconds, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    ?BDB:env_set_timeout(BdbPortHandle, N, bdb_DB_SET_TXN_TIMEOUT),  
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({default_db_config_proplist, Proplist}, StateAndPostponedConfigOptions) when is_list(Proplist) ->
    ?BDB_U:throw_if_not_proplist(Proplist),        % Check well-formedness
    put(default_db_config_proplist, Proplist),     % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({max_num_concurrent_client_procs, N}, StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    put(max_num_concurrent_client_procs, N),       % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({min_limit_concurrent_transactions, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    %% This can only be set before opening the environment.  
    %% The actual limit is likely to be 2 to 3 times the number set here - hence the 'minimum' tag.
    ?BDB:set_num_valmap_slots(BdbPortHandle, bdb_db_txn, N),
    ?BDB:env_set_tx_max(BdbPortHandle, N),    
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({{initial_num_valmap_slots, ValmapType}, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_num_valmap_slots(BdbPortHandle, ValmapType, N),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({max_locks, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lk_max_locks(BdbPortHandle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_lockers, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lk_max_lockers(BdbPortHandle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_locked_objects, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lk_max_objects(BdbPortHandle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({max_mutexes, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:mutex_set_max(BdbPortHandle, N),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({log_region_size_bytes, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lg_regionmax(BdbPortHandle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({log_file_size_bytes, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lg_max(BdbPortHandle, N),
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({log_buffer_size_bytes, N}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) when is_integer(N), N >= 0 ->
    ?BDB:set_lg_bsize(BdbPortHandle, N),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({replication, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({replication, ReplicationConfigList}, 
                                   {#state{support_procs = SupportProcs} = State, PostponedConfigOptions}) ->

    %% Check that all required parameters are present
    %% (i.e. those that don't have defaults)
    lists:foreach(fun(RequiredKey) -> case proplists:lookup(RequiredKey, ReplicationConfigList) of 
                                          {RequiredKey, _} -> ok;
                                          none             -> throw({missing_required_replication_config, RequiredKey, ReplicationConfigList})
                                      end
                  end,
                  ?REPLICATION_REQUIRED_KEYS),

    %% Clients will probably want to be notified of replication events, so spawn a gen_fsm with a well-known name.
    %% We CANNOT call get_replication_gen_event_proc_name() here as we don't know which variant to call 
    %% (i.e. this port_coordinator might not be registered as ?MODULE). We need to get our dynamic process name.
    GenEventProcessName = dynamic_replication_gen_event_proc_name(),
    {ok, GenEventPid} = gen_event:start_link({local, GenEventProcessName}),
    ?INFO_LOGGER("Started ~w (~w)~n", [GenEventProcessName, GenEventPid]),

    NewSupportProc =
        #support_proc{name      = replication_gen_event, 
                      pid       = GenEventPid,
                      start_fun = fun(_) -> ok end,
                      stop_fun  = fun(#support_proc{pid = Pid}) -> gen_event:stop(Pid) end},

    %% Replication has a bunch of sub-options -- process them.
    lists:foldl(fun(ConfigArg, StateAndPostponedConfigOptions) -> 
                        handle_replication_option_before_env_open(ConfigArg, StateAndPostponedConfigOptions) 
                end, 
                {State#state{support_procs = [NewSupportProc | SupportProcs]}, PostponedConfigOptions},
                ReplicationConfigList);


handle_init_option_before_env_open({distributed_transactions, disabled}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;
handle_init_option_before_env_open({distributed_transactions, DistributedTxnConfigList}, 
                                   {#state{bdb_port_handle = BdbPortHandle,
                                           support_procs   = SupportProcs} = State, PostponedConfigOptions}) ->

    %% Check that all required parameters are present
    %% (i.e. those that don't have defaults)
    lists:foreach(fun(RequiredKey) -> case proplists:lookup(RequiredKey, DistributedTxnConfigList) of 
                                          {RequiredKey, _} -> ok;
                                          none             -> throw({missing_required_distributed_transactions_config, RequiredKey, DistributedTxnConfigList})
                                      end
                  end,
                  ?DISTRIBUTED_TRANSACTIONS_REQUIRED_KEYS),

    %% Distributed transaction support has a bunch of sub-options -- process them.
    %%
    %% Also, we need a support process to resolve any prepared txns that we discover when cleaning up after a client 
    %% (group_valmap_cleanup sends us an untagged event for these).

    %% Note that the stop function for this support process attempts to 
    %% do a graceful shutdown, rather than a hard exit(kill).
    %% see close_environment for why this is important -- ideally we want to 
    %% avoid leaving a txn_abort/commit in progress if possible, as that 
    %% means close_environment cannot close any DB handles, and must 
    %% leave significant work to bdb_stop (which blocks an Erlang VM scheduler thread).
    %% Note that if we don't get a response quickly then we do hard-kill it
    %% (i.e. the above is an optimization, not essential).

    NewSupportProc =
        #support_proc{name      = incrementally_resolve_prepared_txns, 
                      pid       = proc_lib:spawn_link(
                                    fun() -> incrementally_resolve_prepared_txns_process(
                                               BdbPortHandle,
                                               DistributedTxnConfigList) 
                                    end),
                      start_fun = fun(#support_proc{pid = Pid}) -> Pid ! go end,
                      stop_fun  = fun(#support_proc{pid = Pid}) ->
                                          Pid ! {self(), stop},
                                          Timeout = 1000,
                                          receive 
                                              {Pid, stopped}   -> 
                                                  ?INFO_LOGGER("incrementally_resolve_prepared_txns_process ~w has exited gracefully~n", [Pid]),
                                                  ok;
                                              {'EXIT', Pid, Reason} -> 
                                                  ?WARNING_LOGGER("incrementally_resolve_prepared_txns_process ~w has exit with reason ~p~n", [Pid, Reason]),
                                                  ok
                                          after 
                                              Timeout ->
                                                  ?WARNING_LOGGER("hard-aborting incrementally_resolve_prepared_txns_process ~w as it did not gracefully exit within ~w ms~n", [Pid, Timeout]),
                                                  exit(Pid, kill)
                                          end
                                  end},

    lists:foldl(fun(ConfigArg, AccStateAndPostponedConfigOptions) -> 
                        handle_distributed_transactions_option_before_env_open(ConfigArg, AccStateAndPostponedConfigOptions) 
                end, 
                {State#state{support_procs = [NewSupportProc | SupportProcs]},
                 PostponedConfigOptions},
                DistributedTxnConfigList);


handle_init_option_before_env_open({run_txn_funs_on_BDB_event, EventsTxnFunSpecs}, 
                                   StateAndPostponedConfigOptions) when is_list(EventsTxnFunSpecs) ->

    %% Note: BDB_PANIC is not supported (see run_mode_specific_funs_on_BDB_PANIC)
    lists:foreach(fun({bdb_DB_EVENT_REP_MASTER, TxnFunSpecs})      when is_list(TxnFunSpecs) -> is_txn_fun_spec_list(TxnFunSpecs);
                     ({bdb_DB_EVENT_REP_CLIENT, TxnFunSpecs})      when is_list(TxnFunSpecs) -> is_txn_fun_spec_list(TxnFunSpecs);
                     ({bdb_DB_EVENT_REP_NEWMASTER, TxnFunSpecs})   when is_list(TxnFunSpecs) -> is_txn_fun_spec_list(TxnFunSpecs);
                     ({bdb_DB_EVENT_REP_STARTUPDONE, TxnFunSpecs}) when is_list(TxnFunSpecs) -> is_txn_fun_spec_list(TxnFunSpecs);
                     ({bdb_DB_EVENT_WRITE_FAILED, TxnFunSpecs})    when is_list(TxnFunSpecs) -> is_txn_fun_spec_list(TxnFunSpecs)
                  end,
                  EventsTxnFunSpecs),
    put(run_txn_funs_on_BDB_event, EventsTxnFunSpecs),  % don't do get(config_proplist) for every event, as a we might get a lot of PERM_FAILED events
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({run_mode_specific_funs_on_BDB_PANIC, Funs}, 
                                   StateAndPostponedConfigOptions) when is_list(Funs) ->
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({{thread_count_stacksize_qlenlimit, ThreadpoolName}, {NumThreads, StackSizeBytes, QLenLimit}}, 
                                   {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(NumThreads), NumThreads >= 0, is_integer(StackSizeBytes), StackSizeBytes >= 0,is_integer(QLenLimit), QLenLimit >= 0 ->
    ThreadpoolId = case ThreadpoolName of
                       workers      -> ?BDB_THREADPOOL_ID_WORKERS;
                       valmap_stop  -> ?BDB_THREADPOOL_ID_VALMAP_STOP;
                       consume_wait -> ?BDB_THREADPOOL_ID_CONSUME_WAIT;
                       housekeeping -> ?BDB_THREADPOOL_ID_HOUSEKEEPING;
                       lock_detect  -> ?BDB_THREADPOOL_ID_LOCK_DETECT
                   end,
    ?BDB:set_threadpool_params(BdbPortHandle, ThreadpoolId, NumThreads, StackSizeBytes, QLenLimit),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({file_permissions, FilePermissions}, StateAndPostponedConfigOptions) when is_integer(FilePermissions) ->
    put(file_permissions, FilePermissions),        % Store for later use (an immutable cache)
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({env_verbose_types, VerboseTypeList}, {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) ->
    lists:foreach(fun(VerboseType) -> ?BDB:env_set_verbose(BdbPortHandle, VerboseType, 1) end,
                  VerboseTypeList),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({coredump_on_errcall_interval_secs_and_regex, {IntervalSecs, RegexMode, PathPrefix, Regex}}, 
                                   {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(IntervalSecs), IntervalSecs >= 0 ->

    ?BDB:set_write_coredump_on_errcall(BdbPortHandle, IntervalSecs, RegexMode, PathPrefix, Regex),
    StateAndPostponedConfigOptions;


handle_init_option_before_env_open({port_log_directory, _}, StateAndPostponedConfigOptions) ->
    %% handled by init
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({enable_port_logging_before_open_environment, _}, StateAndPostponedConfigOptions) ->
    %% handled by init
    StateAndPostponedConfigOptions;

handle_init_option_before_env_open({enable_port_logging_before_close_environment, _}, StateAndPostponedConfigOptions) ->
    %% handled by terminate
    StateAndPostponedConfigOptions;


%% ... and anything else is a typo
handle_init_option_before_env_open(Other, _StateAndPostponedConfigOptions) ->
    throw( {unrecognized_config_option, Other} ).


%% Handlers for replication sub-options
handle_replication_option_before_env_open({this_hostname_port, {Hostname, TCPIPPort}}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) ->

%%     %% Check that the host-name provided for use by BDB (repmgr) does actually match this host name, or we will have chaos.
%%     %% TODO: is this misguided?   (answer; probably -- hence it is commented out for now)
%%     {ok, ThisHostNameString} = inet:gethostname(),
%%     {ok, HostEnt} = inet:gethostbyname(ThisHostNameString, inet),
%%     true = lists:member(Hostname, HostEnt#hostent.h_aliases),

    ?BDB:repmgr_set_local_site(BdbPortHandle, Hostname, TCPIPPort, []),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({this_site_priority, Priority}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Priority), Priority >= 0 ->

    ?BDB:rep_set_priority(BdbPortHandle, Priority),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({request_retransmit_after_min_max_records, {Min, Max}}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Min), Min > 0, is_integer(Max), Max > 0 ->

    ?BDB:rep_set_request(BdbPortHandle, Min, Max),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({config_flags, Flags}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_list(Flags) ->

    lists:foreach(fun(F) -> ?BDB:rep_set_config(BdbPortHandle, F, 1) end, Flags),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({allow_not_durably_replicated_db_open = Key, enabled = Value}, StateAndPostponedConfigOptions) ->
    put(Key, Value),
    StateAndPostponedConfigOptions;
handle_replication_option_before_env_open({allow_not_durably_replicated_db_open = Key, disabled = Value}, StateAndPostponedConfigOptions) ->
    put(Key, Value),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({peer_hostnames_ports, HostnameTCPIPPortsList}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) ->

    %% TODO: Eventually we want to allow DB_REPMGR_PEER to be set.  (Only one, the last one, is used).
    %% But repmgr BDB v4.6 is going to improve that feature, so we leave it for now.

    %% repmgr_add_remote_site returns an opaque integer 'eid' for each site.
    %% We need to store the mapping of eid -> HostPort, for decoding NEWMASTER events (they contain the eid of the new master)
    RemoteSiteEIDsToHostPorts
        = lists:map(fun({Hostname, TCPIPPort} = HostPort) when is_list(Hostname), is_integer(TCPIPPort) -> 
                            {?BDB:repmgr_add_remote_site(BdbPortHandle, Hostname, TCPIPPort, []), HostPort}
                    end,
                    HostnameTCPIPPortsList),
    put(remote_site_eids_to_hostports, RemoteSiteEIDsToHostPorts),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({initial_role_and_threadcount, {_Role, _NumReplicationThreads}} = Option, 
                                          {State, PostponedConfigOptions}) 
  when is_atom(_Role), is_integer(_NumReplicationThreads), _NumReplicationThreads >= 0 ->

    %% These are the arguments for repmgr_start(), which we cannot call until after env_open has been called.
    %% So we move it to the AfterEnvOpenConfigOptions section.
    {State, [Option | PostponedConfigOptions]};


handle_replication_option_before_env_open({ack_policy, Policy}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_atom(Policy) ->

    ?BDB:repmgr_set_ack_policy(BdbPortHandle, Policy),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({limit_transfer_per_call_bytes, Bytes}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Bytes), Bytes > 0, Bytes < 1024 * 1024 * 1024 ->

    ?BDB:rep_set_limit(BdbPortHandle, 0, Bytes),  % always less than 1GB :)
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({{timeout, acks_microseconds}, Microseconds}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Microseconds), Microseconds >= 0 ->

    ?BDB:rep_set_timeout(BdbPortHandle, bdb_DB_REP_ACK_TIMEOUT, Microseconds),    
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({{timeout, election_milliseconds}, Milliseconds}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Milliseconds), Milliseconds >= 0 ->

    ?BDB:rep_set_timeout(BdbPortHandle, bdb_DB_REP_ELECTION_TIMEOUT, Milliseconds * 1000),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({{timeout, election_retry_milliseconds}, Milliseconds}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Milliseconds), Milliseconds >= 0 ->

    ?BDB:rep_set_timeout(BdbPortHandle, bdb_DB_REP_ELECTION_RETRY, Milliseconds * 1000),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({{timeout, connection_retry_milliseconds}, Milliseconds}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Milliseconds), Milliseconds >= 0 ->

    ?BDB:rep_set_timeout(BdbPortHandle, bdb_DB_REP_CONNECTION_RETRY, Milliseconds * 1000),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({{internal_delay, master_checkpoint_seconds}, Seconds}, 
                                          {#state{bdb_port_handle = BdbPortHandle}, _} = StateAndPostponedConfigOptions) 
  when is_integer(Seconds), Seconds >= 0 ->

    ?BDB:rep_set_master_checkpoint_sleep(BdbPortHandle, Seconds),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open({event_handlers, ModuleNamesArgs}, StateAndPostponedConfigOptions) when is_list(ModuleNamesArgs) ->

    GenEventName = dynamic_replication_gen_event_proc_name(),
    lists:foreach(fun({ModuleName, Args}) ->
                          %%?INFO_LOGGER("init: calling gen_event:add_handler{~w, ~w, ~w)~n", [GenEventName, ModuleName, Args]),
                          %% TODO: we don't call add_sup_handler here as we are installing the *caller's* handler but the 
                          %% gen_event_EXIT message woud come to the port coordinator, not the caller.
                          %% Options:
                          %% 1. call add_sup_handler in the start() before returning to the caller (might miss events)
                          %% 2. add a handle_info clause for gen_event_EXIT and forward it to the process that spawned us (or require that the event_handlers config argument specify a pid to send the message too)                          
                          ok = gen_event:add_handler(GenEventName, ModuleName, Args)
                  end,
                  ModuleNamesArgs),
    StateAndPostponedConfigOptions;


handle_replication_option_before_env_open(Other, _) ->
    throw( {unrecognized_config_option, Other} ).


%% Handlers for distributed_transactions sub-options
handle_distributed_transactions_option_before_env_open({gtm_rg_caller_state, #?CS{}}, StateAndPostponedConfigOptions) ->
    StateAndPostponedConfigOptions;

handle_distributed_transactions_option_before_env_open({gtm_timeout_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    StateAndPostponedConfigOptions;

handle_distributed_transactions_option_before_env_open({resolve_retry_period_milliseconds, N}, StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    StateAndPostponedConfigOptions;

handle_distributed_transactions_option_before_env_open({resolve_max_num_attempts, N}, StateAndPostponedConfigOptions) when is_integer(N), N > 0 ->
    StateAndPostponedConfigOptions;

handle_distributed_transactions_option_before_env_open(Other, _) ->
    throw( {unrecognized_config_option, Other} ).


%% handlers for options delayed until after env_open has been called.
handle_init_option_after_env_open({cache_throttle_io_write_ops_and_sleep_microseconds, {SequentialWriteOps, SleepMicroseconds}}, 
                                  #state{bdb_port_handle = BdbPortHandle} = State) ->
    ?BDB:set_mp_max_write(BdbPortHandle, SequentialWriteOps, SleepMicroseconds),
    State;

handle_init_option_after_env_open({initial_role_and_threadcount, {Role, NumReplicationThreads}}, 
                                  #state{bdb_port_handle = BdbPortHandle} = State) ->
    ?INFO_LOGGER("init: calling repmgr_start~n"),
    %% This may generate an untagged event. If so, we will handle it later in handle_info.
    ?BDB:repmgr_start(BdbPortHandle, NumReplicationThreads, Role),    
    State.


%%--------------------------------------------------------------------
%% Function: handle_call(Request, From, State) -> {reply, Reply, State} |
%%                                      {reply, Reply, State, Timeout} |
%%                                      {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, Reply, State} |
%%                                      {stop, Reason, State}
%% Description: Handling call messages
%%--------------------------------------------------------------------
handle_call(acquire_bdb_port_handle, 
            {FromPid, _RequestTag}, 
            #state{bdb_port_handle      = BdbPortHandle,
                   clients_ets_table_id = ClientsEtsTableId} = State) ->

    %% Limit the number of concurrent users
    %% We track the number of clients ourselves rather than using ets:info(Tab) and extracing {size, N}
    %% as ets:info returns a lot of other junk and we don't want the garbage.

    CurrNumClients = ets:lookup_element(ClientsEtsTableId, curr_num_clients, 2),
    MaxNumClients  = get(max_num_concurrent_client_procs),

    if CurrNumClients >= MaxNumClients ->

            ?WARNING_LOGGER("handle_info: acquire_bdb_port_handle failed with too_many_client_procs~n"),
            %% We include both numbers in the error as in future I want to mak max_num_concurrent_client_procs 
            %% adjustable at runtime, and if someone lowers it then immediately afterwards CurrNumClients may 
            %% actually be bigger than MaxNumClients for a while.
            {reply, {error, {too_many_client_procs, CurrNumClients, MaxNumClients}}, State};

       true ->

            %% We monitor all active clients as we need to call group_valmap_cleanup etc if they crash.
            %% We use monitor and not link, as 
            %%
            %% - There is no need for a bi-directional link.  If the server crashes then 
            %%   terminate() wille explicitly kill all registered clients, and if that 
            %%   doesn't get a chance to happen (e.g. server is hard-killed) then the 
            %%   port will reject new BDB commands.
            %%
            %% - We don't want the caller to be able to break the link by calling unlink.
            %%
            %% - link() was written before exceptions and has awkward failure modes 
            %%   if the client process has already exited when link is called.
            %%   and has the following behaviour.
            %%
            %%   From http://www.erlang.org/ml-archive/erlang-questions/200510/msg00204.html
            %%
            %%   link(Pid) -> error(noproc) if Pid is local and does not exist and
            %%                              caller has process_flag(trap_exit, false)
            %%                true          otherwise
            %%    
            %%   And if the Pid dies, or even was remote and did not exist
            %%   (those cases can not be discriminated), the caller will get
            %%   an exit signal that can be converted to a regular message
            %%   with process_flag(trap_exit, true).
            %%
            %%   Monitor solves this by always appearing to be successful 
            %%   (even if the client is already dead) and just sending a
            %%   normal DOWN message immediately.  i.e. It hides the race condition.

            MonitorRef = erlang:monitor(process, FromPid),

            %% Each client is issued a transiently-unique group-valmap-id (a 32-bit unsigned int).
            %% (We want 2 current clients to not have the same group-id, but we don't guarantee this.
            %%  If they do have the same ID then one client will close/abort the other's cursors/txns,
            %%  but EDTK generation-ids makes that safe (throws exceptions), just extremely undesirable.)
            %%
            %% We increment prev_group_valmap_id and issue the new value to this client.
            %%
            %%  - If we overflow the type range (2^32) then we need to wrap around.
            %%  - We must never issue ?SERVER_GROUP_VALMAP_ID to a client, as it owns the DBENV and all of the DB handles.
            %%    ?SERVER_GROUP_VALMAP_ID is always 0.
            %%
            %% Fortuantely this ETS API does exactly what we want:
            %%
            %%  ets:update_counter(Tab, Key, {Pos,Incr,Threshold,SetValue}) -> Result
            %%  
            %%  Destructively update the object with key Key in the table Tab by adding Incr to the element at the Pos:th position. 
            %%  The new counter value is returned.
            %%  If a Threshold is specified, the counter will be reset to the value SetValue if the following conditions occur: 
            %%  - The Incr is not negative (>= 0) and the result would be greater than (>) Threshold 
            %%  - The Incr is negative (< 0) and the result would be less than (<) Threshold

            NewGroupValmapId = ets:update_counter(ClientsEtsTableId, prev_group_valmap_id, {2, 1, 16#FFFFFFFF, ?SERVER_GROUP_VALMAP_ID + 1}),

            %% Insert the client into the table.
            %% It is an error for a client process to call acquire_bdb_port_handle twice.
            %% Otherwise we would have to track multiple group valmap ids, to cleanup if that process exits,
            %% and we would have to ref-count calls to link before we can unlink in release_port.

            case ets:insert_new(ClientsEtsTableId, {FromPid, MonitorRef, NewGroupValmapId}) of

                false -> 
                    ?WARNING_LOGGER("handle_info: Registered client ~w called acquire_bdb_port_handle again~n", [FromPid]),
                    {reply, {error, caller_already_has_a_bdb_port_handle}, State};

                true ->
                    _NewNumClients = ets:update_counter(ClientsEtsTableId, curr_num_clients, 1),

                    %%?INFO_LOGGER("handle_call(acquire_bdb_port_handle): registered client pid ~w. MonitorRef = ~w, group_valmap_id ~w (new #clients = ~w)~n", [FromPid, MonitorRef, GroupValmapId, _NewNumClients]),

                    {reply, 
                     {ok, BdbPortHandle#bdb_port_handle{group_valmap_id = NewGroupValmapId,
                                                        issued_to_pid   = FromPid}},          % owned by the client
                     State}
            end
    end;


handle_call({get_db_handle, {RequestedDBType, DBId, SpecifiedDBConfigProplist}}, 
            {FromPid, _RequestTag}, 
            #state{db_handle_cache = DbHandleCache} = State) ->

    %% Combine the SpecifiedDBConfigProplist with the port-coordinator's default config.
    %% 'safe' here means an internal try/catch, so we don't kill the server if the client passes something 
    %% that is totally malformed.

    case ?BDB_U:safe_resolve_config_proplist(SpecifiedDBConfigProplist, 
                                             get(default_db_config_proplist)) of

        {ok, NewDBConfigProplist} ->
            %% Is this database already open>
            case dict:find(DBId, DbHandleCache) of

                {ok, {DB, ExistingDBType, ExistingDBConfigProplist}} ->

                    %% We need to check that DBType and DBConfigPropList are compatible
                    %% (i.e. Another process may have opened this database with a differnt type, or a different page-size etc.)
                    case RequestedDBType of 

                        ExistingDBType ->
                            %% Database is the correct type.  
                            %% If the supplied config is compatible with the existing config, then return 
                            %% the handle, otherwise return an error
                            check_db_config_for_open_database(DBId, DB, NewDBConfigProplist, ExistingDBConfigProplist, ExistingDBType, FromPid, State);

                        bdb_DB_UNKNOWN ->
                            %% The caller doesn't know what type it is, and doesn't care.
                            %% e.g. They might have discovered it via list_all_open_databases(), 
                            %% or by filelib:fold_files("*.db") on the env_dir.
                            check_db_config_for_open_database(DBId, DB, NewDBConfigProplist, ExistingDBConfigProplist, ExistingDBType, FromPid, State);

                        _ ->
                            ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~w to open existing DB handle (~p) with wrong type (existing: ~w, new ~w)~n", [FromPid, DBId, ExistingDBType, RequestedDBType]),
                            {reply, {error, {mismatched_db_type, ExistingDBType, RequestedDBType}}, State}
                    end;

                error ->
                    %% We have not yet opened a database with this filename.
                    %% The database may or may not already exist on disk.
                    %% If it doesn't exist on disk, we create it.
                    open_unknown_database(RequestedDBType, 
                                          DBId, 
                                          NewDBConfigProplist, 
                                          FromPid, 
                                          State)
            end;

        {error, Error} ->
            ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~w to open DB (~p) with malformed config ~p (error: ~p)~n", [FromPid, DBId, SpecifiedDBConfigProplist, Error]),
            {reply, {error, {bad_db_config, Error, SpecifiedDBConfigProplist}}, State}

    end;


handle_call(release_bdb_port_handle, 
            {FromPid, _RequestTag}, 
            #state{clients_ets_table_id = ClientsEtsTableId} = State) ->

    %% We use dict:find instead of dict:fetch as a badly behaving client 
    %% should not be able to kill the server.
    case ets:lookup(ClientsEtsTableId, FromPid) of

        [{_FromPid, MonitorRef, _GroupValmapId}] ->

            %% Note: we don't use the 'flush' option to demonitor.
            %% That option can be used to purge any DOWN message
            %% for this client which may already be sitting in the
            %% message queue.  This would mean that handle_info(DOWN) would 
            %% not need to worry about receiving DOWN messages from 
            %% processes not in clients_ets_table.
            %% However, the cost is blocking longer here to re-scan the entire 
            %% message queue.  On balance I think it will be faster to just 
            %% process the messages in order and handle the DOWN case where 
            %% the client has already been removed.

            erlang:demonitor(MonitorRef),

            ets:delete(ClientsEtsTableId, FromPid),
            _NewNumClients = ets:update_counter(ClientsEtsTableId, curr_num_clients, -1),

            %%?INFO_LOGGER("handle_call(release_bdb_port_handle): From client process ~w (monitor ref ~w, group_valmap_id ~w already cleaned on client side) (new #clients ~w)~n", [FromPid, MonitorRef, GroupValmapId, _NewNumClients]),

            {reply, 
             {ok, ok},  % tuple required as client call is inside ?THROW_ERRORS
             State};

        [] ->
            ?INFO_LOGGER("handle_call(release_bdb_port_handle): From NON-client pid ~w~n", [FromPid]),
            %% This used to be possible before R11B as
            %% unlink/demonitor was asynchronous, so if a client
            %% called release_bdb_port_handle correctly (causing the
            %% above unlink) but then exited, we might still get an
            %% exit signal.  With R11B unlink/demonitor are still
            %% asynchronous but all subsequent exit events/DOWN
            %% messages from that process to this one are discarded.
            %% So this error case should really only happen if the
            %% caller is misbehaving.  So we return an error to throw
            %% in the caller.
            {reply, {error, caller_does_not_have_a_bdb_port_handle}, State}
    end;


handle_call(list_all_open_databases, _From,  #state{db_handle_cache = DbHandleCache} = State) ->
    {reply, 
     {ok, dict:fold(fun({DBFilename, []}, _, Acc) -> [DBFilename | Acc];
                       (DBId, _, Acc)             -> [DBId | Acc]
                    end,
                    [],
                    DbHandleCache)},
     State};


handle_call(get_config, _From, State) ->
    {reply, {ok, get(config_proplist)}, State};


handle_call(start_checkpointer_if_enabled, _From, State) ->
    {reply, {ok, ok}, server_start_checkpointer_if_enabled(State)};


handle_call(stop_checkpointer_if_running, _From, State) ->
    {reply, {ok, ok}, server_stop_checkpointer_if_running(State)};


handle_call({close_unused_database, DBId}, 
            {FromPid, _RequestTag}, 
            #state{bdb_port_handle = BdbPortHandle,
                   db_handle_cache = DbHandleCache} = State) ->

    ?INFO_LOGGER("handle_call(close_unused_database): From client process ~w. DBId = ~300p~n", [FromPid, DBId]),

    case dict:find(DBId, DbHandleCache) of

        {ok, {DB, _ExistingDBType, _ExistingDBConfigProplist}} ->

            try
                ?BDB:db_close(BdbPortHandle, DB, []),

                %% Remove the DB handle from our cache
                {reply, 
                 {ok, ok},
                 State#state{db_handle_cache = dict:erase(DBId, DbHandleCache)}}

            catch
                Class:Error ->
                    %% The BDB documentation for db_close says: 
                    %%  "The DB handle may not be accessed again after DB->close is called, regardless of its return."
                    {reply, 
                     {ok, {db_close_failed, {Class, Error}}}, 
                     State#state{db_handle_cache = dict:erase(DBId, DbHandleCache)}}
            end;

        error ->
            %% We have not yet opened a database with this filename 
            {reply, 
             {error, {unknown_database, DBId}},
             State}
    end.


%%--------------------------------------------------------------------
%% Function: handle_cast(Msg, State) -> {noreply, State} |
%%                                      {noreply, State, Timeout} |
%%                                      {stop, Reason, State}
%% Description: Handling cast messages
%%--------------------------------------------------------------------
handle_cast(stop, State) ->
    {stop, normal, State}.


%%--------------------------------------------------------------------
%% Function: handle_info(Info, State) -> {noreply, State} |
%%                                       {noreply, State, Timeout} |
%%                                       {stop, Reason, State}
%% Description: Handling all non call/cast messages
%%
%% - DOWN signals from any registered clients
%% - EXIT signals from support_procs
%% - All port replies when we are in pipe-mode
%% - All port replies to non_block calles from this process 
%%   (e.g. group_valmap_cleanup_non_block)
%%--------------------------------------------------------------------
handle_info({ErlangPort, {data, Bytes}}, 
            #state{bdb_port_handle = #bdb_port_handle{erlang_port = ErlangPort} = BdbPortHandle} = State) -> 

    %% This is a raw message from the pipe-driver. 
    %% We convert it to a term and forward it to the client process that sent the corresponding command.
    %%
    %% [Related sidebar: This server process itself occasionally calls blocking functions in the 
    %% ?BDB module (e.g. group_valmap_cleanup).  
    %% All such functions implicitly call ?BDB:receive_reply().
    %% While waiting for it's specific tagged reply, receive_reply() is 
    %% forced to accept all pipe-driver byte messages (to examine the tags).
    %% Its behaviour is to convert the message to term form and
    %%  - if ther tag matches, return the message
    %%  - if the tag does NOT match then extracts the ClientPid from the tag 
    %%    and forwards the message to that process.  Then recurse to wait for 
    %%    the requested message.
    %% i.e. ?BDB:receive_reply() implicitly participates in the 
    %% pipe-driver forwarding mechanism on behalf of this (or any other) server 
    %% that handles EDTK port messages in pipe-driver mode.
    %% As pipe replies are arriving all the time, this is certain to happen 
    %% eventually when we are blocking for the reply to a group_valmap_cleanup.
    %% Alternatively, servers would have to completely avoid the blocking ?BDB functions 
    %% (so avoiding ever calling receive_reply even indirectly).
    %% i.e. Servers would be restricted to using the non-blocking commands,
    %% and would have to keep the tags of all outstanding messages in 
    %% State, and handle them as they arrive asynchronously. (Much more complex.)

    {noreply,
     case ?BDB:pipe_driver_bytes_to_term(ErlangPort, Bytes) of
         {ErlangPort, TagBin, ok}         = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, ok, {_}}    = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, ok, {external_term, BinTerm}} -> 
                                                     forward_pipedriver_reply_to_client(TagBin, {ErlangPort, TagBin, ok, binary_to_term(BinTerm)}, State);
         {ErlangPort, TagBin, ok, _}      = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, error, {_}} = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);
         {ErlangPort, TagBin, error, _}   = Reply -> forward_pipedriver_reply_to_client(TagBin, Reply, State);

         %% Untagged events may be generated by the driver (e.g. BDB event notifications).
         %% They should be sent to the port's connected process -- i.e. self().
         %% The handler either exits or returns a new State.
         {ErlangPort, bdb_untagged_event, Event} ->  handle_untagged_event(Event, State);

         %% 'untagged' errors can happen in pipe-mode too (see below)
         {ErlangPort, bdb_untagged_error, _} = Msg -> exit({bdb_port_error, {BdbPortHandle, Msg}});

         %% Anything else is a serious error (e.g. malformed messages).
         Other -> exit({bdb_bad_pipe_driver_reply, {BdbPortHandle, Other}})
     end};


handle_info({ErlangPort, bdb_untagged_event, Event}, #state{bdb_port_handle = #bdb_port_handle{erlang_port = ErlangPort}} = State) -> 
    {noreply, handle_untagged_event(Event, State)};


handle_info({'DOWN', MonitorRef, process, FromPid, Reason}, 
            #state{bdb_port_handle      = #bdb_port_handle{group_valmap_cleanup_receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup} = BdbPortHandle,
                   clients_ets_table_id = ClientsEtsTableId} = State) -> 

    %% A registered client crashed, so we need to call group_valmap_cleanup for it.
    %%
    %% Imprtant notes:
    %%
    %% 1. Currently we call group_valmap_cleanup in *blocking* mode.  That may need to 
    %%    abort transactions, which involves disk IO and possibly replication.
    %%    So this may block for siginificant time.
    %%    So if a LOT of registered clients start crashing then port-coordinator 
    %%    could be quite unresponsive while it processes all of the group-valmap-cleanup 
    %%    operations.  
    %%    Instead we could either spawn a process to call group-valmap-cleanup,
    %%    or call the non_block variationg and remember the tag.  The former 
    %%    would require tracking all of the spawned pids in State, so that we can
    %%    wait for them to finish during clean shutdown (that's a safety issue 
    %%    -- not optional).  The latter (using the non_block variant) would 
    %%    require storing the tags in State and doing handle_info for the 
    %%    replies. And in pipe-mode, when we receive all replies, that is more 
    %%    work (check every message's tag against the list of outstanding 
    %%    group-valmap cleanups).
    %%
    %%    Also, see point #2 below about the safety guarantees we must provide.
    %%     
    %%    Anyway, it remains to be proven that we need this.  Indeed if 
    %%    large numbers of client processes are crashing then throttling the 
    %%    rate at which port-coordinator issues new port handles is probably 
    %%    a good thing.  My main concern is lack of responsiveness to shutdown.
    %%    But remember:
    %%     
    %%    Important #2: normal successful client shutdown (i.e. calls to 
    %%    release_bdb_port_handle) does NOT cause this problem as we 
    %%    arrange for the client process to call group_valamap_cleanup,
    %%    not the server (the server merely un-registers the client).
    %%     
    %%    Important #2: Also, most applications will be using ?BDB_H:do_txn
    %%    or ?BDB_RG:do_txn_on_master/any_site in order to get automatic 
    %%    retry on deadlock.  Those functions catch any crashes and 
    %%    do 'clean' shutdown (release_bdb_port_handle) on the client side.
    %%     
    %%    So in most applications the port-coordinator should rarely see 
    %%    an EXIT signal from  registered client, which is why I have not 
    %%    optimized for that case.
    %%
    %% 2. The client might have crashed (receive_reply_timeout) while calling 
    %%    group_valmap_cleanup itself (e.g. release_bdb_port_handle does this 
    %%    on the client side, and ?BDB_H:do_txn calls release_bdb_port_handle 
    %%    even if the client TxnFun crashes -- i.e. not only when it returns 
    %%    normally).
    %%
    %%    Therefore, when we call group_valmap_cleanup here, it might 
    %%    find that the earlier call is **still running**.
    %%
    %%    We CANNOT de-register the client (remove it from clients_ets_table)
    %%    until the original group_valmap_cleanup has finished, because
    %%    if the port-coordinator shuts down with a 'lost' in-progress call,
    %%    it may believe it is safe to call db_close when that 'lost' call is 
    %%    still trying to abort transactions, and that will irrevocably corrupt 
    %%    the database.
    %%    Therefore, we must block here until group_valmap_cleanup really completes.
    %%    (If we implement either of the non-blocking options described in 
    %%    point #1 then we would need similar handling before we altered 
    %%    clients_ets_table).
    %%
    %%    Currently such a second overlapping call will throw 
    %%          {bdb_error, valmap_group_being_cleaned}.
    %%    So we catch that exception, and retry (poll) until we stop 
    %%    getting that exception, or until we decide to give up 
    %%    (we will wait for group_valmap_cleanup_receive_reply_timeout).
    %%
    %%    Note that we are in a gen_server callback, and gen_server might 
    %%    timeout, or catch any exception/exit that we throw and then call 
    %%    terminate (infact the latter definitely happens).
    %%    Therefore close_environment MUST also be prepared to block on
    %%    {bdb_error, valmap_group_being_cleaned}, and throw if the 
    %%    condition does not clear.  (If close_environment throws then 
    %%    it will avoid calling db_close explicitly, which is all that matters),
    %%    the Erlang port will be closed, and bdb_stop() in the C driver 
    %%    code will do appropriate waiting.  So this is safe.
    %%
    %%    Implementation note: we could have arranged for the C code 
    %%    to block the second call to group_valmap_cleanup instead 
    %%    of throwing {bdb_error, valmap_group_being_cleaned}.
    %%    That would be less code here, and arguably cleaner.
    %%    However I think it is practically worse
    %%    - less visibility into what is happening
    %%    - ties up 2 threads in the critical valmap_stop threadpool
    %%      instead of one.
    %%    This situation should be extremely rare, so I think 
    %%    that the polling solution is best.
    %%
    %% 3. Of course, this call to group_valmap_cleanup might hit 
    %%    a receive_reply_timeout.  In that case, gen_server 
    %%    will call terminate, and again it is important that 
    %%    close_environment correctly handle 
    %%    {bdb_error, valmap_group_being_cleaned}.

    %% Note that release_bdb_port_handle does _not_ use the 'flush' option to erlant:demonitor
    %% (to avoid re-scanning the message queue).  So we _may_ very rarely see a DOWN message for 
    %t a process that is nolonger in clients_ets_table, so we have to handle that.

    case ets:lookup(ClientsEtsTableId, FromPid) of

        [{FromPid, MonitorRef, GroupValmapId}] ->

            BdbPortHandleForGroupValmapCleanup = BdbPortHandle#bdb_port_handle{receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup},

            ?INFO_LOGGER("handle_info: Received DOWN for client ~w (~w, ~w). Cleaning group_valmap_id ~w~n", [FromPid, MonitorRef, Reason, GroupValmapId]),
            try 
                ?BDB:group_valmap_cleanup(BdbPortHandleForGroupValmapCleanup, GroupValmapId)
            of
                0            -> ok;
                NumResources -> ?INFO_LOGGER("handle_info: Received DOWN for client (~w) and cleaned ~w resources for group_valmap_id ~w~n", [FromPid, NumResources, GroupValmapId])
            catch
                {bdb_error, valmap_group_being_cleaned} ->
                    %% Do periodic polling (**which blocks the
                    %% port-coordinator**) until this exception stops
                    %% happening (should be fast), as described above.
                    %% If this function returns normally then
                    %% everything is fine.  If the exception does not
                    %% stop, this function will call exit.
                    poll_until_group_valmap_cleanup_is_complete(BdbPortHandle, GroupValmapId, ReceiveReplyTimeoutForGroupValmapCleanup)
            end,

            ets:delete(ClientsEtsTableId, FromPid),
            _NewNumClients = ets:update_counter(ClientsEtsTableId, curr_num_clients, -1),

            {noreply, State};

        [] ->
            %% Client must have been unregistered via release_bdb_port_handle 
            %% but a DOWN message arrived anyway (see comments above).
            %% Just ignore this.
            {noreply, State}
    end;


%% Note: the is_pid() test is in the following clause is important, as in pipe-mode we might also receive exit messages from the port (that is handled later).
handle_info({'EXIT', FromPid, Reason}, #state{support_procs = SupportProcs}) when is_pid(FromPid) ->

    %% It must be one of our support processes
    %%
    %% (note: if we ever get badmatch here it probably means that we
    %% have removed an entry from State#state.support_procs but
    %% *without* unlinking that process first -- and that process
    %% later crashed)

    {value, #support_proc{name = Name}} 
        = lists:keysearch(FromPid, #support_proc.pid, SupportProcs),

    %% If a support process dies we assume that the error must be serious, so we die too.
    ?ERROR_LOGGER("handle_info: Received EXIT signal from support process ~w ~w (reason ~w)~n", [Name, FromPid, Reason]),
    exit({support_process_exited, {Name, Reason}});


handle_info({'EXIT', ErlangPort, _} = Msg, 
            #state{bdb_port_handle = #bdb_port_handle{erlang_port = ErlangPort} = BdbPortHandle}) -> 
    %% From docs for erlang:open_port : 
    %% - During use of a port opened using {spawn, Name}, errors arising when sending messages to it are reported to the owning process using signals of the form {'EXIT', Port, PosixCode}
    exit({bdb_port_error, {BdbPortHandle, Msg}});


handle_info({ErlangPort, _} = Msg, 
            #state{bdb_port_handle = #bdb_port_handle{erlang_port = ErlangPort} = BdbPortHandle}) -> 
    %% From docs for erlang:open_port : 
    %% - if the 'exit_status' option is used then:  when the external process connected to the port exits, a message of the form {Port,{exit_status,Status}} is sent to the connected process, where Status is the exit status of the external process. If the program aborts, on Unix the same convention is used as the shells do (i.e., 128+signal).
    %% - if the 'eof' option is used then:          the port will not be closed at the end of the file and produce an exit signal. Instead, it will remain open and a {Port, eof} message will be sent to the process holding the port.
    exit({bdb_port_error, {BdbPortHandle, Msg}});


handle_info(Unrecognized, _State) ->
    exit({unrecognized_message, Unrecognized}).


%%--------------------------------------------------------------------
%% Function: terminate(Reason, State) -> void()
%% Description: This function is called by a gen_server when it is about to
%% terminate. It should be the opposite of Module:init/1 and do any necessary
%% cleaning up. When it returns, the gen_server terminates with Reason.
%% The return value is ignored.
%%--------------------------------------------------------------------
terminate(Reason, #state{bdb_port_handle = BdbPortHandle} = State) ->

    ?INFO_LOGGER("terminate: (reason: ~p)~n", [Reason]),

    kill_all_support_procs_and_clients_and_close_environment(Reason, State),

    ?INFO_LOGGER("terminate: closing port~n"),
    normal = ?BDB:shutdown(BdbPortHandle),

    %% IMPORTANT: in pipe-mode shutdown is apparently asynchronous with pipe-main.
    %% That's why we use DB_REGISTER in pipe-mode (the port_coordinator may restart
    %% while the previous pipe-main is still closing the environment).

    ?INFO_LOGGER("terminate: success~n"),
    ok.


%%--------------------------------------------------------------------
%% Func: code_change(OldVsn, State, Extra) -> {ok, NewState}
%% Description: Convert process state when code is changed
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.


%%--------------------------------------------------------------------
%%% Internal functions
%%--------------------------------------------------------------------

%%
%% Helpers for handle_info
%%

forward_pipedriver_reply_to_client(TagBin, Reply, State) ->
    %% (Remember, these messages may ALSO be forwarded by ?BDB:receive_reply)
    case binary_to_term(TagBin) of 
        {_Ref, ClientPid} when is_reference(_Ref), is_pid(ClientPid) -> 
            ClientPid ! Reply,
            State
    end.


handle_untagged_event({need_to_resolve_prepared_txn, <<_TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, #bdb_db_txn{index = _TxnValmapIdx} = Txn}, 
                      #state{support_procs = SupportProcs} = State) -> 

    %%?INFO_LOGGER("handle_untagged_event: sending dist txn ~w (valmap index ~w) to incrementally_resolve_prepared_txns~n", [_TxnCounter, _TxnValmapIdx]),
    {value, #support_proc{pid = ResolverPid}} = lists:keysearch(incrementally_resolve_prepared_txns, #support_proc.name, SupportProcs),
    ResolverPid ! {need_to_resolve, DistributedTxnGUID, Txn},
    State;


handle_untagged_event({sleepycat_event, bdb_DB_EVENT_PANIC} = Event, State) -> 

    ?ERROR_LOGGER("handle_untagged_event: Got DB_EVENT_PANIC : hard-killing all clients and then exiting~n"),

    %% We don't publish this event via replication_gen_event, as exiting will kill the gen_event process anyway.
    %% We don't call do_run_txn_funs_on_BDB_event as we are about to exit.

    Config = get(config_proplist),
    PortStartFunc = get_value_for_key(port_start_function, Config),
    Funs          = get_value_for_key(PortStartFunc,
                                      get_value_for_key(run_mode_specific_funs_on_BDB_PANIC, Config)),
    lists:foreach(fun(Fun) -> Fun(State) end, Funs),

    %% If the Funs all complete, we must still exit.
    exit(Event);


handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_MASTER = EventType}, 
                      #state{bdb_port_handle = #bdb_port_handle{port_coordinator_name = PortCoordName} = BdbPortHandle} = State) -> 

    %% "The replication group of which this site is a member has just elected the local site as its master."
    ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_MASTER (we have been elected master)~n"),

    %% See comments in init().  If both replication and distributed transactions are enabled 
    %% then a new master must always resolve any prepared txns before continuing.
    %% This requires that a master GTM is available, or it will block (retrying) indefinitely.

    case get_value_for_key(distributed_transactions, get(config_proplist)) of

        disabled ->
            ok;

        DistributedTxnConfigList ->
            %% We may need to run txn_recover here.  However, clients might be trying to start new transactions.
            %% BDB does catch that and returns errors for txn_begin.   (I've requested that Sleepcat change this 
            %% and allow txn_begin in this circumstance -- to improve write-availablity of the master.
            %% txn_begin is safe here as the new master will still have the locks owned by the unresolved txn, 
            %% -- unlike after normal (startup-time) recovery (when there are unresolved transactions discovered 
            %% by txn_recover), as normal recovery does not acquire locks (it is single-threaded).
            %%
            %% TODO: this blocks the port-coordinator from handling normal requests like 
            %% acquire/release_bdb_port_handle and get_db_handle, as well as OTP system requests.
            %% The 'normal' requests aren't much use anyway as BDB locks out the application 
            %% until all txns are resolved. But if Sleepycat allowed us to do non-committed reads 
            %% in this state, we might have better availability.
            %% So should we spawn a process to call do_distributed_transaction_recovery()
            %% (and call gen_event:notify from that process when it is done -- i.e. this site is
            %% certainly not a useful master until that process completes successfully).

            ?INFO_LOGGER("handle_untagged_event: Calling do_distributed_transaction_recovery after DB_EVENT_REP_MASTER~n"),
            do_distributed_transaction_recovery(BdbPortHandle, bdb_DB_FIRST, DistributedTxnConfigList),
            ?INFO_LOGGER("handle_untagged_event: do_distributed_transaction_recovery finished (after DB_EVENT_REP_MASTER)~n")
    end,

    %% Send notification to other parties AFTER txn_recover is finished.
    gen_event:notify(get_replication_gen_event_pid(State), {PortCoordName, EventType}),

    %% And in parallel, run some callbacks if desired
    do_run_txn_funs_on_BDB_event(PortCoordName, EventType),

    State;


handle_untagged_event({sleepycat_event, bdb_DB_EVENT_REP_NEWMASTER = EventType, MasterEnvId}, 
                      #state{bdb_port_handle = #bdb_port_handle{port_coordinator_name = PortCoordName}} = State) -> 

    %% "The replication group of which this site is a member has just elected a new master; the local site is not the new master."

    RemoteSiteEIDsToHostPorts = get(remote_site_eids_to_hostports),
    case lists:keysearch(MasterEnvId, 1, RemoteSiteEIDsToHostPorts) of

        {value, {_, MasterHostPort}} ->
            ?INFO_LOGGER("handle_untagged_event: Got DB_EVENT_REP_NEWMASTER (site ~500p is now master)~n", [MasterHostPort]),
            gen_event:notify(get_replication_gen_event_pid(State), {PortCoordName, EventType, MasterHostPort}),

            %% In parallel, run some callbacks if desired
            do_run_txn_funs_on_BDB_event(PortCoordName, EventType, MasterHostPort),

            State;       

        false ->
            exit({unrecognized_site_with_NEWMASTER_event, MasterEnvId, RemoteSiteEIDsToHostPorts})
    end;


%% All of the others currently follow a standard pattern:
%%  bdb_DB_EVENT_REP_CLIENT
%%  bdb_DB_EVENT_REP_STARTUPDONE
%%  bdb_DB_EVENT_REP_PERM_FAILED
%%  bdb_DB_EVENT_WRITE_FAILED
handle_untagged_event({sleepycat_event, EventType}, 
                      #state{bdb_port_handle = #bdb_port_handle{port_coordinator_name = PortCoordName}} = State) -> 

    ?INFO_LOGGER("handle_untagged_event: Got ~w~n", [EventType]),

    %% Publish event to all interested parties
    gen_event:notify(get_replication_gen_event_pid(State), {PortCoordName, EventType}),

    %% In parallel, run some callbacks if desired
    do_run_txn_funs_on_BDB_event(PortCoordName, EventType),

    State.


is_txn_fun_spec_list(TxnFunSpecs) ->

    lists:foreach(fun({Module, Function, _FixedArgs, NumDeadlockRetries}) 
                     when is_atom(Module), is_atom(Function), is_integer(NumDeadlockRetries), NumDeadlockRetries >= 0 ->
                          %% check that there is a public function of this name and arity 2
                          true = erlang:function_exported(Module, Function, 2)
                  end,
                  TxnFunSpecs).


do_run_txn_funs_on_BDB_event(PortCoordName, EventType) ->
    do_run_txn_funs_on_BDB_event(PortCoordName, EventType, []).

do_run_txn_funs_on_BDB_event(PortCoordName, EventType, EventArgs) ->

    EventsTxnFuns = get(run_txn_funs_on_BDB_event),

    case lists:keysearch(EventType, 1, EventsTxnFuns) of
        false -> 
            %%?INFO_LOGGER("do_run_txn_funs_on_BDB_event: no txn funcs to execute for ~w~n", [EventType]);
            ok;

        {value, {_, TxnFunSpecs}} -> 
            %% We pass the port-coordinator name rather than pid to get clearer logging
            lists:foreach(
              fun({Module, Function, FixedArgs, NumDeadlockRetries} = TxnFunSpec) ->
                      ?INFO_LOGGER("do_run_txn_funs_on_BDB_event: running: ~w, ~500p~n", [EventType, TxnFunSpec]),
                      %% Do the TxnFuns in parallel 
                      %% (Danger; if we don't spawn_link then they might live forever.
                      %%  Howver, they can choose to monitor the port-coordinator.)
                      spawn(
                        fun() ->
                                TxnFun = {Module, Function, {EventArgs, FixedArgs}},
                                ?INFO_LOGGER("do_run_txn_funs_on_BDB_event (~w, ~w, ~500p): worker starting~n", [PortCoordName, EventType, TxnFun]),
                                try 
                                    %% TODO: pass NumDeadlockRetries from somewhere
                                    ?BDB_H:do_txn(PortCoordName, TxnFun, NumDeadlockRetries)
                                of
                                    Result ->
                                        ?INFO_LOGGER("do_run_txn_funs_on_BDB_event (~w, ~w, ~500p): returned ~p~n", [PortCoordName, EventType, TxnFun, Result])
                                catch
                                    Class:Reason -> 
                                        ?INFO_LOGGER("do_run_txn_funs_on_BDB_event (~w, ~w, ~500p): failed with ~w:~p~n", [PortCoordName, EventType, TxnFun, Class, Reason])
                                end
                        end)
              end,
              TxnFunSpecs)    
    end.


%%
%% Helpers for get_db_handle_with_db_config
%%

check_db_config_for_open_database(DBId, DB, NewDBConfigProplist, ExistingDBConfigProplist, ExistingDBType, FromPid, State) ->

    case find_mismatched_db_config_entries(NewDBConfigProplist, ExistingDBConfigProplist) of 
        [] ->
            %%?INFO_LOGGER("handle_call(get_db_handle): Database ~500p is already open, so returning existing DB handle (~500p) to client pid ~w~n", [DBId, DB, FromPid]),
            %% We don't reference-count DB handles, as there is no way to make it accurate.
            {reply, {ok, DB}, State};

        MismatchList ->
            ?WARNING_LOGGER("handle_call(get_db_handle): Attempt by client pid ~w to open existing database (~500p) with same type (~w) but different config (mismatches: ~500p)~n", [FromPid, DBId, ExistingDBType, MismatchList]),
            {reply, {error, {mismatched_db_config, MismatchList}}, State}
    end.

%% Returns a list of incompaible config entries 
%%     [{Key, MismatchedExistingValue, NewValue}, ...]
%%  or [] if the lists are compatible.
%%
%% Policy: the new list must be a subset of the existing list
%% i.eAccMismatchList. all config entries in the new list must be present in the existing list 
%% and must be identical, but the existing list may have additional 
%% config entries that are not present in the new list
find_mismatched_db_config_entries(NewList, ExistingList) ->

    lists:foldl(fun({db_open_flags, NewDbOpenFlags}, AccMismatchList) ->
                        %% Special-case handling for the flags:
                        %% We need to ignore any DB_CREATE in NewDbOpenFlags.
                        %%
                        %% To make this fast, the db_open_flags in ExistingList
                        %% have been sorted and stripped of bdb_DB_CREATE.
                        %% So we can just do the same to NewDbOpenFlags and then compare

                        NormalizedNewDbOpenFlags = lists:sort(NewDbOpenFlags) -- [bdb_DB_CREATE],

                        case get_value_for_key(db_open_flags, ExistingList) of 
                            NormalizedNewDbOpenFlags -> 
                                AccMismatchList;
                            MismatchedExistingValue ->
                                [{db_open_flags, MismatchedExistingValue, NormalizedNewDbOpenFlags} | AccMismatchList]
                        end;

                   ({Key, NewValue}, AccMismatchList)  ->
                        case proplists:lookup(Key, ExistingList) of
                            {Key, NewValue}                ->  AccMismatchList;     % ok
                            {Key, MismatchedExistingValue} ->  [{Key, MismatchedExistingValue, NewValue} 
                                                                | AccMismatchList]; % mismatch
                            none                           ->  AccMismatchList      % ok
                        end
                end,
                [],       % initial AccMismatchList
                NewList).


%% When opening a database with DB_UNKNOWN we must re-construct its config
convert_db_stat_to_db_config_proplist(bdb_DB_BTREE, DbStat) ->
    [{page_size_bytes,             get_value_for_key(bt_pagesize, DbStat)},
     {set_btree_min_keys_per_page, get_value_for_key(bt_minkey,   DbStat)},
     {set_fixed_record_size_bytes, get_value_for_key(bt_re_len,   DbStat)},
     {set_fixed_record_pad_byte,   get_value_for_key(bt_re_pad,   DbStat)}
    ];

convert_db_stat_to_db_config_proplist(bdb_DB_HASH, DbStat) ->
    [{page_size_bytes,             get_value_for_key(hash_pagesize, DbStat)}
    ];

convert_db_stat_to_db_config_proplist(bdb_DB_QUEUE, DbStat) ->
    [{page_size_bytes,                get_value_for_key(qs_pagesize,   DbStat)},
     {set_queue_extent_size_db_pages, get_value_for_key(qs_extentsize, DbStat)},
     {set_fixed_record_size_bytes,    get_value_for_key(qs_re_len,     DbStat)},
     {set_fixed_record_pad_byte,      get_value_for_key(qs_re_pad,     DbStat)}
     ].


%% This is called if we haven't yet opened a database with name Filename.
%% The database may or may not already exist on disk.
%% If it doesn't exist, we create it.
%% Returns a valid gen_server:handle_call return value : {reply, Reply, State}
open_unknown_database(RequestedDBType, 
                      {Filename, DatabaseNameWithinFile} = DBId, 
                      NewDBConfigProplist, 
                      FromPid, 
                      #state{bdb_port_handle = BdbPortHandle} = State) ->

    %% As DB handles are shared across all clients, we call db_create with the server's GroupValmapId
    %% (which is a hidden argument in BdbPortHandle)
    try ?BDB:db_create(BdbPortHandle, []) of

        DB -> 
            %% Do any optional configuring of the database BEFORE we open it.

            %% Calling the BDB configure functions must be done in a try-catch, to avoid 
            %% killing the server if the client passed bad config values.
            try
                %%?INFO_LOGGER("handle_call(get_db_handle): Configuring DB (~300p) for client pid ~w (config: ~p)~n", [DBId, FromPid, NewDBConfigProplist]),
                lists:foreach(fun(ConfigArg) -> handle_db_config_option(BdbPortHandle, DB, ConfigArg) end,
                              NewDBConfigProplist),

                %% If OptionTxn is non-void it belongs to the **client** process. That's OK.
                %%
                %% If the caller provided a bad filename, or a filename that exists but is the wrong DBType
                %% then this will fail and we don't want to kill the server.
                %% (Note: DB_READ_UNCOMMITTED simply *allows* dirty-reads later; it does not mean that all reads are dirty).

                %% We keep the list of flags sorted to make comparison easier when opening 
                %% other handles to this same database (such requests must be rejected if the flags differ 
                %% **apart** from the presence/absence of DB_CREATE, which is not a property of the database or handle,
                %% but is an instruction to db_open).
                SortedDbOpenFlags = lists:sort( get_value_for_key(db_open_flags, NewDBConfigProplist) ),

                try ?BDB:db_open(BdbPortHandle, 
                                 DB, 
                                 void, 
                                 Filename, 
                                 DatabaseNameWithinFile,
                                 RequestedDBType, 
                                 SortedDbOpenFlags,
                                 get(file_permissions)) of
                    ok ->
                        finish_successful_db_open(RequestedDBType, DBId, NewDBConfigProplist, SortedDbOpenFlags, DB, State)
                catch
                    {bdb_error, bdb_DB_REP_NOTPERM = Reason} = ErrorReply ->

                        %% BDB returned success but our wrapper got a PERM_FAILED event.
                        %% The client may or may not be concerned about this.  
                        %% Currently we only let clients choose via server configuration, 
                        %% not on a per-case basis.
                        case get(allow_not_durably_replicated_db_open) of
                            enabled ->
                                ?WARNING_LOGGER("handle_call(get_db_handle): got ~w (**but configured to continue anyway**) when opening DB (~500p) for client pid ~w ~n", [Reason, DBId, FromPid]),
                                finish_successful_db_open(RequestedDBType, DBId, NewDBConfigProplist, SortedDbOpenFlags, DB, State);

                            disabled ->
                                ?ERROR_LOGGER("handle_call(get_db_handle): Failed (~p) to open DB (~500p) for client pid ~w ~n", [Reason, DBId, FromPid]),
                                ?BDB:db_close(BdbPortHandle, DB, []),
                                ?INFO_LOGGER("handle_call(get_db_handle): closed DB handle after failed (~p) to open DB (~500p) for client pid ~w ~n", [Reason, DBId, FromPid]),
                                {reply, {error, ErrorReply}, State}
                        end;

                    {bdb_error, Reason} = ErrorReply ->
                        ?ERROR_LOGGER("handle_call(get_db_handle): Failed (~p) to open DB (~500p) for client pid ~w ~n", [Reason, DBId, FromPid]),
                        ?BDB:db_close(BdbPortHandle, DB, []),
                        {reply, {error, ErrorReply}, State}
                end

            catch
                throw:Exception -> 
                    ?ERROR_LOGGER("handle_call(get_db_handle): Exception ~p configuring or opening database (~500p) for client pid ~w (db config: ~p)~n", [Exception, DBId, FromPid, NewDBConfigProplist]),
                    ?BDB:db_close(BdbPortHandle, DB, []),
                    {reply, {error, {bad_db_config, Exception, NewDBConfigProplist}}, State};

                error:Exception -> 
                    ?ERROR_LOGGER("handle_call(get_db_handle): Error ~p configuring or opening database (~500p) for client pid ~w (db config: ~p)~n", [Exception, DBId, FromPid, NewDBConfigProplist]),
                    ?BDB:db_close(BdbPortHandle, DB, []),
                    {reply, {error, {bad_db_config, Exception, NewDBConfigProplist}}, State}
            end

    catch
        {bdb_error, _} = Error -> 
            ?ERROR_LOGGER("handle_call(get_db_handle): db_create failed ~p~n", [Error]),
            {reply, {error, {db_create_failed, Error}}, State}
    end.

% private helper
finish_successful_db_open(RequestedDBType,
                          DBId,
                          NewDBConfigProplist,
                          SortedDbOpenFlags,
                          DB,
                          #state{bdb_port_handle = BdbPortHandle,
                                 db_handle_cache = DbHandleCache} = State) ->

    {ActualDBType, StoreDBConfigProplist}
        = case RequestedDBType of

              bdb_DB_UNKNOWN ->
                  %% We need to read its type and on-disk config (and reconstruct its 
                  %% db_config_proplist from that on-disk information), so that future 
                  %% attempts to open this handle can be checked for incompatible type/options

                  DbStat = ?BDB:db_stat(BdbPortHandle, DB, void, bdb_DB_FAST_STAT),
                  DBType = get_value_for_key(db_type, DbStat),

                  %% db_open_flags is properties of the DB handle, not of the physical database,
                  %% so we can't reconstruct it from the on-disk config.
                  %% Note that we don't have to remove DB_CREATE, as it is illegal with DB_UNKNOWN.

                  {DBType, [{db_open_flags, SortedDbOpenFlags} 
                            | convert_db_stat_to_db_config_proplist(DBType, DbStat)]};

              _ ->
                  {RequestedDBType, 
                   %% Remove DB_CREATE
                   lists:keyreplace(db_open_flags, 
                                    1, 
                                    NewDBConfigProplist,
                                    {db_open_flags, SortedDbOpenFlags -- [bdb_DB_CREATE]})}
          end,

    %%?INFO_LOGGER("handle_call(get_db_handle): Successfully opened database ~500p (new DB handle ~500p, type=~w, config=~500p) for client pid ~w~n", [DBId, DB, ActualDBType, StoreDBConfigProplist, FromPid]),
    {reply, 
     {ok, DB}, 
     State#state{db_handle_cache = dict:store(DBId, 
                                              {DB, ActualDBType, StoreDBConfigProplist},
                                              DbHandleCache)}}.


handle_db_config_option(_Port, _DB, {db_open_flags, Flags}) when is_list(Flags) ->
    %% This is handled when we call db_open
    ok;

handle_db_config_option(Port, DB, {page_size_bytes, N}) when is_integer(N), N > 0 ->
    ?BDB:db_set_pagesize(Port, DB, N);

handle_db_config_option(Port, DB, {set_flags, Flags}) when is_list(Flags) ->
    ?BDB:db_set_flags(Port, DB, Flags);

handle_db_config_option(Port, DB, {set_btree_min_keys_per_page, MinKey}) when is_integer(MinKey), MinKey > 0 ->
    ?BDB:db_set_bt_minkey(Port, DB, MinKey);

handle_db_config_option(Port, DB, {set_fixed_record_size_bytes, NumBytes}) when is_integer(NumBytes), NumBytes > 0 ->
    ?BDB:db_set_re_len(Port, DB, NumBytes);

handle_db_config_option(Port, DB, {set_fixed_record_pad_byte, ByteValue}) when is_integer(ByteValue), ByteValue >= 0, ByteValue =< 16#FF ->
    ?BDB:db_set_re_pad(Port, DB, ByteValue);

handle_db_config_option(Port, DB, {set_queue_extent_size_db_pages, NumDBPages}) when is_integer(NumDBPages), NumDBPages > 0 ->
    ?BDB:db_set_q_extentsize(Port, DB, NumDBPages);

handle_db_config_option(_, _, DbConfigOption) ->
    throw({unrecognized_db_config_option, DbConfigOption}).



%%
%% Helper to cleanly close environment.  Will throw or exit if it has problems.
%%

kill_all_support_procs_and_clients_and_close_environment(
  TerminateReason,
  #state{bdb_port_handle         = #bdb_port_handle{receive_reply_timeout                      = NormalReceiveReplyTimeout,
                                                    group_valmap_cleanup_receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup,
                                                    group_valmap_id                            = SERVER_GROUP_VALMAP_ID} = BdbPortHandle,
         clients_ets_table_id = ClientsEtsTableId,
         db_handle_cache      = DbHandleCache,
         support_procs        = SupportProcs}) ->

    ConfigProplist = get(config_proplist),

    ?BDB:adjust_port_logging(BdbPortHandle, 
                             enable, 
                             get_value_for_key(enable_port_logging_before_close_environment, ConfigProplist)),

    %% Theoretically, we don't actually need to do anything, as when the connected 
    %% process dies, the port will close and will do valmap cleanup.
    %% (If the gen_server crashes then we rely on all that auto-cleanup).
    %%
    %% i.e. The simplest possible explicit cleanup is ...
    %%
    %%  normal = ?BDB:shutdown(BdbPortHandle).
    %%
    %% ... and even that is not actually necessary (as the the VM
    %% will close the port automatically when the port-coordinator 
    %% process exits).
    %%
    %% However, as port shutdown (in linked-driver mode) blocks the 
    %% entire Erlang VM scheduler until it is finished, it *IS* beneficial 
    %% to to do the cleanup steps here, as it has much less impact on 
    %% the VM's scheduler thread.
    %%
    %% Note that in pipe-mode we DON'T bother to continue forwarding reply
    %% messages (for any already-in-progress commands) to the client processes 
    %% that sent those commands. But that is fine as we are about 
    %% to force all client processes to exit.
    %% (We do get replies to commands that _we_ issue here, because 
    %% ?BDB:receive_reply contains code to do that.)
    %% And note that pipe-main does not know that we are terminating 
    %% until we actually close the Erlang port.
    

    %% Ensure that any *future* calls to txn_checkpoint don't do a master-sleep.
    %% This call doesn't affect any in-progress checkpoint, but the checkpointer 
    %% support-proc might start a new one, and we will explicitly call txn_checkpoint 
    %% before closing databases.)
    ?BDB:rep_set_master_checkpoint_sleep(BdbPortHandle, 0),

    %% Disable any throttling of dirty-page IO, to make sure that 
    %% any in-progress calls to txn_checkpoint, memp_trickle, or memp_sync
    %% complete as quickly as possible (ideally BDB would have an 
    %% API to abort those operations).    
    ?BDB:set_mp_max_write(BdbPortHandle, 0, 0),

    %% Interrupt any in-progress internal BDB calls to interruptable_os_sleep 
    %% (e.g. txn_checkpoint on master) 
    %% so that all such calls finish as quickly as possible.
    %% Once we have called this, _all_ subsequent internal BDB calls to 
    %% interruptable_os_sleep will return immediately without sleeping.
    ?BDB:env_interrupt_os_sleep(BdbPortHandle),

    %% Enable automatic deadlock detection as we are about to shutdown 
    %% support processes and client processes are still running at this point
    %% so they might hit BDB deadlocks.
    %% (For the same reason, if the periodic deadlock-detector support process 
    %% is running then we allow it to stay running until group_valmap_cleanup
    %% has completed on ALL client group-valmap-ids.  Currently the 
    %% periodic deadlock-detector is *required* to avoid certain kinds of 
    %% threadpool deadlock -- see TODO file.  Also the bdb_stop 
    %% code in the driver runs the explicit deadlock-detector if 
    %% it blocks for too long waiting for cursors or txns to finish.)

    catch ?BDB:set_lk_detect(BdbPortHandle, bdb_DB_LOCK_DEFAULT),

    %% First, tell all of the support processes to stop *APART* from the 
    %% periodic deadlock-detector if it is running (see comment above).
    %%
    %% We call a support-proc-specific 'stop' function, rather than always using 
    %% exit(Pid,kill), as the process may be executing a BDB command 
    %% and we might want the BDB command to finish before it exits (rather than
    %% killing the process while is waiting for the port reply) so
    %% that the exit message really does mean that no BDB commands are
    %% being executed that were started by that process.  
    %% This happens to not be totally essential at the time of writing, as the support
    %% processes only use the DBENV handle (so it will have an incremented 
    %% refcount if a process is still using it).  It will become important 
    %% if any support process in the future opens a transaction or cursor 
    %% (e.g. a de-fragmenter using DB->compact), so we implement it correctly now.
    %%
    %% As with all terminate() functions, we cannot rely on EXIT signals for 
    %% support procs as we may have already eaten such a message (it might 
    %% be why we are terminating). So we need to monitor them.

    MostSupportProcMonitorRefs
        = tell_support_procs_to_gracefully_exit( lists:keydelete(deadlock_detector, 
                                                                 #support_proc.name, 
                                                                 SupportProcs) ),

    %% Now kill all of our registered client processes, and 
    %% run group_valmap_cleanup on each of their valmap group ids.
    %% We want to do this in parallel to minimize shutdown down, 
    %% so we use the non_block variant of group_valmap_cleanup.
    %% (That is less painful than spawning a cleanup process for 
    %% each one, as in pipe mode the replies all still come to this 
    %% server process, so we would have to forward them).

    BdbPortHandleForGroupValmapCleanup = BdbPortHandle#bdb_port_handle{receive_reply_timeout = ReceiveReplyTimeoutForGroupValmapCleanup},

    CleanGroupValmapTags = 
        ets:foldl(fun({ClientPid, ClientMonitorRef, ClientGroupValmapId}, CleanGroupValmapTags) ->

                          %% We don't need the DOWN messages from the clients we are about to kill
                          %% (They aren't harmful -- we just want to avoid accumulating lots of msgs
                          %% in the msg queue.  So we don't bother using the 'flush' option.)[
                          %%
                          %% (Note: if there was a risk of the client catching our exit signal and running 
                          %% group_valmap_cleanup itself (e.g. like ?BDB_H:do_txn) then we _would_ want to 
                          %% wait for the DOWN signal from the client before calling group_valmap_cleanup
                          %% here, to avoid possible valmap_group_being_cleaned errors (we do handle those 
                          %% but want to minimize them). However, we use exit(Pid,kill) to kill the client,
                          %% which the client cannot catch/prevent/delay.)

                          erlang:demonitor(ClientMonitorRef),

                          ?INFO_LOGGER("close_environment: killing client pid ~w (ref ~w) and cleaning group_valmap_id ~w~n", [ClientPid, ClientMonitorRef, ClientGroupValmapId]),
                          exit(ClientPid, kill),

                          [{ClientPid, ClientGroupValmapId, ?BDB:group_valmap_cleanup_non_block(BdbPortHandleForGroupValmapCleanup, 
                                                                                                ClientGroupValmapId)}
                           | CleanGroupValmapTags];

                     ({curr_num_clients,     _}, CleanGroupValmapTags) -> CleanGroupValmapTags;
                     ({prev_group_valmap_id, _}, CleanGroupValmapTags) -> CleanGroupValmapTags
                  end,
                  [], 
                  ClientsEtsTableId),

    %% We cannot close Db handles until all cursors and transactions are closed.
    %% So we *must* block until all of the above parallel operations finish, 
    %%
    %% IMPORTANT: this code must correctly handle group_valmap_cleanup
    %% failing with {bdb_error, valmap_group_being_cleaned}.
    %% i.e. We CANNOT proceed to close any DB handles until
    %% we know for certain that all client group valmaps have 
    %% finished being cleaned up.
    %% See handle_info('EXIT') for details.
    %%
    %% If any of the group_valmap_cleanup operations fail any any other way, then we 
    %% *cannot* safely close Db handles, so we just allow the exception to propagate 
    %% which will kill the process and close the port. Then the port's 'stop'
    %% function will impose the correct ordering on all valmaps that are left.
    %% (Note that if we do close the port under those circumstances, other group_valmap_cleanup 
    %% operations may still be in-progress. The port's _stop() function can handle that.)

    lists:foreach(fun({ClientPid, ClientGroupValmapId, Tag}) ->
                          ?INFO_LOGGER("close_environment: waiting for response to group_valmap_cleanup_non_block for client ~w, group-id ~w.~n", [ClientPid, ClientGroupValmapId]),

                          NumResources = 
                              try 
                                  ?BDB:receive_group_valmap_cleanup_reply(BdbPortHandleForGroupValmapCleanup, Tag)
                              catch
                                  {bdb_error, valmap_group_being_cleaned} ->
                                      %% Do periodic polling until this exception stops happening, 
                                      %% as described in handle_info('EXIT').
                                      %% If this function returns normally then everything is fine.
                                      %% If the exception does not stop, it will exit() (and gen_server will all terminate).
                                      poll_until_group_valmap_cleanup_is_complete(BdbPortHandle, ClientGroupValmapId, ReceiveReplyTimeoutForGroupValmapCleanup)
                              end,

                          ?INFO_LOGGER("close_environment: got reply for group_valmap_cleanup_non_block for client ~w, group-id ~w (freed ~w resources).~n", [ClientPid, ClientGroupValmapId, NumResources])
                  end,
                  CleanGroupValmapTags),

    %% All valmap groups belonging to client processs have been cleaned up.
    %% (but remember that if we are using distributed transactions, any prepared txns 
    %% will not have been automatically aborted -- see below for how we handle that).

    %% Now we need to consider whether to flush the page-cache, which is really 
    %% any issue of speed-of-shutdown vs speed-of-recovery.
    %%
    %% We obviously want fast shutdown, but we don't want to _always_ force a long 
    %% recovery period.  Also, we want to correctly support non-transactional 
    %% environments, which will be irrevocably corrupted
    %%
    %% There are 2 related issues:
    %%
    %% 1. Should we do a final checkpoint?  (Obviously this is only a
    %%    factor for transactional environments so we only consider it
    %%    ifcheckpoints are actually enabled.)  
    %%
    %%    There may or may not be a checkpoint in progress right now.
    %%    But even if there _is_ one in progress, it might have been
    %%    started some time ago, so a signficant number of dirty pages
    %%    might have accumulated in the cache that will not be flushed
    %%    by that in-progress checkpoint.
    %%
    %% 2. When we call db_close, should we pass the DB_NOSYNC flag?
    %%    Without that flag, db_close will flush any dirty pages for
    %%    that database. For large databases that can take a long
    %%    time.  
    %%
    %%    For transactional environments, that is purely wasted
    %%    work/time -- it *won't*Q reduce recovery time as there
    %%    would still need to be a checkpoint record after the
    %%    db_close records.
    %%
    %%    But for non-transactional environments/databases, we CANNOT 
    %%    safely use DB_NOSYNC (without a prior explicit call to db_sync)
    %%    as that would likely leave the database irrevocably corrupt
    %%    (there is no recovery mechanism).
    %%
    %% The current policy & rationale:
    %%
    %%   - If this is a 'clean' shutdow (TerminateReason = 'normal') then 
    %%     we do a checkpoint here.  This means that shutdown takes longer 
    %%     but recovery will be fast.  The assumption is that a 'clean' 
    %%     shutdown is instigated in a controlled way, usually by a human 
    %%     operator, so taking longer over shutdown to minimize recovery 
    %%     is both acceptable and desired.
    %%
    %%   - If this is not a clean shutdown (TerminateReason = anything
    %%     other than normal) then we don't do a checkpoint here.
    %%     This pushes more time to recovery, but the benefit is that
    %%     we call env_close as soon as possible.  That is important
    %%     when shutting down the master site, as we want the remaining sites 
    %%     to elect a new master as soon as possible -- i.e. in parallel with 
    %%     shutdown & restart of the ex-master.  But the current BDB-HA
    %%     Replication Manager code has no explicit way to disengage a site 
    %%     from the replication group (i.e. to force an election by removing 
    %%     the master site).  Currently that can only be done by calling env_close 
    %%     or by killing the entire OS process (in our case, the Erlang VM 
    %%     if we are in linked in mode).
    %%     So by not doing a checkpoint, we allow env_close to be called as 
    %%     soon as possible, and a new master to be elected as soon as possible.
    %%
    %%     There was an additional issue here that is fixed by patch 20070511.01.
    %%     As the master is required to close all DB handles before it can 
    %%     disengage from the group by calling env_close, the db-close records
    %%     are logged and replicated.  The DB_NOSYNC flag was not replicated, 
    %%     so this was causing replicas to call db_sync which significantly slowed 
    %%     the election. The patch makes replicas nolonger call db_sync in 
    %%     response to such records.
    %%
    %%   - Later, before calling db_close we first call db_get_transactional().
    %%
    %%     If the DB handle is transactional then we do specify DB_NOSYNC.
    %%     In 'clean' shutdown the final checkpoint will have already flushed      
    %%     all databases so DB_NOSYNC is irrelevant.  
    %%     And in non-clean shutdown, DB_NOSYNC ensures rapid shutdown and 
    %%     prevents wasted work of flushing db pages without placing a checkpoint 
    %%     record after them.
    %%
    %%     If the DB handle is non-transactional then we do sync the database 
    %%     and so avoid corrupting the file (unless we crash of course -- that 
    %%     is the risk of using a non-transactional database).
    %%
    %%   - The 'emergency' shutdown code in bdb_stop() follows the above policy.
    %%     i.e. It does not do a checkpoint (because that code only has work to 
    %%     do if something has gone badly wrong and this port-coordinator shutdown 
    %%     code did not finish executing), and the 'valmap cleanup' function for 
    %%     DB handles tests db->get_transactional() and passes DB_NOSYNC iff the 
    %%     handle is transactional.

    case TerminateReason of

        normal ->

            case get_value_for_key(checkpoint_seconds_and_min_kb, ConfigProplist) of
                disabled ->
                    ok;
                {_PeriodSeconds, MinKBytes} ->
                    %% Note that any configured checkpoint_receive_reply_timeout_milliseconds
                    %% is likely to be *vastly* too long as it will include compensation for 
                    %% {internal_delay, master_checkpoint}, and for aggressive mp_max_write on a 
                    %% large cache, neither of which can happen here.
                    %% However, the checkpoint can still take serious time, so we can't use 
                    %% the default receive_reply_timeout.
                    %% Actually, our best guess is the configured time for 
                    %% env_open_receive_reply_timeout_milliseconds,
                    %% as that includes time for truly dirty shutdown (i.e. an entire cache 
                    %% worth of dirty pages).
                    EnvOpenReceiveReplyTimeout = get_value_for_key(env_open_receive_reply_timeout_milliseconds, ConfigProplist),
                    ?INFO_LOGGER("close_environment: starting final checkpoint as TerminateReason is ~w (using env_open_receive_reply_timeout_milliseconds of ~w)~n", [TerminateReason, EnvOpenReceiveReplyTimeout]),
                    ?BDB:txn_checkpoint(BdbPortHandle#bdb_port_handle{receive_reply_timeout = EnvOpenReceiveReplyTimeout}, MinKBytes, 0, []),
                    ?INFO_LOGGER("close_environment: final checkpoint finished~n")
            end;
        _ ->
            ?INFO_LOGGER("close_environment: NOT doing final checkpoint as TerminateReason is ~w~n", [TerminateReason])
    end,
    
    %% Before we can close the DB handles we need to be **100% certain** that there are no open cursors or txns, 
    %% *and* no in-progress operations using any DB handles (i.e. direct db_get/put)
    %% *and* no possibility that any new such operations might be started.
    %% Otherwise, closing the DB handles will irrevocably corrupt the entire environment 
    %% (i.e. running recovery will fail due to log inconsistency).
    %% So it is definitely worth being paranoid on this.
    %% Circumstances that could cause this:
    %%
    %%  - Any bug in the registration of clients, or in the kill-all-clients-and-wait-for-all-exits code above, 
    %%    which leaves a process running and performing BDB operations
    %%
    %%  - Ignorance of the documented restrictions by client processes -- e.g. a process legitimately acquires a 
    %%    bdb_port_handle but then pass it to another process to use, and hack the handle's issued_to_pid field 
    %%    so that berkeley_db_drv functions will accept it, even though it is invisible & unknown to the port-coordinator.
    %%    Note that such rogue processes might have opened cursors, txns and even private DB handles
    %%    which the port-coordinator has no knowledge of.
    %%
    %% Neither of these have ever happened in practice, but it could be catastrophic if they did.
    %%
    %% There are a various ways to tackle this; e.g. 
    %%
    %% First, to ensure that there are no current or future operations that use ANY cursor, txn, or DB handle:
    %% 
    %%  - add a BDB driver API that installsa filter in outputv()
    %%    which rejects all but a safe subset of commands (at minimum db_close and env_close), and 
    %%    add another command that waits for all in-progress operations to finish.
    %% or
    %%  - shrink the size *and* queue-lengths of the worker threadpools to 0, and wait for that to 
    %%    complete.  This simultaneously prevents any new operations, and waits for 
    %%    all in-progress operations to finish, and doesn't require any new C code.
    %%    It also further reduces time spent in bdb_stop after terminate exits.
    %%    Obviously we still need to allow db_close/env_close etc to run
    %%    So this relies on the fact that those commands are assigned to the 'valmap_stop' 
    %%    threadpool.  That's a fairly subtle relationship, but it is still less work 
    %%    than the install-filter options.
    %%    IMPORTANT: bdb_stop MUST be left with some threads to use for shutdown (e.g. if 
    %%    terminate crashes leaving open DB handles after we shrink the threadpools.
    %%    This is currently achieved by having bdb_stop use the VALMAP_STOP threadpool.)
    %%
    %% We currently use the shrink-threadpools approach.
    %%
    %% Second, to ensure that there are 
    %%   i.   NO open cursors 
    %%   ii.  NO open txns (apart from any unresolved txns), 
    %%   iii. Any DB handles NOT owned by the port-coordinator's group-valmap-id (always 0)
    %%        This means that a rogue process created a private DB handle, that the port-coordinator doesn't know about.
    %%   iv.  (Paranoia) if ANY open valmap is in-use (even prepared txns)
    %%
    %% We add a small C command to check this. The command returns an error (causing us to throw) if it finds any problem, 
    %% which allows bdb_stop to run and clean up the mess.  
    %%
    %% If the command returns success then it is safe to close the DB handles that we know about, 
    %% and then close the DbEnv.  (We don't ask the C command to close the DB handles or environment, as 
    %% it may take significant time and we don't want to block the Erlang thread, 
    %% or write yet another C cleanup function.)
    %%
    %% (Note: step (iv) in the above checks allows us to use group_valmap_cleanup on the server's group id (0)
    %% to close the environment, rather than calling env_close.
    %% If we didn't do that check then we would want to call env_close, as that command 
    %% has code to check that NO other valmap slots of any type are still open.)

    %% Step #1: shrink basic worker threadpools
    %%
    %% we DON'T do VALMAP_STOP, as we need it
    %% we DON'T do LOCK_DETECT yet (we do it later), as the deadlock_detector support proc 
    %% must run until all normal operations have finished

    ?BDB:set_threadpool_params(BdbPortHandle, ?BDB_THREADPOOL_ID_WORKERS, 0, 64*1024, 0),
    ?BDB:set_threadpool_params(BdbPortHandle, ?BDB_THREADPOOL_ID_CONSUME_WAIT, 0, 64*1024, 0),
    ?BDB:set_threadpool_params(BdbPortHandle, ?BDB_THREADPOOL_ID_HOUSEKEEPING, 0, 64*1024, 0),

    repeat_until_true(
      fun() ->
              case  {get_value_for_key(num_threads, ?BDB:get_threadpool_info(BdbPortHandle, ?BDB_THREADPOOL_ID_WORKERS)),
                     get_value_for_key(num_threads, ?BDB:get_threadpool_info(BdbPortHandle, ?BDB_THREADPOOL_ID_CONSUME_WAIT)),
                     get_value_for_key(num_threads, ?BDB:get_threadpool_info(BdbPortHandle, ?BDB_THREADPOOL_ID_HOUSEKEEPING))}
              of
                  {0, 0, 0} -> 
                      true;
                  Other -> 
                      ?WARNING_LOGGER("close_environment: waiting for WORKERS,CONSUME_WAIT,HOUSEKEEPING threadpools to empty: ~w~n", [Other]),
                      timer:sleep(100),
                      false
              end
      end),

    %% Now it is really is safe to stop the periodic deadlock-detector 
    %% (which may or may not be running of course)
    DeadlockDetectorMonitorRefs 
        = case lists:keysearch(deadlock_detector, #support_proc.name, SupportProcs) of
              false        -> [];
              {value, PDD} -> tell_support_procs_to_gracefully_exit([PDD])
          end,

    %% Now wait for ALL support-procs to exit.
    %%
    %% IMPORTANT: if any of the support procs crash or are hard-killed they might leave BDB operations in progress.
    %%
    %%  - checkpointer/cache_trickle might leave txn_checkpoint and/or memp_trickle operations in the housekeeping threadpool
    %%  - deadlock_detector might have left lock_detect operation in the lock_detect threadpool
    %%  - incrementally_resolve_prepared_txns might leave an txn_abort/commit operation in progress in the valmap_stop threadpool
    %%
    %% The first one (checkpoint/memp_trickle) MUST be finished by now as we have shrunk that threadpool to 0.
    %% The second one (lock_detect) will be taken care of in a moment, as we are about to shrink the lock_detect threadpool to 0.
    %% The third one (incrementally_resolve_prepared_txns) will be caught by the check_safe_to_close_db_handles() function.
    %% It won't usually happen as that support proc is given a chance to exit cleanly, and is only hard-killed if it doesn't 
    %% respond to that request.

    wait_for_support_procs_to_exit(DeadlockDetectorMonitorRefs ++ MostSupportProcMonitorRefs),

    %% Now wait for the deadlock detector to stop
    ?BDB:set_threadpool_params(BdbPortHandle, ?BDB_THREADPOOL_ID_LOCK_DETECT, 0, 64*1024, 0),
    repeat_until_true(
      fun() ->
              case get_value_for_key(num_threads, ?BDB:get_threadpool_info(BdbPortHandle, ?BDB_THREADPOOL_ID_LOCK_DETECT)) of
                  0 -> 
                      true;
                  Other -> 
                      ?WARNING_LOGGER("close_environment: waiting for LOCK_DETECT threadpool to empty: ~w~n", [Other]),
                      timer:sleep(100),
                      false
              end
      end),

    %% Now throw if there are ANY open cursors, ANY non-prepared txns, ANY DB handles that are not owned by the port-coordinator,
    %% or ANY valmap (of any type, inc. dbenv) that is in-use (either shared or exclusive).
    try ?BDB:check_is_safe_for_port_coordinator_to_close_db_handles(BdbPortHandle)
    catch
        SafetyError ->
            ?ERROR_LOGGER("EXITING as check_is_safe_for_port_coordinator_to_close_db_handles failed with ~p~n", [SafetyError]),
            exit(failed_check_is_safe_for_port_coordinator_to_close_db_handles)
    end,

    %% Note that, if a client had a prepared (unresolved) transaction, then it is *still open*.
    %% (group_valmap_cleanup will have sent us an untagged event about it, but NOT aborted it.)
    %%
    %% In future versions of BDB (>v4.5) it might be safe to close DB handles here.
    %% But as of BDB v4.5, we CANNOT (normally) close DB handles if there are any prepared (unresolved) transactions 
    %% still open.  This is because DB::close logs a record that makes the log unrecoverable if there is an open txn before it.
    %%
    %% Sleepycat have given us a patch to allow this (it prevents DB::close from logging), but currently it must be explicitly 
    %% activated (SR [#14785])
    %% Sleepycat promise to support this officially at some point (perhaps via a different mechanism, or automatically).

    case get_value_for_key(distributed_transactions, ConfigProplist) of
        disabled ->
            0;
        _ ->
            ?INFO_LOGGER("close_environment: Checking for prepared transactions~n"),

            %% We have a special BDB driver command for this.
            %% It counts the number of prepared-but-unresolved txns, and if there are any,
            %% it activates the hack provided by Sleepycat (SR [#14785]) that makes it safe 
            %% to close DB handles with open transactions, by disabling logging of DB::close.
            %% mike.ubell@oracle.com says:
            %%
            %%   You asked what is the down side of not logging
            %%   database close on the master.  Along with not logging
            %%   the close we also do not free the "log_fileid" which
            %%   is a unique number that maps the fileid in log
            %%   records to the database that they correspond to.
            %%   This mapping is used both by the transaction abort
            %%   code and the recovery code.
            %%    
            %%   When recovery is run it will write close records for
            %%   all those databases which were not closed before the
            %%   restart and permit the fileids to be freed (and
            %%   reused).  If recovery is never run then the fileids
            %%   cannot be reused and are leaked.  Since there is an
            %%   array indexed by fileid to do the mapping this will
            %%   eventually burn a significant chunk of memory.
            %%    
            %% Sleepycat do plan to support this functionality officially, although the mechanism 
            %% may change (ideally it would be automatic -- then this command could be removed).
            %%
            %% The command was written speicifically for the current circumstances.  Therefore 
            %% if it finds any transactions still in use, it returnx a badarg error (as there shouldn't 
            %% be any).
            %%
            %% Note: originally this was implemented by checking to see if we have any 'need to resolve' 
            %% untagged event messages in our message queue -- group_valmap_cleanup on client group ids 
            %% will generate these if the client has an unresolved transaction when it was killed.
            %% However, that's a bit of a pain to do for 2 reasons
            %%   1. In pipe mode, the messages need to be converted first (easy to dobut more code).
            %%   2. It fails to count any unresolved txns that were being processed by the 'incremental resolve'
            %%      support process.
            %% Rather than track all of the unresolved txns in State, it's easier just to call some C code to 
            %% do it, especially as if there _are_ any, we need to call some C code to activate the hack.

            NumUnresolvedTxns = ?BDB:make_db_close_safe_if_there_are_unresolved_transactions(BdbPortHandle),
            ?INFO_LOGGER("close_environment: There are ~w prepared-but-unresolved transactions ~n", [NumUnresolvedTxns])
    end,

    ?INFO_LOGGER("close_environment: closing DB handles~n"),

    %% Close all Db handles. 
    %%
    %% Important: we loop over the db_handle_cache and close each handle explicitly rather than doing group_valmap_cleanup 
    %% on the server's group-id, for 2 reasons
    %%
    %% 1.  If group_valmap_cleanup attempts to close a DB handle and finds that there is a prepared-but-unresolved txn open, 
    %%     it HARD-ABORTS the OS process.  group_valmap_cleanup must assume that the process will continue to run -- that 
    %%     clients may attempt to abort or commit the prepared txns, so aborting the process is the only safe course.
    %%
    %% 2.  Closing DB handles is potentially expensive, and while group_valmap_cleanup would not block the Erlang VM 
    %%     (as it uses an async thread), it closes each handle sequentially.  Here we can do them in parallel.
    %%
    %% See comments above on why we call db_get_transactional to decide whether to pass DB_NOSYNC.

    DbCloseTags = dict:fold(fun(DBId, {DB, _ExistingDBType, _ExistingDBConfigProplist}, DbCloseTags) ->

                                    %% (Note: db_get_transactional is a non-async operation (blocks Erlang VM, but is 
                                    %% just a single flag test), so it doesn't matter that we have already shrunk the 
                                    %% worker threadpool to 0.)
                                    Flags = case ?BDB:db_get_transactional(BdbPortHandle, DB) of 
                                                1 -> [bdb_DB_NOSYNC];
                                                0 -> []
                                            end,

                                    ?INFO_LOGGER("close_environment: closing DB ~300p (with Flags ~w)~n", [DBId, Flags]),
                                    Tag = ?BDB:db_close_non_block(BdbPortHandle, DB, Flags),

                                    [{DBId, Tag} | DbCloseTags]
                            end,
                            [], 
                            DbHandleCache),

    lists:foreach(fun({_DBId, Tag}) ->
                          ?INFO_LOGGER("close_environment: waiting for response to db_close_non_block for database ~300p~n", [_DBId]),
                          ?BDB:receive_db_close_reply(BdbPortHandle, Tag),
                          ?INFO_LOGGER("close_environment: got reply for db_close_non_block for ~300p~n", [_DBId])
                  end,
                  DbCloseTags),

    %% Close the DbEnv handle.  
    %% It should be the only thing still open, so we do this indirectly via group_valmap_cleanup
    %% on the server's group-valmap-id.
    %% (TODO: calling env_close explicitly would check that all other valmap slots of all types 
    %% are empty, via a hack inserted by berkeley_db.xml. But we have already called 
    %% check_is_safe_for_port_coordinator_to_close_db_handles, which checks that all valmaps 
    %% are empty or owned by ?SERVER_GROUP_VALMAP_ID, so we don't need to do that.)
    %%
    %% It is safe to do this even if NumUnresolvedTxns /= 0.

    %% If replication is enabled, closing the environment might block for the duration of an election.

    PortHandleWithEnvCloseTimeout =
        case get_value_for_key(replication, ConfigProplist) of 
            disabled ->
                %% Replication is disabled, so we use the normal timeout.
                BdbPortHandle;

            ReplicationConfigProplist ->
                ElectionTimeoutMs = get_value_for_key({timeout, election_milliseconds}, ReplicationConfigProplist),
                %%?INFO_LOGGER("close_environment: Replication is enabled so allowing an additional election_milliseconds timeout (~w) when closing environment (total timeout ~w)~n", [ElectionTimeoutMs, ElectionTimeoutMs + NormalReceiveReplyTimeout]),
                BdbPortHandle#bdb_port_handle{receive_reply_timeout = ElectionTimeoutMs + NormalReceiveReplyTimeout}
        end,

    NumServerResources = ?BDB:group_valmap_cleanup(PortHandleWithEnvCloseTimeout, 
                                                   SERVER_GROUP_VALMAP_ID),
    ?INFO_LOGGER("close_environment: cleaned server group valmap id ~w (freed ~w resources)~n", [SERVER_GROUP_VALMAP_ID, NumServerResources]),

    ok.


repeat_until_true(Fun) ->
    case Fun() of 
        true  -> ok;
        false -> repeat_until_true(Fun)
    end.


%% Helper -- see handle_info('EXIT'), and close_environment.
poll_until_group_valmap_cleanup_is_complete(_BdbPortHandle, _GroupValmapId, 0) ->
    %% Give up
    throw({bdb_error, valmap_group_being_cleaned});

poll_until_group_valmap_cleanup_is_complete(BdbPortHandle, GroupValmapId, Timeout) ->
    StartTime = now(),
    ?INFO_LOGGER("poll_until_group_valmap_cleanup_is_complete: cleaning group_valmap_id ~w failed with valmap_group_being_cleaned. Retrying in 1 second.~n", [GroupValmapId]),
    timer:sleep(1000),
    try 
        ?BDB:group_valmap_cleanup(BdbPortHandle, GroupValmapId)
    catch
        {bdb_error, valmap_group_being_cleaned} ->
            poll_until_group_valmap_cleanup_is_complete(BdbPortHandle, GroupValmapId, reduced_timeout(Timeout, StartTime))
    end.

reduced_timeout(CurrTimeout, StartTime) ->
    case CurrTimeout - ?BDB_U:now_diff_milliseconds(now(), StartTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.



%%
%% Support-processes
%%

%% Returns NamedMonitorRefs : [{ProcName, MonitorRef}, ...]
tell_support_procs_to_gracefully_exit(SupportProcs) ->

    lists:foldl(fun(#support_proc{name = Name, pid = Pid, stop_fun = StopFun} = SupportProc, Acc) ->

                        ?INFO_LOGGER("stopping support proc ~w (pid ~w)~n", [Name, Pid]),
                        MonitorRef = erlang:monitor(process, Pid),
                        unlink(Pid), % Important: we nolonger want the exit signal (handle_info will terminate if it receives it)

                        StopFun(SupportProc),

                        [{Name, MonitorRef} | Acc]
                end,
                [],
                SupportProcs).

%% Requires return value from tell_support_procs_to_gracefully_exit/1
wait_for_support_procs_to_exit(NamedMonitorRefs) ->
    lists:foreach(fun({ProcName, MonitorRef}) ->
                          receive
                              {'DOWN', MonitorRef, process, _, Reason} -> 
                                  ?INFO_LOGGER("support proc ~w exited with: ~w~n", [ProcName, Reason])
                          end
                  end,
                  NamedMonitorRefs).


%% Checkpointint and cache-trickling
%%
%% We support several options:
%%
%%   - checkpointing
%%
%%   - cache_trickling
%%
%%   - cache_trickling and checkpointing, both in parallel 
%%     (i.e. using two BDB threads, and up to *twice* rate of dirty-page IO 
%%     set by set_mp_max_write)
%%
%%   - cache_trickling and checkpointing, alternating in a single 
%%     process (so never more than set_mp_max_write rate of dirty-page IO).
%%     ** With replication, txn_checkpoint sleeps for a long time 
%%     ** (e.g. a minute) to allow replicas to finish their dirty-page IO,
%%     ** so this option might not do enough dirty-page IO to avoid 
%%     ** imapcting txns.
%%
%% cache_trickle should be a frequent operation (ideally continuous),
%% used to keep enough pages free in the page-cache that a transaction
%% rarely has to flush a dirty page to disk to free a page.
%%
%% checkpointing is used to bound recovery time, and involves flushing 
%% all dirty pages with LSNs before the current (checkpoint LSN).
%% Checkpoints might take tens of seconds.
%%
%% If both checkpointing and cache_trickle are enabled and
%% do_parallel_trickle_and_checkpoint is *disabled* then we start a
%% single process. The process assumes that cache-trickle is much more
%% frequent than checkpointing, so it does cache-trickles (with the
%% specified period between them) until it discovers that the time
%% since the last checkpoint is bigger than CheckpointPeriodSeconds,
%% when it does a checkpoint instead.
%%
%% When replication is enabled, the best strategy is likely to be:
%%
%%   - set do_parallel_trickle_and_checkpoint to 'enabled'
%%
%%   - set mp_max_write to HALF the maximum allowable rate of dirty-page IO
%%     (unfortunately when txn_checkpoint is sleeping waiting for 
%%     replicas then the rate of IO will drop to half, but it's better 
%%     than nothing)
%%
%%   - run cache-trickling on ALL sites almost continuously with a high 'keep clean' percentage
%%
%%   - run checkpoints almost almost continuously on MASTER ONLY, to bound recovery 
%%     ?BDB_RG uses run_txn_funs_on_BDB_event to call stop_checkpointer_if_running
%%     and start_checkpointer_if_enabled as appropriate.
%%
%%     Checkpoints should only be run on the Master as the master sends 
%%     messages to all all replicas when txn_checkpoint is called
%%     (and when it finishes) so that the replicas do checkpoints in 
%%     parallel (as much as possible) with the master.  If the replicas 
%%     call txn_checkpoint themselves then that would constitute a 3rd 
%%     thread doing dirty-page IO on replicas.
%%
%%     Implementation detail: to avoid the mess of actually stopping and starting
%%     support-processes, these methods simply send messages to 
%%     checkpointer and/or cache_trickler to tell them to forget about 
%%     checkpointing (or to start doing it again).
%%

%% This helper is used by init() to start the initial configuration
init_checkpointer_and_or_cache_trickler(State) ->

    Config = get(config_proplist),
    CheckpointReceiveReplyTimeout = get_value_for_key(checkpoint_receive_reply_timeout_milliseconds, Config),

    case {get_value_for_key( checkpoint_seconds_and_min_kb,          Config),
          get_value_for_key( cache_trickle_milliseconds_and_percent, Config)} of

        {disabled, disabled} ->
            ?INFO_LOGGER("start_checkpointer_and_or_cache_trickler: both checkpointer and cache_trickler are disabled~n"),
            State;

        {CheckpointerArgs, disabled} ->
            ?INFO_LOGGER("start_checkpointer_and_or_cache_trickler: starting ONLY checkpointer~n"),
            start_checkpointer_process(CheckpointerArgs, CheckpointReceiveReplyTimeout, State);

        {disabled, TricklerArgs} ->
            ?INFO_LOGGER("start_checkpointer_and_or_cache_trickler: starting ONLY cache_trickler~n"),
            start_cache_trickler_process(TricklerArgs, CheckpointReceiveReplyTimeout, State);

        {CheckpointerArgs, TricklerArgs} ->
            case get_value_for_key(do_parallel_trickle_and_checkpoint, Config) of 
                enabled ->
                    ?INFO_LOGGER("start_checkpointer_and_or_cache_trickler: starting BOTH checkpointer and cache_trickler (in parallel)~n"),
                    StateTmp = start_checkpointer_process(CheckpointerArgs, CheckpointReceiveReplyTimeout, State),
                    start_cache_trickler_process(TricklerArgs, CheckpointReceiveReplyTimeout, StateTmp);

                disabled ->
                    ?INFO_LOGGER("start_checkpointer_and_or_cache_trickler: starting cache_trickler_checkpointer (sequential)~n"),
                    start_cache_trickler_checkpointer_process(TricklerArgs, CheckpointerArgs, CheckpointReceiveReplyTimeout, State)
            end
    end.


start_cache_trickler_process({TricklePeriodMilliseconds, TrickleCleanPercent},
                             CheckpointReceiveReplyTimeout,
                             #state{bdb_port_handle = BdbPortHandle,
                                    support_procs   = SupportProcs} = State) ->

    %% Note that the support processes should NOT call
    %% acquire_bdb_port_handle() [as all client processes must] as
    %% that would register them as clients of the server, and
    %% terminate would kill them, rather than shutting them down
    %% cleanly.  So we pass our instance of bdb_port_handle, and hack
    %% the issued_to_pid member in process startup.

    %% This support proc is shutdown with exit(kill), not sending a
    %% stop message, because checkpoints might run for a very long
    %% time.  This is safe because the process doesn't open any db
    %% handles, cursors or txns.  When the environment is closed, the
    %% in-progress checkpoint/trickle will be holdint a shared
    %% ref-count on the db_env handle, so dbenv->close won't be called
    %% until it finishes.

    NewSupportProc =
        #support_proc{name      = cache_trickler, 
                      pid       = proc_lib:spawn_link(
                                    fun() -> 
                                            cache_trickler_process(
                                              BdbPortHandle, 
                                              TricklePeriodMilliseconds, 
                                              TrickleCleanPercent,
                                              CheckpointReceiveReplyTimeout)
                                    end),
                      start_fun = fun(_) -> ok end, % no-op
                      stop_fun  = fun(#support_proc{pid = Pid}) -> 
                                          exit(Pid, kill) % safe to kill this while memp_trickle is still running (and we can close DBs etc in parallel)
                                  end},
    State#state{support_procs = [NewSupportProc | SupportProcs]}.

cache_trickler_process(PortCoordBdbPortHandle, PeriodMilliseconds, Percent, CheckpointReceiveReplyTimeout) ->
    ?INFO_LOGGER("cache_trickler: starting (period: ~w milliseconds, percent: ~w, receive_reply_timeout: ~w)~n", [PeriodMilliseconds, Percent, CheckpointReceiveReplyTimeout]),
    cache_trickler_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                                               receive_reply_timeout = CheckpointReceiveReplyTimeout}, 
                        PeriodMilliseconds, 
                        Percent).

cache_trickler_loop(BdbPortHandle, PeriodMilliseconds, Percent) ->

    try 
        ?BDB:memp_trickle(BdbPortHandle, Percent)
    %%of
    %%    _NumFlushedPages ->
    %%    ?INFO_LOGGER("cache_trickler: Successful (flushed ~w pages). Sleeping ~w seconds~n", [NumFlushedPages, PeriodMilliseconds]),
    catch
        {bdb_error, bdb_DB_REP_LOCKOUT} = Err ->
            %% If this flag is being used:
            %%  {replication, 
            %%    [{config_flags, [bdb_DB_REP_CONF_NOWAIT]}]}
            %% then this error might be returned:
            ?INFO_LOGGER("cache_trickler: got ~w~n", [Err]);

        {bdb_error, eagain} = Err ->
            %% During failover, replicas might return this:
            ?INFO_LOGGER("cache_trickler: got ~w~n", [Err]);

        Class:Reason ->
            exit({memp_trickle, {Class, Reason, erlang:get_stacktrace()}})
    end,

    receive
        Other -> exit({cache_trickler, {unexpected_msg, Other}})
    after
        PeriodMilliseconds -> cache_trickler_loop(BdbPortHandle, PeriodMilliseconds, Percent)
    end.


start_checkpointer_process({CheckpointPeriodSeconds, CheckpointMinKBytes},
                           CheckpointReceiveReplyTimeout,
                           #state{bdb_port_handle = BdbPortHandle,
                                  support_procs   = SupportProcs} = State) ->
    Fun = fun() -> 
                  checkpointer_process(
                    BdbPortHandle, 
                    CheckpointPeriodSeconds, 
                    CheckpointMinKBytes,
                    CheckpointReceiveReplyTimeout)
          end,
    NewSupportProc =
        #support_proc{name      = checkpointer, 
                      pid       = proc_lib:spawn_link(Fun),
                      start_fun = fun(_) -> ok end,       % no-op
                      stop_fun  = fun(#support_proc{pid = Pid}) -> 
                                          exit(Pid, kill) % safe to kill this while memp_trickle/txn_checkpoint is still running (and we can close DBs etc in parallel)
                                  end},
    State#state{support_procs = [NewSupportProc | SupportProcs]}.

checkpointer_process(PortCoordBdbPortHandle, PeriodSeconds, MinKBytes, ReceiveReplyTimeout) ->
    ?INFO_LOGGER("checkpointer: starting (period: ~w seconds, min kbytes: ~w, receive_reply_timeout: ~w)~n", [PeriodSeconds, MinKBytes, ReceiveReplyTimeout]),
    checkpointer_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                                             receive_reply_timeout = ReceiveReplyTimeout}, 
                      PeriodSeconds, 
                      MinKBytes).

checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes) ->

    try
        %%?INFO_LOGGER("checkpointer: starting checkpoint~n"),
        ?BDB:txn_checkpoint(BdbPortHandle, MinKBytes, 0, [])
        %%?INFO_LOGGER("checkpointer: finished checkpoint, sleeping ~w seconds~n", [PeriodSeconds]),
    catch
        %% If this flag is being used:
        %%  {replication, 
        %%    [{config_flags, [bdb_DB_REP_CONF_NOWAIT]}]}
        %% then this error might be returned:
        {bdb_error, bdb_DB_REP_LOCKOUT} = Err ->
            ?INFO_LOGGER("checkpointer: got ~w~n", [Err]);
        Class:Reason ->
            exit({txn_checkpoint, {Class, Reason, erlang:get_stacktrace()}})
    end,

    receive
        stop_checkpointing_for_now ->
            ?INFO_LOGGER("checkpointer: told to suspend checkpointing~n"),
            checkpointer_no_checkpointing_loop(BdbPortHandle, PeriodSeconds, MinKBytes);
        
        start_checkpointing_for_now -> checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes);
        Other                       -> exit({checkpointer, {unexpected_msg, Other}})
    after
        PeriodSeconds * 1000        -> checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes)
    end.

checkpointer_no_checkpointing_loop(BdbPortHandle, PeriodSeconds, MinKBytes) ->
    receive
        start_checkpointing_for_now -> 
            ?INFO_LOGGER("checkpointer: told to resume checkpointing after suspension~n"),
            checkpointer_loop(BdbPortHandle, PeriodSeconds, MinKBytes);

        stop_checkpointing_for_now -> checkpointer_no_checkpointing_loop(BdbPortHandle, PeriodSeconds, MinKBytes);
        Other                      -> exit({checkpointer, {unexpected_msg, Other}})
    end.
        


start_cache_trickler_checkpointer_process({TricklePeriodMilliseconds, TrickleCleanPercent},
                                          {CheckpointPeriodSeconds, CheckpointMinKBytes},
                                          CheckpointReceiveReplyTimeout,
                                          #state{bdb_port_handle = BdbPortHandle,
                                                 support_procs   = SupportProcs} = State) ->
    Fun = fun() -> 
                  cache_trickler_checkpointer_process(
                    BdbPortHandle, 
                    TricklePeriodMilliseconds, 
                    TrickleCleanPercent,
                    CheckpointPeriodSeconds, 
                    CheckpointMinKBytes, 
                    CheckpointReceiveReplyTimeout)
          end,
    NewSupportProc =
        #support_proc{name      = checkpointer, 
                      pid       = proc_lib:spawn_link(Fun),
                      start_fun = fun(_) -> ok end, % no-op
                      stop_fun  = fun(#support_proc{pid = Pid}) -> 
                                          exit(Pid, kill) % safe to kill this while memp_trickle/txn_checkpoint is still running (and we can close DBs etc in parallel)
                                  end},
    State#state{support_procs = [NewSupportProc | SupportProcs]}.

cache_trickler_checkpointer_process(PortCoordBdbPortHandle, 
                                    CacheTricklePeriodMilliseconds, 
                                    CacheTrickleCleanPercent,
                                    CheckpointPeriodSeconds, 
                                    CheckpointMinKBytes, 
                                    ReceiveReplyTimeout) ->
    ?INFO_LOGGER("cache_trickler_checkpointer: starting (cache_trickle_period_milliseconds: ~w, cache_trickle_clean_percent: ~w, checkpoint_period_seconds: ~w, checkpoint_min_kbytes: ~w, receive_reply_timeout: ~w)~n", [CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodSeconds, CheckpointMinKBytes, ReceiveReplyTimeout]),
    cache_trickler_checkpointer_loop(
      PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                             receive_reply_timeout = ReceiveReplyTimeout}, 
      CacheTricklePeriodMilliseconds, 
      CacheTrickleCleanPercent,
      CheckpointPeriodSeconds * 1000, 
      CheckpointMinKBytes,
      now()). % LastCheckpointTime -- but we don't actually run a checkpoint at startup, because recovery normally takes care of that

cache_trickler_checkpointer_loop(BdbPortHandle, 
                                 CacheTricklePeriodMilliseconds, 
                                 CacheTrickleCleanPercent,
                                 CheckpointPeriodMilliseconds, 
                                 CheckpointMinKBytes,
                                 LastCheckpointTime) ->

    Now = now(),
    NextCheckpointTime = 
        case ?BDB_U:now_diff_milliseconds(Now, LastCheckpointTime) of

            N when N < CheckpointPeriodMilliseconds ->

                try
                    _NumFlushedPages = ?BDB:memp_trickle(BdbPortHandle, CacheTrickleCleanPercent)
                    %%?INFO_LOGGER("cache_trickler_checkpointer: cache_trickle flushed ~w pages to clean ~w%. Sleeping ~w milliseconds~n", [NumFlushedPages, CacheTrickleCleanPercent, CacheTricklePeriodMilliseconds]),
                catch
                    {bdb_error, bdb_DB_REP_LOCKOUT} = Err ->
                        ?INFO_LOGGER("cache_trickler_checkpointer: memp_trickle got ~w~n", [Err]);
                    {bdb_error, eagain} = Err ->
                        ?INFO_LOGGER("cache_trickler_checkpointer: memp_trickle got ~w~n", [Err]);
                    Class:Reason ->
                        exit({memp_trickle, {Class, Reason, erlang:get_stacktrace()}})
                end,

                LastCheckpointTime;

            _ ->
                try
                    %%?INFO_LOGGER("cache_trickler_checkpointer: starting checkpoint~n"),
                    ?BDB:txn_checkpoint(BdbPortHandle, CheckpointMinKBytes, 0, [])
                    %%?INFO_LOGGER("cache_trickler_checkpointer: finished checkpoint.~n"),
                catch
                    {bdb_error, bdb_DB_REP_LOCKOUT} = Err ->
                        ?INFO_LOGGER("cache_trickler_checkpointer: txn_checkpoint got ~w~n", [Err]);
                    Class:Reason ->
                        exit({txn_checkpoint, {Class, Reason, erlang:get_stacktrace()}})
                end,

                %% We set NextCheckpointTime to the 'Now' from **before** the checkpoint was started,
                %% because the purpose of checkpoints is to bound recovery time.  
                %% If checkpoints are taking longer than the speicified period 
                %% (quite likely) then we _do_ want to run them continuously. We don't want to leave 
                %% a full period between the end of the previous checkpoint and the 
                %% start of the next, as clearly the application is writing data 
                %% very fast, and so will accumulate a lot before the next checkpoint.
                %% (Note that new data written after a checkpoint starts is potentially
                %% not part of that checkpoint.)
                Now
        end,

    receive
        stop_checkpointing_for_now ->
            ?INFO_LOGGER("cache_trickler_checkpointer: told to suspend checkpointing~n"),
            cache_trickler_checkpointer_only_trickle_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds,  CheckpointMinKBytes,
                                                          NextCheckpointTime);

        start_checkpointing_for_now ->
            cache_trickler_checkpointer_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds, CheckpointMinKBytes, 
                                             NextCheckpointTime);

        Other -> exit({cache_trickler_checkpointer, {unexpected_msg, Other}})
    after
        %% We don't bother working out if LastCheckpointTime is closer than CacheTricklePeriodMilliseconds 
        %% -- we just assume that the latter is much smaller than CheckpointPeriodMilliseconds so it is approximately
        %% ok to sleep for the smaller value.
        CacheTricklePeriodMilliseconds ->
            cache_trickler_checkpointer_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds, CheckpointMinKBytes, 
                                             NextCheckpointTime)
    end.

cache_trickler_checkpointer_only_trickle_loop(BdbPortHandle, 
                                              CacheTricklePeriodMilliseconds, 
                                              CacheTrickleCleanPercent,
                                              CheckpointPeriodMilliseconds, 
                                              CheckpointMinKBytes,
                                              LastCheckpointTime) ->

    try
        _NumFlushedPages = ?BDB:memp_trickle(BdbPortHandle, CacheTrickleCleanPercent)
        %%?INFO_LOGGER("cache_trickler_checkpointer_only_trickle: Successful (flushed ~w pages). Sleeping ~w seconds~n", [NumFlushedPages, CacheTricklePeriodMilliseconds]),
    catch
        {bdb_error, bdb_DB_REP_LOCKOUT} = Err->
            ?INFO_LOGGER("cache_trickler_checkpointer: memp_trickle got ~w~n", [Err]);
        {bdb_error, eagain} = Err ->
            ?INFO_LOGGER("cache_trickler_checkpointer: memp_trickle got ~w~n", [Err]);
        Class:Reason ->
            exit({memp_trickle, {Class,Reason, erlang:get_stacktrace()}})
    end,

    receive
        start_checkpointing_for_now ->
            ?INFO_LOGGER("cache_trickler_checkpointer: told to resume checkpointing after suspension~n"),
            cache_trickler_checkpointer_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds, CheckpointMinKBytes, LastCheckpointTime);

        stop_checkpointing_for_now -> 
            cache_trickler_checkpointer_only_trickle_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds, CheckpointMinKBytes, LastCheckpointTime);

        Other -> exit({cache_trickler_checkpointer, {unexpected_msg, Other}})
    after
        CacheTricklePeriodMilliseconds ->
            cache_trickler_checkpointer_only_trickle_loop(BdbPortHandle, CacheTricklePeriodMilliseconds, CacheTrickleCleanPercent, CheckpointPeriodMilliseconds, CheckpointMinKBytes, LastCheckpointTime)
    end.


%% These helpers are exposed as part of the public API (to be used via run_txn_funs_on_BDB_event)
%% return: State
server_start_checkpointer_if_enabled(State) ->

    case list_running_support_procs_with_names([checkpointer, cache_trickler_checkpointer], State) of
        [#support_proc{pid = Pid}] -> 
            Pid ! start_checkpointing_for_now;
        _ -> 
            ?INFO_LOGGER("start_checkpointer_if_enabled: nothing to do~n")
    end,
    State.

server_stop_checkpointer_if_running(State) ->

    case list_running_support_procs_with_names([checkpointer, cache_trickler_checkpointer], State) of
        [] ->
            ?INFO_LOGGER("stop_checkpointer_if_running: nothing to do~n");
        [#support_proc{pid = Pid}] -> 
            Pid ! stop_checkpointing_for_now
    end,
    State.


list_running_support_procs_with_names(Names, #state{support_procs = SupportProcs}) ->
    lists:filter(fun(#support_proc{name = Name}) -> lists:member(Name, Names) end,
                 SupportProcs).




%%
%% Other support_procs
%%

deadlock_detector_process(PortCoordBdbPortHandle, InitialDelayMilliseconds, PeriodMilliseconds, Policy, ReceiveReplyTimeout) ->
    %% this process does have a start_fun to delay it's first operation.
    receive go -> 
            ?INFO_LOGGER("deadlock_detector_process starting (initial delay ~w milliseconds, period: ~w milliseconds, receive_reply_timeout: ~w)~n", [InitialDelayMilliseconds, PeriodMilliseconds, ReceiveReplyTimeout]),
            deadlock_detector_loop(PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid         = self(),
                                                                          receive_reply_timeout = ReceiveReplyTimeout}, 
                                   PeriodMilliseconds, 
                                   Policy,
                                   InitialDelayMilliseconds) % initial delay is longer to let startup settle down, or we occasionally get receive_reply_timeout
    end.

deadlock_detector_loop(BdbPortHandle, PeriodMilliseconds, Policy, ReceiveReplyTimeout) ->
    receive
        stop  -> ?INFO_LOGGER("deadlock_detector exiting~n");
        Other -> exit({deadlock_detector, {unexpected_msg, Other}})
    after
        ReceiveReplyTimeout ->
            try ?BDB:lock_detect(BdbPortHandle, [], Policy) of
                0 -> 
                    ok;
                NumBrokenLocks ->
                    ?INFO_LOGGER("deadlock_detector_process: broke ~w locks~n", [NumBrokenLocks])
            catch
                 %% If this flag is being used:
                 %%  {replication, 
                 %%    [{config_flags, [bdb_DB_REP_CONF_NOWAIT]}]}
                 %% then this error might be returned:
                 {bdb_error, bdb_DB_REP_LOCKOUT} = Err ->
                    ?INFO_LOGGER("deadlock_detector_process: got ~w~n", [Err]);
                 Class:Reason ->
                    exit({lock_detect, {Class,Reason, erlang:get_stacktrace()}})
            end,
            deadlock_detector_loop(BdbPortHandle, PeriodMilliseconds, Policy, PeriodMilliseconds)
    end.


%% This calls txn_commit and txn_abort on prepared transactions that
%% were once owned by now-dead clients (group_valmap_cleanup sends an
%% untagged event to the port coordinator for these, and the port
%% coordinator sends them to us.)
%%
%% Note: if commit or abort fail due to being selected to resolve a
%% deadlock then we die, and kill the entire port_coordinator.  There
%% is nothing else we can do -- if we lose this txn handle, we have to
%% run recovery so we can run txn_recover.
incrementally_resolve_prepared_txns_process(PortCoordBdbPortHandle, DistributedTxnConfigList) ->
    receive go -> 
            ?INFO_LOGGER("incrementally_resolve_prepared_txns starting~n"),

            #?CS{persist_read_fun = GTM_RgCallerStatePersistReadFun} = 
                get_value_for_key(gtm_rg_caller_state, DistributedTxnConfigList),

            incrementally_resolve_prepared_txns_loop(
              PortCoordBdbPortHandle#bdb_port_handle{issued_to_pid = self()}, 
              GTM_RgCallerStatePersistReadFun,
              DistributedTxnConfigList)
    end.

incrementally_resolve_prepared_txns_loop(BdbPortHandle, GTM_RgCallerStatePersistReadFun, DistributedTxnConfigList) ->
    receive
        {FromPid, stop} -> 
            FromPid ! {self(), stopped},
            ?INFO_LOGGER("incrementally_resolve_prepared_txns exiting~n");

        {need_to_resolve, <<_:?DB_XIDDATASIZE/binary>> = DistributedTxnGUID, Txn} ->

            %% While we were processing the last batch (or stuck in
            %% resolve_distributed_txns_with_retry waiting for the GTM
            %% to come online), several need_to_resolve messages may have arrived.
            %% As resolving them involves a round-trip to the GTM, we want to accumulate a batch 
            %% rather than doing each one individually,            
            ListDistributedTxnGUIDsTxns = accumulate_batch_to_resolve([{DistributedTxnGUID, Txn}]),

            %% TxnGUIDs is the list that is actually sent to the GTM.
            %% As there may be more than one local client participating in the distributed transaction,
            %% is is possible that we may be asked to resolve the same GUID more than once.
            %% We don't need to send more than one copy of each GUID to the GTM,
            %% so we remove duplicates.
            TxnGUIDs = lists:usort( lists:map(fun({TxnGUID, #bdb_db_txn{}}) -> TxnGUID end, 
                                              ListDistributedTxnGUIDsTxns) ),

            %%?INFO_LOGGER("incrementally_resolve_prepared_txns batch ~p~nGUIDS ~p~n", [ListDistributedTxnGUIDsTxns, TxnGUIDs]),

            resolve_distributed_txns_with_retry(BdbPortHandle, 
                                                ListDistributedTxnGUIDsTxns, 
                                                TxnGUIDs,
                                                GTM_RgCallerStatePersistReadFun, 
                                                DistributedTxnConfigList, 
                                                1), % AttemptNumber

            incrementally_resolve_prepared_txns_loop(BdbPortHandle, GTM_RgCallerStatePersistReadFun, DistributedTxnConfigList);        

        Other -> exit({incrementally_resolve_prepared_txns, {unexpected_msg, Other}})
    end.

accumulate_batch_to_resolve(Acc) ->
    receive
        {need_to_resolve, <<_:?DB_XIDDATASIZE/binary>> = DistributedTxnGUID, Txn} ->
            accumulate_batch_to_resolve([{DistributedTxnGUID, Txn} | Acc])
    after
        %% TODO: we could wait for a few milliseconds, to see if any more arrive.
        %% However that is potentially dangerous -- if they keep arriving in dribs and 
        %% drabs we might hold on to transactions that need resolving for significant 
        %% time, rather than resolving them expediently. So we'd need to limit how 
        %% long we wait in total between resolving each batch.
        %% So for now I'm just going to process whatever is already in the queue,
        %% and not wait at all.
        0 ->
            Acc
    end.


%% TxnFuns to call the Global Transaction Manager in the style of a 'weakly replicated application' 
%% (see comments at top of berkeley_db_global_transaction_manager.erl 
%% and for txn_fun_gtm_* in berkeley_db_replication_group_helpers.erl)
txn_fun_gtm_resolve_distributed_txn(_BdbPort, DistributedTxnGUID) ->

    {ReplicationGroupName, SiteName} = get(txn_fun_replication_context),
    GTM_ServerName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(ReplicationGroupName, 
                                                                             SiteName),
    ?BDB_GTM:resolve(GTM_ServerName, DistributedTxnGUID).


%% Helper function for distributed transaction support.
%% TxnRecoverFlags should be bdb_DB_FIRST when called from outside.
%% Returns nothing useful. Exits or throws if there is a problem.
%%
%% If replication is disabled this is called after env_open.
%% If replication is enabled this is called when DB_EVENT_REP_MASTER is received.
do_distributed_transaction_recovery(BdbPortHandle, TxnRecoverFlags, DistributedTxnConfigList) ->
    
    #?CS{persist_read_fun = GTM_RgCallerStatePersistReadFun} 
        = get_value_for_key(gtm_rg_caller_state, DistributedTxnConfigList),

    case ?BDB:txn_recover(BdbPortHandle, TxnRecoverFlags) of
        [] ->
            ?INFO_LOGGER("do_distributed_transaction_recovery: Nothing more to do.~n"),
            ok;

        ListDistributedTxnGUIDsTxns ->
            %%?INFO_LOGGER("do_distributed_transaction_recovery: Need to resolve ~p~n", [ListDistributedTxnGUIDsTxns]),

            %% Check the results from txn_recover.
            %% e.g. If it could not allocate memory it puts an atom in the list
            TxnGUIDs =
                lists:map(fun({<<_:?DB_XIDDATASIZE/binary>> = DistributedTxnGUID, #bdb_db_txn{}}) -> 
                                  DistributedTxnGUID;
                             (_) ->
                                  exit({txn_recover_returned_malformed_results, ListDistributedTxnGUIDsTxns})
                          end, 
                          ListDistributedTxnGUIDsTxns),

            %% Ask GTM for a decision on all of the TxnGUIDs.
            %% If we cannot contact the master GTM then this blocks (doing periodic retries) until it does succeed.
            resolve_distributed_txns_with_retry(BdbPortHandle, 
                                                ListDistributedTxnGUIDsTxns, 
                                                TxnGUIDs, 
                                                GTM_RgCallerStatePersistReadFun, 
                                                DistributedTxnConfigList, 
                                                1), % AttemptNumber

            %% Repeat until txn_recover returns an empty list
            do_distributed_transaction_recovery(BdbPortHandle, bdb_DB_NEXT, DistributedTxnConfigList)
    end.


resolve_distributed_txns_with_retry(BdbPortHandle, 
                                    ListDistributedTxnGUIDsTxns, 
                                    TxnGUIDs, 
                                    GTM_RgCallerStatePersistReadFun, 
                                    DistributedTxnConfigList, 
                                    AttemptNumber) ->

    %% GTM state might have changed
    GTM_RgCallerState = GTM_RgCallerStatePersistReadFun(),

    ?INFO_LOGGER("resolve_distributed_txns_with_retry: Asking GTM to resolve ~p~n",
                 [lists:map(fun({<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>>, #bdb_db_txn{index = TxnValmapIdx}}) -> 
                                    {TxnCounter, TxnValmapIdx} end,
                            ListDistributedTxnGUIDsTxns)]),

    try
        ?BDB_RG_H:do_txn_on_master(GTM_RgCallerState, 
                                   {?MODULE, txn_fun_gtm_resolve_distributed_txn, TxnGUIDs}, 
                                   0,                                                                       % NumDeadlocRetries is 0 as this TxnFun only calls the GTM, it does not do a BDB transaction itself
                                   get_value_for_key(gtm_timeout_milliseconds, DistributedTxnConfigList))
    of
        {ListTxnResolutions, _GTM_RgCallerState2} ->

            %%?INFO_LOGGER("resolve_distributed_txns_with_retry: Got resolutions from GTM ~p~n", [ListTxnResolutions]),

            %% Loop over GTM results and call txn_commit or txn_abort on each txn handle that 
            %% is involved in that GUID.  
            %% Important: if there are more than one local clients participating in the distributed txn
            %% then there may be more than 1 transaction handle for each GUID.
            %%
            %% When using replication, a new master will call this code **before** it starts it's 
            %% probe_to_sync_replicas thread (attempts to get out of the waiting_for_durability state).
            %% It is therefore possible (even likely) that txn_commit/txn_abort will throw 
            %% NOTPERM errors.   We catch and discard these.  We *have* to resolve all prepared 
            %% txns before exiting from this function, and the txns will be made permanent 
            %% before the master is exposed to clients (transitions to the 'available' state).
            %% And if the txns are not made durable, and this master fails, then the next 
            %% recovered master will simply find the prepared version and repeat this code (ask 
            %% the GTM for a decision).
            

            lists:foreach(
              fun({ok, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, committed}}) -> 
                      lists:foreach(fun(#bdb_db_txn{index = TxnValmapIdx} = Txn) -> 
                                            ?INFO_LOGGER("resolve_distributed_txns_with_retry: GTM says to commit dist. txn ~w, so committing txn valmap idx ~w~n", [TxnCounter, TxnValmapIdx]),
                                            try ?BDB:txn_commit(BdbPortHandle, Txn, []) 
                                            catch
                                                {bdb_error, bdb_DB_REP_NOTPERM} -> ok
                                            end
                                    end,
                                    find_txn_handles_for_GUID(DistributedTxnGUID, ListDistributedTxnGUIDsTxns, ListTxnResolutions));

                 ({ok, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>> = DistributedTxnGUID, aborted}}) -> 
                      lists:foreach(fun(#bdb_db_txn{index = TxnValmapIdx} = Txn) -> 
                                            ?INFO_LOGGER("resolve_distributed_txns_with_retry: GTM says to abort dist. txn ~w, so aborting txn valmap idx ~w~n", [TxnCounter, TxnValmapIdx]),
                                            try ?BDB:txn_abort(BdbPortHandle, Txn) 
                                            catch
                                                {bdb_error, bdb_DB_REP_NOTPERM} -> ok
                                            end
                                    end,
                                    find_txn_handles_for_GUID(DistributedTxnGUID, ListDistributedTxnGUIDsTxns, ListTxnResolutions));

                 ({error, {<<TxnCounter:64, 0:((?DB_XIDDATASIZE*8)-64)>>, _}} = Error) -> 
                      %% This is a catstrophic error
                      ?ERROR_LOGGER("resolve_distributed_txns_with_retry: GTM:resolve returned error ~p for dist. txn ~w. Exiting!~n", [Error, TxnCounter]),
                      exit({global_transaction_manager_resolve_returned_error, Error})
              end,
              ListTxnResolutions),

            ?INFO_LOGGER("resolve_distributed_txns_with_retry: done current batch~n")
    catch
        Exception ->

            RetryPeriod = get_value_for_key(resolve_retry_period_milliseconds, DistributedTxnConfigList),
            ?ERROR_LOGGER("resolve_distributed_txns_with_retry: Attempt #~w failed with ~p when attempting to resolve ~p using GTM ~p~n", [AttemptNumber, Exception, ListDistributedTxnGUIDsTxns, GTM_RgCallerState]),

            case get_value_for_key(resolve_max_num_attempts, DistributedTxnConfigList) of

                MaxNumAttempts when AttemptNumber < MaxNumAttempts ->
                    timer:sleep(RetryPeriod),
                    ?INFO_LOGGER("resolve_distributed_txns_with_retry: Retrying in ~w milliseconds~n", [RetryPeriod]),
                    resolve_distributed_txns_with_retry(BdbPortHandle, 
                                                        ListDistributedTxnGUIDsTxns, 
                                                        TxnGUIDs, 
                                                        GTM_RgCallerStatePersistReadFun, 
                                                        DistributedTxnConfigList, 
                                                        AttemptNumber + 1);

                MaxNumAttempts ->
                    ?ERROR_LOGGER("resolve_distributed_txns_with_retry: Hit retry limit. Exiting~n"),
                    exit({resolve_distributed_txns_hit_retry_limit, MaxNumAttempts, RetryPeriod})
            end
   end.

find_txn_handles_for_GUID(DistributedTxnGUID, ListDistributedTxnGUIDsTxns, ListTxnResolutions) ->
    case [Txn || {GUID, Txn} <- ListDistributedTxnGUIDsTxns, GUID =:= DistributedTxnGUID] of
        [] -> exit({global_transaction_manager_resolve_results_are_missing_an_entry, ListDistributedTxnGUIDsTxns, ListTxnResolutions});
        L  -> L
    end.



%%
%% General helper functions
%%

%% This works for assoc-lists like config_proplist and the top-level of our BDB records.
%% We use lists:keysearch is it is a bif (faster), rather than proplists:lookup.
get_value_for_key(Key, AssocList) ->
    case lists:keysearch(Key, 1, AssocList) of
        {value, {Key, Value}} -> Value;
        false                 -> throw({missing_required_key, Key, AssocList})
    end.

dynamic_replication_gen_event_proc_name() ->
    {registered_name, PortCoordinatorProcessName} = process_info(self(), registered_name),
    list_to_atom(atom_to_list(PortCoordinatorProcessName) ++ ?REPLICATION_GEN_EVENT_PROCESS_NAME_SUFFIX).

get_port_coordinator_proc_name() ->
    {registered_name, Pid} = process_info(self(), registered_name),
    Pid.

get_replication_gen_event_pid(#state{support_procs = SupportProcs}) ->
    {value, #support_proc{pid = Pid}} = lists:keysearch(replication_gen_event, #support_proc.name, SupportProcs),
    Pid.


