
-module('berkeley_db_test').

%% This includes all of Scott's original tests, plus tests for the new features 
%% (normal & distributed transactions, replication, 'statistics' APIs etc etc).
%%
%% This file also contains some basic performance & stress tests (commented out 
%% as they take a long time to run).  But most stress & fault-injection testing 
%% (which was extensive) was done against a server application that uses 
%% this driver, and which is not included here.
%%
%% The 'TODO' file describes a few missing tests.
%%
%% CAVEAT: some tests (e.g. replication tests) contain timeouts which
%% may vary by host type, or even across test runs.  I've tried to
%% make them fairly robust, but you may occasionally get false
%% failures; e.g. 'no_master' exceptions, or an assert that there were
%% unexpected messages in a message queue when the test ends (I find
%% it very useful to assert this at the end of tests), or sometimes
%% the replication tests elect an unexpected site as Master (doesn't
%% necessarily show a problem in the code).  
%% If this happens then sometimes the tests (especially the replication
%% tests) may appear to continue running after the test driver process
%% crashes due to the test failure. This can be spotted by the lack of
%% progress, and probably repetitive pattern of logging, or stalled
%% logging (simply nothing new happening).  If you break the test and
%% grep the log for an error report (line beginning with "** exited:")
%% you will find the reason.  If this happens a lot on your machine,
%% searh for timer:sleep() calls in this file and try increasing the
%% timeouts.
%%
%% Running the tests:
%%
%% The old 'make regression' mechanism still works (but see comments in this 
%% file and in the 'TODO' file about problems with 'erl -s module function').  
%%
%% To capture logs, run the tests like this:
%%
%%    rm regression.out; make regression 2>&1 | tee regression.out
%%
%% To debug a problem with the driver I usually use tracing.
%% The pman utility that comes with Erlang/OTP makes this very easy:
%%
%%    - write a test that reproudces the problem
%%
%%    - start a node/shell that is ready to load the test and driver, e.g.
%%
%%         rm regression.out; make start_debug_node1 2>&1 | tee regression.out
%%    
%%    - start a 'suspended' test, e.g:
%%
%%         Pid = spawn_link(fun() -> receive go -> berkeley_db_test:my_new_test() end end).
%%
%%    - Run pman:start().  
%%
%%      [First time, go to File->Options. Enable everything; trace on send+receive+functions+events, 
%%       inherit on 'All spawns' and 'All links', and trace-to-file.]
%%
%%    - In the pman GUI, select the Pid of the suspended test (to begin tracing).
%%
%%    - start the test with  "Pid ! go."
%%
%% The regression.log and test trace usually show what is wrong.
%%
%% For internal driver problems you may want to enable verbose logging.
%% Add lines like this to the test, after opening/acquiring the BDB port.
%%   
%%       ?BDB:adjust_port_logging(BdbPort, enable, [all]),
%%       ?BDB:adjust_driver_log_level(BdbPort, enable, [all]),  % often not necessary
%%


-define(BDB,       berkeley_db_drv).
-define(BDB_H,     berkeley_db_helpers).
-define(BDB_U,     berkeley_db_utilities).
-define(BDB_PC,    berkeley_db_port_coordinator).
-define(BDB_RG,    berkeley_db_replication_group).
-define(BDB_RG_H,  berkeley_db_replication_group_helpers).
-define(BDB_SEQ,   berkeley_db_sequence_server).
-define(BDB_GTM,   berkeley_db_global_transaction_manager).
-define(BDB_MSG_Q, berkeley_db_msg_queue).

-include("berkeley_db_replication_group_helpers.hrl").
-define(CS,        berkeley_db_replication_group_caller_state).

-include("berkeley_db_drv.hrl").
-include("berkeley_db_msg_queue.hrl").
-include("berkeley_db_replication_group_internal.hrl").

-compile(export_all).   % For test purposes only.

-export([regression/0, leak_test/0, example_replication_site/3, example_distributed_txn/1]).
-export([performance_test_1/1]).

%% these have to be exported to be used with ?BDB_RG:do_txn_on_master
-export([txn_fun_write_many_large_txn_and_return_before_committing/2]).
-export([test_txn_fun_put_1/2, test_txn_fun_crash_in_a_txn/2, test_txn_fun_exit_in_a_txn/2, test_txn_fun_return_with_open_resources/2, test_txn_fun_get/2]).
-export([deadlock_do_txn_on_master_txn1/2, deadlock_do_txn_on_master_txn2/2, deadlock_do_txn_on_master_verify/2]).
-export([txn_fun_distributed_txn_participant_group_A/2, txn_fun_distributed_txn_participant_group_B/2, txn_fun_distributed_txn_participant_throw/2]).
-export([txn_fun_report_which_site_executes/2]).

%% gen_event callbacks (for test_replication_election)
-export([init/1, handle_event/2, terminate/2]).

-define(AT, io:format("~w At line ~w~n", [self(), ?LINE])).   %% remove with regex    ^ *\?AT, *^J


%% trap_exit handling.  
%% These tests may be executed in different ways 
%%   1. "erl -s berkeley_db_test regression" (e.g. 'make regression' does this - the normal way of running the tests)
%%   2. From the shell with berkeley_db_test:regression().
%%   3. From a spawned process (possibly spawned from the shell)
%%
%% In #1,       trap_exit is ***TRUE*** by default for some reason.  (VERY BAD!)
%% In #2 and #3 trap_exit is false by default (as expected)
%%
%% So for consistency, we DISABLE trap_exits here (so a crashing process will make the test fail).
%% (An alternative would be to always spawn a process to run the tests -- that would give us a shell 
%% to examine things if the test process hung. But we would have to synchronize with it, as we often do
%%  erl -s <some test> -s erlang halt, so the VM would exit immediately if the main process did not block.).
%%
%% If individual tests need to ensure that trap_exit is true (e.g. if a spawned process is expected to crash, and we don't 
%% want it to kill the test driver) then then should preserve the original state of trap_exit, and restore it afterwards.
%%
%% If tests want to ensure that spawned processes have executed normally, they should use erlang:monitor(). 

regression() ->

    io:format("\nSetting trap_exit to FALSE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, false),

    %% This returns {{Year, Month, Day}, {Hour, Minute, Second}}
    %% Year = Month = Day = Hour = Minute = Second = int()
    TestStartTime = erlang:localtime(),

    ok = regression2(start),
    io:format("\nAll regression tests via linked-in style PASSED.\n"),

    ok = regression2(start_pipe),
    io:format("\nAll regression tests via pipe style PASSED.\n"),

    io:format("\nAll regression tests PASSED.\n"),

    io:format("\nRestoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    false = process_flag(trap_exit, OldTrapExitSetting),

    [] = list_coredump_files_created_since(TestStartTime),

    ok.

%%
%% Note: all of test calls to env_open and db_open do NOT set DB_THREAD 
%% because the drive code forces it (as the XML file declares that those 
%% valmap types support concurrent use).
%%

%%
%% In pipe-mode, The erlang call to ?BDB:shutdown(P) returns before pipe-main has finished executing _stop().
%% So when Erlang then immediately calls ?BDB:start() on the same environment, we get chaos (environment panics,
%% possibly crashes)
%%
%% We can use DB_REGISTER to prevent that:
%%    DB_REGISTER uses a mutex in the environment to prevent that. 
%%    DB_REGISTER also detects if recovery was needed and not specified (so we always specify it).
%%  
%% HOWEVER, DB_REGISTER prevents a single OS process from opening more than one DbEnv handle.
%% That breaks the "each Erlang process uses it's own port instance" model
%% So we can't force DB_REGISTER in the driver code -- the Erlang side has to specifiy it 
%% if it wants it.
%% All env_open calls in this test suite (that do not specify DB_PRIVATE) specify
%% DB_REGISTER and DB_RECOVER.
%%

regression2(StartFunc) ->

    io:format("\nPerform berkeley_db_utilties:test()\n"),
    ?BDB_U:test(),

    io:format("\nPerform various environment tests\n"),
    Port1 = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(Port1, enable, [all]),
    %% TODO This currently crashes the driver
    %%ok = ?BDB:env_remove(Port1, "", []),
    normal = ?BDB:shutdown(Port1),

    Port2 = ?BDB:StartFunc(),
    %%BDBVersion = ?BDB:db_version(Port2),
    ok = env_tests(Port2),
    normal = ?BDB:shutdown(Port2),

    io:format("\nPerform DB table operations tests\n"),
    Port3 = ?BDB:StartFunc(),
    ok = ?BDB:env_open(Port3, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644),
    %%?BDB:adjust_port_logging(Port3, enable, [all]),
    ok = db_ops_tests(Port3),
    ok = db_ops_tests2(Port3),
    ok = ?BDB:env_close(Port3, []),
    normal = ?BDB:shutdown(Port3),

    io:format("\nPerform DB simple cursor tests\n"),
    Port4 = ?BDB:StartFunc(),
    ok = ?BDB:env_open(Port4, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644),
    ok = cursor_tests(Port4),
    normal = ?BDB:shutdown(Port4),

    io:format("\nPerform DB lock tests\n"),
    Port5 = ?BDB:StartFunc(),
    ok = lock_tests1(Port5),
    normal = ?BDB:shutdown(Port5),
    Port6 = ?BDB:StartFunc(),
    ok = lock_tests2(Port6),
    normal = ?BDB:shutdown(Port6),

    ok = test_get_set_threadpool_info(StartFunc),

    ok = test_get_set_num_valmap_slots(StartFunc),

    ok = test_basic_transactions(StartFunc),

    ok = test_basic_deadlock_handling(StartFunc),

    ok = test_recovery(StartFunc),

    ok = test_attempt_reuse_valmap_tuples_after_closing_them(StartFunc),

    ok = test_nested_transactions(StartFunc),

    ok = test_stop_with_nested_transactions(StartFunc),

    %% All tests below this point assume that multiple async worker threads are available
    ok = test_concurrent_access_to_threadsafe_valmaps(StartFunc),

    ok = test_group_valmap_cleanup(StartFunc),

    ok = test_consume_wait(StartFunc),

    ok = test_use_of_one_port_instance_from_multiple_erlang_processes(StartFunc),

    ok = test_basic_port_coordinator(StartFunc),

    ok = test_environment_locking(StartFunc),

    %% Not yet fully implemented (it does full tests but relies on a human to look at the output; it really needs to check checkpoint-LSN and dirty-pages-trickled stats)
    ok = test_basic_port_coordinator_checkpoint_and_cache_trickle(StartFunc),

    ok = test_port_coordinator_threadpool_deadlock(StartFunc),

    ok = test_port_coordinator_multiple_client_processes(StartFunc),

    ok = test_port_coordinator_client_cleanup(StartFunc),

    %% Not yet fully implemented
    %%ok = test_port_coordinator_overlapping_group_valmap_cleanup(StartFunc),

    ok = test_port_coordinator_shutdown_with_active_clients(StartFunc),

    %% This one takes significant time and doesn't test much, so it is off by default
    %%ok = stress_test_allocation_of_binaries(StartFunc),

    ok = test_sequence_server(StartFunc),

    ok = test_msg_queue(StartFunc, 4000), % arg is QRecordLen (large means payload fits in DB_QUEUE but lots of wasted space)
    ok = test_msg_queue(StartFunc, 16),   % small means payload must be split across DB_QUEUE and DB_BTREE

    ok = test_basic_replication(StartFunc),

    ok = test_replication_election(StartFunc),

    ok = test_global_transaction_manager(StartFunc),

    ok = test_basic_distributed_transactions_stop_with_prepared_but_not_in_use_txns(StartFunc),

    ok = test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns(StartFunc),

    ok = test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_SAME_group(StartFunc),

    %% This test is no-longer executed as the condition it checks for is illegal and dangerous but no-longer explicitly tested for
    %% (as the only current check is extremely conservative and became overly restrictive)
    %% See comments in berkely_db.xml (search for 'num_unresolved_txns_owned_by_this_group')
    %%ok = test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_DIFFERENT_group(StartFunc),

    ok = test_basic_distributed_transactions_prepare_finds_group_cleanup_in_progress(StartFunc),

    ok = test_port_coordinator_distributed_transactions_clean_shutdown_with_prepared_txns(StartFunc),

    ok = test_port_coordinator_distributed_transactions_kill_with_prepared_txns(StartFunc),

    ok = test_port_coordinator_distributed_transactions_client_dies_with_prepared_committed_txn(StartFunc),

    %% Not yet implemented
    %%ok = test_port_coordinator_distributed_transactions_client_dies_with_prepared_aborted_txn(StartFunc),
    %%ok = test_port_coordinator_distributed_transactions_client_dies_with_prepared_not_aborted_or_commited_txn(StartFunc),

    ok = test_replication_group_start_and_spawn_respawn(StartFunc),

    ok = test_replication_group_start_sites_no_respawn(StartFunc),

    ok = test_do_txn_on_master_retry_on_deadlock(StartFunc),

    ok = test_db_compact(StartFunc),

    ok = test_replicated_do_distributed_transaction(StartFunc),

    ok = test_replicated_do_distributed_transaction_with_failover(StartFunc),

    ok = test_replication_do_txn_redirect(StartFunc),

    ok = test_replication_perm_failed_events_limit(StartFunc),

    ok = test_replication_waiting_for_durability(StartFunc),

    ok = test_write_coredump(StartFunc),

    ok = test_write_errcall_coredump(StartFunc),

    ok = test_local_do_txn_error_handling(StartFunc),

    ok = test_clone_bdb_port_handle(StartFunc),

    case StartFunc of
        start ->
            ok = test_replication_remove_txn_logs(StartFunc);
        start_pipe ->
            io:format("TODO: test_replication_remove_txn_logs behaves differently in start_pipe mode~n")
    end,

    case StartFunc of
        start ->
            io:format("TODO: test_panic_port_coordinator kills the VM in linked-in mode, so it is not run automatically~n");
        start_pipe ->
            ok = test_panic_port_coordinator(StartFunc)
    end,

    ok = test_driver_stop_while_running_valmap_stop_commands(StartFunc),

    ok = test_c_get_multiple(StartFunc),

    ok = test_env_interrupt_os_sleep(StartFunc),

    ok.


list_coredump_files_created_since(SinceTime) ->    
    filelib:fold_files(".", 
                       "core.*", % regex, not a file system wildcard
                       false,    % not recursive
                       fun(FileName, Acc) ->

                               %% This returns {{Year,Month,Day},{Hour,Min,Sec}} which is directly comparable to erlang:localtime().
                               FileModifiedTime = filelib:last_modified(FileName),

                               if FileModifiedTime >= SinceTime -> [FileName | Acc];
                                  true                          -> Acc
                               end
                       end,
                       []).


%% These macros may be used multiple times in the same function, so they each need to be in their own scope,
%% or we get "Result unsafe" errors at compile time.  Using an immediately-applied lambda solves the problem.
%% TODO: reduce the number of special-cases; this became a bit silly.
-define(EXPECT_THROW(ExpectedException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     ExpectedException -> ok
                 end
         end)()).

-define(EXPECT_EXIT(ExpectedException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_exit, ?LINE, Result})
                 catch
                     exit:ExpectedException -> ok
                 end
         end)()).

-define(EXPECT_BDB_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {bdb_error, ExpectedSubException} -> ok
                 end
        end)()).

-define(EXPECT_BDB_HELPER_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {bdb_helper_error, ExpectedSubException} -> ok
                 end
        end)()).

-define(EXPECT_PORT_COORD_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {bdb_port_coordinator_error, ExpectedSubException} -> ok
                 end
         end)()).

-define(EXPECT_GTM_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {bdb_gtm_error, ExpectedSubException} -> ok
                 end
         end)()).

-define(EXPECT_RG_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {bdb_replication_group_error, ExpectedSubException} -> ok
                 end
         end)()).

-define(EXPECT_RG_SITE_DOWN_ERROR(SiteLocator, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, {bdb_replication_group_error, {site_is_down, SiteLocator, some_reason}}, Result})
                 catch
                     {bdb_replication_group_error, {site_is_down, SiteLocator, _}} -> ok
                 end
         end)()).

-define(EXPECT_RG_NO_SITES_ERROR(FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, {bdb_replication_group_error, {no_sites_available, #?CS{}}, Result}})
                 catch
                     {bdb_replication_group_error, {no_sites_available, #?CS{}}} -> ok
                 end
         end)()).

-define(EXPECT_TXN_FUN_ERROR(ExpectedSubException, FunCall),
        (fun() ->
                 try FunCall of
                     Result -> throw({expected_throw, ?LINE, Result})
                 catch
                     {txn_fun_error, ExpectedSubException} -> ok
                 end
         end)()).

-define(EXPECT_RECEIVE_REPLY_TIMEOUT(ReceiveReplyFuncName, BdbPort, TagBin),
        (fun() ->
                 try ?BDB:ReceiveReplyFuncName(BdbPort, TagBin) of
                     Result -> throw({expected_receive_reply_timeout, ?LINE, {BdbPort, TagBin}, Result})
                 catch
                     exit:{bdb_receive_reply_timeout, {BdbPort, TagBin}} -> ok
                 end
         end)()).


env_tests(P) ->
    %% We only support one environment per port
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_dbenv}, ?BDB:env_create(P, [])),

    ok = ?BDB:env_set_cachesize(P, 0, 50 * 1024 * 1024, 1),	 % #gigabytes, #bytes, #fragments

    {GB, Bytes, 1} = ?BDB:env_get_cachesize(P),
    %% get_cache_size may return a larger amount.
    if (GB * 1024 * 1024 * 1024) + Bytes < (50 * 1024 * 1024) ->
	    throw(env_set_cachesize_error);
       true -> ok
    end,

    DefaultMaxTxns = ?BDB:env_get_tx_max(P),
    if DefaultMaxTxns < 20 -> throw(env_get_cachesize_error);
       true                -> ok
    end,
    ok = ?BDB:env_set_tx_max(P, ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS),
    NewMaxTxns = ?BDB:env_get_tx_max(P),
    if NewMaxTxns < ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS -> throw(env_set_cachesize_error);
       true                                              -> ok
    end,


    %% DB_ENV->set_data_dir must be called before DB_ENV->open
    ok = ?BDB:env_set_data_dir(P, "."),
    ok = ?BDB:env_set_shm_key(P, 42),
    ok = ?BDB:env_open(P, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644),
    ok = env_tests_misc(P),


    %% We've set various weird env flags, don't use this environment now!
    ok = ?BDB:env_close(P, []),

    %% cnewcom: This was in Scott's original tests, where it was causing undefined behaviour as the handle is already closed.
    %% Valmap generation-ids make this safe now.
    ?EXPECT_BDB_ERROR(badarg, ?BDB:env_close(P, [])),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

env_tests_misc(P) ->
    %% XXX Won't test DB_ENV->set_errfile yet.

    ok = ?BDB:env_set_flags(P, [bdb_DB_NOLOCKING, bdb_DB_NOMMAP], 1),

    ok = ?BDB:env_set_tas_spins(P, 42),
    ok = ?BDB:env_set_tmp_dir(P, "."),
    ok = ?BDB:env_set_verbose(P, bdb_DB_VERB_WAITSFOR, 1),
    _Bin1 = ?BDB:db_strerror(P, 1),
    {_Major, _Minor, _Patch, _Bin2} = ?BDB:db_version(P),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

db_ops_tests(P) ->
    DB1 = ?BDB:db_create(P, []),

    %% Can't do these without an opened table.
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB1, void, "foo-key", [])),
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_fd(P, DB1)),
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_get_byteswapped(P, DB1)),
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_get_type(P, DB1)),

    %% Open a database on-disk
    Name1 = "regression0.db",
    file:delete(Name1),
    ok = ?BDB:db_open(P, DB1, void, Name1, "", bdb_DB_BTREE, [bdb_DB_CREATE], 8#644),

    %% Open an in-memory database
    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "", "", bdb_DB_HASH, [bdb_DB_CREATE], 0),

    %% In the v4.0 driver this apparently worked (incorrectly -- it's an in-memory database).
    %% In v4.4.20 it totally crashes the driver (no errcall output or anything)
    %% Sleepcat will fix the crash : SR [#14157]
    %%?EXPECT_BDB_ERROR(_, ?BDB:db_fd(P, DB2)),

    bdb_DB_BTREE = ?BDB:db_get_type(P, DB1),
    bdb_DB_HASH = ?BDB:db_get_type(P, DB2),

    ok = ?BDB:db_close(P, DB2, []),
    %% cannot close same handle twice
    ?EXPECT_BDB_ERROR(badarg, ?BDB:db_close(P, DB2, [])),

    %% Try again, now that the table is open.
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_del(P, DB1, void, "foo-key", [])),
    _ = ?BDB:db_fd(P, DB1),
    _ = ?BDB:db_get_byteswapped(P, DB1),

    %% Simple put & get ops
    Key1s = "key1",
    Key1b = list_to_binary(Key1s),
    Val1 = <<"Hello, world!">>,
    ok = ?BDB:db_put(P, DB1, void, Key1b, "data1", []),
    ok = ?BDB:db_put(P, DB1, void, Key1b, Val1, []),
    ?EXPECT_BDB_ERROR(bdb_DB_KEYEXIST, ?BDB:db_put(P, DB1, void, Key1b, "data1-blahblah", [bdb_DB_NOOVERWRITE])),
    {Key1b, Val1} = ?BDB:db_get(P, DB1, void, Key1s, <<>>, []),
    Val1 = ?BDB:db_get_data(P, DB1, void, Key1s, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, "key0", <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, "key2", <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get_data(P, DB1, void, "key2", [])),
    _ = ?BDB:db_sync(P, DB1, []),
    
    ok = ?BDB:db_close(P, DB1, []),

    %% Create a database, verify it's there, 
    %% add a couple of records and then truncate the entire database, then
    %% rename it, verify it's renamed
    DB4a = ?BDB:db_create(P, []),
    Name4a = "regression-before.db",
    Name4b = "regression-after.db",
    ok = ?BDB:db_open(P, DB4a, void, Name4a, "", bdb_DB_BTREE, [bdb_DB_CREATE], 8#644),
    _ = file:read_file_info(Name4a),
    ok = ?BDB:db_put(P, DB4a, void, "foo", "foodata", []),
    ok = ?BDB:db_put(P, DB4a, void, "bar", "bardata", []),
    2 = ?BDB:db_truncate(P, DB4a, void, []),
    ok = ?BDB:db_close(P, DB4a, []),
    DB4b = ?BDB:db_create(P, []),
    ok = ?BDB:db_rename(P, DB4b, Name4a, "", Name4b, []),
    {error, _} = file:read_file_info(Name4a),
    {ok, _} = file:read_file_info(Name4b),

    %% Remove the renamed database file, then verify it's gone
    DB3b = ?BDB:db_create(P, []),
    ok = ?BDB:db_remove(P, DB3b, Name4b, "", []),
    {error, enoent} = file:read_file_info(Name4b),

    %% QQQ No simple test for db_upgrade, alas.
    
    %% QQQ I'm seeing weird things happen with db_verify.  I don't know
    %% much about how DB->verify is supposed to work, but it doesn't
    %% look healthy, and I don't know if it's my fault, so I'm going to
    %% ignore the matter for now.  When I run it, I see several:
    %% "beam in free(): warning: modified (chunk-) pointer" errors.
    %% Hrm, I'll comment this out for now....
    %%
    %DB5 = ?BDB:db_create(P, []),
    %_ = ?BDB:db_verify(P, DB5, Name1, "", "", []),
    %TmpFile5a = "Verify-output",
    %file:delete(TmpFile5a),
    %ok = ?BDB:db_verify(P, DB5, Name1, "", TmpFile5a, []),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

db_ops_tests2(P) ->
    DB6 = ?BDB:db_create(P, []),

    TmpFile0 = "TmpErrFile",
    file:delete(TmpFile0),
    ok = ?BDB:db_set_errfile(P, DB6, TmpFile0, "w"),
    _ = file:read_file_info(TmpFile0),
    ok = ?BDB:db_close(P, DB6, []),
    file:delete(TmpFile0),

    DB7 = ?BDB:db_create(P, []),
    %% DB->set_cachesize will fail in a shared environment
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_set_cachesize(P, DB7, 0, 256*1024, 0)),
    ok = ?BDB:db_set_flags(P, DB7, [bdb_DB_DUP, bdb_DB_DUPSORT, bdb_DB_REVSPLITOFF]),
    ok = ?BDB:db_set_pagesize(P, DB7, 4096),
    ok = ?BDB:db_set_bt_minkey(P, DB7, 64),
    ok = ?BDB:db_close(P, DB7, []),

    DB8 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_re_delim(P, DB8, $:),
    ok = ?BDB:db_set_re_len(P, DB8, 256),
    ok = ?BDB:db_set_re_pad(P, DB8, 16#20),
    ok = ?BDB:db_set_re_source(P, DB8, "/etc/passwd"),
    ok = ?BDB:db_close(P, DB8, []),

    DB9 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_h_ffactor(P, DB9, 0),
    ok = ?BDB:db_set_h_nelem(P, DB9, 1024*1024),
    ok = ?BDB:db_set_lorder(P, DB9, 4321),
    ok = ?BDB:db_close(P, DB9, []),

    DB10 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_q_extentsize(P, DB10, 100),
    ok = ?BDB:db_close(P, DB10, []),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

cursor_tests(P) ->
    %% Open an in-memory database
    DB2 = ?BDB:db_create(P, []),
    %% Create an in-memory table
    ok = ?BDB:db_set_flags(P, DB2, [bdb_DB_DUP, bdb_DB_DUPSORT]),
    ok = ?BDB:db_open(P, DB2, void, "", "", bdb_DB_BTREE, [bdb_DB_CREATE], 0),

    %% Put records into the table so that it isn't completely empty.
    Key1s = "key1",
    Key1b = list_to_binary(Key1s),
    Val1 = <<"Hello, world!">>,
    Key2s = "key2",
    Key2b = list_to_binary(Key2s),
    Val2 = <<"Greetings, planet!">>,
    Key3s = "key3",
    Key3b = list_to_binary(Key3s),
    Val3 = <<"I'm not long for this world, er, table">>,
    %% Add records out of sort order.
    ok = ?BDB:db_put(P, DB2, void, Key2b, Val1, []),
    ok = ?BDB:db_put(P, DB2, void, Key2s, Val2, []), % bin, list no matter
    ok = ?BDB:db_put(P, DB2, void, Key1b, Val1, []),
    ok = ?BDB:db_put(P, DB2, void, Key3s, Val3, []),

    C1 = ?BDB:db_cursor(P, DB2, void, []),

    {Key1b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_FIRST]),
    1 = ?BDB:c_count(P, C1, []),
    {Key1b, Val1} = ?BDB:c_get(P, C1, Key1b, <<>>, [bdb_DB_SET]),
    1 = ?BDB:c_count(P, C1, []),
    {Key2b, Val2} = ?BDB:c_get(P, C1, Key2b, <<>>, [bdb_DB_SET]),
    2 = ?BDB:c_count(P, C1, []),
    {Key2b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_NEXT_DUP]),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_NEXT_DUP])),
    
    {Key3b, Val3} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_LAST]),
    {Key2b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_PREV]),
    {Key2b, Val2} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_PREV]),
    {Key1b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_PREV]),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_PREV])),
    {Key1b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_CURRENT]),
    {Key2b, Val2} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_NEXT]),

    C2 = ?BDB:c_dup(P, C1, []),
    {Key2b, Val1} = ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_NEXT]),
    % C2 is currently uninitialized
    ?EXPECT_BDB_ERROR(einval, ?BDB:c_get(P, C2, <<>>, <<>>, [bdb_DB_CURRENT])),
    % DB_NEXT on an uninitialized cursor will return the first key/value
    {Key1b, Val1} = ?BDB:c_get(P, C2, <<>>, <<>>, [bdb_DB_NEXT]),
    {Key2b, Val2} = ?BDB:c_get(P, C2, <<>>, <<>>, [bdb_DB_NEXT]),
    ok = ?BDB:c_close(P, C2),

    {Key2b, Val2} = ?BDB:c_get(P, C1, Key2b, <<>>, [bdb_DB_SET]),
    ok = ?BDB:c_put(P, C1, Key2b, Val2, bdb_DB_CURRENT), % Same stuff
    NewVal2 = "Blarf",
    NewVal2b = list_to_binary(NewVal2),
    ok = ?BDB:c_put(P, C1, Key2b, NewVal2, bdb_DB_KEYFIRST),
    ?EXPECT_BDB_ERROR(bdb_DB_KEYEXIST, ?BDB:c_put(P, C1, Key2b, NewVal2b, bdb_DB_NODUPDATA)),
    ?EXPECT_BDB_ERROR(einval, ?BDB:c_put(P, C1, Key2b, "Can't clobber current when using sorting func", bdb_DB_CURRENT)),

    {Key3b, Val3} = ?BDB:c_get(P, C1, Key3b, <<>>, [bdb_DB_SET]),
    ok = ?BDB:c_del(P, C1, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, C1, Key3b, <<>>, [bdb_DB_SET])),

    ok = ?BDB:c_close(P, C1),

    ok = ?BDB:db_close(P, DB2, []),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

lock_tests2(P) ->
    ok = ?BDB:set_lk_conflicts(P, <<0, 0, 0, 0>>, 2),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),
    ok = ?BDB:set_lk_max_lockers(P, 500),
    ok = ?BDB:set_lk_max_locks(P, 500),
    ok = ?BDB:set_lk_max_objects(P, 500),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.
    
lock_tests1(P) ->
    ok = ?BDB:env_open(P, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK], 8#644),
    0 = ?BDB:lock_detect(P, [], bdb_DB_LOCK_DEFAULT),
    LockID0 = ?BDB:lock_id(P),
    LockID1 = ?BDB:lock_id(P),
    LockID2 = ?BDB:lock_id(P),
    _LockID3 = ?BDB:lock_id(P),
    _LockID4 = ?BDB:lock_id(P),
    ok = ?BDB:lock_id_free(P, LockID0),
    ?EXPECT_BDB_ERROR(einval, ?BDB:lock_id_free(P, LockID0)),
    
    Lock0 = ?BDB:lock_get(P, LockID1, [], "Foo lock0!", bdb_DB_LOCK_READ),
    _Lock_1 = ?BDB:lock_get(P, LockID2, [], "Foo lock0!", bdb_DB_LOCK_READ),
    ok = ?BDB:lock_put(P, Lock0),
    ?EXPECT_BDB_ERROR(einval, ?BDB:lock_put(P, Lock0)),

    %% QQQ I don't understand why this fails:
    %% "my_debug_errcall: (null): DB_ENV->lock_get: invalid lock mode 2"
    %% It does the same thing for any lockmode "greater" than DB_LOCK_READ:
    %% write, iwrite, iread, and iwr.
    %Lock2 = ?BDB:lock_get(P, LockID3, 0, "Foo writelock", bdb_DB_LOCK_WRITE),

    LockStat = ?BDB:lock_stat(P, []),
    {value, {st_maxlockers, 1000}} = lists:keysearch(st_maxlockers, 1, LockStat),

    MutexStat = ?BDB:mutex_stat(P, []),
    {value, {st_mutex_inuse, 100}} = lists:keysearch(st_mutex_inuse, 1, MutexStat),

    ok = ?BDB:env_close(P, []),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_get_set_threadpool_info(StartFunc) ->

    io:format("\nPerforming test_get_set_threadpool_info\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    %%?BDB:adjust_port_logging(P, enable, [all]),

    %% I'm tired of re-coding this test when the default threadcounts in berkeley_db.xml are changed.
    %% I just had to reduce the default threadcounts from 30 to 1 (see berkeley_db.xml for reason)
    %% So lets just set the sizes here.
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_WORKERS, 30, 64*1024, 1000),
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 30, 64*1024, 2147483647), % INT_MAX is default max_qlen

    timer:sleep(100),

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_WORKERS},
     {curr_queue_len,             0},
     {curr_idle_threads,          30},
     {num_threads,                30},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            1000}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_WORKERS),

    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_WORKERS, 5, 64*1024, 2000),

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_WORKERS},
     {curr_queue_len,             NumPills1},       % we don't know the absolute length, but we know that the only enqueued items are poison pills
     {curr_idle_threads,          NumThreads1},
     {num_threads,                NumThreads1},
     {curr_enqueued_poison_pills, NumPills1},
     {queue_len_limit,            2000}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_WORKERS),

    5 = NumThreads1 - NumPills1,                  % we do know this relationship

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_VALMAP_STOP},
     {curr_queue_len,             0},
     {curr_idle_threads,          30},
     {num_threads,                30},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            2147483647}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_VALMAP_STOP),

    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 35, 64*1024, 100000),
    timer:sleep(1),

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_VALMAP_STOP},
     {curr_queue_len,             0},
     {curr_idle_threads,          35},
     {num_threads,                35},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            100000}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_VALMAP_STOP),


    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_CONSUME_WAIT},
     {curr_queue_len,             0},
     {curr_idle_threads,          0},
     {num_threads,                0},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            0}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_CONSUME_WAIT),


    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_HOUSEKEEPING},
     {curr_queue_len,             0},
     {curr_idle_threads,          2},
     {num_threads,                2},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            10}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING),

    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING, 0, 64*1024, 18),   % *can* set 0 threads (operations on this pool will be queued up until a thread is added)

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_HOUSEKEEPING},
     {curr_queue_len,             NumPills5},
     {curr_idle_threads,          NumThreads5},
     {num_threads,                NumThreads5}, % we don't know the exact number
     {curr_enqueued_poison_pills, NumPills5},   % we don't know the exact number
     {queue_len_limit,            18}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING),

    0 = NumThreads5 - NumPills5,                    % we do know this relationship

    %%io:format("\nTest enforcement of queue length limit\n"),
    timer:sleep(50),
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING, 2, 64*1024, 0), % some threads but zero-length queue
    ?EXPECT_BDB_ERROR({edtk_queue_len_limit_reached, 3}, ?BDB:txn_checkpoint(P, 0, 0, [])),

    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING, 2, 64*1024, 1),
    ok = ?BDB:txn_checkpoint(P, 0, 0, []),

    %% Range checks on threadpool-id
    ?EXPECT_BDB_ERROR(badarg, ?BDB:get_threadpool_info(P, -1)),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_threadpool_params(P, -1, 5, 64*1024, 1000)),

    ?EXPECT_BDB_ERROR(badarg, ?BDB:get_threadpool_info(P, 5)),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_threadpool_params(P, 5, 5, 64*1024, 1000)),

    normal = ?BDB:shutdown(P),	

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_get_set_num_valmap_slots(StartFunc) ->

    io:format("\nPerforming test_get_set_num_valmap_slots\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    %%?BDB:adjust_port_logging(P, enable, [all]),

    ?BDB_INITIAL_NUM_VALMAP_DBENV_SLOTS  = ?BDB:get_num_valmap_slots(P, bdb_dbenv),
    %%io:format("Cannot reduce #slots if one of them is in use~n"),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_dbenv, 0)),

    ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_db),

    %%io:format("Grow~n"),
    ok = ?BDB:set_num_valmap_slots(P, bdb_db, 2 * ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS ),
    2 * ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_db),

    %%io:format("Shrink~n"),
    ok = ?BDB:set_num_valmap_slots(P, bdb_db, ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS ),
    ?BDB_INITIAL_NUM_VALMAP_DB_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_db),

    ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    ?BDB_INITIAL_NUM_VALMAP_CURSOR_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_cursor),

    %%io:format("Not allowed to shrink to 0 even if none are in use~n"),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_db_txn, 0)),
    ?BDB_INITIAL_NUM_VALMAP_DB_TXN_SLOTS = ?BDB:get_num_valmap_slots(P, bdb_db_txn),

%%     %%io:format("OK to shrink to 0 if non are in-use~n"),
%%     ?BDB:adjust_port_logging(P, enable, [all]),
%%     ok  = ?BDB:set_num_valmap_slots(P, bdb_db_txn, []),
%%     0 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
%%     ?EXPECT_BDB_ERROR(enomem, ?BDB:txn_begin(P, void, [])),

    %%io:format("Growing it allows us to allocate again~n"),
    ok  = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1),
    1 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    _Txn1 = ?BDB:txn_begin(P, void, []),

    %%io:format("... up to the new limit~n"),
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_db_txn}, ?BDB:txn_begin(P, void, [])),

    %%io:format("And growing it yet again fixes it again~n"),
    ok  = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 4),
    4 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    Txn2 = ?BDB:txn_begin(P, void, []),
    Txn3 = ?BDB:txn_begin(P, void, []),
    Txn4 = ?BDB:txn_begin(P, void, []),
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_db_txn}, ?BDB:txn_begin(P, void, [])),

    %%io:format("Cannot reduce #slots if one of them is in use~n"),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1)),
    4 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    
    %%io:format("Freeing the top one allows us to shrink by exactly 1~n"),
    ok = ?BDB:txn_abort(P, Txn4),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1)),
    ok  = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 3),
    3 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_db_txn}, ?BDB:txn_begin(P, void, [])),

    %%io:format("Freeing a lower one does NOT allow us to shrink by 1~n"),
    ok = ?BDB:txn_abort(P, Txn2),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_db_txn, 2)),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1)),
    3 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),

    %%io:format("Can still use that free middle slot~n"),
    Txn2a = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:txn_abort(P, Txn2a),
    Txn2b = ?BDB:txn_begin(P, void, []),
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_db_txn}, ?BDB:txn_begin(P, void, [])),
    ok = ?BDB:txn_abort(P, Txn2b),

    %%io:format("Freeing top one above a hole allow us to shrink by N~n"),
    ok = ?BDB:txn_abort(P, Txn3),
    ok = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1),
    1 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    ?EXPECT_BDB_ERROR({no_free_valmap_slots, bdb_db_txn}, ?BDB:txn_begin(P, void, [])),

    %%io:format("group_valmap_cleanup frees bunches (of all types)~n"),
    ok = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 10),
    10 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),
    _Txn2c = ?BDB:txn_begin(P, void, []),
    _Txn3c = ?BDB:txn_begin(P, void, []),
    _Txn4c = ?BDB:txn_begin(P, void, []),
    %% Txns 1, 2c, 3c, 4c and DbEnv = 5 resources
    5 = ?BDB:group_valmap_cleanup(P, P#bdb_port_handle.group_valmap_id),
    ok = ?BDB:set_num_valmap_slots(P, bdb_db_txn, 1),
    1 = ?BDB:get_num_valmap_slots(P, bdb_db_txn),

%%     io:format("Can reduce dbenv to 0 after it has been closed~n"),
%%     ok = ?BDB:set_num_valmap_slots(P, bdb_dbenv, 0),
%%     0 = ?BDB:get_num_valmap_slots(P, bdb_dbenv),

    normal = ?BDB:shutdown(P),	

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_transactions(StartFunc) ->	 
    io:format("\nPerforming test_basic_transactions\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(P, enable, [all]),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),

    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    Key1 = <<"key1">>,
    Val1 = <<"val1">>,

    Key2 = <<"-key2">>,
    Val2 = <<"-val2">>,


    %%io:format("\nTest commit\n"),
    Txn1 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_put(P, DB1, Txn1, Key1, Val1, []),
    ok = ?BDB:txn_commit(P, Txn1, []),
    {Key1, Val1} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),
    

    %%io:format("\nTest abort\n"),
    Txn2 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_del(P, DB1, Txn2, Key1, []),
    ok = ?BDB:db_put(P, DB2, Txn2, Key2, Val2, []),
    ok = ?BDB:txn_abort(P, Txn2),
    {Key1, Val1} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB2, void, Key2, <<>>, [])),

    %%io:format("\nnow commit the same transaction & verify that the result is different\n"),
    Txn3 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_del(P, DB1, Txn3, Key1, []),
    ok = ?BDB:db_put(P, DB2, Txn3, Key2, Val2, []),
    ok = ?BDB:txn_commit(P,  Txn3, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, Key1, <<>>, [])),
    {Key2, Val2} = ?BDB:db_get(P, DB2, void, Key2, <<>>, []),

    % cleanup test state
    1 = ?BDB:db_truncate(P, DB2, void, []),


    CNewVal1 = <<"CNewval1">>,
    CNewKey2 = <<"CNewkey2">>,
    CNewVal2 = <<"CNewval2">>,

    %%io:format("\nTest transactional cursor (with commit)\n"),

    %% Init test state
    ok = ?BDB:db_put(P, DB2, void, Key2, Val2, []),

    Txn4 = ?BDB:txn_begin(P, void, []),

    C1 = ?BDB:db_cursor(P, DB1, Txn4, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_FIRST])),
    ok = ?BDB:c_put(P, C1, Key1, CNewVal1, bdb_DB_KEYFIRST),

    C2 = ?BDB:db_cursor(P, DB2, Txn4, []),
    {Key2, Val2} = ?BDB:c_get(P, C2, Key2, <<>>, [bdb_DB_SET]),   % RELIES on result of previous test
    ok = ?BDB:c_put(P, C2, <<>>, CNewVal1, bdb_DB_CURRENT),	   % overwrite value at current pos
    ok = ?BDB:c_put(P, C2, CNewKey2, CNewVal2, bdb_DB_KEYFIRST), % insert another key

    ok = ?BDB:c_close(P, C2),
    ok = ?BDB:c_close(P, C1),
    ok = ?BDB:txn_commit(P, Txn4, []),

    CheckC1 = ?BDB:db_cursor(P, DB1, void, []),
    {Key1, CNewVal1} = ?BDB:c_get(P, CheckC1, <<>>, <<>>, [bdb_DB_FIRST]),
    ok = ?BDB:c_close(P, CheckC1),

    CheckC2 = ?BDB:db_cursor(P, DB2, void, []),
    {Key2, CNewVal1} = ?BDB:c_get(P, CheckC2, <<>>, <<>>, [bdb_DB_FIRST]),
    {CNewKey2, CNewVal2} = ?BDB:c_get(P, CheckC2, <<>>, <<>>, [bdb_DB_NEXT]),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, CheckC2, <<>>, <<>>, [bdb_DB_NEXT])),
    ok = ?BDB:c_close(P, CheckC2),

    % cleanup test state
    1 = ?BDB:db_truncate(P, DB1, void, []),
    2 = ?BDB:db_truncate(P, DB2, void, []),


    %%io:format("\nTest transactional cursor (with abort)\n"),

    %% Init test state
    ok = ?BDB:db_put(P, DB2, void, Key2, Val2, []),

    Txn5 = ?BDB:txn_begin(P, void, []),

    C3 = ?BDB:db_cursor(P, DB1, Txn5, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, C3, <<>>, <<>>, [bdb_DB_FIRST])),
    ok = ?BDB:c_put(P, C3, Key1, CNewVal1, bdb_DB_KEYFIRST),

    ok = ?BDB:set_mp_max_write(P, 5, 30000),
    {5, 30000} = ?BDB:get_mp_max_write(P),

    3 = ?BDB:memp_trickle(P, 100),

    C4 = ?BDB:db_cursor(P, DB2, Txn5, []),
    {Key2, Val2} = ?BDB:c_get(P, C4, Key2, <<>>, [bdb_DB_SET]),   % RELIES on result of previous test
    ok = ?BDB:c_put(P, C4, <<>>, CNewVal1, bdb_DB_CURRENT),	   % overwrite value at current pos
    ok = ?BDB:c_put(P, C4, CNewKey2, CNewVal2, bdb_DB_KEYFIRST), % insert another key


    %% child txns just to test txn_stat status = TXN_COMMITTED/ABORTED
    Txn5cA = ?BDB:txn_begin(P, Txn5, []),
%%     Txn5cAID = ?BDB:txn_id(P, Txn5cA),
    %%io:format("Txn5cAID ~w~n", [Txn5cID]),
    ok = ?BDB:db_put(P, DB1, Txn5cA, Key1, Val1, []),
    ok = ?BDB:txn_abort(P, Txn5cA),

    Txn5cC = ?BDB:txn_begin(P, Txn5, []),
%%     Txn5cCID = ?BDB:txn_id(P, Txn5cC),
    %%io:format("Txn5cCID ~w~n", [Txn5cID]),
    ok = ?BDB:db_put(P, DB2, Txn5cC, Key1, Val1, []),
    ok = ?BDB:txn_commit(P, Txn5cC, []),   % but parent will eventually be aborted anyway

    TxnStat = ?BDB:txn_stat(P, []),
    {value, {active_txns, ActiveTxns}} = lists:keysearch(active_txns, 1, TxnStat),
    %%io:format("ActiveTxns ~p~n", [ActiveTxns]),

    Txn5ID = ?BDB:txn_id(P, Txn5),
    %%io:format("Txn5ID ~w~n", [Txn5ID]),

    {value, {Txn5ID, ActiveTxn5}} = lists:keysearch(Txn5ID, 1, ActiveTxns),
    %%io:format("ActiveTxn5 ~p~n", [ActiveTxn5]),
    {value, {status, bdb_TXN_RUNNING}} = lists:keysearch(status, 1, ActiveTxn5),
    {value, {xid, not_prepared}} = lists:keysearch(xid, 1, ActiveTxn5),
    
%%     {value, {Txn5cAID, ActiveTxn5cA}} = lists:keysearch(Txn5cAID, 1, ActiveTxns),
%%     %%io:format("ActiveTxn5cA ~p~n", [ActiveTxn5c~]),
%%     {value, {parent_txn_id, Txn5ID}} = lists:keysearch(parent_txn_id, 1, ActiveTxn5cA),
%%     {value, {status, bdb_TXN_ABORTED}} = lists:keysearch(status, 1, ActiveTxn5cA),
%%     {value, {xid, not_distributed}} = lists:keysearch(xid, 1, ActiveTxn5cA),

%%     {value, {Txn5cCID, ActiveTxn5cA}} = lists:keysearch(Txn5cCID, 1, ActiveTxns),
%%     %%io:format("ActiveTxn5cA ~p~n", [ActiveTxn5c~]),
%%     {value, {parent_txn_id, Txn5ID}} = lists:keysearch(parent_txn_id, 1, ActiveTxn5cA),
%%     {value, {status, bdb_TXN_COMMITTED}} = lists:keysearch(status, 1, ActiveTxn5cA),
%%     {value, {xid, not_distributed}} = lists:keysearch(xid, 1, ActiveTxn5cA),

    {value, {st_time_ckp, ChkTime}} = lists:keysearch(st_time_ckp, 1, TxnStat),
    ElapsedTime = ?BDB_U:now_diff_milliseconds(now(), ChkTime),
    if ElapsedTime > 2000 -> throw({bad_st_time_ckp, ElapsedTime});
       true               -> ok
    end,


    ok = ?BDB:c_close(P, C4),
    ok = ?BDB:c_close(P, C3),
    ok = ?BDB:txn_abort(P, Txn5),

    CheckC3 = ?BDB:db_cursor(P, DB1, void, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, CheckC3, <<>>, <<>>, [bdb_DB_FIRST])),
    ok = ?BDB:c_close(P, CheckC3),

    CheckC4 = ?BDB:db_cursor(P, DB2, void, []),
    {Key2, Val2} = ?BDB:c_get(P, CheckC4, <<>>, <<>>, [bdb_DB_FIRST]),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:c_get(P, CheckC4, <<>>, <<>>, [bdb_DB_NEXT])),
    ok = ?BDB:c_close(P, CheckC4),

    % cleanup test state
    1 = ?BDB:db_truncate(P, DB2, void, []),

    ok = ?BDB:txn_checkpoint(P, 0, 0, []),


    LogStat = ?BDB:log_stat(P, []),
    {value, {st_cur_file, 1}} = lists:keysearch(st_cur_file, 1, LogStat),


    ["log.0000000001"]       = ?BDB:log_archive(P, [bdb_DB_ARCH_LOG]),
    ["test1.db", "test2.db"] = ?BDB:log_archive(P, [bdb_DB_ARCH_DATA]),
    []                       = ?BDB:log_archive(P, [bdb_DB_ARCH_REMOVE]),

    ok = ?BDB:log_flush(P),


    MempStat = ?BDB:memp_stat(P, []),
    {value, {st_page_trickle,3}} = lists:keysearch(st_page_trickle, 1, MempStat),

    {value, {file_stats, FileStats}} = lists:keysearch(file_stats, 1, MempStat),
    %%io:format("FileStats ~p~n", [FileStats]),

    {value, {"test1.db", Test1DbFileStats}} = lists:keysearch("test1.db", 1, FileStats),
    %%io:format("Test1DbFileStats ~p~n", [Test1DbFileStats]),
    {value, {st_cache_miss, 2}} = lists:keysearch(st_cache_miss, 1, Test1DbFileStats),

    {value, {"test2.db", Test2DbFileStats}} = lists:keysearch("test2.db", 1, FileStats),
    %%io:format("Test2DbFileStats ~p~n", [Test2DbFileStats]),
    {value, {st_cache_miss, 3}} = lists:keysearch(st_cache_miss, 1, Test2DbFileStats),


    DbStatBtree = ?BDB:db_stat(P, DB1, void, bdb_DB_READ_COMMITTED),
    {value, {db_type,bdb_DB_BTREE}} = lists:keysearch(db_type, 1, DbStatBtree),
    {value, {bt_levels,1}} = lists:keysearch(bt_levels, 1, DbStatBtree),


    DbStatHash = ?BDB:db_stat(P, DB2, void, bdb_DB_READ_COMMITTED),
    {value, {db_type,bdb_DB_HASH}} = lists:keysearch(db_type, 1, DbStatHash),
    {value, {hash_buckets,2}} = lists:keysearch(hash_buckets, 1, DbStatHash),


    ok = ?BDB:db_close(P, DB1, []),
    ok = ?BDB:db_close(P, DB2, []),

    ok = ?BDB:env_close(P, []),
    normal = ?BDB:shutdown(P),


    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_deadlock_handling(StartFunc) ->	 
    io:format("\nPerforming test_basic_deadlock_handling\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(P, enable, [all]),

    %% I just had to reduce the default threadcounts from 30 to 1 (see berkeley_db.xml for reason)
    %% This test requires more than that.
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_WORKERS, 30, 64*1024, 1000),
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 30, 64*1024, 1000),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),

    %% set a deterministic deadlock resolution policy, so we can predict the result
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_MINLOCKS),

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB1a = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1a, void, "test1a.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    Key1 = <<"key1">>,
    Val1 = <<"val1">>,

    Key2 = <<"-key2">>,
    Val2 = <<"-val2">>,


    %% Txn1 locks a page in DB1 and a page in DB1a
    Txn1 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_put(P, DB1, Txn1, Key1, Val1, []),
    ok = ?BDB:db_put(P, DB1a, Txn1, Key1, Val1, []),

    %% Txn2 locks a page in DB2
    Txn2 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_put(P, DB2, Txn2, Key2, Val2, []),

    %% Now Txn1 attempts to acquire the lock held by Txn2 (and blocks, so we have to use non_block variant, or spawn a process)
    Tag = ?BDB:db_put_non_block(P, DB2, Txn1, Key2, Val1, []), 

    %% Now Txn2 attempts to acquire the lock held by Txn1, which is a deadlock
    %% Txn2 has fewer locks, so it loses under the MINLOCKS policy
    ?EXPECT_BDB_ERROR(bdb_DB_LOCK_DEADLOCK, ?BDB:db_put(P, DB1, Txn2, Key1, Val2, [])),
    ok = ?BDB:txn_abort(P, Txn2),

    %% Now Txn1's non-blocking operation can complete
    ok = ?BDB:receive_db_put_reply(P, Tag),

    ok = ?BDB:txn_commit(P, Txn1, []),

    %% Txn1 was committed...
    {Key1, Va1} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),
    {Key1, Va1} = ?BDB:db_get(P, DB1a, void, Key1, <<>>, []),
    {Key2, Va1} = ?BDB:db_get(P, DB2, void, Key2, <<>>, []),

    ok = ?BDB:db_close(P, DB1, []),
    ok = ?BDB:db_close(P, DB1a, []),
    ok = ?BDB:db_close(P, DB2, []),

    ok = ?BDB:env_close(P, []),
    normal = ?BDB:shutdown(P),


    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_recovery(StartFunc) ->    
    io:format("\nPerforming test_recovery\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    ok = env_open_with_retry(P, Dir, [bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),
    DB = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    PutLoop = fun(0, _) ->
                      ok;
                 (N, PutLoop) ->
                      ok = ?BDB:db_put(P, DB, void, integer_to_list(N), "misc-junk", []),
                      PutLoop(N - 1, PutLoop)
              end,
    PutLoop(500, PutLoop),

    io:format("\ntest_recovery : panicking environment...\n"),
    %%?BDB:adjust_port_logging(P, enable, [all]),
    ok = ?BDB:env_set_flags(P, [bdb_DB_PANIC_ENVIRONMENT], 1),
    wait_for_sleepycat_failure_event(StartFunc, P, bdb_DB_EVENT_PANIC),
    
    normal = ?BDB:shutdown(P),

    io:format("\ntest_recovery : restart without running recovery (fails)\n"),
    P1 = ?BDB:StartFunc(),
    ?EXPECT_BDB_ERROR(bdb_DB_RUNRECOVERY, env_open_with_retry(P1, Dir, [bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE)),
    wait_for_sleepycat_failure_event(StartFunc, P1, bdb_DB_EVENT_PANIC),
    normal = ?BDB:shutdown(P1),
    %% ... and that did not fix the problem
    P1a = ?BDB:StartFunc(),
    ?EXPECT_BDB_ERROR(bdb_DB_RUNRECOVERY, env_open_with_retry(P1a, Dir, [bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE)),
    wait_for_sleepycat_failure_event(StartFunc, P1a, bdb_DB_EVENT_PANIC),
    normal = ?BDB:shutdown(P1a),

    io:format("\ntest_recovery : restart with run recovery (succeeds)\n"),
    P2 = ?BDB:StartFunc(),
    ok = ?BDB:env_set_verbose(P2, bdb_DB_VERB_RECOVERY, 1),
    ok = env_open_with_retry(P2, Dir, [bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P2, bdb_DB_LOCK_DEFAULT),
    DB2 = ?BDB:db_create(P2, []),
    ok = ?BDB:db_open(P2, DB2, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),

    GetLoop = fun(0, _) ->
                      ok;
                 (N, GetLoop) ->
                      Key = list_to_binary(integer_to_list(N)),
                      {Key, <<"misc-junk">>} = ?BDB:db_get(P2, DB2, void, Key, <<>>, []),
                      GetLoop(N - 1, GetLoop)
              end,
    GetLoop(500, GetLoop),

    normal = ?BDB:shutdown(P2),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

wait_for_sleepycat_failure_event(StartFunc, BdbPortHandle, EventCode) ->
    case StartFunc of 
        start ->
             ok;
        start_pipe ->
            convert_pipe_messages_until_no_more_after_N_milliseconds(BdbPortHandle, 1000)
    end,
    receive
        {ErlangPort, bdb_untagged_event, {sleepycat_event, EventCode}} when ErlangPort == BdbPortHandle#bdb_port_handle.erlang_port ->
            ok
    after
        1000 ->
            {messages, []} = process_info(self(), messages),
            throw(timeout_waiting_for_event)
    end.


test_attempt_reuse_valmap_tuples_after_closing_them(StartFunc) ->
    io:format("\nPerforming test_attempt_reuse_valmap_tuples_after_closing_them\n"),

    P = ?BDB:StartFunc(),

    ok = env_open_with_retry(P, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    %% use & then release the first txn valmap
    Txn1 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:txn_commit(P, Txn1, []),

    %% overwrite Txn1's old (now null) slot with an new txn
    Txn2 = ?BDB:txn_begin(P, void, []),
    %% now attempt to use Txn1's expired valmapt tuple to refer to the (now valid) slot occupied by Txn2
    %% (this is the dangerous case that per-slot generation-ids are designed to solve)
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_commit(P, Txn1, [])),

    ok = ?BDB:txn_abort(P, Txn2),

    %% attempt to reuse Txn1's expired valmap (after it has been closed and is null)
    %% (i.e. the same command that worked fine when it was open)
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_commit(P, Txn1, [])),
  
    normal = ?BDB:shutdown(P),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_nested_transactions(StartFunc) ->    
    io:format("\nPerforming test_nested_transactions\n"),
 
    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    BaseP = ?BDB:StartFunc(),

    ok = env_open_with_retry(BaseP, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(BaseP, bdb_DB_LOCK_DEFAULT),

    DB1 = ?BDB:db_create(BaseP, []),
    ok = ?BDB:db_open(BaseP, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(BaseP, []),
    ok = ?BDB:db_open(BaseP, DB2, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB3 = ?BDB:db_create(BaseP, []),
    ok = ?BDB:db_open(BaseP, DB3, void, "test3.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    %% Get a unique group id for tests
    P = BaseP#bdb_port_handle{group_valmap_id = BaseP#bdb_port_handle.group_valmap_id + 1},

    %%io:format("\nTest in-order commit\n"),
    Key1 = <<"Nestedkey1">>,
    Val1 = <<"Nestedval1">>,
    TxnP1 = ?BDB:txn_begin(P, void, []),
    TxnC1 = ?BDB:txn_begin(P, TxnP1, []),

    %%io:format("put/get on a parent with unresolved children is an error\n"),
    ?EXPECT_BDB_ERROR(eperm, ?BDB:db_put(P, DB1, TxnP1, Key1, "garbage", [])),

    ok = ?BDB:db_put(P, DB1, TxnC1, Key1, Val1, []),
    ok = ?BDB:txn_commit(P, TxnC1, []),
    ok = ?BDB:txn_commit(P, TxnP1, []),
    {Key1, Val1} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),
    ?AT,
    0 = ?BDB:group_valmap_cleanup(P, P#bdb_port_handle.group_valmap_id),
    ?AT,

    io:format("\nTest out-of-order commit (child is implicitly committed)\n"),
    ?BDB:adjust_port_logging(P, enable, [valmap_stop]),
    Key2 = <<"-Nestedkey2">>,
    Val2 = <<"-Nestedval2">>,
    TxnP2 = ?BDB:txn_begin(P, void, []),
    TxnC2 = ?BDB:txn_begin(P, TxnP2, []),
    ok = ?BDB:db_put(P, DB1, TxnC2, Key2, Val2, []),
    %% intentionally no commit of C2!
    ok = ?BDB:txn_commit(P, TxnP2, []),
    {Key2, Val2} = ?BDB:db_get(P, DB1, void, Key2, <<>>, []),
    %% should not be able to use child valmap now
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P, TxnC2)),
    %% and the child txn should not need to be aborted
    ?AT,
    0 = ?BDB:group_valmap_cleanup(P, P#bdb_port_handle.group_valmap_id),
    ?AT,

    io:format("\nTest out-of-order abort (child is implicitly aborted)\n"),
    Key2a = <<"=-Nestedkey2a">>,
    Val2a = <<"=-Nestedval2a">>,
    TxnP2a = ?BDB:txn_begin(P, void, []),
    TxnC2a = ?BDB:txn_begin(P, TxnP2a, []),
    ok = ?BDB:db_put(P, DB1, TxnC2a, Key2a, Val2a, []),
    %% intentionally no commit/abort of C2!
    ok = ?BDB:txn_abort(P, TxnP2a),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, Key2a, <<>>, [])),
    % should not be able to use child valmap now
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P, TxnC2a)),
    ?AT,
    0 = ?BDB:group_valmap_cleanup(P, P#bdb_port_handle.group_valmap_id),
    ?AT,
    ?BDB:adjust_port_logging(P, disable, [valmap_stop]),

    %%io:format("\nTest commit child, abort parent\n"),
    Key3 = <<"-Nestedkey3">>,
    Val3 = <<"-Nestedval3">>,
    TxnP3 = ?BDB:txn_begin(P, void, []),
    TxnC3 = ?BDB:txn_begin(P, TxnP3, []),
    ok = ?BDB:db_put(P, DB1, TxnC3, Key3, Val3, []),
    ok = ?BDB:txn_commit(P, TxnC3, []),
    ok = ?BDB:txn_abort(P, TxnP3),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, Key3, <<>>, [])),

    %%io:format("\nTest multiple children and grandchildren (comitting via top-level parent)\n"),
    Key4a = <<"-Nestedkey4a">>,
    Val4a = <<"-Nestedval4a">>,
    Key4b = <<"key4b">>,
    Val4b = <<"val4b">>,
    Key4c = <<"key4c">>,
    Val4c = <<"val4c">>,
    TxnP4 = ?BDB:txn_begin(P, void, []),

    TxnC4a = ?BDB:txn_begin(P, TxnP4, []),
    ok = ?BDB:db_put(P, DB1, TxnC4a, Key4a, Val4a, []),

    TxnC4b = ?BDB:txn_begin(P, TxnP4, []),
    TxnC4bi = ?BDB:txn_begin(P, TxnC4b, []),
    ok = ?BDB:db_put(P, DB2, TxnC4bi, Key4b, Val4b, []),

    TxnC4bii = ?BDB:txn_begin(P, TxnC4b, []),
    ok = ?BDB:db_put(P, DB3, TxnC4bii, Key4c, Val4c, []),
    ok = ?BDB:txn_abort(P, TxnC4bii),
    ok = ?BDB:txn_commit(P, TxnP4, []), % implicitly commits C4a, C4b, C4bi
    {Key4a, Val4a} = ?BDB:db_get(P, DB1, void, Key4a, <<>>, []),
    {Key4b, Val4b} = ?BDB:db_get(P, DB2, void, Key4b, <<>>, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB3, void, Key4c, <<>>, [])),
    % should not be able to use child valmap now
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P, TxnC4a)),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P, TxnC4b)),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P, TxnC4bi)),

    % cleanup test state
    _ = ?BDB:db_truncate(BaseP, DB1, void, []),
    _ = ?BDB:db_truncate(BaseP, DB2, void, []),
    _ = ?BDB:db_truncate(BaseP, DB3, void, []),

    ok = ?BDB:db_close(BaseP, DB1, []),
    ok = ?BDB:db_close(BaseP, DB2, []),
    ok = ?BDB:db_close(BaseP, DB3, []),

    ok = ?BDB:env_close(BaseP, []),
    normal = ?BDB:shutdown(BaseP),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_stop_with_nested_transactions(StartFunc) ->    
    io:format("\nPerforming test_stop_with_nested_transactions\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB3 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB3, void, "test3.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    Key1a = <<"will-be-commited-key">>,
    Val1a = <<"rhubarb">>,
    ok = ?BDB:db_put(P, DB3, void, Key1a, Val1a, []),

    %%io:format("\nCreate parents & children and then close driver, to force stop() to process them\n"),
    Key4a = <<"-Nestedkey4a">>,
    Val4a = <<"-Nestedval4a">>,
    Key4b = <<"key4b">>,
    Val4b = <<"val4b">>,
    Key4c = <<"key4c">>,
    Val4c = <<"val4c">>,
    TxnP4 = ?BDB:txn_begin(P, void, []),

    TxnC4a = ?BDB:txn_begin(P, TxnP4, []),
    ok = ?BDB:db_put(P, DB1, TxnC4a, Key4a, Val4a, []),
    ok = ?BDB:txn_commit(P, TxnC4a, []),
 
    TxnC4b = ?BDB:txn_begin(P, TxnP4, []),
    TxnC4bi = ?BDB:txn_begin(P, TxnC4b, []),
    ok = ?BDB:db_put(P, DB2, TxnC4bi, Key4b, Val4b, []),

    TxnC4bii = ?BDB:txn_begin(P, TxnC4b, []),
    ok = ?BDB:db_put(P, DB3, TxnC4bii, Key4c, Val4c, []),
    ok = ?BDB:txn_commit(P, TxnC4bii, []),

    %%io:format("\nenabling logging\n"),
    ok = ?BDB:adjust_port_logging(P, enable, [valmap_start, valmap_stop, group_valmap_cleanup_outputv, group_valmap_cleanup_work]),
    ok = ?BDB:adjust_driver_logging(P, enable, [shutdown]),

    %%%io:format("\nCalling shutdown to abort open TxnP4 open child/grandchild TxnC4b, TxnC4bi, and committed child/grandchild TxnC4a,TxnC4bii\n"),
    normal = ?BDB:shutdown(P),

    %%io:format("\nOpening new port to verify that operations were aborted\n"),
    P2 = ?BDB:StartFunc(),

    ok = env_open_with_retry(P2, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P2, bdb_DB_LOCK_DEFAULT),

    DB1a = ?BDB:db_create(P2, []),
    ok = ?BDB:db_open(P2, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2a = ?BDB:db_create(P2, []),
    ok = ?BDB:db_open(P2, DB2a, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB3a = ?BDB:db_create(P2, []),
    ok = ?BDB:db_open(P2, DB3a, void, "test3.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    {Key1a, Val1a} = ?BDB:db_get(P2, DB3a, void, Key1a, <<>>, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB1a, void, Key4a, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2a, void, Key4b, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB3a, void, Key4c, <<>>, [])),

    % cleanup test state
    _ = ?BDB:db_truncate(P2, DB1a, void, []),
    _ = ?BDB:db_truncate(P2, DB2a, void, []),
    _ = ?BDB:db_truncate(P2, DB3a, void, []),

    ok = ?BDB:db_close(P2, DB1a, []),
    ok = ?BDB:db_close(P2, DB2a, []),
    ok = ?BDB:db_close(P2, DB3a, []),

    ok = ?BDB:env_close(P2, []),
    normal = ?BDB:shutdown(P2),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


env_open_with_retry(Port, Dir, Flags, Mode, CallerLineNum) ->
    try ?BDB:env_open(Port, Dir, Flags, Mode)
    catch
        {bdb_error, eagain} ->
            io:format("env_open ~w returned EAGAIN; retrying in 100ms", [CallerLineNum]),
            timer:sleep(100),
            env_open_with_retry(Port, Dir, Flags, Mode, CallerLineNum)
    end.

test_concurrent_access_to_threadsafe_valmaps(StartFunc) ->    
    io:format("\nPerforming test_concurrent_access_to_threadsafe_valmaps\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    %%case StartFunc of
    %%    start_pipe -> ?BDB:adjust_port_logging(P, enable, [all]);
    %%    start      -> do_nothing
    %%end,

    Loop = fun(0, CmdTags, _) ->
                   CmdTags;
              (NumOperationsTodo, CmdTags, Loop) ->
                   %%io:format("Doing put (~w to go)\n", [NumOperationsTodo-1]),
                   Loop(NumOperationsTodo - 1, 
                        [{?BDB:db_put_non_block(P, DB1, void, integer_to_list(NumOperationsTodo), "junk", []), receive_db_put_reply}
                         |CmdTags], 
                        Loop)
              end,

    %% enqueue some concurrent operations
    NumOperationsTodo = 50,   % >1 will fail until Db handles support concurrent use
    CmdTags = Loop(NumOperationsTodo, [], Loop),

    receive_tagged_port_replies(P#bdb_port_handle{receive_reply_timeout=500}, CmdTags),

    %%print_all_keys_and_values(P, DB1),

    ok = ?BDB:db_close(P, DB1, []),
    ok = ?BDB:db_close(P, DB2, []),
    ok = ?BDB:env_close(P, []),
    normal = ?BDB:shutdown(P),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_group_valmap_cleanup(StartFunc) ->
    
    io:format("\nPerforming test_group_valmap_cleanup\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),

    %% I just had to reduce the default threadcounts from 30 to 1 (see berkeley_db.xml for reason)
    %% This test requires more than that.
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_WORKERS, 30, 64*1024, 1000),
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 30, 64*1024, 1000),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    %% Remember: if group_valmap_cleanup needs to close a DB handle and there are ANY open txns, it will HARD-ABORT the OS process.

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB3 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB3, void, "test3.db", "", bdb_DB_HASH, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),


    %%io:format("\nChildren must have same group_id as parent\n"),
    TxnP1 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=1}, void, []),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=2}, TxnP1, [])),
    TxnC1 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=1}, TxnP1, []),
    Key1 = <<"key1">>,
    Val1 = <<"val1">>,
    ok = ?BDB:db_put(P, DB1, TxnC1, Key1, Val1, []),
    ok = ?BDB:txn_commit(P, TxnP1, []),
    {Key1, Val1} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),

    %%io:format("\nCleanup closes all valmaps open for that group-id, but no others\n"),
    TxnP2 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=2}, void, []),
    TxnC2 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=2}, TxnP2, []),
    Key2 = <<"key2">>,
    Val2 = <<"val2">>,
    ok = ?BDB:db_put(P, DB2, TxnC2, Key2, Val2, []),

    Key3a = <<"key3a">>,
    Val3a = <<"val3a">>,
    ok = ?BDB:db_put(P, DB3, void, Key3a, Val3a, []),

    TxnP3 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=3}, void, []),
    TxnC3 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=3}, TxnP3, []),
    Key3b = <<"key3b">>,
    Val3b = <<"val3b">>,
    ok = ?BDB:db_put(P, DB3, TxnC3, Key3b, Val3b, []),

    ?BDB:adjust_port_logging(P, enable, [group_valmap_cleanup_outputv, group_valmap_cleanup_work]),
    2 = ?BDB:group_valmap_cleanup(P, 3),
    ?BDB:adjust_port_logging(P, disable, [group_valmap_cleanup_outputv, group_valmap_cleanup_work]),

    ok = ?BDB:txn_commit(P, TxnP2, []),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_commit(P, TxnP3, [])),

    {Key2, Val2} = ?BDB:db_get(P, DB2, void, Key2, <<>>, []),

    %%io:format("\nAfter group cleanup is finished, can reuse that group_id\n"),
    %% (Failure due to reusing a group_id that is *currently* being cleaned up, is tested below)
    _TxnP2a = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=3}, void, []),

    {Key3a, Val3a} = ?BDB:db_get(P, DB3, void, Key3a, <<>>, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB3, void, Key3b, <<>>, [])),

%%     case StartFunc of
%%         start      -> ?BDB:adjust_port_logging(P, enable, [all]);
%%         start_pipe -> do_nothing
%%     end,

    %%io:format("\nGroup cleanup of in-progress start operations AND simultaneous stop\n"),
    Loop = fun(0, CmdTags, _) ->
                   CmdTags;
              (NumOperationsTodo, CmdTags, Loop) ->
                   Tag = case NumOperationsTodo rem 4 of
                       0 -> {?BDB:txn_begin_non_block(P#bdb_port_handle{group_valmap_id=2}, void, []),            receive_txn_begin_reply};
                       1 -> {?BDB:db_put_non_block(P, DB2, void, integer_to_list(NumOperationsTodo), "junk", []), receive_db_put_reply};
                       2 -> {?BDB:txn_begin_non_block(P#bdb_port_handle{group_valmap_id=3}, void, []),            receive_txn_begin_reply};
                       3 -> {?BDB:db_put_non_block(P, DB3, void, integer_to_list(NumOperationsTodo), "junk", []), receive_db_put_reply}
                   end,
                   Loop(NumOperationsTodo - 1, [Tag|CmdTags], Loop)
              end,

    % Using 8 means that we see RESERVED and in-use entries AND some of the in-use entries have refcounts>1
    % Using 80 found a bug in that the highest txn (#31) was skipped during cleanup as it was believed to be a child.
    NumOperationsTodo = 30,   % >1 will fail unless Db handles support concurrent use
    LoopCmdTags = Loop(NumOperationsTodo, [], Loop),
    %%?BDB:adjust_port_logging(P, enable, [all]),
    CmdTags = [{?BDB:group_valmap_cleanup_non_block(P, 2), receive_group_valmap_cleanup_reply} | LoopCmdTags],

    case StartFunc of
        start ->          
            %%io:format("\nCannot reuse a group_id while it is being cleaned-up\n"),
            %% IMPORTANT This test will FAIL if 0 threads are available (either VM threads or private threadpool, it doesn't matter),
            %% as the group_valmap_cleanup above is done synchronously, so has already finished (even though it was non-blocking mode)
            ?EXPECT_BDB_ERROR(valmap_group_being_cleaned, 
                              ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=2}, void, []));
        start_pipe -> 
            %% In Pipe mode we get a hard abort if we close the port while replies are still coming in.
            receive_tagged_port_replies(P#bdb_port_handle{receive_reply_timeout=500}, CmdTags)
    end,

    normal = ?BDB:shutdown(P),

    %% We DON'T get replies to all the messages as the port is closed before the are executed -- but some probably made it through.
    %% So empty our message buffer.

    timer:sleep(1000),
    {messages, []} = process_info(self(), messages),


    %%io:format("\nTest multiple overlapping calls to cleanup the same group\n"),
    %% This can happen:
    %%    1. release_bdb_port_handle() calls group_valmap_cleanup() directly (from client process) before telling the port-coordinator that it is done.
    %%    2. If the client crashes before telling the port-coordinator that it is done, the port-coordinator's  handle_info({'EXIT', ?.})    will call group-valmap-cleanup again.   
    %%        [e.g. If the first call to group-valmap-cleanup takes an absurdly long time (>20s) so that the client crashes due to a gen_server timeout, then this would happen.
    %%         In that case the much more serious problem is why is the first group-valmap-cleanup command taking far too long?   Is the valmap_stop threadpool overloaded? It currently has 10 threads, same as the worker pool)
    %% or
    %%    1. release_bdb_port_handle() calls group_valmap_cleanup() directly.
    %%    2. If port_coordinator tries to shutdown, then port_coordinator also calls group_valmap_cleanup()  [as the client process has not un-registered yet]
    %% or:
    %%     1. BDB_H:do_txn() gets a deadlock and calls group_valmap_cleanup() [so it can retry the txn]
    %%     2. do_txn() crashes, or port_coordinator tries to shutdown, and port_coordinator calls group_valmap_cleanup()

    ensure_empty_dir(Dir),

    P2 = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(P2, disable, [all]),
    ok = env_open_with_retry(P2, Dir, [bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),

    NumDBs = 100,
    ok  = ?BDB:set_num_valmap_slots(P2, bdb_db, 100),
    lists:foreach(fun(_) -> ?BDB:db_create(P2#bdb_port_handle{group_valmap_id=4}, []) end,       lists:seq(1, NumDBs)),

    NumTxns = 100,
    ok  = ?BDB:set_num_valmap_slots(P2, bdb_db_txn, 100),
    lists:foreach(fun(_) -> ?BDB:txn_begin(P2#bdb_port_handle{group_valmap_id=4}, void, []) end, lists:seq(1, NumTxns)),

    timer:sleep(1), % yied, to ensure that the next two instructions aren't split by premptive scheduling
    OverlapTag1 = ?BDB:group_valmap_cleanup_non_block(P2, 4),
    ?EXPECT_BDB_ERROR(valmap_group_being_cleaned,
                      ?BDB:group_valmap_cleanup(P2, 4)),

    case ?BDB:receive_group_valmap_cleanup_reply(P2, OverlapTag1) of 
        N when N == NumDBs + NumTxns ->
            ok;
        Other -> throw({unexpected_at_line, ?LINE, Other})
    end,
    normal = ?BDB:shutdown(P2),

    timer:sleep(100),
   {messages, []} = process_info(self(), messages),

    ok.


test_consume_wait(StartFunc) ->
    
    io:format("\nPerforming test_consume_wait\n"),

    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),

    %% By default the driver has 0 threads in pool #2 (the one used by consume-wait)
    %% as it is a rarely used feature
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_CONSUME_WAIT, 5, 64*1024, 5),

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),
    ok = ?BDB:env_set_flags(P, [bdb_DB_AUTO_COMMIT], 1),
    DB1 = ?BDB:db_create(P#bdb_port_handle{group_valmap_id=1}, []),

    %% We are actually going to put stuff so these configurations are necessary (otherwise fixed-length is 0)
    %% And they must be done before the database is first created.
    RecordSize = 16, %% bytes
    ok = ?BDB:db_set_re_len(P, DB1, RecordSize),  %% arg is bytes
    ok = ?BDB:db_set_re_pad(P, DB1, 32),         %% arg is byte value (default is 32, i.e. space character)
    ok = ?BDB:db_set_q_extentsize(P, DB1, 100),  %% arg is #db pages, not bytes

    ok = ?BDB:db_open(P, DB1, void, "test_queue.db", "", bdb_DB_QUEUE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    %% a non-void transaction is required
    ?EXPECT_BDB_ERROR(badarg, ?BDB:db_get(P#bdb_port_handle{receive_reply_timeout=infinity}, DB1, void, <<>>, <<>>, [bdb_DB_CONSUME_WAIT])),

    %% receive_reply_timeout must be infinity (use txn_set_timeout and a periodic deadlock detector to set a real timeout)
    Txn0 = ?BDB:txn_begin(P, void, []),
    ?EXPECT_BDB_ERROR({db_get_consume_wait_receive_reply_timeout_not_infinity, P},
                      ?BDB:db_get(P, DB1, Txn0, <<>>, <<>>, [bdb_DB_CONSUME_WAIT])),
    ok = ?BDB:txn_commit(P, Txn0, []),

    DbStatQueue = ?BDB:db_stat(P, DB1, void, bdb_DB_READ_COMMITTED),
    {value, {db_type,bdb_DB_QUEUE}} = lists:keysearch(db_type, 1, DbStatQueue),
    {value, {qs_re_len,16}} = lists:keysearch(qs_re_len, 1, DbStatQueue),
 
    %% shutdown with race condition does not hang  
    %% (this test once caught a bug where stop called unblock_all_consume_wait but invoke had not started it yet, 
    %% so shutdowno *did* hang)
    Txn1 = ?BDB:txn_begin(P#bdb_port_handle{group_valmap_id=1}, void, []),
    _Tag = ?BDB:db_get_non_block(P, DB1, Txn1, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),   %% non_block ignores receive_reply_timeout, so it doesn't have to be infinity
    normal = ?BDB:shutdown(P),

    P2 = ?BDB:StartFunc(),

    %% I just had to reduce the default threadcounts from 30 to 1 (see berkeley_db.xml for reason)
    %% This test requires more than that.
    ok = ?BDB:set_threadpool_params(P2, ?BDB_THREADPOOL_ID_WORKERS, 30, 64*1024, 1000),
    ok = ?BDB:set_threadpool_params(P2, ?BDB_THREADPOOL_ID_VALMAP_STOP, 30, 64*1024, 1000),

    ok = ?BDB:set_threadpool_params(P2, ?BDB_THREADPOOL_ID_CONSUME_WAIT, 5, 64*1024, 5),
    ok = env_open_with_retry(P2, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:env_set_flags(P2, [bdb_DB_AUTO_COMMIT], 1),
    ok = ?BDB:set_lk_detect(P2, bdb_DB_LOCK_DEFAULT),
    DB2 = ?BDB:db_create(P2#bdb_port_handle{group_valmap_id=1}, []),
    ok = ?BDB:db_open(P2, DB2, void, "test_queue.db", "", bdb_DB_QUEUE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),


    %% shutdown without race condition (consume_wait is allowed to start before shutdown) does not hang  
    Txn2 = ?BDB:txn_begin(P2#bdb_port_handle{group_valmap_id=1}, void, []),
    _Tagok = ?BDB:db_get_non_block(P2, DB2, Txn2, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    timer:sleep(100),
    normal = ?BDB:shutdown(P2),

    P3 = ?BDB:StartFunc(),

    %% I just had to reduce the default threadcounts from 30 to 1 (see berkeley_db.xml for reason)
    %% This test requires more than that.
    ok = ?BDB:set_threadpool_params(P3, ?BDB_THREADPOOL_ID_WORKERS, 30, 64*1024, 1000),
    ok = ?BDB:set_threadpool_params(P3, ?BDB_THREADPOOL_ID_VALMAP_STOP, 30, 64*1024, 1000),

    ok = ?BDB:set_threadpool_params(P3, ?BDB_THREADPOOL_ID_CONSUME_WAIT, 5, 64*1024, 5),
    ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    ok = ?BDB:set_lk_detect(P3, bdb_DB_LOCK_DEFAULT),
    ok = ?BDB:env_set_flags(P3, [bdb_DB_AUTO_COMMIT], 1),
    DB3 = ?BDB:db_create(P3#bdb_port_handle{group_valmap_id=1}, []),
    ok = ?BDB:db_open(P3, DB3, void, "test_queue.db", "", bdb_DB_QUEUE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    %%  - lock timeouts (e.g. when using lock_vec to unblock a db_get(CONSUME_WAIT) 
    %%    should cause DB_LOCK_NOTGRANTED, not DB_LOCK_DEADLOCK (so we can distinguish between the two)
    ok = ?BDB:env_set_flags(P3, [bdb_DB_TIME_NOTGRANTED], 1),

    %% group_valmap_cleanup with race condition does not hang 
    %%?BDB:adjust_port_logging(P3, enable, [all]),
    Txn3 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=2}, void, []),
    Tag1 = ?BDB:db_get_non_block(P3, DB3, Txn3, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    timer:sleep(1),
    1 = ?BDB:group_valmap_cleanup(P3, 2),
    %% expect the db_get to return an error (consume-wait was interrupted)
    %% note that we only get DB_LOCK_NOTGRANTED instead of DB_LOCK_DEADLOCK because of the above env_set_flags call.
    ?EXPECT_BDB_ERROR(bdb_DB_LOCK_NOTGRANTED, ?BDB:receive_db_get_reply(P3#bdb_port_handle{receive_reply_timeout=500}, Tag1)),
    %% And the txn will have been aborted by the group_valmap_cleanup
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P3, Txn3)),


    %% group_valmap_cleanup without race condition does not hang 
    Txn4 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=3}, void, []),
    Tag2 = ?BDB:db_get_non_block(P3, DB3, Txn4, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    timer:sleep(100),
    1 = ?BDB:group_valmap_cleanup(P3, 3),
    %% expect the db_get to return badarg
    ?EXPECT_BDB_ERROR(bdb_DB_LOCK_NOTGRANTED, ?BDB:receive_db_get_reply(P3#bdb_port_handle{receive_reply_timeout=500}, Tag2)),
    %% And the txn will have been aborted by the group_valmap_cleanup
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P3, Txn4)),


    %% commiting the transaction does consume the values (we don't see them again)
    NewKey1 = ?BDB:db_put_append(P3, DB3, void, "data1", []),
    NewKey2 = ?BDB:db_put_append(P3, DB3, void, "DATA2", []),

    Txn5 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=1}, void, []),
    {NewKey1, <<"data1", _Padding1/binary>>} = ?BDB:db_get(P3#bdb_port_handle{receive_reply_timeout=infinity}, DB3, Txn5, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    {NewKey2, <<"DATA2", _Padding2/binary>>} = ?BDB:db_get(P3#bdb_port_handle{receive_reply_timeout=infinity}, DB3, Txn5, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    ok = ?BDB:txn_commit(P3, Txn5, []),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<>>, <<>>, [bdb_DB_CONSUME])),


    %% aborting the txn replaces the consumed value (so we will see it again next time)
    NewKey1a = ?BDB:db_put_append(P3, DB3, void, "2daattaa1", []),
    NewKey2a = ?BDB:db_put_append(P3, DB3, void, "2DATA2", []),

    Txn6 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=1}, void, []),
    {NewKey1a, Val1a} = ?BDB:db_get(P3#bdb_port_handle{receive_reply_timeout=infinity}, DB3, Txn6, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    <<"2daattaa1", _Padding3/binary>> = Val1a,
    {NewKey2a, Val2a} = ?BDB:db_get(P3#bdb_port_handle{receive_reply_timeout=infinity}, DB3, Txn6, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    <<"2DATA2", _Padding4/binary>> = Val2a,
    ok = ?BDB:txn_abort(P3, Txn6),
    {NewKey1a, Val1a} = ?BDB:db_get(P3, DB3, void, <<>>, <<>>, [bdb_DB_CONSUME]),
    {NewKey2a, Val2a} = ?BDB:db_get(P3, DB3, void, <<>>, <<>>, [bdb_DB_CONSUME]),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<>>, <<>>, [bdb_DB_CONSUME])),


    %%io:format("It is an error (but safe) to use the same txn while the consume-wait is blocked\n"),
    Txn7 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=1}, void, []),
    Tag3 = ?BDB:db_get_non_block(P3, DB3, Txn7, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    %% try to receive the msg (expect timeout)
    PortWithTimeout500 = P3#bdb_port_handle{receive_reply_timeout=500},
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag3),

    %% It is an error to use the txn, as it is in-use
    ?EXPECT_BDB_ERROR(valmap_in_use, ?BDB:txn_commit(P3, Txn7, [])),

    %% put something into the queue
    NewKey3 = ?BDB:db_put_append(P3, DB3, void, "moredata", []),

    %% receive the msg
    {NewKey3, <<"moredata", _Padding5/binary>>} = ?BDB:receive_db_get_reply(P3#bdb_port_handle{receive_reply_timeout=500}, Tag3),

    ok = ?BDB:txn_commit(P3, Txn7, []),
    %% try to consume the same value again (a polling consume, not a blocking consume-wait)
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<>>, <<>>, [bdb_DB_CONSUME])),


    %% We cannot start more consume-wait operations than we have threads in the threadpool
    ok = ?BDB:set_threadpool_params(P3, 2, 2, 64*1024, 10),   % pool #2, 2 threads, queue length limit 10
    Txn10 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=10}, void, []),
    Tag10 = ?BDB:db_get_non_block(P3, DB3, Txn10, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    Txn11 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=11}, void, []),
    Tag11 = ?BDB:db_get_non_block(P3, DB3, Txn11, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    Txn12 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=12}, void, []),
    %%?BDB:adjust_port_logging(P3, enable, [all]),
    Tag12 = ?BDB:db_get_non_block(P3, DB3, Txn12, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    %% First two should be hung, but the third one should have returned immediately as it couldn't find a thread
    ?EXPECT_BDB_ERROR({edtk_threadpool_too_busy, ?BDB_THREADPOOL_ID_CONSUME_WAIT}, ?BDB:receive_db_get_reply(PortWithTimeout500, Tag12)),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag10),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag11),

    %% Unblock one of them with a message
    NewKey4 = ?BDB:db_put_append(P3, DB3, void, "jdmoredata", []),
    StillStuck1 = try ?BDB:receive_db_get_reply(PortWithTimeout500, Tag10 ) of
                      {NewKey4, <<"jdmoredata", _Padding10/binary>>} -> 
                          Tag11  % still blocked
                  catch
                      exit:{bdb_receive_reply_timeout, {PortWithTimeout500, Tag10}} -> 
                          try ?BDB:receive_db_get_reply(PortWithTimeout500, Tag11) of
                              {NewKey4, <<"jdmoredata", _Padding11/binary>>} ->
                                  Tag10 % still block
                          catch
                              exit:{bdb_receive_reply_timeout, {PortWithTimeout500, Tag11}} -> 
                                  throw(both_should_not_still_be_blocked)
                          end
                  end,

    %% There is a free thread now
    Txn13 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=13}, void, []),
    Tag13 = ?BDB:db_get_non_block(P3, DB3, Txn13, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag13),

    %% The remaining first one is still stuck
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, StillStuck1),

    %% And all threads are full again
    Txn14 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=14}, void, []),
    Tag14 = ?BDB:db_get_non_block(P3, DB3, Txn14, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    ?EXPECT_BDB_ERROR({edtk_threadpool_too_busy, ?BDB_THREADPOOL_ID_CONSUME_WAIT}, ?BDB:receive_db_get_reply(PortWithTimeout500, Tag14)),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag13),

    %% The remaining first one is still stuck
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, StillStuck1),

    %% Unblock one with group_valmap_cleanup
    1 = ?BDB:group_valmap_cleanup(P3, 13),

    ?EXPECT_BDB_ERROR(bdb_DB_LOCK_NOTGRANTED, ?BDB:receive_db_get_reply(PortWithTimeout500, Tag13)),
    %% And the txn will have been aborted by the group_valmap_cleanup
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P3, Txn13)),

    %% There is a free thread now
    Txn15 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=15}, void, []),
    Tag15 = ?BDB:db_get_non_block(P3, DB3, Txn15, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),
    %% The others are still stuck
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, StillStuck1),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag15),

    %% And all threads are full again
    Txn16 = ?BDB:txn_begin(P3#bdb_port_handle{group_valmap_id=16}, void, []),
    Tag16 = ?BDB:db_get_non_block(P3, DB3, Txn16, <<>>, <<>>, [bdb_DB_CONSUME_WAIT]),

    ?EXPECT_BDB_ERROR({edtk_threadpool_too_busy, ?BDB_THREADPOOL_ID_CONSUME_WAIT}, ?BDB:receive_db_get_reply(PortWithTimeout500, Tag16)),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, StillStuck1),
    ?EXPECT_RECEIVE_REPLY_TIMEOUT(receive_db_get_reply, PortWithTimeout500, Tag15),

    normal = ?BDB:shutdown(P3),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_use_of_one_port_instance_from_multiple_erlang_processes(StartFunc) ->
    
    io:format("\nPerforming test_use_of_one_port_instance_from_multiple_erlang_processes\n"),

    case StartFunc of 
        start ->
            Dir = "txn_tests_env",
            ensure_empty_dir(Dir),

            P = ?BDB:StartFunc(),

            ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
            ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

            PrimaryValmapGroupId = 0,
            SpawnedValmapGroupId = 1,
            DB1 = ?BDB:db_create(P#bdb_port_handle{group_valmap_id=PrimaryValmapGroupId}, []),
            ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
            %% spawned processes uses a different db purely to avoid self-deadlock on a bdb page lock, which would be an irrelevant 
            %% side effect of this very artificial test (real applications should never pass Txn or Cursor handles between processes)
            DB2 = ?BDB:db_create(P#bdb_port_handle{group_valmap_id=PrimaryValmapGroupId}, []),
            ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

            PrimaryPid = self(),
            SpawnedPid = spawn_link(fun() -> receive have_set_up_monitor -> ok end,
                                             SpawnedPort=P#bdb_port_handle{issued_to_pid=self()},
                                             ok = ?BDB:db_put(SpawnedPort, DB2, void, "KeyFromSpawnedProcess", <<"val-from-spawned-process">>, []),

                                             %% io:format("\ntest_use_of_one_port_instance_from_multiple_erlang_processes: spawned db_put returned OK\n"),
                                             Txn = ?BDB:txn_begin(SpawnedPort#bdb_port_handle{group_valmap_id=SpawnedValmapGroupId}, void, []),
                                             ok = ?BDB:db_put(SpawnedPort, DB2, Txn, "TxnKeyFromSpawnedProcess2", <<"txn-val-from-spawned-process2">>, []),

                                             PrimaryPid ! {open_txn_handle, Txn},
                                             %% io:format("\ntest_use_of_one_port_instance_from_multiple_erlang_processes: send msg and about to block for exit msg\n"),
                                             receive
                                                 exit -> ok
                                             end
                                    end),

            %% The tests run with trap_exit = false, so if we want to know if a process exits normally we need to monitor it
            MonitorRefSpawned = erlang:monitor(process, SpawnedPid),
            SpawnedPid ! have_set_up_monitor,

            ok = ?BDB:db_put(P, DB1, void, "KeyFromPrimaryProcess", <<"val-from-primary-process">>, []),
            %%io:format("\ntest_use_of_one_port_instance_from_multiple_erlang_processes: primary db_put returned OK\n"),

            OpenTxn = receive
                          {open_txn_handle, Txn} -> 
                              %%io:format("\ntest_use_of_one_port_instance_from_multiple_erlang_processes: primary received txn handle\n"),
                              Txn;
                          Other -> throw({unexpected_msg, Other})
                      end,

            %%?BDB:adjust_port_logging(P, enable, [all]),

            %% It is still open (and the uncommitted modification is still present)
            {<<"TxnKeyFromSpawnedProcess2">>, <<"txn-val-from-spawned-process2">>} = ?BDB:db_get(P, DB2, OpenTxn, <<"TxnKeyFromSpawnedProcess2">>, <<>>, []),

            %% But we can close the group from a different process and so abort the txn
            1 = ?BDB:group_valmap_cleanup(P, SpawnedValmapGroupId),
            ?EXPECT_BDB_ERROR(badarg, ?BDB:db_get(P, DB2, OpenTxn, <<"TxnKeyFromSpawnedProcess2">>, <<>>, [])),  % generation-id has changed
            
            %% The other modifications from both processes are still present
            {<<"KeyFromPrimaryProcess">>, <<"val-from-primary-process">>} = ?BDB:db_get(P, DB1, void, <<"KeyFromPrimaryProcess">>, <<>>, []),
            {<<"KeyFromSpawnedProcess">>, <<"val-from-spawned-process">>} = ?BDB:db_get(P, DB2, void, <<"KeyFromSpawnedProcess">>, <<>>, []),
            ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB2, void, <<"TxnKeyFromSpawnedProcess2">>, <<>>, [])),

            SpawnedPid ! exit,
            normal = ?BDB:shutdown(P),
            receive
                {'DOWN', MonitorRefSpawned, process, SpawnedPid, normal} -> ok
            end;

        start_pipe -> io:format("TODO: test_use_of_one_port_instance_from_multiple_erlang_processes in pipe-driver mode\n")

    end,

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_port_coordinator(StartFunc) ->

    io:format("\nPerforming test_basic_port_coordinator\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(
                        Dir, 
                        [{port_start_function, StartFunc},
                         {checkpoint_receive_reply_timeout_milliseconds, 20 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},
                         {max_num_concurrent_client_procs, 2},
                         {default_db_config_proplist, [{db_open_flags,  [bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT]},
                                                       {page_size_bytes, 16 * 1024}, {set_btree_min_keys_per_page, 2}]}]),

    %% See comments at the top of the file on why tests are expected to preserve the old value of trap_exit 
    %% (if they complete normally -- i.e. I don't bother putting it in a try/after as we don't execute subsequent tests if 
    %% one crashes; a real test driver should do that).
    io:format("test_basic_port_coordinator: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),

    P = ?BDB_PC:acquire_bdb_port_handle(),

    %% It is an error to call acquire_bdb_port_handle twice from the same process
    ?EXPECT_PORT_COORD_ERROR(caller_already_has_a_bdb_port_handle, ?BDB_PC:acquire_bdb_port_handle()),

    %% Check enforcement of limit on number of clients
    TestDriver = self(),
    %% this one should succeed (and consume the last available place)
    C2 = spawn_link(fun() -> 
                            _ = ?BDB_PC:acquire_bdb_port_handle(),
                            receive exit_now -> ok end
                    end),
    %% This one should fail
    C3a = spawn_link(fun() ->
                           ?EXPECT_PORT_COORD_ERROR({too_many_client_procs, 2, 2}, ?BDB_PC:acquire_bdb_port_handle())
                   end),
    receive {'EXIT', C3a, normal} -> ok end,
    %% Tell C2 to release its handle (indirectly, via exiting -- port-coordinator sees exit and cleans up)
    C2 ! exit_now,
    receive {'EXIT', C2, normal} -> ok end,
    %% Now this one should succeed
    C2b = spawn_link(fun() ->
                             P2b = ?BDB_PC:acquire_bdb_port_handle(),
                             TestDriver ! ok_C2b,
                             receive release_now -> ok end,
                             ?BDB_PC:release_bdb_port_handle(P2b),
                             TestDriver ! released_C2b,
                             receive exit_now -> ok end
                     end),
    receive ok_C2b -> ok end,
    %% And the next one should fail
    C3b = spawn_link(fun() ->
                           ?EXPECT_PORT_COORD_ERROR({too_many_client_procs, 2, 2}, ?BDB_PC:acquire_bdb_port_handle())
                   end),
    receive {'EXIT', C3b, normal} -> ok end,
    %% Tell C2b to release its handle explicitly
    C2b ! release_now,
    receive released_C2b -> ok end,
    %% And this one should succeed now
    C2c = spawn_link(fun() -> 
                           _ = ?BDB_PC:acquire_bdb_port_handle()
                   end),
    receive {'EXIT', C2c, normal} -> ok end,
    %% cleanup
    C2b ! exit_now,
    receive {'EXIT', C2b, normal} -> ok end,

    %% Get a new DB handle
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

    %% Get an existing DB handle (should get the same handle not a new one)
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

    %% Open same database with different type
    ?EXPECT_PORT_COORD_ERROR({mismatched_db_type, bdb_DB_BTREE, bdb_DB_QUEUE}, 
                             ?BDB_PC:get_db_handle(P, bdb_DB_QUEUE, "test_coord.db")),

    %% Open same database with different but compatible config (a subset of the default)
    DB1 = ?BDB_PC:get_db_handle_with_db_config(P, bdb_DB_BTREE, "test_coord.db", [{page_size_bytes, 16384}]),

    %% Name must be a list not a binary
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:get_db_handle(P, bdb_DB_QUEUE, <<"test_coord.db">>)),

    %% Open same database with different and incompatible config
    ?EXPECT_PORT_COORD_ERROR({mismatched_db_config, [{page_size_bytes, 16384, 4096}]}, 
                             ?BDB_PC:get_db_handle_with_db_config(P,
                                                                  bdb_DB_BTREE, 
                                                                  "test_coord.db", 
                                                                  [{page_size_bytes, 4096}])),

    %% Try completely malformed config (not a proper list)
    ?EXPECT_PORT_COORD_ERROR({bad_db_config, malformed_proplist, {page_size_bytes, 16384}}, 
                             ?BDB_PC:get_db_handle_with_db_config(P,
                                                                  bdb_DB_BTREE, 
                                                                  "test_coord1.db", 
                                                                  {page_size_bytes, 16384})),

    ?EXPECT_PORT_COORD_ERROR({bad_db_config, 
                              {unrecognized_db_config_option, {no_such_option, 16384}}, 
                              [{set_btree_min_keys_per_page,2}, 
                               {page_size_bytes,16384}, 
                               {db_open_flags,[bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT]},
                               {no_such_option,16384}]}, 
                             ?BDB_PC:get_db_handle_with_db_config(P,
                                                                  bdb_DB_BTREE, 
                                                                  "test_coord1.db", 
                                                                  [{no_such_option, 16384}])),

    %% set_flags requires should take a list of flagmask atoms
    ?EXPECT_PORT_COORD_ERROR({bad_db_config, 
                              function_clause, 
                              [{set_btree_min_keys_per_page, 2}, 
                               {page_size_bytes,16384}, 
                               {db_open_flags,[bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT]},
                               {set_flags, [12]}]},
                             ?BDB_PC:get_db_handle_with_db_config(P,
                                                                  bdb_DB_BTREE, 
                                                                  "test_coord2.db", 
                                                                  [{set_flags, [12]}])),

    %% A file with a single database cannot contain another database
    ?EXPECT_PORT_COORD_ERROR({bdb_error, einval}, 
                             ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_coord.db", "another"})),

    ["test_coord.db"] = lists:sort( ?BDB_PC:list_all_open_databases(P) ),

    %% Test close_unused_database
    %% Names must be lists not binaries
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:close_unused_database(P, <<"test_coord.db">>)),
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:close_unused_database(P, {<<"test_coord.db">>, ""})),
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:close_unused_database(P, {"test_coord.db", <<"">>})),
    ok = ?BDB_PC:close_unused_database(P, "test_coord.db"),

    %% now using DB1 should result in badarg, as the valmap slot's generation-id will have changed
    ?EXPECT_BDB_ERROR(badarg,
                      ?BDB:db_put(P, DB1, void, "should-fail-key", <<"should-fail-record">>, [])),

    %% and port_coordinator should have forgotten the handle
    [] = ?BDB_PC:list_all_open_databases(P),
    DB1ReOpened = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),
    if DB1 =:= DB1ReOpened -> throw(close_unused_database_did_not_forget_handle);
       true                -> ok
    end,
    ["test_coord.db"] = lists:sort( ?BDB_PC:list_all_open_databases(P) ),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, 
                      ?BDB:db_get(P, DB1ReOpened, void, "should-fail-key", <<>>, [])),

    %% And close_unused_database on a database that we have not yet opened, should fail
    ?EXPECT_PORT_COORD_ERROR({unknown_database, {"unknown.db", ""}},
                             ?BDB_PC:close_unused_database(P, "unknown.db")),    


    %% Open multiple databases in same physical file
    DB1a = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_multi.db", "first"}),
    DB1b = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_multi.db", "second"}),
    if DB1a == DB1b ->
            throw(different_databases_within_same_file_return_same_handle);
       true -> ok
    end,

    %% Make sure we can get the existing handles again
    DB1a = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_multi.db", "first"}),
    DB1b = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_multi.db", "second"}),

    %% Names must be lists not binaries
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {<<"test_multi.db">>, "first"})),
    {'EXIT', {function_clause, _}} = (catch ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, {"test_multi.db", <<"first">>})),

    [{"test_multi.db", "first"},
     {"test_multi.db", "second"},
     "test_coord.db"
     ] = lists:sort( ?BDB_PC:list_all_open_databases(P) ),

    %% If we ignore the port coordinator and open the multi-database file directly using DB_UNKNOWN, we should get 
    %% a meta-data database with names of databases
    %%
    %%  "It is not an error to open a file that contains multiple
    %%  databases without specifying a database name, however the
    %%  database type should be specified as DB_UNKNOWN and the
    %%  database must be opened read-only. The handle that is returned
    %%  from such a call is a handle on a database whose key values
    %%  are the names of the databases stored in the database file and
    %%  whose data values are opaque objects. No keys or data values
    %%  may be modified or stored using this database handle.
    DB_meta = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB_meta, void, "test_multi.db", "", bdb_DB_UNKNOWN, [bdb_DB_RDONLY], 0),
    ok = verify_all_keys(P, DB_meta, [<<"first">>,  <<"second">>]),
    ok = ?BDB:db_close(P, DB_meta, []),

    %% If the main test process opens the port then it must release it, or the port controller server will 
    %% kill the test process when it terminates.
    ok = ?BDB_PC:release_bdb_port_handle(P),

    %% Can't release it twice
    ?EXPECT_PORT_COORD_ERROR(caller_does_not_have_a_bdb_port_handle, ?BDB_PC:release_bdb_port_handle(P)),

    %% As we are trapping exits we DO receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next  test might get 'already_started' when it tries to start the server.
    receive
        {'EXIT', ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    throw_if_any_exit_msgs(),  % must be done BEFORE restoring trap_exit
    io:format("test_basic_port_coordinator: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_environment_locking(StartFunc) ->

    io:format("\nPerforming test_environment_locking\n"),

    %% The locking scheme in env_open/env_close is designed to achieve these properties:
    %%
    %% 1. Prevent multiple nodes accessing the same environment simultaneously
    %% 2. Prevent different port-coordinators in the same node from accessing the same environment simultaneously
    %% 3. Prevent overlapping shutdown/restart of the same port-coordinator from accessing the same environment simultaneously
    %% 4. Allows safe use of DB_PRIVATE in both linked-in and pipe modes
    %%    (i.e. only a single OS process, with a single DB_ENV handle may access the environment at any time.)

    %% We can test 2 here easily, and perhaps 3.
    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid1} = ?BDB_PC:start_link(port_coord1, Dir, [{port_start_function, StartFunc}]),

    %% we should not be able to create another port-coordinator on the same environment

    %% The failure of env_open occurs in ?BDB_PC:init().  Even though init() is synchronous with start_link, it is run 
    %% in the spawned process, so we get an EXIT signal IN ADDITION to the error return from start_link
    %% user_default:dbgon(?BDB_PC),
    io:format("test_environment_locking: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),
    ?AT,
    {error, bdb_environment_is_locked} = ?BDB_PC:start_link(second_port_coord, Dir, [{port_start_function, StartFunc}]),
    ?AT,
    receive {'EXIT', _SomePid, bdb_environment_is_locked} -> ok end,
    ?AT,
    io:format("test_environment_locking: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),
    ?AT,
    undefined = whereis(second_port_coord),
    ?AT,

    %% Use ServerPid1 to show that it wasn't affected by the attempt to hijack it's environment
    ok_result = ?BDB_H:do_txn(ServerPid1, {?MODULE, test_txn_fun_put_1, {"key1", "committed1", ok_result}}, 0),

    %% Shutdown ServerPid1
    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef1 = erlang:monitor(process, ServerPid1),
    ?BDB_PC:stop(ServerPid1),
    receive {'DOWN', MonitorRef1, process, ServerPid1, normal} -> ok end,
    
    %% Then start second_port_coord, and show that it works.
    {ok, ServerPid2} = ?BDB_PC:start_link(second_port_coord, Dir, [{port_start_function, StartFunc}]),
    ServerPid2 = whereis(second_port_coord),
    {<<"key1">>, <<"committed1">>} =?BDB_H:do_txn(ServerPid2, {?MODULE, test_txn_fun_get, {"test.db", "key1"}}, 0),

    %% clean shutdown
    MonitorRef2 = erlang:monitor(process, ServerPid2),
    ?BDB_PC:stop(ServerPid2),
    receive {'DOWN', MonitorRef2, process, ServerPid2, normal} -> ok end,

    ok.


test_basic_port_coordinator_checkpoint_and_cache_trickle(StartFunc) ->

    io:format("\nPerforming test_basic_port_coordinator_checkpoints_and_cache_trickle\n"),

    %% Port coord has separate code for each combination of these, so need to test them all.

    Dir = "txn_tests_env",

    ensure_empty_dir(Dir),
    start_stop_port_coord(Dir, [{port_start_function, StartFunc},
                                {cache_trickle_milliseconds_and_percent,                  disabled},
                                {checkpoint_seconds_and_min_kb,                           disabled},
                                {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},
                                {checkpoint_receive_reply_timeout_milliseconds,           5000},
                                {enable_port_logging_before_open_environment,             [command_invoke_end]}
                               ]),

    ensure_empty_dir(Dir),
    start_stop_port_coord(Dir, [{port_start_function, StartFunc},
                                {cache_trickle_milliseconds_and_percent,                  {100, 20}},
                                {checkpoint_seconds_and_min_kb,                           disabled},
                                {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},
                                {checkpoint_receive_reply_timeout_milliseconds,           5000},
                                {enable_port_logging_before_open_environment,             [command_invoke_end]}
                               ]), 

    start_stop_port_coord(Dir, [{port_start_function, StartFunc},
                                {cache_trickle_milliseconds_and_percent,                  disabled},
                                {checkpoint_seconds_and_min_kb,                           {1, 100}},
                                {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},
                                {checkpoint_receive_reply_timeout_milliseconds,           5000},
                                {enable_port_logging_before_open_environment,             [command_invoke_end]}
                               ]),

    start_stop_port_coord(Dir, [{port_start_function, StartFunc},
                                {cache_trickle_milliseconds_and_percent,                  {100, 20}},
                                {checkpoint_seconds_and_min_kb,                           {1, 100}},
                                {do_parallel_trickle_and_checkpoint,                      enabled},    % PARALLEL
                                {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},
                                {checkpoint_receive_reply_timeout_milliseconds,           5000},
                                {{thread_count_stacksize_qlenlimit, housekeeping},        {2, 32*1024, 10}}, % TWO thread
                                {enable_port_logging_before_open_environment,             [command_invoke_end]}
                               ]),

    start_stop_port_coord(Dir, [{port_start_function, StartFunc},
                                {cache_trickle_milliseconds_and_percent,                  {100, 20}},
                                {checkpoint_seconds_and_min_kb,                           {1, 100}},
                                {do_parallel_trickle_and_checkpoint,                      disabled},   % SERIAL
                                {cache_throttle_io_write_ops_and_sleep_microseconds,      disabled},
                                {checkpoint_receive_reply_timeout_milliseconds,           5000},
                                {{thread_count_stacksize_qlenlimit, housekeeping},        {1, 32*1024, 10}}, % ONE thread
                                {enable_port_logging_before_open_environment,             [command_invoke_end]}
                               ]),
    ok.

start_stop_port_coord(Dir, Config) ->
    ensure_empty_dir(Dir),
    {ok, ServerPid} = ?BDB_PC:start_link(Dir, Config),
    MonitorRef = erlang:monitor(process, ServerPid),

    timer:sleep(5000),

    ok = ?BDB_PC:stop(),
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages).


test_port_coordinator_threadpool_deadlock(StartFunc) ->

    io:format("\nPerforming test_port_coordinator_threadpool_deadlock\n"),

    %% This verifies that the default port-coordinator config (DB_ENV->set_timeout(DB_LOCK_TIMEOUT) and running periodic deadlock detector)
    %% will resolve the kind of threadpool deadlock described in the 'TODO' file (group-valmap-cleanup does not yet handle 
    %% valmaps that are 'in used' but not currently being executed -- i.e. are used in operations that are enqueued in a threadpool 
    %% queue but not yet running, so, if all threads in the threadpool are blocked on locks held by a txn that is 'in use' in an 
    %% operation that is stuck in a queue then we have a threadpool deadlock).
    %% When group_valamp_cleanup is improved to be able to abort txns that are just sitting in threadpool queues, 
    %% the periodic deadlock detector will not be required for this test.

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, 
                                         [{port_start_function, StartFunc},
                                          {spawn_options,       [{min_heap_size, 6000}]},
                                          {default_receive_reply_timeout_milliseconds,      5000},
                                          {deadlock_detector_milliseconds_and_policy,       {1000, 1000, bdb_DB_LOCK_DEFAULT}},
                                          {default_global_txn_timeout_microseconds,         1000 * 1000},
                                          {default_global_lock_timeout_microseconds,        disabled},
                                          {{thread_count_stacksize_qlenlimit, workers},     { 1, 32*1024, 1000}},
                                          {{thread_count_stacksize_qlenlimit, valmap_stop}, {10, 32*1024, 1000}}
                                         ]),

    P = ?BDB_PC:acquire_bdb_port_handle(),
    
    1000 * 1000 = ?BDB:env_get_timeout(P, bdb_DB_SET_TXN_TIMEOUT),
    
    BlockOnTxn = ?BDB:txn_begin(P, void, []),
    DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),
    DB2 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord2.db"),

    %% acquire the lock that we will make all worker threads block against
    ok = ?BDB:db_put(P, DB, BlockOnTxn, "key", "value-initial", []),

    %% We could just create another txn in the same Erlang process, but to 
    %% model the real-world condition will will spawn another client
    TestDriver = self(),
    SpawnedPid = spawn_link(
                   fun() -> 
                           SpawnedP  = ?BDB_PC:acquire_bdb_port_handle(),
                           SpawnedDB = ?BDB_PC:get_db_handle(SpawnedP, bdb_DB_BTREE, "test_coord.db"),

                           %% This will block untl the periodic deadlock detector tells it to give up
                           ?EXPECT_BDB_ERROR(bdb_DB_LOCK_NOTGRANTED, 
                                             ?BDB:db_put(SpawnedP, SpawnedDB, void, "key", "value-spawned", [])),

                           TestDriver ! {self(), got_bdb_DB_LOCK_NOTGRANTED},
                           ok = ?BDB_PC:release_bdb_port_handle(SpawnedP)
                   end),

    timer:sleep(50),  % wait for spawned process to start it's db_put
   
    %% This operation will be stuck in the queue until the periodic deadlock detector 
    %% releases the spawned process' db_put
    ok = ?BDB:db_put(P, DB2, BlockOnTxn, "other-key", "other-value", []),
    ok = ?BDB:txn_commit(P, BlockOnTxn, []),

    {<<"key">>, <<"value-initial">>} = ?BDB:db_get(P, DB, void, <<"key">>, <<>>, []),
    {<<"other-key">>, <<"other-value">>} = ?BDB:db_get(P, DB2, void, <<"other-key">>, <<>>, []),

    receive {SpawnedPid, got_bdb_DB_LOCK_NOTGRANTED} -> ok end,

    %% shutdown
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:release_bdb_port_handle(P),
    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:stop(),

    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> throw({timeout_waiting_for_server_exit, ?AT})
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_port_coordinator_multiple_client_processes(StartFunc) ->

    io:format("\nPerforming test_port_coordinator_multiple_client_processes\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, 
                                         [{port_start_function, StartFunc}, 
                                          {env_cache_size_mb_and_fragments, {50, 1}},
                                          {cache_throttle_io_write_ops_and_sleep_microseconds, {300, 5}},
                                          {do_parallel_trickle_and_checkpoint,                 disabled},    % SEQUENTIAL
                                          {{thread_count_stacksize_qlenlimit, workers},         {10, 64*1024, 1000}},
                                          {{thread_count_stacksize_qlenlimit, valmap_stop},     {10, 64*1024, 1000}},
                                          {{thread_count_stacksize_qlenlimit, consume_wait},     {2, 64*1024,   10}},
                                          {{thread_count_stacksize_qlenlimit, housekeeping},     {1, 64*1024,   10}}  % ONE thread
                                         ]),

    %% Two of these processes will intentionall exit abnormally, and we don't want it to kill the test driver.
    %% (As a small side-benefit, enabling trap_exit means that we don't have to bother with erlang:monitor for the processes 
    %% that are expected to exit normally).
    %% See comments at the top of the file on why tests are expected to preserve the old value of trap_exit 
    %% (if they complete normally -- i.e. I don't bother putting it in a try/after as we don't execute subsequent tests if 
    %% one crashes; a real test driver should do that).
    io:format("test_port_coordinator_multiple_client_processes: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),

    SpawnedPid0 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      ok = ?BDB:db_put(P, DB, void, "0SpawnedKey0", <<"0spawned-val0">>, []),

                                      %% clean shutdown with no open valmaps
                                      ok = ?BDB_PC:release_bdb_port_handle(P)
                             end),

    SpawnedPid1 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      Txn = ?BDB:txn_begin(P, void, []),
                                      ok = ?BDB:db_put(P, DB, Txn, "1SpawnedKey1", <<"1spawned-val1">>, []),
                                      ok = ?BDB:txn_commit(P, Txn, []),
                                      %%ok = ?BDB:db_put(P, DB, void, "1SpawnedKey1", <<"1spawned-val1">>, []),

                                      %% clean shutdown with no open valmaps
                                      ok = ?BDB_PC:release_bdb_port_handle(P)
                             end),

    SpawnedPid2 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      Txn = ?BDB:txn_begin(P, void, []),
                                      ok = ?BDB:db_put(P, DB, Txn, "2SpawnedKey2", <<"2spawned-val2">>, []),
                                      ok = ?BDB:txn_commit(P, Txn, [])
                                      %%ok = ?BDB:db_put(P, DB, void, "2SpawnedKey2", <<"2spawned-val2">>, [])

                                      %% normal exit without shutting down, and with no open valmaps
                             end),

    SpawnedPid3 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      Txn = ?BDB:txn_begin(P, void, []),
                                      ok = ?BDB:db_put(P, DB, Txn, "3SpawnedKey3", <<"3spawned-val3">>, [])

                                      %% normal exit without shutting down, but with open valmaps
                             end),

    SpawnedPid4 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      Txn = ?BDB:txn_begin(P, void, []),
                                      ok = ?BDB:db_put(P, DB, Txn, "4SpawnedKey4", <<"4spawned-val4">>, []),

                                      %% intentional port shutdown but with open valmaps
                                      %% (server does not kill us, but does clean up the valmaps)
                                      try 
                                          ?BDB_PC:release_bdb_port_handle(P)
                                      catch 
                                          {bdb_port_coordinator_error, {release_bdb_port_handle, P, still_owned_resources, 1}} -> 
                                              ok
                                      end
                             end),

    SpawnedPid5 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      Txn = ?BDB:txn_begin(P, void, []),
                                      ok = ?BDB:db_put(P, DB, Txn, "5SpawnedKey5", <<"5spawned-val5">>, []),

                                      %% bad shutdown with open valmaps
                                      erlang:fault(intentional_fault_pid5)
                             end),

    SpawnedPid6 = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),
                                      DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                      ok = ?BDB:db_put(P, DB, void, "6SpawnedKey6", <<"6spawned-val6">>, []),

                                      %% bad shutdown with no open valmaps
                                      erlang:fault(intentional_fault_pid6)
                             end),

    io:format("ServerPid=~p, PrimaryPid=~p, SpawnedPid0=~p, SpawnedPid1=~p, SpawnedPid2=~p, SpawnedPid3=~p, SpawnedPid4=~p, SpawnedPid5=~p, SpawnedPid6=~p~n", [ServerPid, self(), SpawnedPid0, SpawnedPid1, SpawnedPid2, SpawnedPid3, SpawnedPid4, SpawnedPid5, SpawnedPid6]),

    %% Handle the abnormal exits first, so we can disable trapping of exits quickly
    P = ?BDB_PC:acquire_bdb_port_handle(),

    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

    %%io:format("Waiting for SpawnedPid0~n"),
    receive
        {'EXIT', SpawnedPid0, normal} -> %%io:format("Got exit from SpawnedPid0~n"),
                                         {<<"0SpawnedKey0">>, <<"0spawned-val0">>} = ?BDB:db_get(P, DB1, void, <<"0SpawnedKey0">>, <<>>, [])
    end,

    %%io:format("Waiting for SpawnedPid1~n"),
    receive
        {'EXIT', SpawnedPid1, normal} -> %%io:format("Got exit from SpawnedPid1~n"),
                                         {<<"1SpawnedKey1">>, <<"1spawned-val1">>} = ?BDB:db_get(P, DB1, void, <<"1SpawnedKey1">>, <<>>, [])
    end,

    %%io:format("Waiting for SpawnedPid2~n"),
    receive
        {'EXIT', SpawnedPid2, normal} -> %%io:format("Got exit from SpawnedPid2~n"),
                                         {<<"2SpawnedKey2">>, <<"2spawned-val2">>} = ?BDB:db_get(P, DB1, void, <<"2SpawnedKey2">>, <<>>, [])
    end,

    %%io:format("Waiting for SpawnedPid3~n"),
    receive
        {'EXIT', SpawnedPid3, normal} -> %%io:format("Got exit from SpawnedPid3~n"),
                                         ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, <<"3SpawnedKey3">>, <<>>, []))
    end,

    %%io:format("Waiting for SpawnedPid4~n"),
    receive
        {'EXIT', SpawnedPid4, normal} -> %%io:format("Got exit from SpawnedPid4~n"),
                                         ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, <<"4SpawnedKey4">>, <<>>, []))
    end,

    %%io:format("Waiting for SpawnedPid5~n"),
    receive
        {'EXIT', SpawnedPid5, {intentional_fault_pid5, _StackTrace5}} -> 
            %%io:format("Got exit from SpawnedPid5~n"),
            ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, <<"5SpawnedKey5">>, <<>>, []))
    end,

    %%io:format("Waiting for SpawnedPid6~n"),
    receive
        {'EXIT', SpawnedPid6, {intentional_fault_pid6, _StackTrace6}} -> 
            %%io:format("Got exit from SpawnedPid6~n"),
            {<<"6SpawnedKey6">>, <<"6spawned-val6">>} = ?BDB:db_get(P, DB1, void, <<"6SpawnedKey6">>, <<>>, [])
    end,

    %% If the main test process opens the port then it must release it, or the port controller server will 
    %% kill the test process when it terminates.
    ok = ?BDB_PC:release_bdb_port_handle(P),

    %% Need to test do_txn as well, and this is a convenient place
    %% (we don't test it's handling of retry-on-deadlock here -- that's done indirectly by testing 
    %% do_txn_on_master later)
    {<<"0SpawnedKey0">>, <<"0spawned-val0">>} 
        = ?BDB_H:do_txn({?MODULE, test_txn_fun_get, {"test_coord.db", "0SpawnedKey0"}}, 3),

    %% As we are trapping exits we DO receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next  test might get 'already_started' when it tries to start the server.
    receive
        {'EXIT', ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    throw_if_any_exit_msgs(),  % must be done BEFORE restoring trap_exit
    io:format("test_port_coordinator_multiple_client_processes: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_port_coordinator_client_cleanup(StartFunc) ->

    io:format("\nPerforming test_port_coordinator_client_cleanup\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),

    PrimaryPid = self(),
    SpawnedPid = spawn_link(fun() -> receive have_set_up_monitor -> ok end,

                                     P = ?BDB_PC:acquire_bdb_port_handle(),

                                     DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                     OpenTxn = ?BDB:txn_begin(P, void, []),
                                     ok = ?BDB:db_put(P, DB1, OpenTxn, "SpawnedKey", <<"spawned-val">>, []),

%% TODO: Trying to test do_txn here was a mistake
%% but the crash causes the C code to kill the VM with 
%%
%%    bdb port 2247 (berkeley_db_port_coordinator): bdb_private_threadpool_async_thread_main: threadid=28da9bb0 write() failed! (-1, errno=9). Aborting.%%
%%
%% Why?
%%
%%                                      {<<"SpawnedKey">>, <<"spawned-val">>} 
%%                                          = ?BDB_H:do_txn({?MODULE, test_txn_fun_get, {"test_coord.db", "SpawnedKey"}}, 3),

                                     %% By NOT calling release_bdb_port_handle() (and then exiting), we cause the server 
                                     %% to call group_valmap_cleanup on our _GroupValmapId

                                     PrimaryPid ! {open_txn, OpenTxn}
                            end),

    %% The tests run with trap_exit = false, so if we want to know if a process exits normally we need to monitor it
    MonitorRefSpawned = erlang:monitor(process, SpawnedPid),
    SpawnedPid ! have_set_up_monitor,

    %%io:format("\nWaiting for open_txn\n"),
    OpenTxn = receive
                  {open_txn, Txn} -> Txn
              end,

    %%io:format("\nWaiting for EXIT from spawnedpid\n"),
    receive
        {'DOWN', MonitorRefSpawned, process, SpawnedPid, normal} -> 
            %%io:format("\nGot normal DOWN from SpawnedPid\n"),
            ok;

        Anything ->
            %%io:format("\nGot ~p from SpawnedPid\n", [Anything]),
            throw({unexpected_msg, Anything})
    end,    

    P = ?BDB_PC:acquire_bdb_port_handle(),
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_commit(P, OpenTxn, [])),  % generation-id has changed
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P, DB1, void, <<"SpawnedKey">>, <<>>, [])),

    %% If the main test process opens the port then it must release it, or the port controller server will 
    %% kill the test process when it terminates.
    ok = ?BDB_PC:release_bdb_port_handle(P),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            {messages, []} = process_info(self(), messages),  % report an messages that we do have
            throw({timeout_waiting_for_server_exit, ?AT})
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_port_coordinator_overlapping_group_valmap_cleanup(StartFunc) ->

    io:format("\nPerforming test_port_coordinator_overlapping_group_valmap_cleanup\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    NumTxnsToStart = 500,

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds,              5000},
                                               {group_valmap_cleanup_receive_reply_timeout_milliseconds, 1}, % absurdly short
                                               {deadlock_detector_milliseconds_and_policy,               disabled},
                                               {{initial_num_valmap_slots, bdb_db_txn},                  NumTxnsToStart},
                                               {{initial_num_valmap_slots, bdb_db},                      NumTxnsToStart},
                                               {max_mutexes,                                             NumTxnsToStart * 20}, % required for lots of databases
                                               {env_cache_size_mb_and_fragments,                         {1, 1}},
                                               {min_limit_concurrent_transactions,                       NumTxnsToStart}
                                              ]),
    Spawner = self(),
    _SpawnedPid = spawn_link(
                   fun() ->
                           ?EXPECT_BDB_HELPER_ERROR(
                              {release_bdb_port_handle, {{exit,  {bdb_receive_reply_timeout, _}}, _Stacktrace}},
                              ?BDB_H:do_txn({?MODULE, txn_fun_write_many_large_txn_and_return_before_committing, NumTxnsToStart}, 0)),

                           Spawner ! done
                   end),
    receive done -> ok end,

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            {messages, []} = process_info(self(), messages),  % report an messages that we do have
            throw({timeout_waiting_for_server_exit, ?AT})
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_write_many_large_txn_and_return_before_committing(P, NumTxnsToStart) ->

    LargeRecord = <<0:100000/unit:8>>,
    io:format("Starting ~w txn~n", [NumTxnsToStart]),
    lists:foreach(fun(N) -> 
                          %% Use different DBs to avoid colliding on root-page lock
                          DB = ?BDB_PC:get_db_handle_with_db_config(P, bdb_DB_BTREE, {"test.db", integer_to_list(N)}, []),
                          Txn = ?BDB:txn_begin(P, void, []),
                          ?BDB:db_put(P, DB, Txn, <<"key">>, LargeRecord, [])
                  end,
                  lists:seq(1, NumTxnsToStart)),

    %%?BDB:adjust_port_logging(P, enable, [all]),

    io:format("Returning with ~w open txn~n", [NumTxnsToStart]),
    my_txn_fun_return_value.


test_port_coordinator_shutdown_with_active_clients(StartFunc) ->

    io:format("\nPerforming test_port_coordinator_shutdown_with_active_clients\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    %% We are going to make the port_coordinator kill a process that is spawn_linked by the test driver.
    %% We don't want it to to alsl kill the test driver.
    %% See comments at the top of the file on why tests are expected to preserve the old value of trap_exit 
    %% (if they complete normally -- i.e. I don't bother putting it in a try/after as we don't execute subsequent tests if 
    %% one crashes; a real test driver should do that).
    io:format("test_port_coordinator_shutdown_with_active_clients: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),

    SpawnedPid = spawn_link(fun() -> P = ?BDB_PC:acquire_bdb_port_handle(),

                                     DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test_coord.db"),

                                     Txn    = ?BDB:txn_begin(P, void, []),
                                     _Cursor1 = ?BDB:db_cursor(P, DB1, void, []),
                                     _Cursor2 = ?BDB:db_cursor(P, DB1, Txn, []),

                                     %%?BDB:adjust_port_logging(P, enable, [all]),

                                     %% We don't do release_bdb_port_handle, and the primary process will stop the server.
                                     Loop = fun(Loop) ->
                                                    receive
                                                        {ping, FromPid} -> FromPid ! {pong_from, self()},
                                                                           Loop(Loop);
                                                        Other           -> throw({spawned_got_unrecognised_message, Other})
                                                    end
                                            end,
                                     Loop(Loop)
                            end),

    PingSpawnedPid = fun() -> 
                             SpawnedPid ! {ping, self()},
                             receive 
                                 {pong_from, SpawnedPid}      -> ok;
                                 Other                        -> throw({primary_got_unrecognised_message, Other})
                             after
                                 500 -> throw({timeout, ?AT})
                             end
                     end,
    PingSpawnedPid(),
    PingSpawnedPid(),

    %% This should cause the server to kill the spawned process
    %% As we are trapping exits we DO receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:stop(),

    receive
        {'EXIT', SpawnedPid, killed} -> ok
    end,

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    receive
        {'EXIT', ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    throw_if_any_exit_msgs(), % must be done BEFORE restoring trap_exit
    io:format("test_port_coordinator_shutdown_with_active_clients: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


stress_test_allocation_of_binaries(StartFunc) ->

    case StartFunc of 
        start ->
            io:format("\nPerforming stress_test_allocation_of_binaries\n"),

            Dir = "txn_tests_env",
            ensure_empty_dir(Dir),

            %% Use the port_coordinator so this will work in pipe-mode.
            %% But we share the port and DB handle.
            {ok, ServerPid} = ?BDB_PC:start_link(Dir,
                                                                      [{port_start_function, StartFunc},
                                                                       {{thread_count_stacksize_qlenlimit, workers}, {100, 64*1024, 1000}}
                                                                      ]),
            P = ?BDB_PC:acquire_bdb_port_handle(),
            DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test.db"),

            %% Speed things up a lot
            ok = ?BDB:env_set_flags(P, [bdb_DB_TXN_NOSYNC], 1),

            NumKeys        = 50,
            NumGetProcs    = 1000,
            NumGetsPerProc = 2000,

            Table = ets:new(stress_test_allocation_of_binaries_table, [public]),

            Value = <<"12345">>,
            PutLoop = fun(0, _) ->
                              ok;
                         (N, PutLoop) ->
                              ok = ?BDB:db_put(P, DB, void, integer_to_list(N - 1), Value, []),
                              true = ets:insert(Table, {{N-1}, Value}),
                              PutLoop(N - 1, PutLoop)
                      end,
            %%io:format("\nstress_test_allocation_of_binaries: Doing puts\n"),
            PutLoop(NumKeys, PutLoop),

            OldTrapExitSetting = process_flag(trap_exit, true),

            SpawnGetProcsLoop 
                = fun(0, Pids, _) -> 
                          Pids;
                     (NumProcs, Pids, SpawnGetProcsLoop) -> 
                          SpawnGetProcsLoop(NumProcs - 1, 
                                            [spawn_opt(fun() -> 
                                                               P2 = P#bdb_port_handle{issued_to_pid=self()},
                                                               GetLoop = fun(0, _) ->
                                                                                 %%io:format("stress_test_allocation_of_binaries: GetProc ~p finished.\n", [NumProcs]),
                                                                                 ok;
                                                                            (N, GetLoop) ->
                                                                                 Key = list_to_binary(integer_to_list(N rem NumKeys)),
                                                                                 {Key, Value} = ?BDB:db_get(P2, DB, void, Key, <<>>, [bdb_DB_READ_UNCOMMITTED]),
                                                                                 GetLoop(N - 1, GetLoop)
                                                                         end,
                                                               GetLoop(NumGetsPerProc, GetLoop)
                                                       end,
                                                       [link, {fullsweep_after, 0}, {min_heap_size, 1000}])
                                             | Pids],
                                            SpawnGetProcsLoop)
                  end,
            GetsProcs = SpawnGetProcsLoop(NumGetProcs, [], SpawnGetProcsLoop),
            io:format("\nstress_test_allocation_of_binaries: Spawned ~p 'GetProcs', each doing ~p db_gets.\n", [NumGetProcs, NumGetsPerProc]),

            HitETS = spawn_link(fun() -> Loop = fun(Loop) -> 
                                                        receive
                                                            stop -> io:format("stress_test_allocation_of_binaries: HitETS got stop\n")

                                                        after
                                                            1 ->
                                                                %% insert a new binary
                                                                %%io:format("stress_test_allocation_of_binaries: inserting\n"),
                                                                true = ets:insert(Table, {term_to_binary(erlang:now()), term_to_binary(erlang:now())}),
                                                                %% Read all the binaries (and throw them away)
                                                                %%ets:tab2list(Table),
                                                                Loop(Loop)
                                                        end
                                                end,
                                         Loop(Loop)
                                end),

            %% Wait for exit signals
            lists:foldl(fun(Pid, _) -> receive
                                           {'EXIT', Pid, Reason} -> 
                                               case Reason of 
                                                   normal -> ok;
                                                   Other  -> throw({get_proc_failed, Other})
                                               end
                                       end
                        end,
                        ok,
                        GetsProcs),

            HitETS ! stop,
            case receive {'EXIT', HitETS, Reason} -> Reason end of
                normal -> ok
            end,
            io:format("\nstress_test_allocation_of_binaries: Done total of ~p db_gets in ~p concurrent processes.\n", [NumGetProcs * NumGetsPerProc, NumGetProcs]),
            
            throw_if_any_exit_msgs(), % must be done BEFORE restoring trap_exit
            io:format("stress_test_allocation_of_binaries: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
            true = process_flag(trap_exit, OldTrapExitSetting),

            ets:delete(Table),

            %%?BDB:adjust_port_logging(P, enable, [all]),

            ok = ?BDB_PC:release_bdb_port_handle(P),

            %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
            MonitorRef = erlang:monitor(process, ServerPid),
            ok = ?BDB_PC:stop(),

            %% It takes a few milliseconds to shutdown the server, so we need to wait 
            %% or the next test might get 'already_started' when it tries to start the server.
            receive
                {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
                Other -> throw({unexpected_msg, Other})
            after
                2000 -> 
                    {messages, []} = process_info(self(), messages),  % report an messages that we do have
                    throw(timeout_waiting_for_server_exit)
            end,
            undefined = whereis(?BDB_PC);

        start_pipe ->
            %% Not necessary in pipe mode
            ok
    end,

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


%% This performance-test is designed to explore the overhead of the Erlang<->marshalling layer.
%% It does a fixed number of operations (e.g. 10,000 or 500,000) using a fixed number of 
%% Erlang processes talking to BDB (so each Erlang process does an equal fraction of the total 
%% number of operations).  It repeats that against different number of worker threads.
%%
%% IMPORTANT: this test is NOT representative of real-world BDB application performance.
%% Application performance depends on many things, including 
%%   - size of BDB cache vs size of application data working set (i.e. cache misses for 'hot data')
%%   - use of log-flushing for durability of transactions (this test disables it, otherwise performance is bounded by disk-io)
%%   - use of replication (can allow log-flushing to be disabled safely, but adds network latency; usually a significant win though) 
%%   - amount of lock contention and frequency of deadlocks (depends on application's data-access patters)
%% and other factors -- see the BDB documentation and forums for a lot of discussion.
%%
%% The Erlang driver adds more factors to the performance equation -- e.g. size of 
%% threadpools, use of linked-in mode (StartFunc = 'start') vs pipe-mode (StartFunc = 'start_pipe').
%% 
%% TODO
%% Record all times and get tavg and tp99.9 latency
%% Different page sizes
%% Puts as overwrites (no page splits)
%% Mixed gets/puts
%% Explicit txn_begin/commit
%% Enable log flushing
%% Optimization to remove memcpy/double allocate

performance_test_1([StartFunc]) ->

    %%erlang:system_flag(fullsweep_after, 0),

    NumRunsToAverage        =       1,

    TotalNumOperations      =  10000,   %500000,

    EnvCacheMBFrag          = {30, 1},
    ExtraDbEnvFlags         = [bdb_DB_TXN_NOSYNC],
    DbConfig                = [{db_open_flags,  [bdb_DB_CREATE, bdb_DB_READ_UNCOMMITTED, bdb_DB_AUTO_COMMIT]}
                               %% [{page_size_bytes, 32768}, {set_flags, [bdb_DB_CHKSUM]}],
                              ],
    DbGetFlags              = [bdb_DB_READ_UNCOMMITTED],
    KeySizeBytes            = 30,
    Record                  = <<0:5000/unit:8>>,
    5000 = size(Record),

    ArgsResults
        = lists:map(fun({NumConcurrentProcesses, NumBDBWorkerThreads} = Args) ->
                            {Args, 
                             performance_test_do_multiple_runs(
                               StartFunc, 
                               NumRunsToAverage,
                               NumConcurrentProcesses,
                               TotalNumOperations div NumConcurrentProcesses, % NumOperationsPerProcess
                               NumBDBWorkerThreads,
                               EnvCacheMBFrag,
                               ExtraDbEnvFlags,
                               DbConfig,
                               DbGetFlags,
                               KeySizeBytes,
                               Record)}
                    end,
                    [
                     {	    1,	  1}
                     ,{     5,	  1}
                     ,{    10,	  1}
                     ,{    50,	  1}
                     ,{   100,	  1}
                     ,{   500,	  1}
                     ,{  1000,	  1}
                     ,{  5000,	  1}
                     ,{ 10000,	  1}
%%                      ,{ 50000,	  1}
%%                      ,{100000,    1}
                     ,{     1,	  5}
                     ,{     5,	  5}
                     ,{    10,	  5}
                     ,{    50,	  5}
                     ,{   100,	  5}
                     ,{   500,	  5}
                     ,{  1000,	  5}
                     ,{  5000,	  5}
                     ,{ 10000,	  5}
%%                      ,{ 50000,	  5}
%%                      ,{100000,    5}
                     ,{     1,	 10}
                     ,{     5,	 10}
                     ,{    10,	 10}
                     ,{    50,	 10}
                     ,{   100,	 10}
                     ,{   500,	 10}
                     ,{  1000,	 10}
                     ,{  5000,	 10}
                     ,{ 10000,	 10}
%%                      ,{ 50000,	 10}
%%                      ,{100000,   10}
                     ,{     1,	 50}
                     ,{     5,	 50}
                     ,{    10,	 50}
                     ,{    50,	 50}
                     ,{   100,	 50}
                     ,{   500,	 50}
                     ,{  1000,	 50}
                     ,{  5000,	 50}
                     ,{ 10000,	 50}
%%                      ,{ 50000,	 50}
%%                      ,{100000,   50}
                     ,{     1,  100}
                     ,{     5,  100}
                     ,{    10,  100}
                     ,{    50,  100}
                     ,{   100,  100}
                     ,{   500,  100}
                     ,{  1000,  100}
                     ,{  5000,  100}
                     ,{ 10000,  100}
%%                      ,{ 50000,  100}
%%                      ,{100000,  100}
                    ]), 

    io:format("~nAll times are in microseconds. All metrics are mean over ~w runs.~n~10w ~10w ~10w ~10w ~10w ~11w ~10w ~10w ~10w ~11w~n", 
              [NumRunsToAverage,
               procs,
	       threads,
	       eff_put,
	       tot_dlocks,
	       min_put,
	       max_put,
	       max_dlocks,
	       eff_get,
	       min_get,
	       max_get]),

    lists:foreach(fun( {{NumConcurrentProcesses, NumBDBWorkerThreads}, 
                        [{mean_effective_time_per_put_microseconds, MeanEffectiveTimePerPutMicroseconds}, 
                         {mean_total_num_deadlocks,                 MeanTotalNumDeadlocks},
                         {mean_min_time_per_put_microseconds,       MeanMinTimePerPutMicroseconds}, 
                         {mean_max_time_per_put_microseconds,       MeanMaxTimePerPutMicroseconds}, 
                         {mean_max_num_deadlocks_per_put,           MeanMaxNumDeadlocksPerOperation},
                         {mean_effective_time_per_get_microseconds, MeanEffectiveTimePerGetMicroseconds}, 
                         {mean_min_time_per_get_microseconds,       MeanMinTimePerGetMicroseconds}, 
                         {mean_max_time_per_get_microseconds,       MeanMaxTimePerGetMicroseconds}]} ) ->

                          io:format("~10w ~10w ~10.1f ~10.1f ~10.1f ~11.1f ~10.1f ~10.1f ~10.1f ~11.1f~n", 
                                    [NumConcurrentProcesses, 
                                     NumBDBWorkerThreads, 
                                     MeanEffectiveTimePerPutMicroseconds, 
                                     MeanTotalNumDeadlocks,
                                     MeanMinTimePerPutMicroseconds,
                                     MeanMaxTimePerPutMicroseconds,
                                     MeanMaxNumDeadlocksPerOperation,
                                     MeanEffectiveTimePerGetMicroseconds,
                                     MeanMinTimePerGetMicroseconds,
                                     MeanMaxTimePerGetMicroseconds])
                  end,
                  ArgsResults),
    ArgsResults.

performance_test_do_multiple_runs(StartFunc, 
                                  NumRunsToAverage,
                                  NumConcurrentProcesses, 
                                  NumOperationsPerProcess,
                                  NumBDBWorkerThreads,
                                  EnvCacheMBFrag,
                                  ExtraDbEnvFlags,
                                  DbConfig,
                                  DbGetFlags,
                                  KeySizeBytes,
                                  Record) ->

    RunResults = lists:map(fun(_) -> performance_test_do_one_run(StartFunc, 
                                                                 NumConcurrentProcesses,
                                                                 NumOperationsPerProcess,
                                                                 NumBDBWorkerThreads,
                                                                 EnvCacheMBFrag,
                                                                 ExtraDbEnvFlags,
                                                                 DbConfig,
                                                                 DbGetFlags,
                                                                 KeySizeBytes,
                                                                 Record) 
                           end, 
                           lists:seq(1, NumRunsToAverage)),

    RunResultsGroupedByMetric = ?BDB_U:lists_transpose(RunResults),

    MeansOfMetrics = lists:map(fun(ValuesForOneMetricFromMultipleRuns) -> 
                                       {list_to_atom("mean_" ++ atom_to_list(element(1, hd(ValuesForOneMetricFromMultipleRuns)))),
                                        lists:foldl(fun({_, Value}, Sum) -> 
                                                            Sum + Value
                                                    end, 
                                                    0,
                                                    ValuesForOneMetricFromMultipleRuns) 
                                        / NumRunsToAverage
                                       }
                               end,
                               RunResultsGroupedByMetric),

    MeansOfMetrics.


performance_test_do_one_run(StartFunc, 
                            NumConcurrentProcesses,
                            NumOperationsPerProcess,
                            NumBDBWorkerThreads,
                            EnvCacheMBFrag,
                            ExtraDbEnvFlags,
                            DbConfig,
                            DbGetFlags,
                            KeySizeBytes,
                            Record) ->

    io:format("\nperformance_test_1\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    %% Use the port_coordinator so this will work in pipe-mode.
    %% But we share the port and DB handle.
    {ok, ServerPid} = ?BDB_PC:start_link(
                        Dir,
                        [{port_start_function, StartFunc},
                         {env_cache_size_mb_and_fragments,                    EnvCacheMBFrag},
                         {env_set_flags,                                      [bdb_DB_AUTO_COMMIT, bdb_DB_LOG_AUTOREMOVE, bdb_DB_REGION_INIT, bdb_DB_TIME_NOTGRANTED]},
                         {default_db_config_proplist,                         DbConfig},
                         {max_num_concurrent_client_procs,                    NumBDBWorkerThreads},
                         {min_limit_concurrent_transactions,                  NumBDBWorkerThreads},
                         {auto_deadlock_detect,                               bdb_DB_LOCK_DEFAULT},      % or 'disabled'
                         {max_locks,                                          1000},
                         {max_lockers,                                        1000},
                         {max_locked_objects,                                 1000},
                         {checkpoint_seconds_and_min_kb,                      {30, 100}},             % or 'disabled'
                         {cache_trickle_milliseconds_and_percent,             {15000, 80}},              % or 'disabled'
                         {cache_throttle_io_write_ops_and_sleep_microseconds, disabled},              % disabled is same as {0, anything}
                         {deadlock_detector_milliseconds_and_policy,          disabled},              % independent of auto_deadlock_detect
                         {replication,                                        disabled},              % see below for sub-options
                         {{thread_count_stacksize_qlenlimit, workers},        {NumBDBWorkerThreads, 64*1024, 100000}},
                         {{thread_count_stacksize_qlenlimit, valmap_stop},      {2, 64*1024, 1000}},   % if replication is enabled, these threads block for acks
                         {{thread_count_stacksize_qlenlimit, consume_wait},     {0, 64*1024,   10}},
                         {{thread_count_stacksize_qlenlimit, housekeeping},     {2, 64*1024,   10}}
                        ]),

    P = ?BDB_PC:acquire_bdb_port_handle(),

    case lists:member(bdb_DB_TXN_NOSYNC, ExtraDbEnvFlags) of
        true ->
            io:format("~nperformance_test_1: DISABLING txn-log flushing\n"),
            ok = ?BDB:env_set_flags(P, [bdb_DB_TXN_NOSYNC], 1);
        _ -> ok
    end,

    DB = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test.db"),

    OldTrapExitSetting = process_flag(trap_exit, true),

    %% First 'puts'
    PutProcs = spawn_processes(NumConcurrentProcesses,
                               fun(ProcessNumber) -> 
                                       receive {go, Caller} -> ok end,
                                       P2 = P#bdb_port_handle{receive_reply_timeout=10 * 60 * 1000, issued_to_pid=self()},

                                       KeyBase = ProcessNumber * NumOperationsPerProcess,

                                       Loop = fun(0, MinTime, MaxTime, TotalNumDeadlocks, MaxNumDeadlocksPerOperation, _NumDeadlocksThisOperation, _Loop) ->
                                                      Caller ! {per_operation_stats, 
                                                                self(), 
                                                                {MinTime, MaxTime, TotalNumDeadlocks, MaxNumDeadlocksPerOperation}};

                                                 (OperationNumber, MinTime, MaxTime, TotalNumDeadlocks, MaxNumDeadlocksPerOperation, NumDeadlocksThisOperation, Loop) ->

                                                      Before = erlang:now(),
                                                      try ?BDB:db_put(P2, 
                                                                      DB, 
                                                                      void,
                                                                      <<(KeyBase + (OperationNumber - 1)):(KeySizeBytes)/unit:8>>, 
                                                                      Record, 
                                                                      []) of
                                                          ok ->
                                                              After = erlang:now(),
                                                              Time = timer:now_diff(After, Before),
                                                              Loop(OperationNumber - 1, 
                                                                   if Time < MinTime -> Time;
                                                                      true -> MinTime
                                                                      end,
                                                                   if Time > MaxTime -> Time;
                                                                      true -> MaxTime
                                                                      end,
                                                                   TotalNumDeadlocks,
                                                                   if NumDeadlocksThisOperation > MaxNumDeadlocksPerOperation ->
                                                                           NumDeadlocksThisOperation;
                                                                      true -> MaxNumDeadlocksPerOperation
                                                                   end,
                                                                   0,  % new NumDeadlocksThisOperation
                                                                   Loop)
                                                      catch
                                                          {bdb_error, bdb_DB_LOCK_DEADLOCK} ->
                                                              After = erlang:now(),
                                                              Time = timer:now_diff(After, Before),
                                                              Loop(OperationNumber, 
                                                                   if Time < MinTime -> Time;
                                                                      true -> MinTime
                                                                      end,
                                                                   if Time > MaxTime -> Time;
                                                                      true -> MaxTime
                                                                      end,
                                                                   TotalNumDeadlocks + 1,
                                                                   MaxNumDeadlocksPerOperation,
                                                                   NumDeadlocksThisOperation + 1,
                                                                   Loop)
                                                      end
                                              end,
                                       Loop(NumOperationsPerProcess, 
                                            100 * 1000000,  % initial MinTimeMicroseconds (100 seconds)
                                            0,              % initial MaxTimeMicroseconds
                                            0,              % initial TotalNumDeadlocks
                                            0,              % initial MaxNumDeadlocksPerOperation
                                            0,              % initial NumDeadlocksThisOperation
                                            Loop)
                               end),
    io:format("~nperformance_test_1: Starting ~p conurrent processes each doing ~p db_put (total of ~p operations)\n", 
              [NumConcurrentProcesses, 
               NumOperationsPerProcess,
               NumConcurrentProcesses * NumOperationsPerProcess]),

    {TotalTimePutsMicroseconds, {MinTimePerPutMicroseconds, MaxTimePerPutMicroseconds, TotalNumDeadlocks, MaxNumDeadlocksPerOperation}} 
        = start_linked_processes_and_time_to_completion(
            PutProcs,
            %% This fun is folded over all of the OperationsStats (return values of the test fun above)
            %% from all of the processes, after they complete
            %% It should return a final tuple of OperationsStats
            fun({MinTimeMicroseconds, MaxTimeMicroseconds, TotalNumDeadlocks, MaxNumDeadlocksPerOperation}, 
                {AccMinTimeMicroseconds, AccMaxTimeMicroseconds, AccTotalNumDeadlocks, AccMaxNumDeadlocksPerOperation}) ->
                    
                    {if MinTimeMicroseconds < AccMinTimeMicroseconds -> MinTimeMicroseconds;
                        true -> AccMinTimeMicroseconds
                     end,
                     if MaxTimeMicroseconds > AccMaxTimeMicroseconds -> MaxTimeMicroseconds;
                        true -> AccMaxTimeMicroseconds
                     end,
                     AccTotalNumDeadlocks + TotalNumDeadlocks,
                     if MaxNumDeadlocksPerOperation > AccMaxNumDeadlocksPerOperation -> MaxNumDeadlocksPerOperation;
                        true -> AccMaxNumDeadlocksPerOperation
                     end}
            end,
            %% This is the seed term for the above fun (i.e. initial {AccMinTimeMicroseconds, AccMaxTimeMicroseconds, AccTotalNumDeadlocks, AccMaxNumDeadlocksPerOperation}}
            {100 * 1000000, 0, 0, 0}),

    EffectiveTimePerPutMicroseconds = TotalTimePutsMicroseconds / (NumConcurrentProcesses * NumOperationsPerProcess),

    io:format("performance_test_1: EffectiveTimePerPutMicroseconds = ~p\n", [EffectiveTimePerPutMicroseconds]),
    io:format("performance_test_1: TotalNumDeadlocks               = ~p\n", [TotalNumDeadlocks]),
    io:format("performance_test_1: MinTimePerPutMicroseconds       = ~p\n", [MinTimePerPutMicroseconds]),
    io:format("performance_test_1: MaxTimePerPutMicroseconds       = ~p\n", [MaxTimePerPutMicroseconds]),
    io:format("performance_test_1: MaxNumDeadlocksPerOperation     = ~p\n", [MaxNumDeadlocksPerOperation]),


    %% Now gets
    GetProcs = spawn_processes(NumConcurrentProcesses,
                               fun(ProcessNumber) -> 
                                       receive {go, Caller} -> ok end,
                                       P2 = P#bdb_port_handle{receive_reply_timeout=10 * 60 * 1000, issued_to_pid=self()},
                                       KeyBase = ProcessNumber * NumOperationsPerProcess,
 
                                       Loop = fun(0, MinTime, MaxTime,_Loop) ->
                                                      Caller ! {per_operation_stats, self(), {MinTime, MaxTime}};

                                                 (OperationNumber, MinTime, MaxTime, Loop) ->

                                                      Before = erlang:now(),
                                                      _ = ?BDB:db_get_data(P2, 
                                                                           DB, 
                                                                           void, 
                                                                           <<(KeyBase + (OperationNumber - 1)):(KeySizeBytes)/unit:8>>, 
                                                                           DbGetFlags),
                                                      After = erlang:now(),
                                                      Time = timer:now_diff(After, Before),

                                                      Loop(OperationNumber - 1, 
                                                           if Time < MinTime -> Time;
                                                              true -> MinTime
                                                           end,
                                                           if Time > MaxTime -> Time;
                                                              true -> MaxTime
                                                           end,
                                                           Loop)
                                              end,
                                       Loop(NumOperationsPerProcess, 
                                            100 * 1000000,  % initial MinTimeMicroseconds (100 seconds)
                                            0,              % initial MaxTimeMicroseconds
                                            Loop)
                               end),
    io:format("~nperformance_test_1: Starting ~p conurrent processes each doing ~p db_get(~p) (total of ~p operations)\n", 
              [NumConcurrentProcesses, 
               NumOperationsPerProcess,
               case lists:member(bdb_DB_READ_UNCOMMITTED, DbGetFlags) of
                   true  -> "read-UNcommitted";
                   false -> "read-committed"
               end,
               NumConcurrentProcesses * NumOperationsPerProcess]),

    {TotalTimeGetsMicroseconds, {MinTimePerGetMicroseconds, MaxTimePerGetMicroseconds}} 
        = start_linked_processes_and_time_to_completion(
            GetProcs,
            %% This fun is folded over all of the OperationsStats (return values of the test fun above)
            %% from all of the processes, after they complete
            %% It should return a final tuple of OperationsStats
            fun({MinTimeMicroseconds, MaxTimeMicroseconds}, {AccMinTimeMicroseconds, AccMaxTimeMicroseconds}) ->
                    {if MinTimeMicroseconds < AccMinTimeMicroseconds -> MinTimeMicroseconds;
                        true -> AccMinTimeMicroseconds
                     end,
                     if MaxTimeMicroseconds > AccMaxTimeMicroseconds -> MaxTimeMicroseconds;
                        true -> AccMaxTimeMicroseconds
                     end}
            end,
            %% This is the seed term for the above fun (i.e. initial {AccMinTimeMicroseconds, AccMaxTimeMicroseconds}}
            {100 * 1000000, 0}),
    EffectiveTimePerGetMicroseconds = TotalTimeGetsMicroseconds / (NumConcurrentProcesses * NumOperationsPerProcess),

    io:format("performance_test_1: EffectiveTimePerGetMicroseconds = ~p\n", [EffectiveTimePerGetMicroseconds]),
    io:format("performance_test_1: MinTimePerGetMicroseconds       = ~p\n", [MinTimePerGetMicroseconds]),
    io:format("performance_test_1: MaxTimePerGetMicroseconds       = ~p\n", [MaxTimePerGetMicroseconds]),

    ok = ?BDB:env_stat_print(P#bdb_port_handle{receive_reply_timeout=infinity}, [bdb_DB_STAT_SUBSYSTEM]),
    ok = ?BDB:db_stat_print(P#bdb_port_handle{receive_reply_timeout=infinity}, DB, []),

    BeforeCheckpoint = erlang:now(),
    ok = ?BDB:txn_checkpoint(P#bdb_port_handle{receive_reply_timeout=infinity}, 0, 0, []),    
    AfterCheckpoint = erlang:now(),
    TimeCheckpointMilliseconds = (timer:now_diff(AfterCheckpoint, BeforeCheckpoint) / 1000),
    io:format("~nperformance_test_1: TimeCheckpointMilliseconds = ~p\n", [TimeCheckpointMilliseconds]),

    throw_if_any_exit_msgs(), % must be done BEFORE restoring trap_exit
    true = process_flag(trap_exit, OldTrapExitSetting),

    ok = ?BDB_PC:release_bdb_port_handle(P),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),

    %% Return results as a *list* of tagged tuples, so caller can run a generic average (mean) function
    %% over all of the individual metrics from multiple runs.
    [{effective_time_per_put_microseconds, EffectiveTimePerPutMicroseconds},
     {total_num_deadlocks,                 TotalNumDeadlocks},
     {min_time_per_put_microseconds,       MinTimePerPutMicroseconds}, 
     {max_time_per_put_microseconds,       MaxTimePerPutMicroseconds}, 
     {max_num_deadlocks_per_put,           MaxNumDeadlocksPerOperation},
     {effective_time_per_get_microseconds, EffectiveTimePerGetMicroseconds},
     {min_time_per_get_microseconds,       MinTimePerGetMicroseconds}, 
     {max_time_per_get_microseconds,       MaxTimePerGetMicroseconds}].


spawn_processes(NumConcurrentProcesses, Fun) ->
    SpawnProcsLoop = fun(0, Pids, _) -> 
                             Pids;
                        (ProcessNumber, Pids, SpawnProcsLoop) -> 
                             SpawnProcsLoop(ProcessNumber - 1, 
                                            [spawn_opt(fun() -> Fun(ProcessNumber - 1) end,
                                                       [link 
                                                        ,{fullsweep_after, 0}
                                                        %% ,{min_heap_size, 1000}
                                                       ])
                                             | Pids],
                                            SpawnProcsLoop)
                     end,

%%     io:format("~nperformance_test: Spawning ~p processes\n", [NumConcurrentProcesses]),
%%     Before = erlang:now(),

    ProcPids = SpawnProcsLoop(NumConcurrentProcesses, [], SpawnProcsLoop),

%%     After = erlang:now(),
%%     TimeSpawnPerProcessMicroseconds = timer:now_diff(After, Before) / NumConcurrentProcesses,
%%     io:format("performance_test: TimeSpawnPerProcessMicroseconds = ~p\n", [TimeSpawnPerProcessMicroseconds]),

    ProcPids.


start_linked_processes_and_time_to_completion(ListOfPids, FunFoldOverAllPerOperationStats, SeedOperationStats) ->

    Before = erlang:now(),
    lists:foreach(fun(Pid) -> Pid ! {go, self()} end, ListOfPids),

    %% Wait for exit signals
    lists:foldl(fun(Pid, _) -> receive
                                   {'EXIT', Pid, Reason} -> 
                                       case Reason of 
                                           normal -> ok;
                                           Other  -> throw({process_failed, Other})
                                       end
                               end
                end,
                ok,
                ListOfPids),
    After = erlang:now(),

    {timer:now_diff(After, Before), 
     %% The processes all send us a message with their per-operation statistics.
     %% We collect the messages and fold the supplied FunCombinePerOperationStats over them to produce a final term.
     lists:foldl(FunFoldOverAllPerOperationStats, 
                 SeedOperationStats,
                 lists:map(fun(Pid) -> 
                                   receive {per_operation_stats, Pid, Stats} -> Stats end
                           end, 
                           ListOfPids))}.


test_sequence_server(StartFunc) ->

    io:format("\nPerforming test_sequence_server\n"),

    %% We are going to kill a sequence_server that is spawn_linked by the test driver.
    %% We don't want it to to also kill the test driver.
    %% See comments at the top of the file on why tests are expected to preserve the old value of trap_exit 
    %% (if they complete normally -- i.e. I don't bother putting it in a try/after as we don't execute subsequent tests if 
    %% one crashes; a real test driver should do that).
    io:format("test_sequence_server: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),

    Dir1 = "send_queue_env",
    ensure_empty_dir(Dir1),

    {ok, PortCoordPid1} = ?BDB_PC:start_link(port_coordinator1, 
                                             Dir1, 
                                             [{port_start_function, StartFunc}
                                             ]),

    {ok, SequenceServerPid1} = ?BDB_SEQ:start_link(sequence_server1,
                                                   port_coordinator1,
                                                   "test_sequence.db", 
                                                   "my_sequence_key",
                                                   [{persist_every_N_consumed_values, 50}]),

    0 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1 = ?BDB_SEQ:consume_one_value(sequence_server1),
    2 = ?BDB_SEQ:consume_one_value(sequence_server1),
    ok = ?BDB_SEQ:stop(sequence_server1),
    receive {'EXIT', SequenceServerPid1, normal} -> ok end,
    undefined = whereis(sequence_server1),

    {ok, SequenceServerPid1a} = ?BDB_SEQ:start_link(sequence_server1,
                                                    port_coordinator1,
                                                    "test_sequence.db", 
                                                    "my_sequence_key",
                                                    [{persist_every_N_consumed_values, 1000}]),
    50 = ?BDB_SEQ:consume_one_value(sequence_server1),
    51 = ?BDB_SEQ:consume_one_value(sequence_server1),
    52 = ?BDB_SEQ:consume_one_value(sequence_server1),
    exit(SequenceServerPid1a, kill),
    receive {'EXIT', SequenceServerPid1a, killed} -> ok end,
    undefined = whereis(sequence_server1),

    {error,{bad_return_value,{unrecognized_config_option,{persist_every_N_consumed_values,0}}}} 
        = ?BDB_SEQ:start_link(sequence_server1,
                              port_coordinator1,
                              "test_sequence.db", 
                              "my_sequence_key",
                              [{persist_every_N_consumed_values, 0}]), % zero is illegal

    %% Amazingly, it takes a little while for the process name to unregister, 
    %% and we get an exit message (even though we were not told the pid)
    receive {'EXIT', _, {bad_return_value,{unrecognized_config_option,{persist_every_N_consumed_values,0}}}} -> ok end,
    undefined = whereis(sequence_server1),

    {ok, SequenceServerPid1b} = ?BDB_SEQ:start_link(sequence_server1,
                                                    port_coordinator1,
                                                    "test_sequence.db", 
                                                    "my_sequence_key",
                                                    [{persist_every_N_consumed_values, 1}]), % transaction every time
    1050 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1051 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1052 = ?BDB_SEQ:consume_one_value(sequence_server1),
    exit(SequenceServerPid1b, kill),
    receive {'EXIT', SequenceServerPid1b, killed} -> ok end,
    undefined = whereis(sequence_server1),

    {ok, SequenceServerPid1c} = ?BDB_SEQ:start_link(sequence_server1,
                                                    port_coordinator1,
                                                    "test_sequence.db", 
                                                    "my_sequence_key",
                                                    [{persist_every_N_consumed_values, 10}]),
    1053 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1054 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1055 = ?BDB_SEQ:consume_one_value(sequence_server1),

    %% A second sequence server that happens to use the same DB,Key will interleave 
    %% (i.e. they will never return the same value)
    %% Note that the cached ranges imply that alternating calls to the two servers 
    %% will jump backwards and forwards in absolute ordering.  That is expected.
    %% (This is a non-typical use-case; normally there will only be a single instance 
    %% of each server.  Uniqueness is the critical property.  Also, the ordering 
    %% for two or more servers can be fixed by using PersistEveryNConsumedValues = 1.)
    {ok, SequenceServerPid2} = ?BDB_SEQ:start_link(sequence_server2,
                                                   port_coordinator1,
                                                   "test_sequence.db", 
                                                   "my_sequence_key",
                                                   [{persist_every_N_consumed_values, 5}]),

    1063 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1056 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1064 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1065 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1057 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1058 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1059 = ?BDB_SEQ:consume_one_value(sequence_server1), 
    1066 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1067 = ?BDB_SEQ:consume_one_value(sequence_server2), % #2 finishes range of 5
    1068 = ?BDB_SEQ:consume_one_value(sequence_server2), % #2 begins new range of 5
    1060 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1061 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1062 = ?BDB_SEQ:consume_one_value(sequence_server1), % #1 finishes range 
    1073 = ?BDB_SEQ:consume_one_value(sequence_server1), % #1 starts new range of 10 (1063 was returned by #2 earlier)
    1069 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1074 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1075 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1070 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1071 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1072 = ?BDB_SEQ:consume_one_value(sequence_server2), % #2 finishes range
    1076 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1077 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1078 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1079 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1080 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1081 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1082 = ?BDB_SEQ:consume_one_value(sequence_server1), % #1 finishes range
    1083 = ?BDB_SEQ:consume_one_value(sequence_server2), % #2 starts new range of 5
    1084 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1085 = ?BDB_SEQ:consume_one_value(sequence_server2),
    1088 = ?BDB_SEQ:consume_one_value(sequence_server1), % #1 starts new range of 10
    1089 = ?BDB_SEQ:consume_one_value(sequence_server1),
    1090 = ?BDB_SEQ:consume_one_value(sequence_server1),

    %% As we are trapping exits we DO receive 'EXIT' messages with Reason 'normal'.

    ok = ?BDB_SEQ:stop(sequence_server1),
    receive {'EXIT', SequenceServerPid1c, normal} -> ok end,

    ok = ?BDB_SEQ:stop(sequence_server2),
    receive {'EXIT', SequenceServerPid2, normal} -> ok end,

    ok = ?BDB_PC:stop(port_coordinator1),
    receive {'EXIT', PortCoordPid1, normal} -> ok end,

    throw_if_any_exit_msgs(), % must be done BEFORE restoring trap_exit
    io:format("test_sequence_server: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_msg_queue(StartFunc, QRecordLen) ->

    io:format("\nPerforming test_msg_queue with QRecordLen = ~w\n", [QRecordLen]),


    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),


    %% The receiver's bdb environment does not have to have the same directory path as the senders.
    %% But if the sender and receiver are on the same node they CANNOT use the same BDB environment
    %% as the send and receive queues will be named the same.
    %% There is no need for such a feature, as if they are on the same host/environment the 
    %% queue entries can just be consumed - there is no need for a wasteful sendpump & receivepump
    %% to copy all of the messates.
    %% HOWEVER, the current implementation does not prevent this, and the single database will 
    %% be opened and messates consumed by send-pump, sent to receivepump, which will enqueue them.
    %% So such an arrangement would continuously cycle the queue as fast as possible (breakint 
    %% FIFO ordering too).
    %% To prevent that, we could add a parameter to the create_bdb_msg_queue that gives it's 
    %% purpose (send or receive), and prefix that to the database filename.
    %% (And when sendpump calls global:send it should prefix MSG_Q#bdb_msg_q.name with "receive_"
    %% to send the message to the receiver).

    SenderEnvDir = "send_queue_env",
    ensure_empty_dir(SenderEnvDir),

    ReceiverEnvDir = "receive_queue_env",
    ensure_empty_dir(ReceiverEnvDir),

    QName = "example_queue",


    %%
    %% On the receive node (which may be remote)...
    %%

    %% To avoid distributed deadlocks between the send-pump and the receive pump, 
    %% we use lock timeouts (or transaction timeouts -- I'm not sure which would be better).
    %% So both the sender and receiver need to run a peridoc deadlock detector at 
    %% *at least* the desired timeout (say 2x the timeout period).
    %% As the timeout period can be long, this can be in addition to an automatic deadlock detector 
    %% (but the later is useless for lock/transaction timeouts).
    {ok, ReceiverPortCoordPid} = ?BDB_PC:start_link(receiver_port_coordinator, 
                                                    ReceiverEnvDir, 
                                                    [{port_start_function, StartFunc}, 
                                                     {env_cache_size_mb_and_fragments, {50, 1}},
                                                     {default_global_txn_timeout_microseconds,  ?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS * 1000}, % default is too short
                                                     {default_global_lock_timeout_microseconds, disabled},
                                                     {deadlock_detector_milliseconds_and_policy, 
                                                      {?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS, ?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS, bdb_DB_LOCK_DEFAULT}},
                                                     {{thread_count_stacksize_qlenlimit, consume_wait},  {5, 32*1024, 5}} % receive_pump needs this as default is 0 threads
                                                    ]),

    %% The receive-pump can be created before or after the sink starts de-queuing -- it doesn't matter.

    %% Ideally these two processes should agree on a queue reclen (the last numeric argument).
    %% If they don't agree, the first one to create the database will win.
    %% The reclen does not have to match that of the sender.

    PurgeMsgIdsOlderThanSeconds = 1,

    ReceivePumpPid = ?BDB_MSG_Q:spawn_link_receivepump(receiver_port_coordinator, QName, QRecordLen, PurgeMsgIdsOlderThanSeconds),

    %% ... this example receiver-worker process will send the drained msgs back to us (as normal Erlang messages), 
    %% ... proving that messages make it all the way through
    ExampleSinkPid = spawn_link_example_sink(QName, QRecordLen),



    %%
    %% On the 'sender' node...
    %%

    {ok, SenderPortCoordPid} = ?BDB_PC:start_link(sender_port_coordinator, 
                                                  SenderEnvDir, 
                                                  [{port_start_function, StartFunc}, 
                                                   {env_cache_size_mb_and_fragments, {50, 1}},
                                                   {default_global_txn_timeout_microseconds,  ?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS * 1000}, % default is too short
                                                   {default_global_lock_timeout_microseconds, disabled},
                                                   {deadlock_detector_milliseconds_and_policy, 
                                                    {?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS, ?BDB_MSG_Q_DISTRIBUTED_DEADLOCK_TIMEOUT_MS, bdb_DB_LOCK_DEFAULT}},
                                                   {{thread_count_stacksize_qlenlimit, consume_wait},  {5, 32*1024, 5}} % send_pump needs this as default is 0 threads
                                                  ]),

    {ok, SequenceServerPid} = ?BDB_MSG_Q:start_link_sequence_server(sender_port_coordinator, 
                                                                    QName, 
                                                                    50), % PersistEveryNConsumedValues

    %% The send-pump can be created before or after the enqueuers start enqueing -- it doeesn't matter.

    %% Ideally the enqueuer's and the send-pump should agree on a queue reclen.
    %% If they don't, the first one to create the database will win.
    %% The reclen does not have to match that of the receiver.
    SendPumpPid = ?BDB_MSG_Q:spawn_link_sendpump(sender_port_coordinator, QName, QRecordLen),


    io:format("\ntest__msg_queue: SequenceServerPid=~w, SendPumpPid=~w ReceivePumpPid=~w ExampleSinkPid=~w\n", [SequenceServerPid, SendPumpPid, ReceivePumpPid, ExampleSinkPid]),


    %% This also tries to send a duplicate message -- and we will detect if it makes it through.
    %% ... this returns [{MsgId, Msg},...]
    ExpectedMsgs = enqueue_some_test_messages(QName, QRecordLen), % implicitly uses sender_port_coordinator


    %% example_sink_process sends the messages back to us (as normal Erlang messages)
    wait_for_test_messages(ExpectedMsgs),

    
    %% Wait to observe the old message ids being purged on the reciever side
    %% (TODO: we could then test this by forcing the message-ids back to 0 and re-sending the messages)
    io:format("waiting for expired msg ids to be purged~n"),
    timer:sleep((PurgeMsgIdsOlderThanSeconds + 3) * 1000),

    %% We currently don't have a way to tell these processes to exit gracefully
    %% (they need to accept a msg, which means that SendPump and ExampleSink need to spawn 
    %% another process to call blocking_consume_one_persist_msg, while they wait for 
    %% either a response from that process or a shutdown message).
    %% OR blocking_consume_one_persist_msg should use the 'non_block' variant,
    %% and return (so it's name should change), and the process could then wait 
    %% in 'receive' for the port reply or a message telling it to shutdown
    %% Until then, as they each are connected to their own port instance, killing them 
    %% does the trick (the Erlang VM will close the port).

    %% See comments at the top of the file on why tests are expected to preserve the old value of trap_exit 
    %% (if they complete normally -- i.e. I don't bother putting it in a try/after as we don't execute subsequent tests if 
    %% one crashes; a real test driver should do that).
    io:format("test_msg_queue: Setting trap_exit to TRUE.\n"),
    OldTrapExitSetting = process_flag(trap_exit, true),

    io:format("\nKilling pumps and sink.\n"),
    exit(SendPumpPid,       kill),
    exit(ReceivePumpPid,    kill),
    exit(ExampleSinkPid,    kill),

    %% We will get  {'EXIT',Pid,killed} messages
    wait_for_exit_messages([SendPumpPid, ReceivePumpPid, ExampleSinkPid]),

    throw_if_any_exit_msgs(),  % must be done BEFORE restoring trap_exit

    io:format("test_msg_queue: Restoring trap_exit to ~p.\n", [OldTrapExitSetting]),
    true = process_flag(trap_exit, OldTrapExitSetting),


    io:format("test_msg_queue: monitoring and stopping receiver port coordinator\n"),
    %% As we are not (now) trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    ReceiverMonitorRef = erlang:monitor(process, ReceiverPortCoordPid),
    ok = ?BDB_PC:stop(receiver_port_coordinator),
    receive
        {'DOWN', ReceiverMonitorRef, process, ReceiverPortCoordPid, normal} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_receiver_exit)
    end,
    undefined = whereis(receiver_port_coordinator),

    io:format("test_msg_queue: monitoring and stopping sequence server\n"),
    SequenceServerMonitorRef = erlang:monitor(process, SequenceServerPid),
    ok = ?BDB_MSG_Q:stop_sequence_server(QName),
    receive
        {'DOWN', SequenceServerMonitorRef, process, SequenceServerPid, normal} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_sequence_server_exit)
    end,

    io:format("test_msg_queue: monitoring and stopping sender port coordinator\n"),
    SenderMonitorRef = erlang:monitor(process, SenderPortCoordPid),
    ok = ?BDB_PC:stop(sender_port_coordinator),
    receive
        {'DOWN', SenderMonitorRef, process, SenderPortCoordPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(receiver_port_coordinator),
    undefined = whereis(sender_port_coordinator),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


%%
%% An example worker process that drains the receiver's queue.
%% For our test case, we just send the message back to the process that spawned this one 
%% (which is also the test driver/original sender). That completes the circuit for an 
%% end-to-end test.
%%

spawn_link_example_sink(QName, QRecLen) ->
    TestDriverPid = self(),
    spawn_link(fun() -> example_sink_init(QName, QRecLen, TestDriverPid) end).

example_sink_init(QName, QRecLen, TestDriverPid) ->
    %%io:format("example_sink_process spawning\n"),
    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(receiver_port_coordinator),

    example_sink_loop(?BDB_MSG_Q:create_bdb_msg_q(BdbPortHandle, 
                                                  QName, 
                                                  QRecLen), 
                      TestDriverPid).

example_sink_loop(MSG_Q, TestDriverPid) ->

    Txn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, void, []),

    %%io:format("example_sink_loop blocking\n"),
    {MsgId, MsgFlags, MsgBodyBin} = ?BDB_MSG_Q:blocking_consume_one_msg(MSG_Q, Txn),

    ActualMsg = if ((MsgFlags band ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) =:= ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM) -> 
                        binary_to_term(MsgBodyBin);
                   true -> MsgBodyBin
                end,

    %%io:format("example_sink_loop got MsgId=~w Msg=~w\n", [MsgId, ActualMsg]),
    ok = ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, Txn, []),

    %%io:format("example_sink_loop sending MsgId=~w Msg=~w\n", [MsgId, ActualMsg]),
    TestDriverPid ! {got_test_msg, MsgId, ActualMsg},
    
    example_sink_loop(MSG_Q, TestDriverPid).


%% This function is only separate from test_msg_queue() so we 
%% can set up tracing before we enqueue any messages.
enqueue_some_test_messages(QName, QRecordLen) ->

    io:format("\nenqueue_test_messages: starting\n"),

    %% Enqueuers (workers) do the following...
    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(sender_port_coordinator),

    %% ... once only, to open the channel (see comments 
    MSG_Q = ?BDB_MSG_Q:create_bdb_msg_q(BdbPortHandle, QName, QRecordLen),

    %% Then, as often as desired...
    SomeWorkTxn = ?BDB:txn_begin(MSG_Q#bdb_msg_q.bdb_port_handle, void, []),

    %% ... (here do some real work in this txn)

    %% The 'real work' succeeded, so transactionally enqueue a persistent msg to tell someone about it.
    TestMsg1 = {test1,"stuff",[4],<<"apple">>},
    TestMsg2 = <<"raw binary">>,
    TestMsg3 = [hello2, {world, []}],

    {MsgId1, _QKey1} = ?BDB_MSG_Q:enqueue_term_msg(MSG_Q, SomeWorkTxn, TestMsg1),
    %%io:format("enqueued MsgId=~w QKey=~w Msg=~w\n", [MsgId1, QKey1, TestMsg1]),

    {MsgId2, _QKey2} = ?BDB_MSG_Q:enqueue_raw_binary_msg(MSG_Q, SomeWorkTxn, TestMsg2),
    %%io:format("enqueued MsgId=~w QKey=~w Msg=~w\n", [MsgId2, QKey2, TestMsg2]),

    {MsgId3, _QKey3} = ?BDB_MSG_Q:enqueue_external_term_msg(MSG_Q, SomeWorkTxn, term_to_binary(TestMsg3)),
    %%io:format("enqueued MsgId=~w QKey=~w Msg=~w\n", [MsgId3, QKey3, TestMsg3]),

    ok = ?BDB:txn_commit(MSG_Q#bdb_msg_q.bdb_port_handle, SomeWorkTxn, []),
    io:format("commited send\n"),

    %% ... SendPump should send it any time now.

    %% We can close the enqueuer's port any time after commiting the enqueue - it won't affect SendPump.
    ok = ?BDB_MSG_Q:close_bdb_msg_q(MSG_Q),

    ok = ?BDB_PC:release_bdb_port_handle(BdbPortHandle),

    [{MsgId1, ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM, TestMsg1}, 
     {MsgId2, 0, TestMsg2},
     {MsgId3, ?BDB_MSG_Q_FLAG_IS_EXTERNAL_TERM, TestMsg3}].
    

wait_for_test_messages(ExpectedMsgs) ->

    %% When the ExampleSink receives it, it sends a message back to us, so we know that the test passed.
    Loop = fun([], _Loop) ->
                   receive
                       %% sanity check that we don't receive any unexpected duplicates
                       Msg -> throw({got_extra_msg, Msg})
                   after
                       500 -> io:format("wait_for_test_messages success (got all expected msgs and no extra msgs)~n")
                   end;

              ([{MsgId, _MsgFlags, Msg} | T], Loop) ->
                   %%io:format("wait_for_test_messages expecting Id=~w MsgFlags=~p Msg=~p\n", [MsgId, MsgFlags, Msg]),
                   receive
                       %% Don't do selective receive on the expected msg as we don't want this 
                       %% code to re-order them (i.e. we want to check the actual order they 
                       %% arrive, not change it)
                       WholeMsg ->
                           case WholeMsg of
                               {got_test_msg, MsgId, Msg} -> 
                                   %%io:format("wait_for_test_messages got expected msg Id=~w Msg=~p\n", [MsgId, Msg]),
                                   Loop(T, Loop);
                               Other                      -> 
                                   throw({got_unexpected_msg, Other, was_expecting, {MsgId, Msg}})
                           end
                   end
              end,
    ok = Loop(ExpectedMsgs, Loop).



test_basic_replication(StartFunc) ->

    io:format("\nPerforming test_basic_replication\n"),

    MasterHostPort   = {"localhost", 4000},
    Replica1HostPort = {"localhost", 4010},
    Replica2HostPort = {"localhost", 4020},

    MasterEnvDir = "txn_tests_env",
    ensure_empty_dir(MasterEnvDir),

    {ok, MasterPortCoordPid} = ?BDB_PC:start_link(
                                  master_port_coordinator, 
                                  MasterEnvDir, 
                                  [{port_start_function, StartFunc}, 
                                   {env_cache_size_mb_and_fragments, {5, 1}},
                                   %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                   {replication, [{this_hostname_port,            MasterHostPort},
                                                  {this_site_priority,            2},
                                                  {peer_hostnames_ports,          [Replica1HostPort, Replica2HostPort]},
                                                  {initial_role_and_threadcount,  {bdb_DB_REP_MASTER, 2}},   % would normally be DB_FULL_ELECTION
                                                  {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                  {limit_transfer_per_call_bytes, 1024 * 1024},
                                                  {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                  {{timeout, election_milliseconds},         5 *    1000}, % milliseconds
                                                  {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                  {{timeout, connection_retry_milliseconds}, 5 *    1000},  % milliseconds
                                                  {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]},
                                                  %% this test populates a database on the master, to verify cold bootstrapping.
                                                  {allow_not_durably_replicated_db_open,  enabled}
                                                 ]}
                                  ]),

    receive {master_port_coordinator, bdb_DB_EVENT_REP_MASTER} = M1 -> 
            io:format("Got event ~p~n", [M1]) 
    end,

    MasterBdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(master_port_coordinator),

    %% This LSN may change with future versions of BDB
    RepStat = ?BDB:rep_stat(MasterBdbPortHandle, []),
    {value, {st_next_lsn, {1, 64}}} = lists:keysearch(st_next_lsn, 1, RepStat),

    {1, 64} = ?BDB:rep_stat_next_lsn(MasterBdbPortHandle, []),

    MasterDB = ?BDB_PC:get_db_handle(MasterBdbPortHandle, bdb_DB_BTREE, "test_replication.db"),
    ?AT,
    receive 
        {master_port_coordinator, bdb_DB_EVENT_REP_PERM_FAILED} -> ok 
    after 
        1000 -> throw({expecting_PERM_FAILED, ?AT})
    end,
    ?AT,

    Key   = <<"replicated_key_1">>,
    Value = <<"*replicated*value*one*">>,
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(MasterBdbPortHandle, MasterDB, void, Key, <<>>, [])),

    %% Put something before starting replicas, to verify bootstrapping of cold replicas.
    %% ... this fails to replicate, but is written on the master
    ?EXPECT_BDB_ERROR(bdb_DB_REP_NOTPERM,
                      ?BDB:db_put(MasterBdbPortHandle, MasterDB, void, Key, Value, [])),
    ?AT,
    receive 
        {master_port_coordinator, bdb_DB_EVENT_REP_PERM_FAILED} -> ok 
    after 
        1000 -> throw({expecting_PERM_FAILED, ?AT})
    end,
    ?AT,

    {Key, Val} = ?BDB:db_get(MasterBdbPortHandle, MasterDB, void, Key, <<>>, []),

    Replica1EnvDir = "send_queue_env",
    ensure_empty_dir(Replica1EnvDir),
    {ok, Replica1PortCoordPid} = ?BDB_PC:start_link(
                                    replica1_port_coordinator, 
                                    Replica1EnvDir, 
                                    [{port_start_function, StartFunc}, 
                                     {env_cache_size_mb_and_fragments, {5, 1}},
                                     {env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                     {replication, [{this_hostname_port,            Replica1HostPort},
                                                    {this_site_priority,            1},
                                                    {peer_hostnames_ports,          [Replica2HostPort, MasterHostPort]},
                                                    {initial_role_and_threadcount,  {bdb_DB_REP_CLIENT, 2}},   % would normally be DB_FULL_ELECTION
                                                    {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                    {limit_transfer_per_call_bytes, 1024 * 1024},
                                                    {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                    {{timeout, election_milliseconds},         5 *    1000}, % milliseconds
                                                    {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                    {{timeout, connection_retry_milliseconds}, 5 *    1000},  % milliseconds
                                                    {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                                   ]}
                                    ]),

    receive {replica1_port_coordinator, bdb_DB_EVENT_REP_CLIENT} = M2 -> 
            io:format("Got event ~p~n", [M2]) 
    end,
    Replica1BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(replica1_port_coordinator),

    Replica2EnvDir = "receive_queue_env",
    ensure_empty_dir(Replica2EnvDir),
    {ok, Replica2PortCoordPid} = ?BDB_PC:start_link(
                                    replica2_port_coordinator, 
                                    Replica2EnvDir, 
                                    [{port_start_function, StartFunc}, 
                                     {env_cache_size_mb_and_fragments, {5, 1}},
                                     {env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                     {replication, [{this_hostname_port,            Replica2HostPort},
                                                    {this_site_priority,            1},
                                                    {peer_hostnames_ports,          [Replica1HostPort, MasterHostPort]},
                                                    {initial_role_and_threadcount,  {bdb_DB_REP_CLIENT, 2}},   % would normally be DB_FULL_ELECTION
                                                    {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                    {limit_transfer_per_call_bytes, 1024 * 1024},
                                                    {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                    {{timeout, election_milliseconds},         5 *    1000}, % milliseconds
                                                    {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                    {{timeout, connection_retry_milliseconds}, 5 *    1000}, % milliseconds
                                                    {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                                   ]}
                                    ]),
    receive {replica2_port_coordinator, bdb_DB_EVENT_REP_CLIENT} = M3 -> 
            io:format("Got event ~p~n", [M3]) 
    end,
    Replica2BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(replica2_port_coordinator),

    io:format("Waiting for NEWMASTER~n"),
    receive {replica1_port_coordinator, bdb_DB_EVENT_REP_NEWMASTER, _} = M4 -> 
            io:format("Got event ~p~n", [M4]) 
    end,
    receive {replica2_port_coordinator, bdb_DB_EVENT_REP_NEWMASTER, _} = M5 -> 
            io:format("Got event ~p~n", [M5]) 
    end,

    %% We CANNOT wait for STARTUPDONE here in v4.5 as it is not triggered until the first live txn
    %% So we have to wait for a period of time for bootstrapping to occur (or opening the databases on the replicas will fail).
    io:format("Sleeping while bootstrapping ~n"),
    timer:sleep(2000),

    %%io:format("opening database on replica1~n"),
    Replica1DB = ?BDB_PC:get_db_handle(Replica1BdbPortHandle, bdb_DB_BTREE, "test_replication.db"),
    {Key, Val} = ?BDB:db_get(Replica1BdbPortHandle, Replica1DB, void, Key, <<>>, []),

    %%io:format("opening database on replica2~n"),
    Replica2DB = ?BDB_PC:get_db_handle(Replica2BdbPortHandle, bdb_DB_BTREE, "test_replication.db"),
    {Key, Val} = ?BDB:db_get(Replica2BdbPortHandle, Replica2DB, void, Key, <<>>, []),

    %% Now do a live-replicated put
    Key2   = <<"++replicated_key_2">>,
    Value2 = <<"!!*replicated*value*two*">>,
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(MasterBdbPortHandle, MasterDB, void, Key2, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(Replica1BdbPortHandle, Replica1DB, void, Key2, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(Replica2BdbPortHandle, Replica2DB, void, Key2, <<>>, [])),

    %% This txn is live and causes STARTUPDONE
    ok = ?BDB:db_put(MasterBdbPortHandle, MasterDB, void, Key2, Value2, []),
    {Key2, Val2} = ?BDB:db_get(MasterBdbPortHandle, MasterDB, void, Key2, <<>>, []),

    %% This assumes that we are using DB_REPMGR_ACKS_ALL_PEERS.  If we use DB_REPMGR_ACKS_QUORUM then one of them sometimes fails (for obvious reasons).
    {Key2, Val2} = ?BDB:db_get(Replica1BdbPortHandle, Replica1DB, void, Key2, <<>>, []),
    {Key2, Val2} = ?BDB:db_get(Replica2BdbPortHandle, Replica2DB, void, Key2, <<>>, []),

    io:format("Waiting for STARTUPDONE~n"),
    receive {replica1_port_coordinator, bdb_DB_EVENT_REP_STARTUPDONE} = M6 -> 
            io:format("Got event ~p~n", [M6]) 
    after 
        5000 -> io:format("messages ~p~n", [process_info(self(), messages)])
    end,
    receive {replica2_port_coordinator, bdb_DB_EVENT_REP_STARTUPDONE} = M7 -> 
            io:format("Got event ~p~n", [M7]) 
    end,

    %% Now do a clean shutdown

    ok = ?BDB_PC:release_bdb_port_handle(Replica1BdbPortHandle),
    ok = ?BDB_PC:release_bdb_port_handle(Replica2BdbPortHandle),
    ok = ?BDB_PC:release_bdb_port_handle(MasterBdbPortHandle),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.

    %%io:format("stopping replica1~n"),
    Replica1Monitor = erlang:monitor(process, Replica1PortCoordPid),
    ok = ?BDB_PC:stop(replica1_port_coordinator),

    %%io:format("stopping replica2~n"),
    Replica2Monitor = erlang:monitor(process, Replica2PortCoordPid),
    ok = ?BDB_PC:stop(replica2_port_coordinator),

    %%io:format("stopping master~n"),
    MasterMonitor = erlang:monitor(process, MasterPortCoordPid),
    ok = ?BDB_PC:stop(master_port_coordinator),


    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    lists:foreach(fun({MonitorRef, Pid, Name}) ->
                          receive
                              {'DOWN', MonitorRef,   process, Pid, normal} -> 
                                  %%io:format("Normal exit from ~p~n", [Name]),
                                  ok
                          after
                              10000 -> throw({timeout_waiting_for_server_exit, Name})
                          end
                  end,
                  [{MasterMonitor, MasterPortCoordPid, master}, {Replica1Monitor, Replica1PortCoordPid, replica1}, {Replica2Monitor, Replica2PortCoordPid, replica2}]),

    undefined = whereis(master_port_coordinator),
    undefined = whereis(replica1_port_coordinator),
    undefined = whereis(replica2_port_coordinator),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_replication_election(StartFunc) ->

    io:format("\nPerforming test_replication_election\n"),

    Site1HostPort = {"localhost", 4010},
    Site2HostPort = {"localhost", 4020},
    Site3HostPort = {"localhost", 4000},

    Site1EnvDir = "send_queue_env",
    ensure_empty_dir(Site1EnvDir),
    {ok, Site1PortCoordPid} = ?BDB_PC:start_link(
                                site1_port_coordinator, 
                                Site1EnvDir, 
                                [{port_start_function, StartFunc}, 
                                 {env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                 {env_cache_size_mb_and_fragments, {5, 1}},
                                 {replication, [{this_hostname_port,            Site1HostPort},
                                                {this_site_priority,            1},
                                                {peer_hostnames_ports,          [Site2HostPort, Site3HostPort]},
                                                {initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}},
                                                {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                {limit_transfer_per_call_bytes, 1024 * 1024},
                                                {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                {{timeout, election_milliseconds},         2 *    1000}, % milliseconds
                                                {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                {{timeout, connection_retry_milliseconds}, 5 *    1000}, % milliseconds
                                                {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                               ]}
                                ]),
    %% We get a EVENT_REP_CLIENT events when we start
    receive 
        {site1_port_coordinator, bdb_DB_EVENT_REP_CLIENT} -> ok 
    after 
        1000 -> throw(timeout_waiting_for_EVENT_REP_CLIENT)
    end,

    %% We can call this anytime, even before the election is done.
    Site1BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(site1_port_coordinator),


    %% This will become the master as it's priority is 2
    Site2EnvDir = "receive_queue_env",
    ensure_empty_dir(Site2EnvDir),
    {ok, Site2PortCoordPid} = ?BDB_PC:start_link(
                                site2_port_coordinator, 
                                Site2EnvDir, 
                                [{port_start_function, StartFunc}, 
                                 {env_cache_size_mb_and_fragments, {5, 1}},
                                 {replication, [{this_hostname_port,            Site2HostPort},
                                                {this_site_priority,            2},
                                                {peer_hostnames_ports,          [Site1HostPort, Site3HostPort]},
                                                {initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}},
                                                {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                {limit_transfer_per_call_bytes, 1024 * 1024},
                                                {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                {{timeout, election_milliseconds},         2 *    1000}, % milliseconds
                                                {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                {{timeout, connection_retry_milliseconds}, 5 *    1000},  % milliseconds
                                                {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                               ]}
                                ]),
    %% ... but it starts as a replica, until the election is done
    receive 
        {site2_port_coordinator, bdb_DB_EVENT_REP_CLIENT} -> ok 
    after 
        1000 -> throw(timeout_waiting_for_EVENT_REP_CLIENT)
    end,
    Site2BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(site2_port_coordinator),


    Site3EnvDir = "txn_tests_env",
    ensure_empty_dir(Site3EnvDir),
    {ok, Site3PortCoordPid} = ?BDB_PC:start_link(
                                site3_port_coordinator, 
                                Site3EnvDir, 
                                [{port_start_function, StartFunc}, 
                                 {env_cache_size_mb_and_fragments, {5, 1}},
                                 {replication, [{this_hostname_port,            Site3HostPort},
                                                {this_site_priority,            1},
                                                {peer_hostnames_ports,          [Site1HostPort, Site2HostPort]},
                                                {initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}},
                                                {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                {limit_transfer_per_call_bytes, 1024 * 1024},
                                                {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                {{timeout, election_milliseconds},         2 *    1000}, % milliseconds
                                                {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                {{timeout, connection_retry_milliseconds}, 5 *    1000}, % milliseconds
                                                {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                               ]}
                                ]),
    receive 
        {site3_port_coordinator, bdb_DB_EVENT_REP_CLIENT} -> ok 
    after 
        1000 -> throw(timeout_waiting_for_EVENT_REP_CLIENT)
    end,
    Site3BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(site3_port_coordinator),


    io:format("Waiting for election result~n"),
    receive {site1_port_coordinator, bdb_DB_EVENT_REP_NEWMASTER, _} = M1 -> 
            io:format("Got event ~p~n", [M1]) 
    end,
    receive {site2_port_coordinator, bdb_DB_EVENT_REP_MASTER}    = M2 ->  % site2 has highest priority
            io:format("Got event ~p~n", [M2]) 
    end,
    receive {site3_port_coordinator, bdb_DB_EVENT_REP_NEWMASTER, _} = M3 -> 
            io:format("Got event ~p~n", [M3]) 
    end,
    io:format("Got election result~n"),

    %% Can now access the database on all sites.

    %%timer:sleep(10000),

    %% master
    Site2DB = ?BDB_PC:get_db_handle(Site2BdbPortHandle, bdb_DB_BTREE, "test_replication.db"),

    %%timer:sleep(10000),

    Site1DB = ?BDB_PC:get_db_handle(Site1BdbPortHandle, bdb_DB_BTREE, "test_replication.db"),
    Site3DB = ?BDB_PC:get_db_handle(Site3BdbPortHandle, bdb_DB_BTREE, "test_replication.db"),

    %% Do a put.
    Key2   = <<"++siteted_key_2">>,
    Value2 = <<"!!*siteted*value*two*">>,
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(Site1BdbPortHandle, Site1DB, void, Key2, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(Site2BdbPortHandle, Site2DB, void, Key2, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(Site3BdbPortHandle, Site3DB, void, Key2, <<>>, [])),

    ok = ?BDB:db_put(Site2BdbPortHandle, Site2DB, void, Key2, Value2, []),

    %% currently STARTUPDONE is only sent after the first live transaction
    io:format("Waiting for replicas to sync-up~n"),
    receive {site1_port_coordinator, bdb_DB_EVENT_REP_STARTUPDONE} = M4 -> 
            io:format("Got event ~p~n", [M4]) 
    end,
    receive {site3_port_coordinator, bdb_DB_EVENT_REP_STARTUPDONE} = M5 -> 
            io:format("Got event ~p~n", [M5]) 
    end,
    io:format("Replicas are in-sync~n"),

    {Key2, Val2} = ?BDB:db_get(Site1BdbPortHandle, Site1DB, void, Key2, <<>>, []),
    {Key2, Val2} = ?BDB:db_get(Site2BdbPortHandle, Site2DB, void, Key2, <<>>, []),
    {Key2, Val2} = ?BDB:db_get(Site3BdbPortHandle, Site3DB, void, Key2, <<>>, []),

    %% Show that we can remove handlers
    ok = gen_event:delete_handler(?BDB_PC:get_replication_gen_event_proc_name(site1_port_coordinator),
                                  ?MODULE,
                                  removing_handler),   % this arbitrary value is passed as the first argument to ?MODULE:terminate/2

    ok = ?BDB_PC:release_bdb_port_handle(Site1BdbPortHandle),
    ok = ?BDB_PC:release_bdb_port_handle(Site2BdbPortHandle),
    ok = ?BDB_PC:release_bdb_port_handle(Site3BdbPortHandle),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.

    %%io:format("stopping site1~n"),
    Site1Monitor = erlang:monitor(process, Site1PortCoordPid),
    ok = ?BDB_PC:stop(site1_port_coordinator),

    %%io:format("stopping site2~n"),
    Site2Monitor = erlang:monitor(process, Site2PortCoordPid),
    ok = ?BDB_PC:stop(site2_port_coordinator),

    %%io:format("stopping site3~n"),
    Site3Monitor = erlang:monitor(process, Site3PortCoordPid),
    ok = ?BDB_PC:stop(site3_port_coordinator),


    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next test might get 'already_started' when it tries to start the server.
    lists:foreach(fun({MonitorRef, Pid, Name}) ->
                          receive
                              {'DOWN', MonitorRef,   process, Pid, normal} -> 
                                  %%io:format("Normal exit from ~p~n", [Name]),
                                  ok
                          after
                              10000 -> throw({timeout_waiting_for_server_exit, Name})
                          end
                  end,
                  [{Site3Monitor, Site3PortCoordPid, site3}, {Site1Monitor, Site1PortCoordPid, site1}, {Site2Monitor, Site2PortCoordPid, site2}]),

    undefined = whereis(site1_port_coordinator),
    undefined = whereis(site2_port_coordinator),
    undefined = whereis(site3_port_coordinator),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


%%
%% gen_event callbacks for replication tests (just forward all events to the process that installed the handler)
%%

init({forward_events_as_messages_to, _ForwardToPid} = State) ->
    io:format("~w:init called ~w~n", [?MODULE, State]),
    {ok, State}.

handle_event(Event, {forward_events_as_messages_to, ForwardToPid} = State) ->
    io:format("~w:handle_event received {~w,~w}~n", [?MODULE, Event, State]),
    ForwardToPid ! Event,
    {ok, State}.

%% If the event handler is deleted due to a call to gen_event:delete_handler, gen_event:swap_handler/3 or gen_event:swap_sup_handler/3, Arg is the Args argument of this function call. 
%% Arg={stop,Reason} if the event handler has a supervised connection to a process which has terminated with reason Reason. 
%% Arg=stop if the event handler is deleted because the event manager is terminating. 
%% Arg=remove_handler if the event handler is deleted because another callback function has returned remove_handler or {remove_handler,Reply}. 
%% Arg={error,Term} if the event handler is deleted because a callback function returned an unexpected value Term, or Arg={error,{'EXIT',Reason}} if a callback function failed. 
terminate(Args, State) ->
    io:format("~w:terminate received ~w, ~w~n", [?MODULE, Args, State]),
    ok.



example_replication_site(BdbEnvDir, MyReplicationHostPort, OtherSitesHostsPorts) ->

    io:format("example_replication_site() starting as ~p (connecting to ~p)~n", [MyReplicationHostPort, OtherSitesHostsPorts]),

    {ok, BdbPortCoordPid} = ?BDB_PC:start_link(
                                BdbEnvDir, 
                                [{replication, [{this_hostname_port,            MyReplicationHostPort},
                                                {this_site_priority,            1},
                                                {initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}},
                                                {ack_policy,                    bdb_DB_REPMGR_ACKS_ALL_PEERS},
                                                {peer_hostnames_ports,          OtherSitesHostsPorts},
                                                {limit_transfer_per_call_bytes, 1024 * 1024},
                                                {{timeout, acks_microseconds},             1 * 1000000}, % microseconds
                                                {{timeout, election_milliseconds},         2 *    1000}, % milliseconds
                                                {{timeout, election_retry_milliseconds},   5 *    1000}, % milliseconds
                                                {{timeout, connection_retry_milliseconds}, 5 *    1000}, % milliseconds
                                                {event_handlers, [{?MODULE, {forward_events_as_messages_to, self()}}]}
                                               ]}
                                ]),

    %% (Can call this before or after the election...)
    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(),

    %% We get a EVENT_REP_CLIENT events when we start
    receive {?BDB_PC, bdb_DB_EVENT_REP_CLIENT} -> ok end,
    
    io:format("Waiting for election result~n"),
    receive 
         {?BDB_PC, bdb_DB_EVENT_REP_MASTER} -> 
             io:format("We are now the MASTER~n");

         {?BDB_PC, bdb_DB_EVENT_REP_NEWMASTER, _} -> 
             io:format("We are a REPLICA~n") 
    end,


    %% (here do stuff with BDB...)


    %% shutdown
    ok = ?BDB_PC:release_bdb_port_handle(BdbPortHandle),

    %% BDB-HA currently always gives us an "EOF" error -- just ignore it.
    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, BdbPortCoordPid),
    ok = ?BDB_PC:stop(),
    receive
        {'DOWN', MonitorRef, process, BdbPortCoordPid, normal} -> 
            ok
    after
        2000 -> throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_distributed_transactions_stop_with_prepared_but_not_in_use_txns(StartFunc) ->

    io:format("\nPerforming test_basic_distributed_transactions_stop_with_prepared_but_not_in_use_txns\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P0 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P0, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1 = ?BDB:db_create(P0, []),
    ok = ?BDB:db_open(P0, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    DB2 = ?BDB:db_create(P0, []),
    ok = ?BDB:db_open(P0, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    
    Txn1 = ?BDB:txn_begin(P0, void, []),
    ok = ?BDB:db_put(P0, DB1, Txn1, "key1", "val1", []),
    Txn1GUID = <<1:128/unit:8>>,
    Txn1ID = ?BDB:txn_id(P0, Txn1),   % currently we have to get the id before calling txn_prepare, as CANNOT_AUTO_CLEANUP acts as a lock (until a "stop" function is called)
    %%io:format("Txn1ID ~w~n", [Txn1ID]),

    TxnStat0 = ?BDB:txn_stat(P0, []),
    {value, {active_txns, ActiveTxns0}} = lists:keysearch(active_txns, 1, TxnStat0),
    {value, {Txn1ID, ActiveTxn01}} = lists:keysearch(Txn1ID, 1, ActiveTxns0),
    {value, {status, bdb_TXN_RUNNING}} = lists:keysearch(status, 1, ActiveTxn01),
    {value, {xid, not_prepared}} = lists:keysearch(xid, 1, ActiveTxn01),

    ok = ?BDB:txn_prepare(P0, Txn1, Txn1GUID),

    %% It's illegal to prepare the same txn twice
    %% Presumably BDB would catch this, but outputv catches it first with
    %%    bdb port 1655: bdb_outputv: failed as valmap argument (or an ancestor) is locked : db_txn #0 (CANNOT_AUTO_CLEAN=true)
    ?EXPECT_BDB_ERROR(valmap_in_use, ?BDB:txn_prepare(P0, Txn1, <<50:128/unit:8>>)),

    Txn2 = ?BDB:txn_begin(P0, void, []),
    ok = ?BDB:db_put(P0, DB2, Txn2, "key2", "val2", []),
    Txn2GUID = <<2:128/unit:8>>,
    Txn2ID = ?BDB:txn_id(P0, Txn2),
    %%io:format("Txn2ID ~w~n", [Txn2ID]),
    %% this one has a child
    Txn2c = ?BDB:txn_begin(P0, Txn2, []),
    ok = ?BDB:db_put(P0, DB2, Txn2c, "key2c", "val2c", []),
%%     Txn2cID = ?BDB:txn_id(P0, Txn2c),
    %%io:format("Txn2cID ~w~n", [Txn2cID]),

    %% It's illegal to prepare a child (bdb catches this)
    ?EXPECT_BDB_ERROR(einval, ?BDB:txn_prepare(P0, Txn2c, Txn2GUID)),

    %% Do really prepare the parent
    ok = ?BDB:txn_prepare(P0, Txn2, Txn2GUID),

    io:format("aborting child of a prepared txn~n"),
    ?EXPECT_BDB_ERROR(badarg, ?BDB:txn_abort(P0, Txn2c)),    % will have incorrect generation-id as discard_descendents_of_prepared_txn() has been called


    %%ok = ?BDB:txn_stat_print(P0, []),

    TxnStat = ?BDB:txn_stat(P0, []),
    {value, {active_txns, ActiveTxns}} = lists:keysearch(active_txns, 1, TxnStat),
    %%io:format("ActiveTxns ~p~n", [ActiveTxns]),

    {value, {Txn1ID, ActiveTxn1}} = lists:keysearch(Txn1ID, 1, ActiveTxns),
    {value, {status, bdb_TXN_PREPARED}} = lists:keysearch(status, 1, ActiveTxn1),
    {value, {xid, Txn1GUID}} = lists:keysearch(xid, 1, ActiveTxn1),

    {value, {Txn2ID, ActiveTxn2}} = lists:keysearch(Txn2ID, 1, ActiveTxns),
    {value, {status, bdb_TXN_PREPARED}} = lists:keysearch(status, 1, ActiveTxn2),
    {value, {xid, Txn2GUID}} = lists:keysearch(xid, 1, ActiveTxn2),

%%     {value, {Txn2cID, ActiveTxn2c}} = lists:keysearch(Txn2cID, 1, ActiveTxns),
%%     {value, {status, bdb_TXN_COMMITTED}} = lists:keysearch(status, 1, ActiveTxn2c),
%%     {value, {xid, not_prepared}} = lists:keysearch(xid, 1, ActiveTxn2c),
%%     {value, {parent_txn_id, Txn2ID}} = lists:keysearch(parent_txn_id, 1, ActiveTxn2c),
    

    case StartFunc of
        start ->          
            normal = ?BDB:shutdown(P0),
            ok;

        start_pipe -> 
            os:cmd("killall pipe-main"), 
            receive
                {ErlangPort, {exit_status, 143}} = _ExitMsg when ErlangPort == P0#bdb_port_handle.erlang_port ->
                    %%io:format("Received ~p~n", [ExitMsg]);
                    ok
            end,
            timer:sleep(100),

            %% shutdown should fail now, obviously
            case (catch ?BDB:shutdown(P0)) of
                {'EXIT',{badarg,_}} ->
                    ok;
                Other ->
                    throw({unexpected, Other})
            end
    end,

    %% Verify that the unresolved txn is returned by txn_recover.
    %%io:format("Re-opening environment~n"),
    P2 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P2, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),

    %% This code correctly causes
    %%   my_debug_errcall: (NULL): recovery of prepared but not yet committed transactions is incomplete
    ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

    %%?BDB:adjust_port_logging(P2, enable, [all]),
    [{Txn1GUID, Txn1Recovered}, {Txn2GUID, Txn2Recovered}] = ?BDB:txn_recover(P2, bdb_DB_FIRST),
    0 = Txn1Recovered#bdb_db_txn.index,
    1 = Txn2Recovered#bdb_db_txn.index,
    %%ok = ?BDB:debug(P2, []),

    %% Still can't start a new txn
    ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

    ok = ?BDB:txn_abort(P2, Txn1Recovered),

    %% Still can't start a new txn
    ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

    ok = ?BDB:txn_commit(P2, Txn2Recovered, []),

    %% Can now start a new txn
    NewTxn = ?BDB:txn_begin(P2, void, []),
    DB1aa = ?BDB:db_create(P2, []),
    ok = ?BDB:db_open(P2, DB1aa, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    ok = ?BDB:db_put(P2, DB1aa, NewTxn, "newkey", "newval", []),
    ok = ?BDB:db_close(P2, DB1aa, []),
    ok = ?BDB:txn_commit(P2, NewTxn, []),

    %% safe to shutdown now
    normal = ?BDB:shutdown(P2),

    %% Verify that the txn effects have been commited or rolled back as expected
    P3 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    DB2a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB2a, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB1a, void, "key1", <<>>, [])),             % was aborted
    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P3, DB2a, void, "key2", <<>>, []), % was committed
    {<<"key2c">>, <<"val2c">>} = ?BDB:db_get(P3, DB2a, void, "key2c", <<>>, []), % was committed
    {<<"newkey">>, <<"newval">>} = ?BDB:db_get(P3, DB1a, void, "newkey", <<>>, []), % was committed
    ok = ?BDB:db_close(P3, DB1a, []),
    ok = ?BDB:db_close(P3, DB2a, []),
    normal = ?BDB:shutdown(P3),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns(StartFunc) ->

    io:format("\nPerforming test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P0 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P0, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1 = ?BDB:db_create(P0, []),
    ok = ?BDB:db_open(P0, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    DB2 = ?BDB:db_create(P0, []),
    ok = ?BDB:db_open(P0, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    PTxns = P0#bdb_port_handle{group_valmap_id = 1},

    Txn1 = ?BDB:txn_begin(PTxns, void, []),
    ok = ?BDB:db_put(PTxns, DB1, Txn1, "key1", "val1", []),
    Txn1GUID = <<1:128/unit:8>>,
    ok = ?BDB:txn_prepare(PTxns, Txn1, Txn1GUID),

    Txn2 = ?BDB:txn_begin(PTxns, void, []),
    ok = ?BDB:db_put(PTxns, DB2, Txn2, "key2", "val2", []),
    Txn2GUID = <<2:128/unit:8>>,
    ok = ?BDB:txn_prepare(PTxns, Txn2, Txn2GUID),

    %%?BDB:adjust_port_logging(P0, enable, [all]),
    2 = ?BDB:group_valmap_cleanup(PTxns, PTxns#bdb_port_handle.group_valmap_id),

    receive_need_to_resolve_event(StartFunc, PTxns, Txn1GUID, Txn1, timeout_is_an_error),
    receive_need_to_resolve_event(StartFunc, PTxns, Txn2GUID, Txn2, timeout_is_an_error),

    %% Can start a new txn even though they are unresolve
    NewTxn = ?BDB:txn_begin(P0, void, []),
    DB3 = ?BDB:db_create(P0, []),
    ok = ?BDB:db_open(P0, DB3, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    ok = ?BDB:db_put(P0, DB3, NewTxn, "newkey", "newval", []),
    ok = ?BDB:db_close(P0, DB3, []),
    ok = ?BDB:txn_commit(P0, NewTxn, []),

    %% simulate incremental resolution of the txns
    ok = ?BDB:txn_abort(PTxns, Txn1),
    ok = ?BDB:txn_commit(PTxns, Txn2, []),

    %% safe to shutdown now
    normal = ?BDB:shutdown(P0),

    %% Verify that the txn effects have been commited or rolled back as expected
    P3 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    DB2a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB2a, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    DB3a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB3a, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB1a, void, "key1", <<>>, [])),             % was aborted
    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P3, DB2a, void, "key2", <<>>, []), % was committed
    {<<"newkey">>, <<"newval">>} = ?BDB:db_get(P3, DB3a, void, "newkey", <<>>, []), % was committed
    ok = ?BDB:db_close(P3, DB1a, []),
    ok = ?BDB:db_close(P3, DB2a, []),
    normal = ?BDB:shutdown(P3),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_SAME_group(StartFunc) ->

    io:format("\nPerforming test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_SAME_group\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    %% This will HARD-ABORT the OS process, so we don't want to do it linked mode
    case StartFunc of
        start ->          
            ok;

        start_pipe -> 
            P0 = ?BDB:StartFunc(),
            ok = env_open_with_retry(P0, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),

            PDbsTxns = P0#bdb_port_handle{group_valmap_id = 1},

            DB1 = ?BDB:db_create(PDbsTxns, []),
            ok = ?BDB:db_open(PDbsTxns, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
            DB2 = ?BDB:db_create(PDbsTxns, []),
            ok = ?BDB:db_open(PDbsTxns, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

            Txn1 = ?BDB:txn_begin(PDbsTxns, void, []),
            ok = ?BDB:db_put(PDbsTxns, DB1, Txn1, "key1", "val1", []),
            Txn1GUID = <<1:128/unit:8>>,
            ok = ?BDB:txn_prepare(PDbsTxns, Txn1, Txn1GUID),

            Txn2 = ?BDB:txn_begin(PDbsTxns, void, []),
            ok = ?BDB:db_put(PDbsTxns, DB2, Txn2, "key2", "val2", []),
            Txn2GUID = <<2:128/unit:8>>,
            ok = ?BDB:txn_prepare(PDbsTxns, Txn2, Txn2GUID),

            %%?BDB:adjust_port_logging(P0, enable, [all]),
            case (catch ?BDB:group_valmap_cleanup(PDbsTxns, PDbsTxns#bdb_port_handle.group_valmap_id)) of
                {'EXIT', {bdb_port_error, {PDbsTxns, {ErlangPort, {exit_status, 1}}}}} when ErlangPort == PDbsTxns#bdb_port_handle.erlang_port ->
                    ok
            end,
            timer:sleep(100),

            %% shutdown should fail now, obviously
            case (catch ?BDB:shutdown(P0)) of
                {'EXIT',{badarg,_}} ->
                    ok;
                Other ->
                    throw({unexpected, Other})
            end,

            %% The driver tried to send events before dying -- it MAY have succeeeded.
            receive_need_to_resolve_event(StartFunc, PDbsTxns, Txn1GUID, Txn1, timeout_is_ok),
            receive_need_to_resolve_event(StartFunc, PDbsTxns, Txn2GUID, Txn2, timeout_is_ok),

            %% Verify that the unresolved txn is returned by txn_recover.
            %%io:format("Re-opening environment~n"),
            P2 = ?BDB:StartFunc(),
            ok = env_open_with_retry(P2, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),

            %% This code correctly causes
            %%   my_debug_errcall: (NULL): recovery of prepared but not yet committed transactions is incomplete
            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

            [{Txn1GUID, Txn1Recovered}, {Txn2GUID, Txn2Recovered}] = ?BDB:txn_recover(P2, bdb_DB_FIRST),
            0 = Txn1Recovered#bdb_db_txn.index,
            1 = Txn2Recovered#bdb_db_txn.index,

            %% Still can't start a new txn
            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

            ok = ?BDB:txn_abort(P2, Txn1Recovered),

            %% Still can't start a new txn
            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),

            ok = ?BDB:txn_commit(P2, Txn2Recovered, []),

            %% Can now start a new txn
            NewTxn = ?BDB:txn_begin(P2, void, []),
            DB1aa = ?BDB:db_create(P2, []),
            ok = ?BDB:db_open(P2, DB1aa, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
            ok = ?BDB:db_put(P2, DB1aa, NewTxn, "newkey", "newval", []),
            ok = ?BDB:db_close(P2, DB1aa, []),
            ok = ?BDB:txn_commit(P2, NewTxn, []),

            %% safe to shutdown now
            normal = ?BDB:shutdown(P2),

            %% Verify that the txn effects have been commited or rolled back as expected
            P3 = ?BDB:StartFunc(),
            ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
            DB1a = ?BDB:db_create(P3, []),
            ok = ?BDB:db_open(P3, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
            DB2a = ?BDB:db_create(P3, []),
            ok = ?BDB:db_open(P3, DB2a, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
            ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB1a, void, "key1", <<>>, [])),             % was aborted
            {<<"key2">>, <<"val2">>} = ?BDB:db_get(P3, DB2a, void, "key2", <<>>, []), % was committed
            {<<"newkey">>, <<"newval">>} = ?BDB:db_get(P3, DB1a, void, "newkey", <<>>, []), % was committed
            ok = ?BDB:db_close(P3, DB1a, []),
            ok = ?BDB:db_close(P3, DB2a, []),
            normal = ?BDB:shutdown(P3)
    end,

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


%% This test is no-longer executed as the condition it checks for is illegal and dangerous but no-longer explicitly tested for
%% (as the only current check is extremely conservative and became overly restrictive)
%% See comments in berkely_db.xml (search for 'num_unresolved_txns_owned_by_this_group')
%%
%%test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_DIFFERENT_group(StartFunc) ->
%% 
%%    io:format("\nPerforming test_basic_distributed_transactions_group_cleanup_prepared_but_not_in_use_txns_then_db_handle_in_DIFFERENT_group\n"),
%% 
%%    Dir = "txn_tests_env",
%%    ensure_empty_dir(Dir),
%% 
%%    P0 = ?BDB:StartFunc(),
%%    ok = env_open_with_retry(P0, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),
%% 
%%    PDbs = P0#bdb_port_handle{group_valmap_id = 1},
%% 
%%    DB1 = ?BDB:db_create(PDbs, []),
%%    ok = ?BDB:db_open(PDbs, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
%%    DB2 = ?BDB:db_create(PDbs, []),
%%    ok = ?BDB:db_open(PDbs, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
%% 
%%    PTxns = P0#bdb_port_handle{group_valmap_id = 2},
%% 
%%    Txn1 = ?BDB:txn_begin(PTxns, void, []),
%%    ok = ?BDB:db_put(PTxns, DB1, Txn1, "key1", "val1", []),
%%    Txn1GUID = <<1:128/unit:8>>,
%%    ok = ?BDB:txn_prepare(PTxns, Txn1, Txn1GUID),
%% 
%%    Txn2 = ?BDB:txn_begin(PTxns, void, []),
%%    ok = ?BDB:db_put(PTxns, DB2, Txn2, "key2", "val2", []),
%%    Txn2GUID = <<2:128/unit:8>>,
%%    ok = ?BDB:txn_prepare(PTxns, Txn2, Txn2GUID),
%% 
%%    %% kill the txns but not the DB handles
%%    %%?BDB:adjust_port_logging(P0, enable, [all]),
%%    2 = ?BDB:group_valmap_cleanup(PTxns, PTxns#bdb_port_handle.group_valmap_id),
%% 
%%    %% we should get incremental resolve events
%%    receive_need_to_resolve_event(StartFunc, PTxns, Txn1GUID, Txn1, timeout_is_an_error),
%%    receive_need_to_resolve_event(StartFunc, PTxns, Txn2GUID, Txn2, timeout_is_an_error),
%% 
%%    %% Can start a new txn even though they are unresolve
%%    NewTxn = ?BDB:txn_begin(P0, void, []),
%%    DB3 = ?BDB:db_create(P0, []),
%%    ok = ?BDB:db_open(P0, DB3, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
%%    ok = ?BDB:db_put(P0, DB3, NewTxn, "newkey", "newval", []),
%%    ok = ?BDB:db_close(P0, DB3, []),
%%    ok = ?BDB:txn_commit(P0, NewTxn, []),
%% 
%%    case StartFunc of
%%        start ->          
%%            %% Test group cleanup of DBs with NO open txns
%%            %% This does NOT abort the OS process
%% 
%%            %% simulate incremental resolution of the txns
%%            ok = ?BDB:txn_abort(PTxns, Txn1),
%%            ok = ?BDB:txn_commit(PTxns, Txn2, []),
%% 
%%            2 = ?BDB:group_valmap_cleanup(PDbs, PDbs#bdb_port_handle.group_valmap_id),
%% 
%%            ?EXPECT_BDB_ERROR(badarg, ?BDB:db_put(P0, DB1, void, "newkey2", "newval2", [])),
%%            DB1aa = ?BDB:db_create(PDbs, []),
%%            ok = ?BDB:db_open(PDbs, DB1aa, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
%%            ok = ?BDB:db_put(P0, DB1aa, void, "newkey2", "newval2", []),
%%            
%%            normal = ?BDB:shutdown(P0);
%% 
%%        start_pipe -> 
%%            %% Test group cleanup of DBs WITH open txns
%% 
%%            %% This will HARD-ABORT the OS process, so we don't want to do it linked mode
%%            %%?BDB:adjust_port_logging(P0, enable, [all]),
%%            case (catch ?BDB:group_valmap_cleanup(PDbs, PDbs#bdb_port_handle.group_valmap_id)) of
%%                {'EXIT', {bdb_port_error, {PDbs, {ErlangPort, {exit_status, 1}}}}} when ErlangPort == PDbs#bdb_port_handle.erlang_port ->
%%                    ok
%%            end,
%%            timer:sleep(100),
%% 
%%            %% shutdown should fail now, obviously
%%            case (catch ?BDB:shutdown(P0)) of
%%                {'EXIT',{badarg,_}} ->
%%                    ok;
%%                Other ->
%%                    throw({unexpected, Other})
%%            end,
%% 
%%            %% Verify that the unresolved txn is returned by txn_recover.
%%            io:format("Re-opening environment~n"),
%%            P2 = ?BDB:StartFunc(),
%%            ok = env_open_with_retry(P2, Dir, [bdb_DB_PRIVATE, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_LOG, bdb_DB_INIT_TXN], 8#644, ?LINE),
%%            %%?BDB:adjust_port_logging(P2, enable, [all]),
%% 
%%            %% This code correctly causes
%%            %%   my_debug_errcall: (NULL): recovery of prepared but not yet committed transactions is incomplete
%%            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),
%% 
%%            [{Txn1GUID, Txn1Recovered}, {Txn2GUID, Txn2Recovered}] = ?BDB:txn_recover(P2, bdb_DB_FIRST),
%%            0 = Txn1Recovered#bdb_db_txn.index,
%%            1 = Txn2Recovered#bdb_db_txn.index,
%% 
%%            %% Still can't start a new txn
%%            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),
%% 
%%            ok = ?BDB:txn_abort(P2, Txn1Recovered),
%% 
%%            %% Still can't start a new txn
%%            ?EXPECT_BDB_ERROR(einval, ?BDB:txn_begin(P2, void, [])),
%% 
%%            ok = ?BDB:txn_commit(P2, Txn2Recovered, []),
%% 
%%            %% Can now start a new txn
%%            NewTxn2 = ?BDB:txn_begin(P2, void, []),
%%            DB1aa = ?BDB:db_create(P2, []),
%%            ok = ?BDB:db_open(P2, DB1aa, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
%%            ok = ?BDB:db_put(P2, DB1aa, NewTxn2, "newkey2", "newval2", []),
%%            ok = ?BDB:db_close(P2, DB1aa, []),
%%            ok = ?BDB:txn_commit(P2, NewTxn2, []),
%% 
%%            %% safe to shutdown now
%%            normal = ?BDB:shutdown(P2)
%%    end,
%% 
%%    %% Verify that the txn effects have been commited or rolled back as expected
%%    P3 = ?BDB:StartFunc(),
%%    ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
%%    DB1a = ?BDB:db_create(P3, []),
%%    ok = ?BDB:db_open(P3, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
%%    DB2a = ?BDB:db_create(P3, []),
%%    ok = ?BDB:db_open(P3, DB2a, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
%%    DB3a = ?BDB:db_create(P3, []),
%%    ok = ?BDB:db_open(P3, DB3a, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
%%    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB1a, void, "key1", <<>>, [])),             % was aborted
%%    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P3, DB2a, void, "key2", <<>>, []), % was committed
%%    {<<"newkey">>, <<"newval">>} = ?BDB:db_get(P3, DB3a, void, "newkey", <<>>, []), % was committed
%%    {<<"newkey2">>, <<"newval2">>} = ?BDB:db_get(P3, DB1a, void, "newkey2", <<>>, []), % was committed
%%    ok = ?BDB:db_close(P3, DB1a, []),
%%    ok = ?BDB:db_close(P3, DB2a, []),
%%    normal = ?BDB:shutdown(P3),
%% 
%%    timer:sleep(100),
%%    {messages, []} = process_info(self(), messages),
%%    ok.


test_basic_distributed_transactions_prepare_finds_group_cleanup_in_progress(StartFunc) ->

    io:format("\nPerforming test_basic_distributed_transactions_prepare_finds_group_cleanup_in_progress\n"),

    %% If txn_prepare executes successfully and afterwards finds that CLEANUP_AS_SOON_AS_NOLONGER_IN_USE is set 
    %% then it does NOT auto-abort the txn, and it DOES send a 'need to resolve event'.
    %% This is a race condition.  We force it to happen by removing all threads from the pool used by txn_prepare 
    %% then enqueue the txn_prepare (non-blocking), then call group_valmap_cleanup (outputv will set CLEANUP_AS_SOON_AS_NOLONGER_IN_USE).
    %% Note that group_valmap_cleanup uses the same threadpool (#1) as txn_prepare uses.  This test does not rely on that, apart from 
    %% incidentally verifying that there are 2 queued commands in that threadpool before it creates a thread to allow txn_prepare to be executed.
    
    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB1, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 0, 64*1024, 100),
    timer:sleep(100),  % wait for poison pills to act (threads to be killed), or curr_queue_len won't be 0

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_VALMAP_STOP},
     {curr_queue_len,             0},
     {curr_idle_threads,          0},
     {num_threads,                0},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            100}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_VALMAP_STOP),

    PTxns = P#bdb_port_handle{group_valmap_id = 1},

    Txn1 = ?BDB:txn_begin(PTxns, void, []),
    ok = ?BDB:db_put(PTxns, DB1, Txn1, "key1", "val1", []),
    Txn1GUID = <<1:128/unit:8>>,
    Txn1PrepareTag = ?BDB:txn_prepare_non_block(PTxns, Txn1, Txn1GUID),

    Txn2 = ?BDB:txn_begin(PTxns, void, []),
    ok = ?BDB:db_put(PTxns, DB2, Txn2, "key2", "val2", []),
    Txn2GUID = <<1:128/unit:8>>,
    Txn2PrepareTag = ?BDB:txn_prepare_non_block(PTxns, Txn2, Txn2GUID),

%%     ?BDB:adjust_port_logging(P, enable, [all]),
    %% We must do non-blocking group cleanup too, or it will time-out.
    GroupValmapCleanupTag = ?BDB:group_valmap_cleanup_non_block(PTxns, PTxns#bdb_port_handle.group_valmap_id),

    [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_VALMAP_STOP},
     {curr_queue_len,             3},
     {curr_idle_threads,          0},
     {num_threads,                0},
     {curr_enqueued_poison_pills, 0},
     {queue_len_limit,            100}] = ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_VALMAP_STOP),

    %% Add a thread to allow the prepare and then the group-cleanup to happen
    ok = ?BDB:set_threadpool_params(P, ?BDB_THREADPOOL_ID_VALMAP_STOP, 1, 64*1024, 10),
   
    ok = ?BDB:receive_txn_prepare_reply(P, Txn1PrepareTag),
    ok = ?BDB:receive_txn_prepare_reply(P, Txn2PrepareTag),

    2 = ?BDB:receive_group_valmap_cleanup_reply(P, GroupValmapCleanupTag),
    receive_need_to_resolve_event(StartFunc, PTxns, Txn1GUID, Txn1, timeout_is_an_error),
    receive_need_to_resolve_event(StartFunc, PTxns, Txn2GUID, Txn2, timeout_is_an_error),

    io:format("Got unresolved txn~n"),

    %% Can start a new txn even though they are unresolve
    NewTxn = ?BDB:txn_begin(P, void, []),
    DB3 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB3, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
    ok = ?BDB:db_put(P, DB3, NewTxn, "newkey", "newval", []),
    ok = ?BDB:db_close(P, DB3, []),
    ok = ?BDB:txn_commit(P, NewTxn, []),

    %% simulate incremental resolution of the txna
    ok = ?BDB:txn_abort(PTxns, Txn1),
    ok = ?BDB:txn_commit(PTxns, Txn2, []),

    %% safe to shutdown now
    normal = ?BDB:shutdown(P),

    %% Verify that the txn effects have been commited or rolled back as expected
    P3 = ?BDB:StartFunc(),
    ok = env_open_with_retry(P3, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),
    DB1a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB1a, void, "test1.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    DB2a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB2a, void, "test2.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    DB3a = ?BDB:db_create(P3, []),
    ok = ?BDB:db_open(P3, DB3a, void, "test3.db", "", bdb_DB_BTREE, [bdb_DB_AUTO_COMMIT], 8#644),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB1a, void, "key1", <<>>, [])),             % was aborted
    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P3, DB2a, void, "key2", <<>>, []), % was committed
    {<<"newkey">>, <<"newval">>} = ?BDB:db_get(P3, DB3a, void, "newkey", <<>>, []), % was committed
    normal = ?BDB:shutdown(P3),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

    
    %% Test stop with prepared but in-use txns

    %% Test txn_prepare on a child (should fail)    

    %% Test low-level txn_prepare failing (I don't know how to make this happen -- maybe do it on a child and remove our own test that prevents that
 

receive_need_to_resolve_event(StartFunc, P1, TxnGUID, Txn, TimeoutIsOK) ->
    %% In pipe mode, the event(s) arrive as a pipe mode message and needs to be converted.
    %% There may be multiple events, so convert them until there appear to be no more.
    %% (in normal use, port coordinator will do this automatically)
    case StartFunc of
        start ->          
            ok;
        start_pipe -> 
            convert_pipe_messages_until_no_more_after_N_milliseconds(P1, 1000)
    end,

    %%{io:format("\nWaiting for event ~p ~p\n", [TxnGUID, Txn]), 
    receive
        {Port2, bdb_untagged_event, {need_to_resolve_prepared_txn, TxnGUID, Txn}} = _Msg2 when Port2 == P1#bdb_port_handle.erlang_port -> 
            %%io:format("\nReceived event ~p\n", [Msg2]), 
            ok
    after
        1000 -> 
            case TimeoutIsOK of 
                timeout_is_an_error ->
                    io:format("\n\ntimeout: msgs: ~p\n\n", [process_info(self(), messages)]),
                    %%{messages, []} = process_info(self(), messages),
                    throw(timeout_waiting_for_need_to_resolve_prepared_transaction);
                timeout_is_ok ->
                    ok
            end
    end.

convert_pipe_messages_until_no_more_after_N_milliseconds(BdbPortHandle, Timeout) ->
    receive
        {ErlangPort, {data, Bytes}} when ErlangPort == BdbPortHandle#bdb_port_handle.erlang_port ->
            Msg = ?BDB:pipe_driver_bytes_to_term(ErlangPort, Bytes),
            %%io:format("received (and forwarding) pipe-driver message ~p\n\n", [Msg]),
            self() ! Msg,
            convert_pipe_messages_until_no_more_after_N_milliseconds(BdbPortHandle, Timeout)
    after
        Timeout -> 
            %% timeout means we've received (and converted) all of the port messages we are going to get.
            ok
    end.
    


test_global_transaction_manager(StartFunc) ->

    io:format("\nPerforming test_global_transaction_manager\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),
    
    {ok, PortCoordPid} = ?BDB_PC:start_link(some_port_coordinator,
                                            Dir, 
                                            [{port_start_function, StartFunc}]),    

    {ok, GTMPid} = ?BDB_GTM:start_link(global_transaction_manager, some_port_coordinator),

    D0 = ?BDB_GTM:get_new_distributed_txn_guid(global_transaction_manager),
    D1 = ?BDB_GTM:get_new_distributed_txn_guid(global_transaction_manager),
    D2 = ?BDB_GTM:get_new_distributed_txn_guid(global_transaction_manager),

    io:format("explicit commit\n"),
    {D0, committed} = ?BDB_GTM:commit(global_transaction_manager,D0),
    {D0, committed} = ?BDB_GTM:resolve(global_transaction_manager,D0),
    ?EXPECT_GTM_ERROR({D0, already_committed}, ?BDB_GTM:abort(global_transaction_manager,D0)),
    {D0, committed} = ?BDB_GTM:commit(global_transaction_manager,D0),

    io:format("explicit abort\n"),
    {D1, aborted} = ?BDB_GTM:abort(global_transaction_manager,D1),
    {D1, aborted} = ?BDB_GTM:resolve(global_transaction_manager,D1),
    ?EXPECT_GTM_ERROR({D1, already_aborted}, ?BDB_GTM:commit(global_transaction_manager,D1)),
    {D1, aborted} = ?BDB_GTM:abort(global_transaction_manager,D1),

    io:format("implicit abort\n"),
    {D2, aborted} = ?BDB_GTM:resolve(global_transaction_manager,D2),
    {D2, aborted} = ?BDB_GTM:resolve(global_transaction_manager,D2),
    ?EXPECT_GTM_ERROR({D2, already_aborted}, ?BDB_GTM:commit(global_transaction_manager,D2)),

    io:format("mutltiple resolve\n"),
    [{ok, {D0, committed}}, 
     {ok, {D1, aborted}}, 
     {ok, {D2, aborted}}] = ?BDB_GTM:resolve(global_transaction_manager,[D0, D1, D2]),

    %%io:format("test that it is an error refer to a txn-id that has not yet been issued~n"),
    %% (gtm should probably kill the client)
    %% *** Ah, but we can't check this because it would slow the GTM down signifcantly
    %% ** (it would have to write to disk every time it issued a GUID).

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.

    %% Shutdown server cleanly, and make sure it happens.
    MonitorRefGTM = erlang:monitor(process, GTMPid),
    MonitorRefGTMSeq = erlang:monitor(process, whereis(global_transaction_manager_sequence_server)),
    ok = ?BDB_GTM:stop(global_transaction_manager),    
    receive
        {'DOWN', MonitorRefGTMSeq, process, _, normal} -> ok
    after
        2000 ->  throw(timeout_waiting_for_server_exit_gtm_seq)
    end,
    receive
        {'DOWN', MonitorRefGTM, process, GTMPid, normal} -> ok
    after
        1000 ->  throw(timeout_waiting_for_server_exit_gtm)
    end,
    undefined = whereis(global_tramsaction_manager),
    undefined = whereis(global_transaction_manager_sequence_server),

    PortCoordMonitorRef = erlang:monitor(process, PortCoordPid),
    ok = ?BDB_PC:stop(PortCoordPid),
    receive
        {'DOWN', PortCoordMonitorRef, process, _, normal} -> ok
    after
        2000 ->  throw(timeout_waiting_for_server_exit_gtm_pc)
    end,
    undefined = whereis(some_port_coordinator),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_port_coordinator_distributed_transactions_clean_shutdown_with_prepared_txns(StartFunc) ->
    io:format("\nPerforming test_port_coordinator_distributed_transactions_clean_shutdown_with_prepared_txns\n"),

    test_port_coordinator_distributed_transactions_kill_or_shutdown_with_prepared_txns(
      StartFunc, 
      fun(_PortCoordPid1, _PortCoordPid2) ->
%%               P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
%%               ?BDB:adjust_port_logging(P, enable, [all]),
%%               ok = ?BDB_PC:release_bdb_port_handle(P),
%%               link(PortCoordPid1),
%%               P2 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),
%%               ?BDB:adjust_port_logging(P2, enable, [all]),
%%               ok = ?BDB_PC:release_bdb_port_handle(P2),
%%               link(PortCoordPid2),
              ok = ?BDB_PC:stop(port_coordinator1),
              ok = ?BDB_PC:stop(port_coordinator2)
      end,
      normal),
    
    ok.

test_port_coordinator_distributed_transactions_kill_with_prepared_txns(StartFunc) ->
    io:format("\nPerforming test_port_coordinator_distributed_transactions_kill_with_prepared_txns\n"),

    test_port_coordinator_distributed_transactions_kill_or_shutdown_with_prepared_txns(
      StartFunc, 
      fun(PortCoordPid1, PortCoordPid2) ->
%%               P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
%%               ?BDB:adjust_port_logging(P, enable, [all]),
%%               P2 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),
%%               ?BDB:adjust_port_logging(P2, enable, [all]),
              exit(PortCoordPid1, kill),
              exit(PortCoordPid2, kill)
      end,
      killed),
    ok.

test_port_coordinator_distributed_transactions_kill_or_shutdown_with_prepared_txns(StartFunc, FunToShutdown, ExpectedExitReason) ->

    io:format("\nPerforming test_port_coordinator_distributed_transactions_kill_or_shutdown_with_prepared_txns\n"),

    %% We are going to (indirectly) kill port_coordinator client processes...
    OldTrapExitSetting = process_flag(trap_exit, true),

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    GTMDir = "txn_tests_env",
    ensure_empty_dir(GTMDir),
    GTM_ReplicationGroupName = gtm_group,
    GTM_Site1                = {{site1, node()}, GTMDir, {"localhost", 4000}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                              {this_site_priority, 1}]}]}, 
    GTM_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           GTM_ReplicationGroupName, 
                           [GTM_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, GTM_ReplicationGroupName),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    _GTM_RgPid = ?BDB_RG:start(
                    gtm_group, 
                    GTM_Site1,
                    [], % no other sites, just for simplicity
                    [{common_port_coordinator_config, [{port_start_function, StartFunc}
                                                      ]}]),

    GTMProcName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(GTM_ReplicationGroupName, site1),
    {ok, GTMPid} = ?BDB_GTM:start_link(GTMProcName, ?BDB_RG:get_port_coordinator_proc_name(gtm_group, site1)),

    PortCoordDir1 = "send_queue_env",
    ensure_empty_dir(PortCoordDir1),
    {ok, PortCoordPid1} = ?BDB_PC:start_link(port_coordinator1, 
                                             PortCoordDir1, 
                                             [{port_start_function, StartFunc},
                                              {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                             ]),

    PortCoordDir2 = "receive_queue_env",
    ensure_empty_dir(PortCoordDir2),
    {ok, PortCoordPid2} = ?BDB_PC:start_link(port_coordinator2, 
                                             PortCoordDir2, 
                                             [{port_start_function, StartFunc},
                                              {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                             ]),

    %% Do a distributed txn in a client that can be running when we shutdown the port coordinator

    TxnGUID = ?BDB_GTM:get_new_distributed_txn_guid(GTMProcName),

    Caller = self(),
    ClientPid1 = spawn_link(fun() ->
                                    P1 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
                                    DB1 = ?BDB_PC:get_db_handle(P1, bdb_DB_BTREE, "dist_txn_1.db"),
                                    Txn1 = ?BDB:txn_begin(P1, void, []),
                                    ok = ?BDB:db_put(P1, DB1, Txn1, "key1", "val1", []),
                                    ok = ?BDB:txn_prepare(P1, Txn1, TxnGUID),
                                    Caller ! {have_prepared, self()},
                                    receive never_get_this_message -> ok end 
                            end),

    ClientPid2 = spawn_link(fun() ->
                                    P2 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),
                                    DB2 = ?BDB_PC:get_db_handle(P2, bdb_DB_BTREE, "dist_txn_2.db"),
                                    Txn2 = ?BDB:txn_begin(P2, void, []),
                                    ok = ?BDB:db_put(P2, DB2, Txn2, "key2", "val2", []),
                                    ok = ?BDB:txn_prepare(P2, Txn2, TxnGUID),
                                    Caller ! {have_prepared, self()},
                                    receive never_get_this_message -> ok end 
                            end),

    receive {have_prepared, ClientPid1} -> ok end,
    receive {have_prepared, ClientPid2} -> ok end,

    {TxnGUID, committed} = ?BDB_GTM:commit(GTMProcName, TxnGUID),

    MonitorRef1 = erlang:monitor(process, PortCoordPid1),
    MonitorRef2 = erlang:monitor(process, PortCoordPid2),

    FunToShutdown(PortCoordPid1, PortCoordPid2),

    receive
        {'DOWN', MonitorRef1, process, PortCoordPid1, ExpectedExitReason} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit1)
    end,
    receive
        {'EXIT', PortCoordPid1, ExpectedExitReason} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_port_coord_to_die1)
    end,
    undefined = whereis(port_coordinator1),

    receive
        {'DOWN', MonitorRef2, process, PortCoordPid2, ExpectedExitReason} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit2)
    end,
    receive
        {'EXIT', PortCoordPid2, ExpectedExitReason} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_port_coord_to_die1)
    end,
    undefined = whereis(port_coordinator2),

    %% Now that port_coord monitors clients rather than linking to them,
    %% it is possible (e.g. if port-coord is hard-killed and doesn't not get 
    %% a chance to call terminate) that current clients won't be killed.
    %% This is safe as the BDB erlang port is closing so the will crash if they 
    %% try to issue any more commands.  And bdb_stop() will of course 
    %% do group_valmap_cleanup for them (even though they are still 
    %% running). 
    %% But it means that if FunToShutdown() uses exit(PortCoordPid,kill)
    %% then _we_ need to kill the clients.

    (catch exit(ClientPid1,killed)),
    (catch exit(ClientPid2,killed)),

    receive
        {'EXIT', ClientPid1, killed} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_client1_to_be_killed)
    end,
    receive
        {'EXIT', ClientPid2, killed} -> ok
    after
        2000 -> 
            throw(timeout_waiting_for_client2_to_be_killed)
    end,

    throw_if_any_exit_msgs(),  % must be done BEFORE restoring trap_exit
    true = process_flag(trap_exit, OldTrapExitSetting),


    %% run recovery
    {ok, PortCoordPid1a} = ?BDB_PC:start_link(port_coordinator1a, 
                                              PortCoordDir1, 
                                              [{port_start_function, StartFunc},
                                               {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                              ]),

    {ok, PortCoordPid2a} = ?BDB_PC:start_link(port_coordinator2a, 
                                              PortCoordDir2, 
                                              [{port_start_function, StartFunc},
                                               {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                              ]),

    %% The transaction should have been committed on both nodes
    P1a = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1a),
    DB1a = ?BDB_PC:get_db_handle(P1a, bdb_DB_BTREE, "dist_txn_1.db"),
    {<<"key1">>, <<"val1">>} = ?BDB:db_get(P1a, DB1a, void, "key1", <<>>, []), % was committed

    P2a = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2a),
    DB2a = ?BDB_PC:get_db_handle(P2a, bdb_DB_BTREE, "dist_txn_2.db"),
    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P2a, DB2a, void, "key2", <<>>, []), % was committed


    %% Shutdown server cleanly, and make sure it happens.
    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:release_bdb_port_handle(P1a),
    MonitorRef1a = erlang:monitor(process, PortCoordPid1a),
    ok = ?BDB_PC:stop(port_coordinator1a),
    receive
        {'DOWN', MonitorRef1a, process, PortCoordPid1a, normal} -> ok;
        Other3 -> throw({unexpected_msg, Other3})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit1a)
    end,
    undefined = whereis(port_coordinator1a),

    ok = ?BDB_PC:release_bdb_port_handle(P2a),
    MonitorRef2a = erlang:monitor(process, PortCoordPid2a),
    ok = ?BDB_PC:stop(port_coordinator2a),
    receive
        {'DOWN', MonitorRef2a, process, PortCoordPid2a, normal} -> ok;
        Other4 -> throw({unexpected_msg, Other4})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit2a)
    end,
    undefined = whereis(port_coordinator2a),

    MonitorRefGTM = erlang:monitor(process, GTMPid),
    GTMSeqProcName = list_to_atom(atom_to_list(GTMProcName) ++ "_sequence_server"),  % a private convention
    MonitorRefGTMSeq = erlang:monitor(process, whereis(GTMSeqProcName)),
    ok = ?BDB_GTM:stop(GTMProcName),    
    receive
        {'DOWN', MonitorRefGTMSeq, process, _, normal} -> ok
    after
        2000 ->  throw(timeout_waiting_for_server_exit_gtm_seq)
    end,
    receive
        {'DOWN', MonitorRefGTM, process, GTMPid, normal} -> ok
    after
        1000 ->  throw(timeout_waiting_for_server_exit_gtm)
    end,
    undefined = whereis(global_transaction_manager),
    undefined = whereis(global_transaction_manager_sequence_server),

    io:format("************** stopping gtm group~n"),    
    ok = ?BDB_RG:stop_entire_group({GTM_ReplicationGroupName, site1, node()}),
    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    timer:sleep( case StartFunc of 
                     start      -> 2000; 
                     start_pipe -> 4000 
                 end ),
    GTM_Site1_RgName = ?BDB_RG:get_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_RgName),
    GTM_Site1_PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_PortCoordName),

    ets:delete(RgCallerStateTableId),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


test_port_coordinator_distributed_transactions_client_dies_with_prepared_committed_txn(StartFunc) ->
    
    io:format("\nPerforming test_port_coordinator_distributed_transactions_client_dies_with_prepared_committed_txn\n"),

    %% Spawned clients are going to die, so...
    OldTrapExitSetting = process_flag(trap_exit, true),

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    GTMDir = "txn_tests_env",
    ensure_empty_dir(GTMDir),
    GTM_ReplicationGroupName = gtm_group,
    GTM_Site1                = {{site1, node()}, GTMDir, {"localhost", 4000}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                              {this_site_priority, 1}]}]}, 
    GTM_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           GTM_ReplicationGroupName, 
                           [GTM_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, GTM_ReplicationGroupName),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    _GTM_RgPid = ?BDB_RG:start(
                    gtm_group, 
                    GTM_Site1,
                    [], % no other sites, just for simplicity
                    [{common_port_coordinator_config, [{port_start_function, StartFunc}
                                                      ]}]),

    GTMProcName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(GTM_ReplicationGroupName, site1),
    {ok, GTMPid} = ?BDB_GTM:start_link(GTMProcName, ?BDB_RG:get_port_coordinator_proc_name(gtm_group, site1)),

    PortCoordDir1 = "send_queue_env",
    ensure_empty_dir(PortCoordDir1),
    {ok, PortCoordPid1} = ?BDB_PC:start_link(port_coordinator1, 
                                             PortCoordDir1, 
                                             [{port_start_function, StartFunc},
                                              {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                             ]),

    PortCoordDir2 = "receive_queue_env",
    ensure_empty_dir(PortCoordDir2),
    {ok, PortCoordPid2} = ?BDB_PC:start_link(port_coordinator2, 
                                             PortCoordDir2, 
                                             [{port_start_function, StartFunc},
                                              {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                             ]),

    %% Do a distributed txn
    TxnGUID = ?BDB_GTM:get_new_distributed_txn_guid(GTMProcName),
    Caller = self(),

    ClientPid1 = spawn_link(fun() ->
                                    P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
                                    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "dist_txn_1.db"),
                                    Txn = ?BDB:txn_begin(P, void, []),
                                    ok = ?BDB:db_put(P, DB1, Txn, "key1", "val1", []),
                                    ok = ?BDB:txn_prepare(P, Txn, TxnGUID),                                   
                                    Caller ! {prepared, self()},
                                    receive
                                        committed ->
                                            %% here a correct program should call txn_commit, but we just fall off the end of the function
                                            ok
                                    end
                            end),

    ClientPid2 = spawn_link(fun() ->
                                    P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),
                                    DB2 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "dist_txn_2.db"),
                                    Txn = ?BDB:txn_begin(P, void, []),
                                    ok = ?BDB:db_put(P, DB2, Txn, "key2", "val2", []),
                                    ok = ?BDB:txn_prepare(P, Txn, TxnGUID),
                                    Caller ! {prepared, self()},
                                    receive
                                        committed ->
                                            %% here a correct program should call txn_commit, but we intentionally crash
                                            erlang:error({intentional_crash, self()})
                                    end
                            end),

    ClientPid3 = spawn_link(fun() ->
                                    P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
                                    DB3 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "dist_txn_3.db"),
                                    Txn = ?BDB:txn_begin(P, void, []),
                                    ok = ?BDB:db_put(P, DB3, Txn, "key3", "val3", []),
                                    ok = ?BDB:txn_prepare(P, Txn, TxnGUID),
                                    Caller ! {prepared, self()},
                                    receive
                                        committed ->
                                            %% here a correct program should call txn_commit, but we accidentally (intentionally) disconnect from the port coordinator
                                            try
                                                ?BDB_PC:release_bdb_port_handle(P)
                                            catch
                                                {bdb_port_coordinator_error, {release_bdb_port_handle, P, still_owned_resources, 1}} ->
                                                    ok
                                            end
                                    end
                            end),

    ClientPid4 = spawn_link(fun() ->
                                    P = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),
                                    DB4 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "dist_txn_4.db"),
                                    Txn = ?BDB:txn_begin(P, void, []),
                                    ok = ?BDB:db_put(P, DB4, Txn, "key4", "val4", []),
                                    ok = ?BDB:txn_prepare(P, Txn, TxnGUID),
                                    Caller ! {prepared, self()},
                                    receive
                                        %% The test driver will kill this process
                                        never_get_this_message ->
                                            erlang:halt()
                                    end
                            end),

    lists:foreach(fun(ClientPid) -> receive {prepared, ClientPid} -> ok end end, 
                  [ClientPid1, ClientPid2, ClientPid3, ClientPid4]),

    {TxnGUID, committed} = ?BDB_GTM:commit(GTMProcName, TxnGUID),

    %% tell the clients to die in different ways
    %% the port_coordinators should run incremental recovery
    ClientPid1 ! committed,
    ClientPid2 ! committed,
    ClientPid3 ! committed,
    exit(ClientPid4, kill),

    lists:foreach(fun({ClientPid, intentional_crash}) -> receive {'EXIT', ClientPid, {{intentional_crash, ClientPid}, _}} -> ok end;
                     ({ClientPid, Reason}) -> receive {'EXIT', ClientPid, Reason} -> ok end end, 
                  [{ClientPid1, normal}, 
                   {ClientPid2, intentional_crash}, 
                   {ClientPid3, normal}, 
                   {ClientPid4, killed}]),

    %% wait for it to happen
    timer:sleep(500),

    %% The transaction should have been committed everywhere
    P1 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator1),
    P2 = ?BDB_PC:acquire_bdb_port_handle(port_coordinator2),

    DB1 = ?BDB_PC:get_db_handle(P1, bdb_DB_BTREE, "dist_txn_1.db"),
    {<<"key1">>, <<"val1">>} = ?BDB:db_get(P1, DB1, void, "key1", <<>>, []), % was committed

    DB2 = ?BDB_PC:get_db_handle(P2, bdb_DB_BTREE, "dist_txn_2.db"),
    {<<"key2">>, <<"val2">>} = ?BDB:db_get(P2, DB2, void, "key2", <<>>, []), % was committed

    DB3 = ?BDB_PC:get_db_handle(P1, bdb_DB_BTREE, "dist_txn_3.db"),
    {<<"key3">>, <<"val3">>} = ?BDB:db_get(P1, DB3, void, "key3", <<>>, []), % was committed

    DB4 = ?BDB_PC:get_db_handle(P2, bdb_DB_BTREE, "dist_txn_4.db"),
    {<<"key4">>, <<"val4">>} = ?BDB:db_get(P2, DB4, void, "key4", <<>>, []), % was committed

    throw_if_any_exit_msgs(),  % must be done BEFORE restoring trap_exit
    true = process_flag(trap_exit, OldTrapExitSetting),

    %% Shutdown server cleanly, and make sure it happens.
    ok = ?BDB_PC:release_bdb_port_handle(P1),
    MonitorRef1 = erlang:monitor(process, PortCoordPid1),
    ok = ?BDB_PC:stop(port_coordinator1),
    receive
        {'DOWN', MonitorRef1, process, PortCoordPid1, _} -> ok;
        Other3 -> throw({unexpected_msg, Other3})
    after
        8000 -> 
            throw({?AT, timeout_waiting_for_server_exit})
    end,
    undefined = whereis(port_coordinator1),

    ok = ?BDB_PC:release_bdb_port_handle(P2),
    MonitorRef2 = erlang:monitor(process, PortCoordPid2),
    ok = ?BDB_PC:stop(port_coordinator2),
    receive
        {'DOWN', MonitorRef2, process, PortCoordPid2, _} -> ok;
        Other4 -> throw({unexpected_msg, Other4})
    after
        8000 -> 
            throw({?AT, timeout_waiting_for_server_exit})
    end,
    undefined = whereis(port_coordinator2),

    MonitorRefGTM = erlang:monitor(process, GTMPid),
    GTMSeqProcName = list_to_atom(atom_to_list(GTMProcName) ++ "_sequence_server"),  % a private convention
    MonitorRefGTMSeq = erlang:monitor(process, whereis(GTMSeqProcName)),
    ok = ?BDB_GTM:stop(GTMProcName),    
    receive
        {'DOWN', MonitorRefGTMSeq, process, _, normal} -> ok
    after
        3000 ->  throw({?AT, timeout_waiting_for_server_exit_gtm_seq})
    end,
    receive
        {'DOWN', MonitorRefGTM, process, GTMPid, normal} -> ok
    after
        2000 ->  throw({?AT, timeout_waiting_for_server_exit_gtm})
    end,
    undefined = whereis(global_transaction_manager),
    undefined = whereis(global_transaction_manager_sequence_server),

    io:format("************** stopping gtm group~n"),    
    ok = ?BDB_RG:stop_entire_group({GTM_ReplicationGroupName, site1, node()}),
    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    timer:sleep( case StartFunc of 
                     start      -> 2000; 
                     start_pipe -> 4000 
                 end ),
    GTM_Site1_RgName = ?BDB_RG:get_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_RgName),
    GTM_Site1_PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_PortCoordName),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.


example_distributed_txn(StartFunc) ->

    %% A distributed txn needs a GTM

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    GTMDir = "receive_queue_env",
    ensure_empty_dir(GTMDir),
    GTM_ReplicationGroupName = gtm_group,
    GTM_Site1                = {{site1, node()}, GTMDir, {"localhost", 4000}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                              {this_site_priority, 1}]}]}, 
    GTM_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           GTM_ReplicationGroupName, 
                           [GTM_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, GTM_ReplicationGroupName),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    _GTM_RgPid = ?BDB_RG:start(
                    gtm_group, 
                    GTM_Site1,
                    [], % no other sites, just for simplicity
                    [{common_port_coordinator_config, [{port_start_function, StartFunc}
                                                      ]}]),

    GTMProcName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(GTM_ReplicationGroupName, site1),
    {ok, GTMPid} = ?BDB_GTM:start_link(GTMProcName, ?BDB_RG:get_port_coordinator_proc_name(gtm_group, site1)),

    %% A distributed txn has one leader and any number of participants.
    %% Our example will have one participant.

    %% We start all of the port_coordinators in the test driver process, as we don't want them to be killed if a participant crashes 
    %% (it just confuses things).  i.e. We simulate an OTP application starting the port coordinator.
    LeaderEnvDir = "txn_tests_env",
    ensure_empty_dir(LeaderEnvDir),
    {ok, LeaderBdbPortCoordPid} = ?BDB_PC:start_link(port_coordinator_leader,
                                                      LeaderEnvDir, 
                                                      [{port_start_function, StartFunc},
                                                       {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                                      ]),

    ParticipantEnvDir = "send_queue_env",
    ensure_empty_dir(ParticipantEnvDir),
    {ok, ParticipantBdbPortCoordPid} = ?BDB_PC:start_link(port_coordinator_participant1,
                                                          ParticipantEnvDir, 
                                                          [{port_start_function, StartFunc},
                                                           {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                                          ]),

    LeaderPid = self(),

    OldTrapExitSetting = process_flag(trap_exit, true),

    ExampleParticipantPid = 
        spawn_link(fun() ->
                           BdbPortHandle   = ?BDB_PC:acquire_bdb_port_handle(port_coordinator_participant1),

                           receive
                               {do_txn, LeaderPid, TxnGUID} ->

                                   io:format("Participants has been told to begin work\n"),
                                   DB = ?BDB_PC:get_db_handle(BdbPortHandle, 
                                                              bdb_DB_BTREE, 
                                                              "dist_txn_participant.db"),

                                   Txn = ?BDB:txn_begin(BdbPortHandle, void, []),
                                   ok = ?BDB:db_put(BdbPortHandle, DB, Txn, "key-partiticpant", "val-partiticpant", []),
                                   ok = ?BDB:txn_prepare(BdbPortHandle, Txn, TxnGUID),
                                   io:format("Participant has successfully prepared\n"),

                                   %% We must NOT unilaterally abort or commit now!
                                   %% If we crash, the port_coordinator will ask the GTM for a decision.

                                   %% Tell the leader we are ready
                                   LeaderPid ! {prepared, self(), TxnGUID},

                                   receive
                                       {commit, TxnGUID} ->
                                           io:format("Participant has been told to commit\n"),
                                           ok = ?BDB:txn_commit(BdbPortHandle, Txn, []);

                                       {abort, TxnGUID} ->
                                           io:format("Participant has been told to abort\n"),
                                           ok = ?BDB:txn_abort(BdbPortHandle, Txn)
                                   after
                                       1000 ->
                                           %% If the leader never tells us to commit or abort we can 
                                           %% just crash and the port coordinator will ask the GTM 
                                           %% for a decision
                                           io:format("Participant has timed-out waiting for a commit/abort decision from leader\n"),
                                           throw({participant_timed_out_waiting_for_commit_or_abort, TxnGUID})
                                   end,

                                   io:format("Participant is shutting down cleanly\n"),
                                   ok = ?BDB_PC:release_bdb_port_handle(BdbPortHandle)
                           end
                   end),

    %% Now run the leader.
    example_txn_leader(GTMProcName, ExampleParticipantPid),

    %% Wait for participant to exit before cleaning up
    receive {'EXIT', ExampleParticipantPid, Reason1} -> io:format("Participant has exited with code ~p\n", [Reason1]) end,

    %% As we are trapping exits we DO receive 'EXIT' messages with Reason 'normal'.
    ok = ?BDB_PC:stop(port_coordinator_participant1),
    ok = ?BDB_PC:stop(port_coordinator_leader),

    ok = ?BDB_RG:stop_entire_group({GTM_ReplicationGroupName, site1, node()}),
    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    timer:sleep( case StartFunc of 
                     start      -> 1000; 
                     start_pipe -> 3000 
                 end ),
    GTM_Site1_RgName = ?BDB_RG:get_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_RgName),
    GTM_Site1_PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(GTM_ReplicationGroupName, site1),
    undefined = whereis(GTM_Site1_PortCoordName),

    ok = ?BDB_GTM:stop(GTMProcName),

    timer:sleep(100),
    %% We are expecting port_coordinator and GTM exit messages
    receive {'EXIT', ParticipantBdbPortCoordPid, Reason2} -> io:format("Participant port_coordinator has exited with code ~p\n", [Reason2]) end,
    receive {'EXIT', GTMPid, Reason4} -> io:format("GTM has exited with code ~p\n", [Reason4]) end,
    receive {'EXIT', LeaderBdbPortCoordPid, Reason3} -> io:format("Leader port_coordinator has exited with code ~p\n", [Reason3]) end,

    true = process_flag(trap_exit, OldTrapExitSetting),

    io:format("messages left: ~p~n", [process_info(self(), messages)]),
    {messages, []} = process_info(self(), messages),
    ok.


example_txn_leader(GTMProcName, ExampleParticipantPid) ->

    BdbPortHandle = ?BDB_PC:acquire_bdb_port_handle(port_coordinator_leader),

    %% Get a globally unique ID for the transaction.
    %% This can be done any time before we start calling prepare.
    %% It should be fast, but for performance it could be done in 
    %% parallel (like a promise, or future-ptr) -- I haven't implemented that yet.
    TxnGUID = ?BDB_GTM:get_new_distributed_txn_guid(GTMProcName),

    %% The leader does *not* have to do a local txn itself, but in this example we do.
    %% We can tell the participants to begin and prepare their local txns before or 
    %% after we have prepared our local txn.
    %% The participants txns can be split into any number of message-driver sub-steps.
    %% In this example, we tell them to do some work and (implicitly) prepare the txn,
    %% in parallel with our own local txn.
    io:format("Leader tells participants to begin work\n"),
    ExampleParticipantPid ! {do_txn, self(), TxnGUID},

    %% Do our local (optional) txn.
    DB = ?BDB_PC:get_db_handle(BdbPortHandle, bdb_DB_BTREE, "dist_txn_leader.db"),
    Txn = ?BDB:txn_begin(BdbPortHandle, void, []),
    ok = ?BDB:db_put(BdbPortHandle, DB, Txn, "key-leader", "val-leader", []),
    ok = ?BDB:txn_prepare(BdbPortHandle, Txn, TxnGUID),
    io:format("Leader has prepared its local txn\n"),

    %% Before we can commit we MUST wait for ALL participants (and our local txn) to successfully prepare.
    TxnState = receive

                   {prepared, ExampleParticipantPid, TxnGUID} ->

                       %% If the GTM returns success here, the txn is committed.
                       %% If the GTM returns an error then the txn may or may not be committed.
                       %% (you can find out by calling ?BDB_GTM:resolve(TxnGUID)
                       io:format("Participant has prepared, so leader is asking GTM to commit\n"),
                       {TxnGUID, committed} = ?BDB_GTM:commit(GTMProcName, TxnGUID),

                       %% We must now tell all participants to commit.
                       %% We don't have to wait for them, or worry about them failing.
                       io:format("Global commit succeeded: leader is telling participants to commit\n"),
                       ExampleParticipantPid ! {commit, TxnGUID},

                       io:format("Leader is committing it's own txn\n"),
                       ok = ?BDB:txn_commit(BdbPortHandle, Txn, []),

                       committed;


                   {'EXIT', ExampleParticipantPid, _} ->

                       %% If a participant fails to prepare then the leader can choose to crash, 
                       %% or to explicitly abort.  Here we choose to explicitly abort.
                       io:format("Participant crashed before or after preparing. Leading is choosing to abort the txn\n"),
                       {TxnGUID, aborted} = ?BDB_GTM:abort(GTMProcName, TxnGUID),

                       io:format("Leader has told GTM that txn is aborted.  Leader is aborting it's own txn\n"),
                       ok = ?BDB:txn_abort(BdbPortHandle, Txn),

                       %% If there is more than one participant then we should tell any that have not 
                       %% crashed to abort.
                       io:format("Leader is telling any other surviving participants to abort their local txns\n"),
                       %%ExampleParticipantPid2 ! {commit, TxnGUID};            

                       aborted

               after
                   1000 ->
                       io:format("Leader failed to get a 'prepared' message from a participant: leader decides to crash\n"),
                       %% This is the same as the 'participant crashed' case ->
                       %% the leader can choose to crash, or explicitly abort.
                       %% As we did explicit abort above, here we choose to crash.
                       %% The port_coordinator will ask the GTM for a decision on the transaction, and
                       %% and will abort it.
                       throw({no_prepared_response_from_participant, ExampleParticipantPid})

               end,
    ok = ?BDB_PC:release_bdb_port_handle(BdbPortHandle),
    TxnState.



test_do_txn_on_master_retry_on_deadlock(StartFunc) ->

    io:format("\nPerforming test_do_txn_on_master_rety_on_deadlock\n"),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    {_GoodSitesPids, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            [{{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 3}]}]}, % first master
             {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 2}]}]}, % failover master
             {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
            ], 
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{log_removal_period_seconds,     disabled},
             {common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                               {auto_deadlock_detect, bdb_DB_LOCK_MINLOCKS},  % ** so we can predict outcome */
                                               {deadlock_detector_milliseconds_and_policy, disabled},
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 2}}, 
                                                              {ack_policy,		       bdb_DB_REPMGR_ACKS_ALL_PEERS}
                                                             ]}]},
             %% disable respawing so we can change master by killing it (otherwise it gets respawned so quickly that it wins the election)
             {spawn_other_sites, disabled}]),

    MasterHint1 = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, infinity),

    TestDriverPid = self(),

    Key1 = <<"key1">>,
    Val1 = <<"val1">>,

    Key2 = <<"-key2">>,
    Val2 = <<"-val2">>,

    LocalPid1 = spawn_link(fun() -> 
                                   TestDriverPid ! {self(),
                                                    result,
                                                    ?BDB_RG:do_txn_on_master(
                                                      MasterHint1,
                                                      {?MODULE, deadlock_do_txn_on_master_txn1, {TestDriverPid, Key1, Val1, Key2}},
                                                      0,       % NumDeadlocRetries
                                                      20000)}  % don't wan't gen_server to timeout
                           end),
    LocalPid2 = spawn_link(fun() -> 
                                   TestDriverPid ! {self(), 
                                                    result,
                                                    ?BDB_RG:do_txn_on_master(
                                                      MasterHint1,
                                                      {?MODULE, deadlock_do_txn_on_master_txn2, {TestDriverPid, Key1, Key2, Val2}},
                                                      1,       % Allow a single retry on deadlock
                                                      20000)}  % don't wan't gen_server to timeout
                           end),

    receive {TxnPid1, waiting_before_locking_db2} -> ok end,
    receive {TxnPid2, waiting_before_locking_db1} -> ok end,

    ?AT,
    TxnPid1 ! can_lock_db2,
    ?AT,
    TxnPid2 ! can_lock_db1,  % causes the deadlock
    ?AT,

    %% Txn1 is allows to complete
    receive {LocalPid1, result, Result1} -> ok end,
    ?AT,
    {ok1, MasterHint1} = Result1,
    ?AT,

    %% Txn2 is being retried (and it's pid has not changed)
    receive {TxnPid2, waiting_before_locking_db1} -> ok end,
    ?AT,
    TxnPid2 ! can_lock_db1,
    ?AT,

    %% Txn2 can complete
    receive {LocalPid2, result, Result2} -> ok end,
    ?AT,
    {ok2, MasterHint1} = Result2,
    ?AT,

    %% Check the result now that all txns have finished
    {ok_verified, MasterHint1} = ?BDB_RG:do_txn_on_master(
                                   MasterHint1,
                                   {?MODULE, deadlock_do_txn_on_master_verify, {Key1, Val1, Key2, Val2}}),
    ?AT,

    io:format("************** stopping group~n"),    
    ok = ?BDB_RG:stop_entire_group({test_group, site1, node()}),

    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 15000, ?LINE),

    {messages, []} = process_info(self(), messages),
    ok.

deadlock_do_txn_on_master_txn1(P, {TestDriverPid, Key1, Val1, Key2}) ->
    %% Txn1 locks a page in DB1 and a page in DB1a
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test1.db"),
    DB1a = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test1a.db"),

    Txn1 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_put(P, DB1, Txn1, Key1, Val1, []),
    ok = ?BDB:db_put(P, DB1a, Txn1, Key1, Val1, []),

    %% Waiting until Txn2 locks a page in DB2
    TestDriverPid ! {self(), waiting_before_locking_db2},
    io:format("*******Txn1 waiting for 'can_lock_db2' ~w~n", [self()]),
    receive can_lock_db2 -> ok end,
    io:format("*******Txn1 got 'can_lock_db2'~n"),

    %% Now Txn1 attempts to acquire the lock held by Txn2 (and blocks)
    DB2 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test2.db"),
    ?AT,
    ok = ?BDB:db_put(P, DB2, Txn1, Key2, Val1, []), 
    ?AT,

    %% Here Txn2 has been selected as the deadlock victim and aborted
    ok = ?BDB:txn_commit(P, Txn1, []),
    ok1.

deadlock_do_txn_on_master_txn2(P, {TestDriverPid, Key1, Key2, Val2}) ->

    %% second time through we want to give the other process (well, the thread calling db_put)
    %% a chance to acquire the lock that we release in group-valmap-cleanup when we get the deadlock exception
    %% I guess that just because the other thread is blocked on the lock, it 
    %% is not guaranteed to get it; this process might retry fast enough that it 
    %% gets the lock again. That could only happen if the mutex used by BDB is not FIFO, which 
    %% means it is subject to starvation.  But I am testing on RHEL3 with a back-port of NPTL,
    %% and NPTL mutexes are not fair:
    %% http://groups.google.com/group/linux.kernel/browse_thread/thread/845bc19f72ce9aa2/b7b0fa24db9366c4?lnk=st&q=pthread+mutex+FIFO&rnum=1#b7b0fa24db9366c4
    %% and neither are POSIX standard mutexes:
    %% http://groups.google.com/group/comp.programming.threads/browse_thread/thread/bf8e9d265321ca48/77a521dbff4c8ea2?lnk=st&q=posix+mutex+FIFO&rnum=3#77a521dbff4c8ea2
    timer:sleep(100),

    %% Txn2 locks a page in DB2
    DB2 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test2.db"),
    Txn2 = ?BDB:txn_begin(P, void, []),
    ok = ?BDB:db_put(P, DB2, Txn2, Key2, Val2, []),

    %% First time through we want to wait until Txn2 locks a page in DB2
    %% (second time through this just gets in the way a bit, but that's OK)
    io:format("*******Txn2 waiting for 'can_lock_db1 ~w'~n", [self()]),
    TestDriverPid ! {self(), waiting_before_locking_db1},
    receive can_lock_db1 -> ok end,
    io:format("*******Txn2 got 'can_lock_db1'~n"),

    %% Now Txn2 attempts to acquire the lock held by Txn1, which is a deadlock
    %% Txn2 has fewer locks, so it loses under the MINLOCKS policy
    %% So this will throw the first time, and Txn2 will be aborted by do_txn_on_master and then retried.
    %% It will succeed the second time.
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test1.db"),
    ?AT,
    ok = ?BDB:db_put(P, DB1, Txn2, Key1, Val2, []),
    ?AT,

    ok = ?BDB:txn_commit(P, Txn2, []),
    ok2.

deadlock_do_txn_on_master_verify(P, {Key1, Val1, Key2, Val2}) ->
    %% Txn1 was committed first and then Txn2 was committed (after a retry)
    DB1 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test1.db"),
    DB1a = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test1a.db"),
    DB2 = ?BDB_PC:get_db_handle(P, bdb_DB_BTREE, "test2.db"),
    {Key1, Val2} = ?BDB:db_get(P, DB1, void, Key1, <<>>, []),   % Txn2 changed this last
    {Key1, Val1} = ?BDB:db_get(P, DB1a, void, Key1, <<>>, []),
    {Key2, Val2} = ?BDB:db_get(P, DB2, void, Key2, <<>>, []),   % Txn2 changed this last
    ok_verified.



test_replication_group_start_and_spawn_respawn(StartFunc) ->

    io:format("\nPerforming test_replication_group_start_and_spawn_respawn\n"),

    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    _Pid = ?BDB_RG:start(
                   test_group, 
                   {{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 2}]}]}, 
                   [{{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 1}]}]},
                    {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
                   ],
                   %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
                   %% (those are all set internally)
                   [{log_removal_period_seconds,     disabled},
                    {master_cache_throttle_io,       {1000, 10000}}, % not intended to limit, just test the feature works
                    {common_port_coordinator_config, [{port_start_function, StartFunc},
                                                      {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                                      %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                                      %%{port_log_directory, "."},
                                                      {replication, [{initial_role_and_threadcount, {bdb_DB_REP_FULL_ELECTION, 2}}, 
                                                                     {ack_policy,		    bdb_DB_REPMGR_ACKS_ALL_PEERS}
                                                                    ]}]},
                    %% we DO want respawn
                    {spawn_other_sites, enabled}
                   ]),
    Node = node(),
    [{{site1, Node}, {replica, not_in_sync, unknown}},
     {{site2, Node}, unknown},
     {{site3, Node}, unknown}]
        = ?BDB_RG:get_group_status({test_group, site1, node()}),

    %% (The other 2 sites have not been started yet)


    %% If tests are run with "erl -s berkeley_db_test regression" then throwing here causes
    %%
    %%{"init terminating in do_boot",{{nocatch,forced_error},[{init,start_it,1},{init,start_em,1}]}}
    %%TYPE ASSERTION FAILED, file beam/erl_term.c, line 106: tag_val_def: 0x18
    %%
    %% If find_or_spawn_peers_process is not spawned (or exits before this point) then it does not happen.
    %%
    %%throw(forced_error),

    Site1PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site1),
    Site2PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site2),
    Site3PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site3),

    %% polling for master fails to find one
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 0)),

    %% waiting a very short timeout fails to find a master (and waits accurate time)
    Start = now(),
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 1, 5000)),
    Elapsed = ?BDB_U:now_diff_milliseconds(now(), Start),
    if Elapsed > 500 -> throw({elapsed_time_too_large, Elapsed});
       true -> ok
    end,

    io:format("************** Waiting for a master to be elected~n"),
    MasterHint1 = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 20000),
    io:format("************** GOT first MASTER ~w~n", [MasterHint1]),    

    %% in BDB v4.5 STARTUPDONE requires a new write transaction 
    %% But ?BDB_RG will do 'probe' txns in the background, so can test sync-up before doing any transactions)
    wait_until_all_sites_are_in_sync(MasterHint1, [site1, site2, site3], 30000, ?LINE),

    %% Test direct hit to master
    {{ok_txn_fun, txn_fun_result}, MasterHint1} 
	= ?BDB_RG:do_txn_on_master({test_group, site1, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key1", "committed1", {ok_txn_fun, txn_fun_result}}}),
    io:format("************** done key1~n"),    

    %% Test routing via a replica
    {ok_result, MasterHint1} 
	= ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key2", "committed2", ok_result}}),
    io:format("************** done key2~n"),    

    %% test helper API
    ThisNode = node(),
    ThisNode = ?BDB_RG:get_node_from_master_location_hint(MasterHint1),

    %% polling now should return immediately
    Start2 = now(),
    MasterHint1 = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 30000),
    Elapsed2 = ?BDB_U:now_diff_milliseconds(now(), Start2),
    if Elapsed2 > 400 -> throw({elapsed_time_too_large2, Elapsed2});
       true -> ok
    end,            

    %% Verify replication has happened
    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_any_site({test_group, site1, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),
    {<<"key2">>, <<"committed2">>} = ?BDB_RG:do_txn_on_any_site({test_group, site1, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}}),

    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_any_site({test_group, site2, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),
    {<<"key2">>, <<"committed2">>} = ?BDB_RG:do_txn_on_any_site({test_group, site2, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}}),

    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_any_site({test_group, site3, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),
    {<<"key2">>, <<"committed2">>} = ?BDB_RG:do_txn_on_any_site({test_group, site3, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}}),

    %% check that they have bit-identical contents (only checks the OPEN databases)
    ?AT,
    MD5_1 = ?BDB_H:do_txn(Site1PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 64*1024}, 0),
    ?AT,
    MD5_1 = ?BDB_H:do_txn(Site2PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 90*1024}, 0),
    ?AT,
    MD5_1 = ?BDB_H:do_txn(Site3PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 128*1024}, 0),

    if is_binary(MD5_1) andalso size(MD5_1) == 16 -> ok;
       true -> throw({bad_md5, ?AT})    
    end,

    io:format("************** killing master~n"),    
    %% TODO: also try killing the replication_group pid, and try shutting it (them) down cleanly
    exit(whereis(Site1PortCoordName), kill),

    timer:sleep(1000), % let exit signal propogate to other sites (or they will still have the wrong master location hint)

    io:format("************** Waiting for a NEW master to be elected~n"),    
    MasterHint2 = ?BDB_RG:get_master_location_hint({test_group, site3, node()}, infinity),
    io:format("************** GOT failover MASTER ~w~n", [MasterHint2]),    
    %% site1 sometimes restarts quckly enough to win, and sometimes even site3 wins

    %% We can test sync-up before performing a write txn, as ?BDB_RG has a feature to do such probe-txns in the background for us
    wait_until_all_sites_are_in_sync(MasterHint2, [site1, site2, site3], 30000, ?LINE),

    {ok_result, MasterHint2} 
	= ?BDB_RG:do_txn_on_master(MasterHint2,
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key3", "committed3", ok_result}}),
    io:format("************** done key3~n"),    
    MasterHint1SiteLocator = MasterHint1#master_location_hint.site_locator,
    ?EXPECT_RG_SITE_DOWN_ERROR(MasterHint1SiteLocator,
                               ?BDB_RG:do_txn_on_master(MasterHint1,
                                                        {?MODULE, test_txn_fun_put_1, 
                                                         {"key4", "committed4", ok_result}})),
    io:format("************** done key4~n"),    
    {ok_result, MasterHint2} 
	= ?BDB_RG:do_txn_on_master({test_group, site1, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key5.1", "committed5.1", ok_result}}),
    io:format("************** done key5.1~n"),    
    {ok_result, MasterHint2} 
	= ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key5.2", "committed5.2", ok_result}}),
    io:format("************** done key5.2~n"),    
    {ok_result, MasterHint2} 
	= ?BDB_RG:do_txn_on_master({test_group, site3, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key5.3", "committed5.3", ok_result}}),
    io:format("************** done key5.3~n"),    

    Start3 = now(),
    MasterHint2 = ?BDB_RG:get_master_location_hint({test_group, site3, node()}, infinity),
    Elapsed3 = ?BDB_U:now_diff_milliseconds(now(), Start3),
    if Elapsed3 > 400 -> throw({elapsed_time_too_large3, Elapsed3});
       true -> ok
    end,            

    %% Again, verify replication has happened
    lists:foreach(fun(Site) ->
                          ?AT,
                          {<<"key3">>, <<"committed3">>} = ?BDB_RG:do_txn_on_any_site({test_group, site1, node()},
                                                                                      {?MODULE, test_txn_fun_get, {"test.db", <<"key3">>}}),
                          ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND}, ?BDB_RG:do_txn_on_any_site({test_group, Site, node()},
                                                                                                         {?MODULE, test_txn_fun_get, {"test.db", <<"key4">>}})),
                          {<<"key5.1">>, <<"committed5.1">>} = ?BDB_RG:do_txn_on_any_site({test_group, Site, node()},
                                                                                          {?MODULE, test_txn_fun_get, {"test.db", <<"key5.1">>}}),
                          {<<"key5.2">>, <<"committed5.2">>} = ?BDB_RG:do_txn_on_any_site({test_group, Site, node()},
                                                                                          {?MODULE, test_txn_fun_get, {"test.db", <<"key5.2">>}}),
                          {<<"key5.3">>, <<"committed5.3">>} = ?BDB_RG:do_txn_on_any_site({test_group, Site, node()},
                                                                                          {?MODULE, test_txn_fun_get, {"test.db", <<"key5.3">>}})
                  end,
                  [site1, site2, site3]),

    %% check that they have bit-identical contents (only checks the OPEN databases)
    ?AT,
    MD5_2 = ?BDB_H:do_txn(Site1PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 64*1024}, 0),
    ?AT,
    MD5_2 = ?BDB_H:do_txn(Site2PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 96*1024}, 0),
    ?AT,
    MD5_2 = ?BDB_H:do_txn(Site3PortCoordName, {?BDB_H, txn_fun_md5_all_keys_values_in_all_on_disk_databases, 100*1024}, 0),

    case MD5_2 of 
        MD5_1                              -> throw({md5_should_not_match, ?AT});
        M when is_binary(M), size(M) == 16 -> ok
    end,

    io:format("************** waiting before stopping group~n"),    
    timer:sleep( case StartFunc of 
                     start      -> 2000; 
                     start_pipe -> 6000 
                 end ),

    io:format("************** stopping group~n"),    
    ok = ?BDB_RG:stop_entire_group({test_group, site1, node()}),

    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 15000, ?LINE),

    {messages, []} = process_info(self(), messages),
    ok.


test_replication_group_start_sites_no_respawn(StartFunc) ->

    io:format("\nPerforming test_replication_group_start_sites_no_respawn\n"),

    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),
    %%SMP_Mode = erlang:system_info(smp_support),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    %% The "replication" options per site are entirely optional.  Here we use it to set a preference for which site is master, as it simplifies the test.
    {_GoodSitesPids, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            [{{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 3}]}]}, % first master
	     {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 1}]}]},
             {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
            ], 
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                               %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                               %%{port_log_directory, "."},
                                               {enable_port_logging_before_open_environment, [threadpool_management]},
                                               {do_parallel_trickle_and_checkpoint, disabled},
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_ELECTION, 1}}, 
                                                              {ack_policy,		      bdb_DB_REPMGR_ACKS_QUORUM}, 
                                                              {limit_transfer_per_call_bytes, 100000}
                                                             ]}]}
             %% we leave respawn disables (the default) so we can change master by killing it (otherwise it gets respawned so quickly that it wins the election)
             %%{spawn_other_sites, disabled}
            ]),

%%    PidsMonitorRefs = [{Pid, erlang:monitor(Pid)} || {_Site, Pid} <- GoodSitesPids],
%%    [link(Pid) || {_Site, Pid} <- GoodSitesPids],

    Site1PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site1),
    Site2PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site2),
    Site3PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site3),

    io:format("************** Cannot do write transactions until a master is elected~n"),
    ?EXPECT_RG_ERROR(no_master,
                     ?BDB_RG:do_txn_on_master({test_group, site1, node()},
                                                                    {?MODULE, test_txn_fun_put_1, 
                                                                     {"key0", "not_committed0", dummy_return}})),

    ?EXPECT_RG_ERROR(no_master,
                     ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                                                    {?MODULE, test_txn_fun_put_1, 
                                                                     {"key0", "not_committed0", dummy_return}})),

    io:format("************** Waiting for a master to be elected~n"),    
    #master_location_hint{site_locator = {test_group, site1, _}} = MasterHint1
        = ?BDB_RG:get_master_location_hint({test_group, site2, node()}, 20000),

    %% Test direct hit to master
    {{ok_txn_fun, txn_fun_result}, MasterHint1} 
	= ?BDB_RG:do_txn_on_master({test_group, site1, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key1", "committed1", {ok_txn_fun, txn_fun_result}}}),
    io:format("************** done key1~n"),    

    %% Test routing via a replica
    {ok_result, MasterHint1} 
	= ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key2", "committed2", ok_result}}),
    io:format("************** done key2~n"),    

    {ok_result, MasterHint1} 
	= ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                   {?MODULE, test_txn_fun_put_1, 
                                    {"key2a", "committed2a", ok_result}}),
    io:format("************** done key2a~n"),    

    wait_until_all_sites_are_in_sync(MasterHint1, [site1, site2, site3], 30000, ?LINE),

    %% Test bad ModuleName/FunctionName/ClientDataTerm
    ?EXPECT_TXN_FUN_ERROR({{error, undef}, [{bad_module,test_txn_fun_put_1,[#bdb_port_handle{},{"key2b","not_committed2b",dummy_return}]} | _]},
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {bad_module, test_txn_fun_put_1, 
                                                    {"key2b", "not_committed2b", dummy_return}})),

    ?EXPECT_TXN_FUN_ERROR({{error, undef}, [{?MODULE,bad_function,_} | _]}, 
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {?MODULE, bad_function, 
                                                    {"key2c", "not_committed2c", dummy_return}})),

    ?EXPECT_TXN_FUN_ERROR({{error, function_clause}, _},
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {?MODULE, test_txn_fun_put_1, 
                                                    {"key2d", "bad ClientTermData"}})),
    %% Test use of MasterLocationHint
    %% Test TxnFun crashing
    %% Test group valmap cleanup (crash/exit/throw before txn commit)

    ?EXPECT_TXN_FUN_ERROR({{error, {badmatch, ok}}, _},
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {?MODULE, test_txn_fun_crash_in_a_txn, []})),
    io:format("************** done key3/4/5~n"),    

    ?EXPECT_TXN_FUN_ERROR({{exit, {intentional_exit, here}}, _},
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {?MODULE, test_txn_fun_exit_in_a_txn, []})),
    io:format("************** done key3a/4a/5a~n"),    

    %% Test TxnFun returning with open txns/cursors
    %% do_txn_on_master detects this by a non-zero NumCleanedResources from group_valmap_cleanup.
    ?EXPECT_TXN_FUN_ERROR({return_with_open_resources, 3, {txn_fun_return_value, stuff}},
                          ?BDB_RG:do_txn_on_master(MasterHint1,
                                                   {?MODULE, test_txn_fun_return_with_open_resources, []})),
    io:format("************** done key6/7~n"),    

    %% Verify replication has happened

    P1 = ?BDB_PC:acquire_bdb_port_handle(Site1PortCoordName),
    DB1 = ?BDB_PC:get_db_handle(P1, bdb_DB_BTREE, "test.db"),

    P2 = ?BDB_PC:acquire_bdb_port_handle(Site2PortCoordName),
    DB2 = ?BDB_PC:get_db_handle(P2, bdb_DB_BTREE, "test.db"),

    P3 = ?BDB_PC:acquire_bdb_port_handle(Site3PortCoordName),
    DB3 = ?BDB_PC:get_db_handle(P3, bdb_DB_BTREE, "test.db"),

    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND}, 
                          ?BDB_RG:do_txn_on_any_site({test_group, site1, node()},
                                                     {?MODULE, test_txn_fun_get, {"test.db", "key0"}})),

    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_any_site({test_group, site2, node()},
                                                                {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key0">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key0">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key0">>, <<>>, [])),

    {<<"key1">>, <<"committed1">>} = ?BDB:db_get(P1, DB1, void, <<"key1">>, <<>>, []),
    {<<"key1">>, <<"committed1">>} = ?BDB:db_get(P2, DB2, void, <<"key1">>, <<>>, []),
    {<<"key1">>, <<"committed1">>} = ?BDB:db_get(P3, DB3, void, <<"key1">>, <<>>, []),

    {<<"key2">>, <<"committed2">>} = ?BDB:db_get(P1, DB1, void, <<"key2">>, <<>>, []),
    {<<"key2">>, <<"committed2">>} = ?BDB:db_get(P2, DB2, void, <<"key2">>, <<>>, []),
    {<<"key2">>, <<"committed2">>} = ?BDB:db_get(P3, DB3, void, <<"key2">>, <<>>, []),

    {<<"key2a">>, <<"committed2a">>} = ?BDB:db_get(P1, DB1, void, <<"key2a">>, <<>>, []),
    {<<"key2a">>, <<"committed2a">>} = ?BDB:db_get(P2, DB2, void, <<"key2a">>, <<>>, []),
    {<<"key2a">>, <<"committed2a">>} = ?BDB:db_get(P3, DB3, void, <<"key2a">>, <<>>, []),

    {<<"key3">>, <<"committed3">>} = ?BDB:db_get(P1, DB1, void, <<"key3">>, <<>>, []),
    {<<"key3">>, <<"committed3">>} = ?BDB:db_get(P2, DB2, void, <<"key3">>, <<>>, []),
    {<<"key3">>, <<"committed3">>} = ?BDB:db_get(P3, DB3, void, <<"key3">>, <<>>, []),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key4">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key4">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key4">>, <<>>, [])),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key5">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key5">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key5">>, <<>>, [])),

    {<<"key3a">>, <<"committed3a">>} = ?BDB:db_get(P1, DB1, void, <<"key3a">>, <<>>, []),
    {<<"key3a">>, <<"committed3a">>} = ?BDB:db_get(P2, DB2, void, <<"key3a">>, <<>>, []),
    {<<"key3a">>, <<"committed3a">>} = ?BDB:db_get(P3, DB3, void, <<"key3a">>, <<>>, []),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key4a">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key4a">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key4a">>, <<>>, [])),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key5a">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key5a">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key5a">>, <<>>, [])),

    {<<"key6">>, <<"committed6">>} = ?BDB:db_get(P1, DB1, void, <<"key6">>, <<>>, []),
    {<<"key6">>, <<"committed6">>} = ?BDB:db_get(P2, DB2, void, <<"key6">>, <<>>, []),
    {<<"key6">>, <<"committed6">>} = ?BDB:db_get(P3, DB3, void, <<"key6">>, <<>>, []),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P1, DB1, void, <<"key7">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key7">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key7">>, <<>>, [])),

    ok = ?BDB_PC:release_bdb_port_handle(P1),
    ok = ?BDB_PC:release_bdb_port_handle(P2),
    ok = ?BDB_PC:release_bdb_port_handle(P3),

    %% Test killing MASTER replication group (respawned by peers, and it's port_coord dies and is respawned by new replication_group)

    io:format("**********Killing MASTER replication_group~n"),

    exit(whereis(Site1PortCoordName), kill),
    %% Alternatives:
    %%exit(whereis(?BDB_RG:get_proc_name(test_group, site1)), kill),
    %%ok = ?BDB_PC:stop(Site1PortCoordName),

    io:format("************** Old MasterLocationHint does not work~n"),
    %% (we have to test this immediately after killing the site1 port_coord, as site1 will be respawned and will win the election)
    ThisNode = node(),  % required as EXPECT_RG_SITE_DOWN_ERROR uses the SiteLocator in a pattern (as a macro)
    ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site1, ThisNode},
                               ?BDB_RG:do_txn_on_master(MasterHint1,
                                                        {?MODULE, test_txn_fun_put_1, {"key8", "not_committed8"}})),

    %% same error using an explicit SiteLocator
    ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site1, ThisNode},
                               ?BDB_RG:do_txn_on_master({test_group, site1, node()},
                                                        {?MODULE, test_txn_fun_put_1, {"key8", "not_committed8"}})),

    %% trying to access the down master indirectly via the other sites 
    %% always fails, but the type of error is a race condition
    %% - if the site we ask has seen that the master has failed then we get no_master.
    %% - if the site that we ask still thinks the master is running then we get site_is_down for the (old) *master* site
    %% On my workstation, the former happens if SMP is disabled, and the latter happens if SMP is enabled.
    %% However, even testing if SMP is enabled changes the scheduling.
    %% So to make it predictable we do a small sleep
    timer:sleep(100),

    ?EXPECT_RG_ERROR(no_master,
                     ?BDB_RG:do_txn_on_master({test_group, site2, node()},
                                              {?MODULE, test_txn_fun_put_1, {<<"key9">>, "not_committed9"}})),
    ?EXPECT_RG_ERROR(no_master,
                     ?BDB_RG:do_txn_on_master({test_group, site3, node()},
                                              {?MODULE, test_txn_fun_put_1, {"key10", <<"not_committed10">>}})),

    io:format("************** Waiting for a new master to be elected~n"),    
    #master_location_hint{site_locator = {test_group, _MasterSiteName2, _}} = MasterHint2
        = ?BDB_RG:get_master_location_hint({test_group, site2, node()}, infinity),

    %% we might have learned about the new master, but the other replica might not yet
    %% so give them a chance
    MasterHint2 = ?BDB_RG:get_master_location_hint({test_group, site3, node()}, infinity),
    
% TODO: remove when master waits until it can write durable txns
%%timer:sleep(1000),   
%%wait_until_all_sites_are_in_sync(MasterHint2, [site2, site3], 30000, ?LINE),
?AT, 

    %% hit replica and find new master
    {{ok_txn_fun, new_master_result}, MasterHint2} 
	= ?BDB_RG:do_txn_on_master({test_group, site3, node()},
                                   {?MODULE, test_txn_fun_put_1, {"key11", <<"committed11">>, {ok_txn_fun, new_master_result}}}),
?AT, 

    %% use new MasterLocationHint
    {ok_result1, MasterHint2} 
	= ?BDB_RG:do_txn_on_master(MasterHint2,
                                   {?MODULE, test_txn_fun_put_1, {<<"key12">>, "committed12", ok_result1}}),

?AT, 
    MasterHint2 = ?BDB_RG:get_master_location_hint({test_group, site2, node()}, 3000),
    {ok_result2, MasterHint2} 
	= ?BDB_RG:do_txn_on_master({test_group, site3, node()},
                                   {?MODULE, test_txn_fun_put_1, {"key12a", <<"committed12a">>, ok_result2}}),

?AT, 

    io:format("************** done key11/12~n"),    

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key8">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key8">>, <<>>, [])),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key9">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key9">>, <<>>, [])),

    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P2, DB2, void, <<"key10">>, <<>>, [])),
    ?EXPECT_BDB_ERROR(bdb_DB_NOTFOUND, ?BDB:db_get(P3, DB3, void, <<"key10">>, <<>>, [])),

    {<<"key11">>, <<"committed11">>} = ?BDB:db_get(P2, DB2, void, <<"key11">>, <<>>, []),
    {<<"key11">>, <<"committed11">>} = ?BDB:db_get(P3, DB3, void, <<"key11">>, <<>>, []),

    {<<"key12">>, <<"committed12">>} = ?BDB:db_get(P2, DB2, void, <<"key12">>, <<>>, []),
    {<<"key12">>, <<"committed12">>} = ?BDB:db_get(P3, DB3, void, <<"key12">>, <<>>, []),

    {<<"key12a">>, <<"committed12a">>} = ?BDB:db_get(P2, DB2, void, <<"key12a">>, <<>>, []),
    {<<"key12a">>, <<"committed12a">>} = ?BDB:db_get(P3, DB3, void, <<"key12a">>, <<>>, []),

    
    %% TODO:test start_sites and start with bad config for one site


    io:format("****~n******~n****** stopping group~n"),    
    ok = ?BDB_RG:stop_entire_group({test_group, site2, node()}),

    %% As stop_entire_group is currently asynchronous, we need to wait for some time for all processes to exit (or next test will fail)
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 15000, ?LINE),

%%     {messages, Msgs} = process_info(self(), messages),
%%     lists:foreach(fun(M) -> io:format("pending message = ~p~n", [M]), timer:sleep(500) end, Msgs),

    {messages, []} = process_info(self(), messages),
    ok.

test_txn_fun_put_1(BdbPort, {Key, Value, Return}) ->
    %% NOTE: page size of 64KB is to catch regression of bug in __repmgr_update_consumed which did not handle 
    %% partial receives correctly (on localhost MTU is 16KB so default page size did not hit this)
    DB = ?BDB_PC:get_db_handle_with_db_config(BdbPort, bdb_DB_BTREE, "test.db", [{page_size_bytes, 65536}]),
    ok = ?BDB:db_put(BdbPort, DB, void, Key, Value, []),
    Return.

test_txn_fun_crash_in_a_txn(BdbPort, []) ->
    DB  = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),
    ok	= ?BDB:db_put(BdbPort, DB, void, "key3", "committed3", []),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok	= ?BDB:db_put(BdbPort, DB, Txn, <<"key4">>, <<"not_committed4">>, []),
    crash_with_badmatch = ?BDB:db_put(BdbPort, DB, Txn, <<"key5">>, <<"not_committed5">>, []),
    ok = ?BDB:txn_commit(BdbPort, Txn, []).

test_txn_fun_exit_in_a_txn(BdbPort, []) ->
    DB  = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),
    ok	= ?BDB:db_put(BdbPort, DB, void, "key3a", "committed3a", []),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok	= ?BDB:db_put(BdbPort, DB, Txn, <<"key4a">>, <<"not_committed4a">>, []),
    ok	= ?BDB:db_put(BdbPort, DB, Txn, <<"key5a">>, <<"not_committed5a">>, []),
    exit({intentional_exit, here}),
    ok = ?BDB:txn_commit(BdbPort, Txn, []).

test_txn_fun_return_with_open_resources(BdbPort, []) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, "test.db"),

    Txn1    = ?BDB:txn_begin(BdbPort, void, []),
    Cursor1 = ?BDB:db_cursor(BdbPort, DB, Txn1, []),
    ok	    = ?BDB:c_put(BdbPort, Cursor1, <<"key6">>, <<"committed6">>, bdb_DB_KEYFIRST),
    ok	    = ?BDB:c_close(BdbPort, Cursor1),
    ok	    = ?BDB:txn_commit(BdbPort, Txn1, []),

    Txn2p   = ?BDB:txn_begin(BdbPort, void, []),
    Txn2c   = ?BDB:txn_begin(BdbPort, Txn2p, []),
    Cursor2 = ?BDB:db_cursor(BdbPort, DB, Txn2c, []),
    ok	    = ?BDB:c_put(BdbPort, Cursor2, <<"key7">>, <<"not_committed6">>, bdb_DB_KEYFIRST),
    %% return with 1 cursor and 2 open txns, hence '3' in return value
    {txn_fun_return_value, stuff}.

test_txn_fun_get(BdbPort, {DatabaseName, Key}) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DatabaseName),
    ?BDB:db_get(BdbPort, DB, void, Key, <<>>, []).
    

test_db_compact(StartFunc) ->	 
    io:format("\nPerforming test_db_compact\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    P = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(P, enable, [all]),

    %% Speed things up a lot
    ok = ?BDB:env_set_flags(P, [bdb_DB_TXN_NOSYNC], 1),
    ok = ?BDB:env_set_cachesize(P, 0, 50 * 1024 * 1024, 1),	 % #gigabytes, #bytes, #fragments

    ok = env_open_with_retry(P, Dir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),

    ok = ?BDB:set_lk_detect(P, bdb_DB_LOCK_DEFAULT),

    DB1 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_pagesize(P, DB1, 4096),
    ok = ?BDB:db_open(P, DB1, void, "not-compacted.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_pagesize(P, DB2, 4096),
    ok = ?BDB:db_open(P, DB2, void, "compacted-not-truncated.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB3 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_pagesize(P, DB3, 4096),
    ok = ?BDB:db_open(P, DB3, void, "compacted-truncated.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    DB4 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_pagesize(P, DB4, 4096),
    ok = ?BDB:db_open(P, DB4, void, "half-compacted-truncated.db", "", bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),

    Record = <<0:800/unit:8>>,   % 800 bytes => 4 records per 4k page

    io:format("filling databases~n"),
    PutLoop = fun(Final, N, _) when N > Final ->
                      ok;
                 (Final, N, PutLoop) ->
                      Key = list_to_binary(io_lib:fwrite("~5.5.0w", [N])),
                      %%io:format("inserting key '~p'~n", [binary_to_list(Key)]),
                      ok = ?BDB:db_put(P, DB1, void, Key, Record, []),
                      ok = ?BDB:db_put(P, DB2, void, Key, Record, []),
                      ok = ?BDB:db_put(P, DB3, void, Key, Record, []),
                      ok = ?BDB:db_put(P, DB4, void, Key, Record, []),
                      PutLoop(Final, N + 1, PutLoop)
              end,
    PutLoop(40, 1, PutLoop),  % 1000 leafe-pages worth

    %% force write to disk to create full size
    ok = ?BDB:txn_checkpoint(P, 0, 0, []),

    StatBtree1 = ?BDB:db_stat(P, DB1, void, bdb_DB_READ_COMMITTED),
    %%io:format("db_stat on DB1 ~p~n", [StatBtree1]),
    {value, {bt_nkeys,40}}          = lists:keysearch(bt_nkeys, 1, StatBtree1),
    {value, {bt_ndata,40}}          = lists:keysearch(bt_ndata, 1, StatBtree1),
    {value, {bt_levels,2}}          = lists:keysearch(bt_levels, 1, StatBtree1),
    {value, {bt_int_pg,1}}          = lists:keysearch(bt_int_pg, 1, StatBtree1),
    {value, {bt_leaf_pg,13}}        = lists:keysearch(bt_leaf_pg, 1, StatBtree1),
    {value, {bt_dup_pg,0}}          = lists:keysearch(bt_dup_pg, 1, StatBtree1),
    {value, {bt_over_pg,0}}         = lists:keysearch(bt_over_pg, 1, StatBtree1),
    {value, {bt_empty_pg,0}}        = lists:keysearch(bt_empty_pg, 1, StatBtree1),
    {value, {bt_free,0}}            = lists:keysearch(bt_free, 1, StatBtree1),
    {value, {bt_int_pgfree,3796}}   = lists:keysearch(bt_int_pgfree, 1, StatBtree1),
    {value, {bt_leaf_pgfree,20270}} = lists:keysearch(bt_leaf_pgfree, 1, StatBtree1),

    io:format("deleting alternate keys~n"),
    DeleteAlternateKeysLoop = fun(Final, N, _) when N > Final ->
                                      ok;
                                 (Final, N, DeleteAlternateKeysLoop) ->
                                      Key = list_to_binary(io_lib:fwrite("~5.5.0w", [N])),
                                      %%io:format("deleting key '~p'~n", [binary_to_list(Key)]),
                                      ok = ?BDB:db_del(P, DB1, void, Key, []),
                                      ok = ?BDB:db_del(P, DB2, void, Key, []),
                                      ok = ?BDB:db_del(P, DB3, void, Key, []),
                                      ok = ?BDB:db_del(P, DB4, void, Key, []),
                                      DeleteAlternateKeysLoop(Final, N + 2, DeleteAlternateKeysLoop)
                              end,
    DeleteAlternateKeysLoop(40, 2, DeleteAlternateKeysLoop),

    StatBtree1a = ?BDB:db_stat(P, DB1, void, bdb_DB_READ_COMMITTED),
    %%io:format("stat on DB1 again ~p~n", [StatBtree1a]),
    {value, {bt_nkeys,20}}          = lists:keysearch(bt_nkeys, 1, StatBtree1a),
    {value, {bt_ndata,20}}          = lists:keysearch(bt_ndata, 1, StatBtree1a),
    {value, {bt_levels,2}}          = lists:keysearch(bt_levels, 1, StatBtree1a),
    {value, {bt_int_pg,1}}          = lists:keysearch(bt_int_pg, 1, StatBtree1a),
    {value, {bt_leaf_pg,13}}        = lists:keysearch(bt_leaf_pg, 1, StatBtree1a),
    {value, {bt_dup_pg,0}}          = lists:keysearch(bt_dup_pg, 1, StatBtree1a),
    {value, {bt_over_pg,0}}         = lists:keysearch(bt_over_pg, 1, StatBtree1a),
    {value, {bt_empty_pg,0}}        = lists:keysearch(bt_empty_pg, 1, StatBtree1a),
    {value, {bt_free,0}}            = lists:keysearch(bt_free, 1, StatBtree1a),
    {value, {bt_int_pgfree,3796}}   = lists:keysearch(bt_int_pgfree, 1, StatBtree1a),
    {value, {bt_leaf_pgfree,36590}} = lists:keysearch(bt_leaf_pgfree, 1, StatBtree1a),


    %% First try compaction without truncation

    CompactResult2 = ?BDB:db_compact(P,
                                           DB2, 
                                           void, 
                                           "",           % optional start key
                                           "",           % optional end key ("" means end of database)
                                           0,            % compact_fillpercent (if not zero then amy page less full than this will be considered for compaction)
                                           0,            % implicit_txn_timeout  (if not zero and 'void' txnid then lock timeout in microseconds)
                                           []),          % flags
    %%io:format("compacted DB2 with flags = 0 (i.e. don't truncate) ~p~n", [CompactResult2]),
    {value, {compact_pages_free,8}}         = lists:keysearch(compact_pages_free, 1, CompactResult2),
    {value, {compact_pages_examine,17}}     = lists:keysearch(compact_pages_examine, 1, CompactResult2),
    {value, {compact_levels,0}}             = lists:keysearch(compact_levels, 1, CompactResult2),
    {value, {compact_deadlock,0}}           = lists:keysearch(compact_deadlock, 1, CompactResult2),
    {value, {compact_pages_truncated,0}}    = lists:keysearch(compact_pages_truncated, 1, CompactResult2),
    {value, {compact_end_key, <<"00033">>}} = lists:keysearch(compact_end_key, 1, CompactResult2),

    StatBtree2 = ?BDB:db_stat(P, DB2, void, bdb_DB_READ_COMMITTED),
    %%io:format("stat on DB2 ~p~n", [StatBtree2]),
    {value, {bt_nkeys,20}}         = lists:keysearch(bt_nkeys, 1, StatBtree2),
    {value, {bt_ndata,20}}         = lists:keysearch(bt_ndata, 1, StatBtree2),
    {value, {bt_levels,2}}         = lists:keysearch(bt_levels, 1, StatBtree2),
    {value, {bt_int_pg,1}}         = lists:keysearch(bt_int_pg, 1, StatBtree2),
    {value, {bt_leaf_pg,5}}        = lists:keysearch(bt_leaf_pg, 1, StatBtree2),
    {value, {bt_dup_pg,0}}         = lists:keysearch(bt_dup_pg, 1, StatBtree2),
    {value, {bt_over_pg,0}}        = lists:keysearch(bt_over_pg, 1, StatBtree2),
    {value, {bt_empty_pg,0}}       = lists:keysearch(bt_empty_pg, 1, StatBtree2),
    {value, {bt_free,7}}           = lists:keysearch(bt_free, 1, StatBtree2),
    {value, {bt_int_pgfree,3968}}  = lists:keysearch(bt_int_pgfree, 1, StatBtree2),
    {value, {bt_leaf_pgfree,4030}} = lists:keysearch(bt_leaf_pgfree, 1, StatBtree2),


    %% Now try truncating

    CompactResult3 = ?BDB:db_compact(P,
                                           DB3, 
                                           void,
                                           "", 
                                           "", 
                                           0,
                                           0,
                                           [bdb_DB_FREE_SPACE]),
    %%io:format("compacted DB3 with flags = bdb_DB_FREE_SPACE (i.e. truncate) ~p~n", [CompactResult3]),
    {value, {compact_pages_free,8}}         = lists:keysearch(compact_pages_free, 1, CompactResult3),
    {value, {compact_pages_examine,18}}     = lists:keysearch(compact_pages_examine, 1, CompactResult3),
    {value, {compact_levels,0}}             = lists:keysearch(compact_levels, 1, CompactResult3),
    {value, {compact_deadlock,0}}           = lists:keysearch(compact_deadlock, 1, CompactResult3),
    {value, {compact_pages_truncated,7}}    = lists:keysearch(compact_pages_truncated, 1, CompactResult3),
    {value, {compact_end_key, <<"00033">>}} = lists:keysearch(compact_end_key, 1, CompactResult3),

    ok = ?BDB:txn_checkpoint(P, 0, 0, []),

    StatBtree3 = ?BDB:db_stat(P, DB3, void, bdb_DB_READ_COMMITTED),
    %%io:format("stat on DB3 ~p~n", [StatBtree3]),
    {value, {bt_nkeys,20}}         = lists:keysearch(bt_nkeys, 1, StatBtree3),
    {value, {bt_ndata,20}}         = lists:keysearch(bt_ndata, 1, StatBtree3),
    {value, {bt_levels,2}}         = lists:keysearch(bt_levels, 1, StatBtree3),
    {value, {bt_int_pg,1}}         = lists:keysearch(bt_int_pg, 1, StatBtree3),
    {value, {bt_leaf_pg,5}}        = lists:keysearch(bt_leaf_pg, 1, StatBtree3),
    {value, {bt_dup_pg,0}}         = lists:keysearch(bt_dup_pg, 1, StatBtree3),
    {value, {bt_over_pg,0}}        = lists:keysearch(bt_over_pg, 1, StatBtree3),
    {value, {bt_empty_pg,0}}       = lists:keysearch(bt_empty_pg, 1, StatBtree3),
    {value, {bt_free,1}}           = lists:keysearch(bt_free, 1, StatBtree3),
    {value, {bt_int_pgfree,3968}}  = lists:keysearch(bt_int_pgfree, 1, StatBtree3),
    {value, {bt_leaf_pgfree,4030}} = lists:keysearch(bt_leaf_pgfree, 1, StatBtree3),


    %% Now try partial compact
    %% This does not truncate any space because the range happens to not include the last physical page in the file.
    %% A second full pass (of DB_FREELIST_ONLY) would truncate that space.

    StartKey = "00010",  % list_to_binary(io_lib:fwrite("~5.5.0w", [10])),
    StopKey  = "00030",  % list_to_binary(io_lib:fwrite("~5.5.0w", [30])),
    CompactResult4 = ?BDB:db_compact(P,
                                           DB4, 
                                           void,
                                           StartKey, 
                                           StopKey, 
                                           0,
                                           0,
                                           [bdb_DB_FREE_SPACE]),
    %%io:format("compacted half (~p,~p) DB4 with flags = bdb_DB_FREE_SPACE (i.e. truncate) ~p~n", [StartKey, StopKey, CompactResult4]),
    {value, {compact_pages_free,4}}         = lists:keysearch(compact_pages_free, 1, CompactResult4),
    {value, {compact_pages_examine,11}}     = lists:keysearch(compact_pages_examine, 1, CompactResult4),
    {value, {compact_levels,0}}             = lists:keysearch(compact_levels, 1, CompactResult4),
    {value, {compact_deadlock,0}}           = lists:keysearch(compact_deadlock, 1, CompactResult4),
    {value, {compact_pages_truncated,0}}    = lists:keysearch(compact_pages_truncated, 1, CompactResult4),
    {value, {compact_end_key, <<"00027">>}} = lists:keysearch(compact_end_key, 1, CompactResult4),

    StatBtree4 = ?BDB:db_stat(P, DB4, void, bdb_DB_READ_COMMITTED),
    %%io:format("stat on DB4 ~p~n", [StatBtree4]),
    {value, {bt_nkeys,20}}          = lists:keysearch(bt_nkeys, 1, StatBtree4),
    {value, {bt_ndata,20}}          = lists:keysearch(bt_ndata, 1, StatBtree4),
    {value, {bt_levels,2}}          = lists:keysearch(bt_levels, 1, StatBtree4),
    {value, {bt_int_pg,1}}          = lists:keysearch(bt_int_pg, 1, StatBtree4),
    {value, {bt_leaf_pg,9}}         = lists:keysearch(bt_leaf_pg, 1, StatBtree4),
    {value, {bt_dup_pg,0}}          = lists:keysearch(bt_dup_pg, 1, StatBtree4),
    {value, {bt_over_pg,0}}         = lists:keysearch(bt_over_pg, 1, StatBtree4),
    {value, {bt_empty_pg,0}}        = lists:keysearch(bt_empty_pg, 1, StatBtree4),
    {value, {bt_free,4}}            = lists:keysearch(bt_free, 1, StatBtree4),
    {value, {bt_int_pgfree,3884}}   = lists:keysearch(bt_int_pgfree, 1, StatBtree4),
    {value, {bt_leaf_pgfree,20310}} = lists:keysearch(bt_leaf_pgfree, 1, StatBtree4),


    %% check physical file sizes
    ok = ?BDB:txn_checkpoint(P, 0, 0, []),
    
    NotCompactedSize          = filelib:file_size(Dir ++ "/not-compacted.db"),
    CompactedNotTruncatedSize = filelib:file_size(Dir ++ "/compacted-not-truncated.db"),
    CompactedTruncatedSize    = filelib:file_size(Dir ++ "/compacted-truncated.db"),

    io:format("NotCompactedSize = ~w~nCompactedTruncatedSize = ~w~nCompactedNotTruncatedSize = ~w~n", [NotCompactedSize, CompactedTruncatedSize, CompactedNotTruncatedSize]),    

    io:format("TODO: file is truncated with flags = 0~n"),    
    %%%CompactedNotTruncatedSize = NotCompactedSize,

    if CompactedTruncatedSize == NotCompactedSize -> throw(failed_to_truncate);
       true -> ok
    end,


    %% On the database that we compacted but did not truncate, try the DB_FREELIST_ONLY mode
    %% (collate freelist and truncate, but don't compact)

    io:format("trying DB2 with FREELIST~n"),
    %% Check flags parameter by pasting this into berkeley_db_drv.c immediately before call to c->i.db->compact
    %%         fprintf(stderr, "%s: %s: threadid=%lx FLAGS %xd\n", c->desc->port_label, __FUNCTION__, pthread_self(), c->i.flags);
    CompactResult2a = ?BDB:db_compact(P,
                                            DB2, 
                                            void,
                                            "", 
                                            "", 
                                            0,
                                            0,
                                            [bdb_DB_FREELIST_ONLY]),
    %%io:format("compacted DB2 with flags = bdb_DB_FREELIST_ONLY (i.e. truncate existing free pages only) ~p~n", [CompactResult2a]),
    %% in FREELIST mode the results are all 0
    {value, {compact_pages_free,0}}                = lists:keysearch(compact_pages_free, 1, CompactResult2a),
    {value, {compact_pages_examine,0}}             = lists:keysearch(compact_pages_examine, 1, CompactResult2a),
    {value, {compact_levels,0}}                    = lists:keysearch(compact_levels, 1, CompactResult2a),
    {value, {compact_deadlock,0}}                  = lists:keysearch(compact_deadlock, 1, CompactResult2a),
    {value, {compact_pages_truncated,1}}           = lists:keysearch(compact_pages_truncated, 1, CompactResult2a),
    {value, {compact_end_key, <<>>}}               = lists:keysearch(compact_end_key, 1, CompactResult2a),

    ok = ?BDB:txn_checkpoint(P, 0, 0, []),

    StatBtree2a = ?BDB:db_stat(P, DB2, void, bdb_DB_READ_COMMITTED),
    %%io:format("stat on DB2 again ~p~n", [StatBtree2a]),
    {value, {bt_nkeys,20}}           = lists:keysearch(bt_nkeys, 1, StatBtree2a),
    {value, {bt_ndata,20}}           = lists:keysearch(bt_ndata, 1, StatBtree2a),
    {value, {bt_levels,2}}           = lists:keysearch(bt_levels, 1, StatBtree2a),
    {value, {bt_int_pg,1}}           = lists:keysearch(bt_int_pg, 1, StatBtree2a),
    {value, {bt_leaf_pg,5}}          = lists:keysearch(bt_leaf_pg, 1, StatBtree2a),
    {value, {bt_dup_pg,0}}           = lists:keysearch(bt_dup_pg, 1, StatBtree2a),
    {value, {bt_over_pg,0}}          = lists:keysearch(bt_over_pg, 1, StatBtree2a),
    {value, {bt_empty_pg,0}}         = lists:keysearch(bt_empty_pg, 1, StatBtree2a),
    {value, {bt_free,6}}             = lists:keysearch(bt_free, 1, StatBtree2a),     % should be 0
    {value, {bt_int_pgfree,3968}}    = lists:keysearch(bt_int_pgfree, 1, StatBtree2a),
    {value, {bt_leaf_pgfree,4030}}   = lists:keysearch(bt_leaf_pgfree, 1, StatBtree2a),


    %% Read entire databases and compare them (make sure we didn't lose any keys)
    %% TODO: use something other than 0's for record, so we can detect corrupt records.
    ReferenceKeysValues = get_all_keys_and_values(P, DB1),
    ReferenceKeysValues = get_all_keys_and_values(P, DB2),
    ReferenceKeysValues = get_all_keys_and_values(P, DB3),
    ReferenceKeysValues = get_all_keys_and_values(P, DB4),

    ok = ?BDB:db_close(P, DB1, []),
    ok = ?BDB:db_close(P, DB2, []),
    ok = ?BDB:db_close(P, DB3, []),
    ok = ?BDB:db_close(P, DB4, []),

    ok = ?BDB:env_close(P, []),
    normal = ?BDB:shutdown(P),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.



test_replicated_do_distributed_transaction(StartFunc) ->

    io:format("\nPerforming test_replicated_do_distributed_transaction\n"),

    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    GTMDir = "txn_tests_env",
    ensure_empty_dir(GTMDir),
    GTM_ReplicationGroupName = gtm_group,
    GTM_Site1                = {{site1, node()}, GTMDir, {"localhost", 4000}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                              {this_site_priority, 1}]}]}, 
    GTM_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           GTM_ReplicationGroupName, 
                           [GTM_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, GTM_ReplicationGroupName),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    _GTM_RgPid = ?BDB_RG:start(
                    gtm_group, 
                    GTM_Site1,
                    [], % no other sites, just for simplicity
                    [{common_port_coordinator_config, [{port_start_function, StartFunc}
                                                      ]}]),

    GTMProcName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(GTM_ReplicationGroupName, site1),
    {ok, GTMPid} = ?BDB_GTM:start_link(GTMProcName, ?BDB_RG:get_port_coordinator_proc_name(gtm_group, site1)),


    GroupA_Dir = "send_queue_env",
    ensure_empty_dir(GroupA_Dir),
    GroupA_Site1 = {{site1, node()}, GroupA_Dir, {"localhost", 4010}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},                                                                                           {this_site_priority, 1}]}]}, 
    _GroupA_RgPid = ?BDB_RG:start(groupA, 
                                  GroupA_Site1,
                                  [], % no other sites, just for simplicity
                                  [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                                                     {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                                                    ]}]),

    GroupB_Dir = "receive_queue_env",
    ensure_empty_dir(GroupB_Dir),    %% TODO: setting this to GroupA_Dir caused hangs in driver stop for this environment every 2nd run (after txn_recovery ran)
    GroupB_Site1 = {{site1, node()}, GroupB_Dir, {"localhost", 4020}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                           {this_site_priority, 1}]}]}, 
    _GroupB_RgPid = ?BDB_RG:start(groupB, 
                                  GroupB_Site1,
                                  [], % no other sites, just for simplicity
                                  [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                                                     {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]}
                                                                    ]}]),

    %% Wait for masters
    timer:sleep(200),
    _GroupA_MLH = ?BDB_RG:get_master_location_hint({groupA, site1, node()}, infinity),
    _GroupB_MLH = ?BDB_RG:get_master_location_hint({groupB, site1, node()}, infinity),

    %% Do replicated distributed txn
    GroupA_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           groupA, 
                           [GroupA_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, groupA),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    GroupB_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           groupB, 
                           [GroupB_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, groupB),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    %% We use different txn_funs just to show that that feature works
    {TxnGUID, PartitionResults} 
        = ?BDB_RG_H:do_distributed_txn_unordered(
             GTM_RgCallerState, 
             [{"GroupA", GroupA_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_group_A, {"test.db", "key1"}}},
              {"GroupB", GroupB_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_group_B, {"test.db", "key1"}}}],
             3,     % NumDeadlocRetries
             5000), % Timeout

    ?DB_XIDDATASIZE = size(TxnGUID),
    group_A_result      = proplists:get_value("GroupA", PartitionResults),
    {group_B_result, 2} = proplists:get_value("GroupB", PartitionResults),
    2 = length(PartitionResults),

    %% TODO:
    %% In R11B-2 if we cause an undef exit here we get a segv.  
    %% In R11B-1 we got a TYPE-ASSERTION from the VM
    GroupA_Site1_PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(groupA, site1),
    GroupB_Site1_PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(groupB, site1),

    {<<"key1">>, <<"group_A_value">>} 
        = ?BDB_H:do_txn(GroupA_Site1_PortCoordName,
                        {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}},
                        0),

    {<<"key1">>, <<"group_B_value">>} 
        = ?BDB_H:do_txn(GroupB_Site1_PortCoordName,
                        {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}},
                        0),


    %% Force some participants to fail

    try
       ?BDB_RG_H:do_distributed_txn_unordered(
          GTM_RgCallerState, 
          [{"GroupA", GroupA_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_group_A, {"test.db", "key2"}}},
           {"GroupB", GroupB_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_throw,   {"test.db", "key2"}}}],
          3,     % NumDeadlocRetries
          5000), % Timeout
        throw({should_have_thrown, ?AT})
    catch 
        {berkeley_db_distributed_txn_error, {some_participants_failed, 
                                             TxnGUID2, 
                                             [{"GroupB",{txn_fun_error,{intended_throw, payload}}}]}} when size(TxnGUID2) =:= ?DB_XIDDATASIZE ->
            ok
    end,

    try
       ?BDB_RG_H:do_distributed_txn_unordered(
          GTM_RgCallerState, 
          [{"GroupA", GroupA_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_throw,   {"test.db", "key2"}}},
           {"GroupB", GroupB_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_group_B, {"test.db", "key2"}}}],
          3,     % NumDeadlocRetries
          5000), % Timeout
        throw({should_have_thrown, ?AT})
    catch 
       {berkeley_db_distributed_txn_error, {some_participants_failed, 
                                            TxnGUID3, 
                                            [{"GroupA",{txn_fun_error,{intended_throw, payload}}}]}} when size(TxnGUID3) =:= ?DB_XIDDATASIZE ->
            ok
    end,

    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND},
                          ?BDB_H:do_txn(GroupA_Site1_PortCoordName,
                                        {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}},
                                        0)),
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND},
                          ?BDB_H:do_txn(GroupB_Site1_PortCoordName,
                                        {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}},
                                        0)),


    %% Shutdown
    MonitorRefGTM = erlang:monitor(process, GTMPid),
    GTMSeqProcName = list_to_atom(atom_to_list(GTMProcName) ++ "_sequence_server"),  % a private convention
    MonitorRefGTMSeq = erlang:monitor(process, whereis(GTMSeqProcName)),
    ok = ?BDB_GTM:stop(GTMProcName),    
    receive
        {'DOWN', MonitorRefGTMSeq, process, _, normal} -> ok
    after
        2000 ->  throw(timeout_waiting_for_server_exit_gtm_seq)
    end,
    receive
        {'DOWN', MonitorRefGTM, process, GTMPid, normal} -> ok
    after
        1000 ->  throw(timeout_waiting_for_server_exit_gtm)
    end,
    undefined = whereis(global_transaction_manager),
    undefined = whereis(global_transaction_manager_sequence_server),

    timer:sleep(1),
    io:format("************** stopping replication groups~n"),    
    ok = ?BDB_RG:stop_entire_group({GTM_ReplicationGroupName, site1, node()}),
    ok = ?BDB_RG:stop_entire_group({groupA, site1, node()}),
    ok = ?BDB_RG:stop_entire_group({groupB, site1, node()}),

    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    wait_until_all_sites_are_down(GTM_ReplicationGroupName, [site1], 15000, ?LINE),
    wait_until_all_sites_are_down(groupA, [site1], 15000, ?LINE),
    wait_until_all_sites_are_down(groupB, [site1], 15000, ?LINE),

    ets:delete(RgCallerStateTableId),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_distributed_txn_participant_group_A(BdbPort, {DistTxnInfo, {DatabaseName, Key}}) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DatabaseName),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok = ?BDB:db_put(BdbPort, DB, Txn, Key, "group_A_value", []),
    ?BDB_RG_H:prepare_and_commit_distributed_txn(BdbPort, Txn, group_A_result, DistTxnInfo).
    
txn_fun_distributed_txn_participant_group_B(BdbPort, {DistTxnInfo, {DatabaseName, Key}}) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DatabaseName),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok = ?BDB:db_put(BdbPort, DB, Txn, Key, "group_B_value", []),
    ?BDB_RG_H:prepare_and_commit_distributed_txn(BdbPort, Txn, {group_B_result, 2}, DistTxnInfo).

txn_fun_distributed_txn_participant_throw(BdbPort, {DistTxnInfo, {DatabaseName, Key}}) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DatabaseName),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok = ?BDB:db_put(BdbPort, DB, Txn, Key, "aborted_value", []),
    throw({intended_throw, payload}),
    ?BDB_RG_H:prepare_and_commit_distributed_txn(BdbPort, Txn, {group_B_result, 2}, DistTxnInfo).


test_replicated_do_distributed_transaction_with_failover(StartFunc) ->

    io:format("\nPerforming test_replicated_do_distributed_transaction_with_failover\n"),

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    GTMDir = "txn_tests_env",
    ensure_empty_dir(GTMDir),
    GTM_ReplicationGroupName = gtm_group,
    GTM_Site1                = {{site1, node()}, GTMDir, {"localhost", 4000}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                              {this_site_priority, 1}]}]}, 
    GTM_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           GTM_ReplicationGroupName, 
                           [GTM_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, GTM_ReplicationGroupName),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    _GTM_RgPid = ?BDB_RG:start(
                    gtm_group, 
                    GTM_Site1,
                    [], % no other sites, just for simplicity
                    [{common_port_coordinator_config, [{port_start_function, StartFunc}
                                                      ]}]),

    GTMProcName = ?BDB_GTM:get_proc_name_from_replication_group_site_name(GTM_ReplicationGroupName, site1),
    {ok, GTMPid} = ?BDB_GTM:start_link(GTMProcName, ?BDB_RG:get_port_coordinator_proc_name(gtm_group, site1)),


    GroupA_Site1_Dir = "send_queue_env",
    ensure_empty_dir(GroupA_Site1_Dir),
    GroupA_Site1 = {{site1, node()}, GroupA_Site1_Dir, {"localhost", 4010}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}}
                                                                                           ]}]}, 
    GroupA_Site2_Dir = "send_queue_env2",
    ensure_empty_dir(GroupA_Site2_Dir),
    GroupA_Site2 = {{site2, node()}, GroupA_Site2_Dir, {"localhost", 4012}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_CLIENT, 2}}
                                                                                           ]}]}, 
    GroupA_Site3_Dir = "send_queue_env3",
    ensure_empty_dir(GroupA_Site3_Dir),
    GroupA_Site3 = {{site3, node()}, GroupA_Site3_Dir, {"localhost", 4014}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_CLIENT, 2}}
                                                                                           ]}]}, 
    {_GroupAPids, []} = ?BDB_RG:start_sites(groupA, 
                                            [GroupA_Site1, 
                                             GroupA_Site2,
                                             GroupA_Site3],
                                            [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                                                               %%{port_log_directory, "."},
                                                                               %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                                                               {env_open_receive_reply_timeout_milliseconds,     30000},
                                                                               {default_receive_reply_timeout_milliseconds,      30000},
                                                                               {deadlock_detector_milliseconds_and_policy,       {30000, 30000, bdb_DB_LOCK_DEFAULT}},
                                                                               {checkpoint_receive_reply_timeout_milliseconds,   30000},
                                                                               {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]},
                                                                               {replication, 
                                                                                [%%{config_flags, [bdb_DB_REP_CONF_NOWAIT]},
                                                                                 {ack_policy,	bdb_DB_REPMGR_ACKS_QUORUM},
                                                                                 {this_site_priority, 1},
                                                                                 {allow_not_durably_replicated_db_open, disabled}
                                                                                ]}
                                                                              ]}]),

    GroupB_Dir = "receive_queue_env",
    ensure_empty_dir(GroupB_Dir),
    GroupB_Site1 = {{site1, node()}, GroupB_Dir, {"localhost", 4020}, [{replication, [{initial_role_and_threadcount, {bdb_DB_REP_MASTER, 2}},
                                                                                      {this_site_priority, 1}]}]}, 
    _GroupB_RgPid = ?BDB_RG:start(groupB, 
                                  GroupB_Site1,
                                  [], % no other sites, just for simplicity
                                  [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                                                     %%{port_log_directory, "."},
                                                                     %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                                                     {env_open_receive_reply_timeout_milliseconds,     30000},
                                                                     {default_receive_reply_timeout_milliseconds,      30000},
                                                                     {deadlock_detector_milliseconds_and_policy,       {30000, 30000, bdb_DB_LOCK_DEFAULT}},
                                                                     {checkpoint_receive_reply_timeout_milliseconds,   30000},
                                                                     {distributed_transactions, [{gtm_rg_caller_state, GTM_RgCallerState}]},
                                                                     {replication, 
                                                                      [%%{config_flags, [bdb_DB_REP_CONF_NOWAIT]}
                                                                      ]}
                                                                    ]}]),

    %% Do replicated distributed txn
    GroupA_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           groupA, 
                           [GroupA_Site1, 
                            GroupA_Site2, 
                            GroupA_Site3], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, groupA),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    GroupB_RgCallerState = ?BDB_RG_H:initialize_caller_state(
                           groupB, 
                           [GroupB_Site1], 
                           5000, % IgnoreDownSitesForMilliseconds,
                           fun() -> 
                                   [V] = ets:lookup(RgCallerStateTableId, groupB),
                                   V
                           end,
                           fun(RgCallerState) -> ets:insert(RgCallerStateTableId, RgCallerState) end),

    %% Wait for masters
    _GTM_MLH   = ?BDB_RG_H:refresh_master_location_hint(GTM_RgCallerState, infinity),

    GroupA_RgCallerState2 = ?BDB_RG_H:refresh_master_location_hint(GroupA_RgCallerState, infinity),
    GroupA_MLH2           = ?BDB_RG_H:get_MLH_from_RgCallerState(GroupA_RgCallerState2),
    {groupA, site1, _}    = ?BDB_RG:get_site_locator_from_master_location_hint(GroupA_MLH2),

    _GroupB_RgCallerState2 = ?BDB_RG_H:refresh_master_location_hint(GroupB_RgCallerState, infinity),

    {TxnGUID, ParticipantResults} = 
        ?BDB_RG_H:do_distributed_txn_unordered(
           GTM_RgCallerState, 
           [{"GroupA", GroupA_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_kill_master_before_commit, {"test.db", "key_A"}}},
            {"GroupB", GroupB_RgCallerState, {?MODULE, txn_fun_distributed_txn_participant_group_B,                   {"test.db", "key_B"}}}],
           3,     % NumDeadlocRetries
           5000), % Timeout
?AT,
    [{"GroupA", group_A_result},   % even though the original master didn't commit, it _did_ prepare, so we do get a result, not an exception
     {"GroupB", {group_B_result, 2}}] = lists:sort(ParticipantResults),
?AT,
    %% Show that failover occurred
    wait_until_all_sites_are_down(groupA, [site1], 15000, ?LINE),
?AT,
    _GroupA_RgCallerState3 = ?BDB_RG_H:refresh_master_location_hint(GroupA_RgCallerState2, infinity),
?AT,
    %% Show that the txn was committed despite the failover after txn_prepare on GroupA
    %% (because the failover occurred _after_ GroupA told the coordinator that it had prepared)
    {TxnGUID, committed} = ?BDB_GTM:resolve(GTMPid, TxnGUID),

?AT,
    {<<"key_B">>, <<"group_B_value">>} =
        ?BDB_RG:do_txn_on_site({groupB, site1, node()},
                               {?MODULE, test_txn_fun_get, {"test.db", <<"key_B">>}},
                               0,
                               20000),
?AT,
    {<<"key_A">>, <<"killed_master_value">>} =
        ?BDB_RG:do_txn_on_site({groupA, site2, node()},
                               {?MODULE, test_txn_fun_get, {"test.db", <<"key_A">>}},
                               0,
                               20000),
?AT,
    {<<"key_A">>, <<"killed_master_value">>} =
        ?BDB_RG:do_txn_on_site({groupA, site3, node()},
                               {?MODULE, test_txn_fun_get, {"test.db", <<"key_A">>}},
                               0,
                               20000),
?AT,

    %% Shutdown
    MonitorRefGTM = erlang:monitor(process, GTMPid),
    GTMSeqProcName = list_to_atom(atom_to_list(GTMProcName) ++ "_sequence_server"),  % a private convention
    MonitorRefGTMSeq = erlang:monitor(process, whereis(GTMSeqProcName)),
    ok = ?BDB_GTM:stop(GTMProcName),    
    receive
        {'DOWN', MonitorRefGTMSeq, process, _, normal} -> ok
    after
        2000 ->  throw(timeout_waiting_for_server_exit_gtm_seq)
    end,
    receive
        {'DOWN', MonitorRefGTM, process, GTMPid, normal} -> ok
    after
        1000 ->  throw(timeout_waiting_for_server_exit_gtm)
    end,
    undefined = whereis(global_transaction_manager),
    undefined = whereis(global_transaction_manager_sequence_server),

    io:format("************** stopping replication groups~n"),    
    ok = ?BDB_RG:stop_entire_group({GTM_ReplicationGroupName, site1, node()}),
    ok = ?BDB_RG:stop_entire_group({groupA, site2, node()}),
    ok = ?BDB_RG:stop_entire_group({groupB, site1, node()}),

    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    wait_until_all_sites_are_down(GTM_ReplicationGroupName, [site1], 15000, ?LINE),
    wait_until_all_sites_are_down(groupA, [site1, site2], 15000, ?LINE),
    wait_until_all_sites_are_down(groupB, [site1], 15000, ?LINE),

    ets:delete(RgCallerStateTableId),

    timer:sleep(100),

    %% The local participant worker for groupA spawned by do_distributed_txn 
    %% will send this message to the process that called do_distributed_txn (the coordinator)
    %% if do_txn_on_master hits a problem.  When the txn_fun calls ?BDB_RG:stop() for groupA 
    %% then the port-coordinator is shutdown, which kills all registered clients.
    %% So do_txn_no_throw gets an exit signal
    receive 
        {"GroupA", do_txn_on_master_failed, TxnGUID, {bdb_replication_group_error, no_master}} ->
            ok
    after 
        0 -> ok
    end,     
?AT,
    %% There *might* also be this message:
    receive
        {"GroupA",prepared,TxnGUID,_WorkerPid,group_A_result} ->
            ok
    after 
        0 -> 
            ok
    end,
?AT,
    %% There *might* also be this message:
    receive
            {"GroupA",do_txn_on_master_failed,TxnGUID,{bdb_replication_group_error,{do_txn_on_master,timeout,groupA}}} ->
            ok
    after 
        0 -> 
            ok
    end,
?AT,

    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_distributed_txn_participant_kill_master_before_commit(BdbPort, {DistTxnInfo, {DatabaseName, Key}}) ->
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DatabaseName),
    Txn = ?BDB:txn_begin(BdbPort, void, []),
    ok = ?BDB:db_put(BdbPort, DB, Txn, Key, "killed_master_value", []),
    prepare_and_kill_master_before_commit(BdbPort, Txn, group_A_result, DistTxnInfo).

%% This is a hacked minimal version of ?BDB_RG_H:prepare_and_commit_distributed_txn
prepare_and_kill_master_before_commit(#bdb_port_handle{port_coordinator_name = PortCoordName,
                                                       group_valmap_id       = GroupValmapId} = BdbPort, 
                                      Txn, TxnFunResult, {ParticipantTag, TxnGUID, DistTxnCoordPid, Timeout}) ->

    ok = ?BDB:txn_prepare(BdbPort, Txn, TxnGUID),

    %% CRITICAL: we CANNOT now unilaterally abort or commit.
    %% We have to get an explicit instruction.

    %% Tell the coordinator that we are waiting for instructions.
    DistTxnCoordPid ! {ParticipantTag, prepared, <<_TxnCounter:64, 0:((128*8)-64)>> = TxnGUID, self(), TxnFunResult},

    receive
        %% These are asynchronous messages -- there's no 
        %% need to reply (although when we return the 
        %% result will be propagated back as the TxnFunResult
        {ParticipantTag, commit, TxnGUID} -> 
            io:format("prepare_and_kill_master: (group valmap id ~w) told to commit ~w, so killing port coord ~w~n", [GroupValmapId, _TxnCounter, port_coordinator_name]),

            %% Instead of calling txn_commit and then returning 'prepare_and_commit_finished'
            %% we force a failover and then sleep (we are a client of the master port-coord, so we will be killed in it's terminate function)
            ok = ?BDB_PC:stop(PortCoordName),

            timer:sleep(20000),
            throw(prepare_and_kill_master_was_not_killed);

        Other ->
            throw({prepare_and_kill_master, unexpected_msg, Other})
    after
        Timeout ->
            throw({berkeley_db_distributed_txn_error,
                   {timeout_waiting_for_decision, TxnGUID, DistTxnCoordPid, Timeout}})
    end.


test_replication_do_txn_redirect(StartFunc) ->

    io:format("\nPerforming test_replication_do_txn_redirect\n"),

    %% This test is timing sensitive, so we don't want accumulated cruft from any previous tests to interfere
    garbage_collect(),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    RgAllSites
        = [{{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 3}]}]}, % first master
           {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 2}]}]}, % second master
           {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
          ], 

    %% The "replication" options per site are entirely optional.  Here we use it to set a preference for which site is master, as it simplifies the test.
    {_GoodSitesPids, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            RgAllSites,
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_ELECTION, 1}}, 
                                                              {ack_policy,		      bdb_DB_REPMGR_ACKS_ALL_PEERS}, 
                                                              {limit_transfer_per_call_bytes, 100000}
                                                             ]}]},
             %% disable respawing
             {spawn_other_sites, disabled}]),

    io:format("************** Waiting for a master to be elected~n"),    
    #master_location_hint{site_locator = {test_group, _MasterSiteName1, _}} = MasterHint1
        = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 20000),

    io:format("************** Got master #1 ~p~n", [_MasterSiteName1]),    

    %% We don't care who is master, but we do need 2 in-sync replicas
    wait_until_all_sites_are_in_sync(MasterHint1, [site1, site2, site3], 30000, ?LINE),

    RgCallerStateTableId = ets:new(rg_caller_state, [public, {keypos, #?CS.replication_group_name}]),

    IgnoreDownSitesForMilliseconds = 6000,

    TestDriver = self(),
    RgCallerState1 = ?BDB_RG_H:initialize_caller_state(
                       test_group, 
                       RgAllSites, 
                       IgnoreDownSitesForMilliseconds,
                       fun() -> 
                               [V] = ets:lookup(RgCallerStateTableId, test_group),
                               V
                       end,
                       fun(RgCallerState) -> 
                               io:format("RgCallerState being written to persistent store: ~p~n", [RgCallerState]),
                               TestDriver ! {new_rg_caller_state, RgCallerState},
                               ets:insert(RgCallerStateTableId, RgCallerState) 
                       end),

    %% The first write should send itself
    receive {new_rg_caller_state, RgCallerState1} -> ok end,
    ?AT,
    [{test_group, site1, _}, {test_group, site2, _}, {test_group, site3, _}] = lists:sort(RgCallerState1#?CS.up_site_locators),
    ?AT,
    [] = lists:sort(RgCallerState1#?CS.down_sites),

    %% do_txn_on_any_site uses random:uniform().
    %% We want predictable test results, so seed the random number generator with a constant
    %% (If we use {0, 0, 0} as the seed then it always returns the same number!!!)
    %%
    %% I was never able to get perfectly deterministic test results, 
    %% but at least this gives a good chance of being able to repro.
    random:seed(0, 0, 1),

    %% Should use all 3 sites approximately equally
    ?AT,
    [site1, site2, site3] = lists:usort(
                              lists:map(fun(_) -> 
                                                {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState1, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                                Result
                                        end,
                                        lists:seq(1, 30))),

    %% the next site is supposed to be site3
    Site3RgName = ?BDB_RG:get_proc_name(test_group, site3),
    exit(whereis(Site3RgName), kill),
    wait_until_all_sites_are_down(test_group, [site3], 30000, ?LINE),
    
    %% site2 is chosen instead
    [site1, site2] = lists:usort(
                       lists:map(fun(_) -> 
                                         {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState1, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                         Result
                                 end,
                                 lists:seq(1, 30))),

    %% RgCallerState should have changed (and the write function will send it to us)
    RgCallerState2 = receive {new_rg_caller_state, R2} -> R2 after 0 -> throw({no_caller_state_2, ?AT}) end,
    ?AT,
    [{test_group, site1, _}, {test_group, site2, _}] = lists:sort(RgCallerState2#?CS.up_site_locators),
    ?AT,
    [{{test_group, site3, _}, _}] = lists:sort(RgCallerState2#?CS.down_sites),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    Site2RgName = ?BDB_RG:get_proc_name(test_group, site2),
    exit(whereis(Site2RgName), kill),
    wait_until_all_sites_are_down(test_group, [site2], 30000, ?LINE),

    [site1] = lists:usort(
                lists:map(fun(_) -> 
                                  {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState2, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                  Result
                          end,
                          lists:seq(1, 30))),
    
    %% RgCallerState should have changed (and the write function will send it to us)
    RgCallerState3 = receive {new_rg_caller_state, R3} -> R3 after 0 -> throw({no_caller_state_3, ?AT}) end,
    ?AT,
    [{test_group, site1, _}] = lists:sort(RgCallerState3#?CS.up_site_locators),
    ?AT,
    [{{test_group, site2, _}, _}, {{test_group, site3, _}, _}] = lists:sort(RgCallerState3#?CS.down_sites),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    ?AT,
    [site1] = lists:usort(
                lists:map(fun(_) -> 
                                  {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState3, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                  Result
                          end,
                          lists:seq(1, 20))),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    %% Trying OLD caller_state should NOT write a new persistent copy, as 
    ?AT,
    [site1] = lists:usort(
                lists:map(fun(_) -> 
                                  {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState1, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                  Result
                          end,
                          lists:seq(1, 20))),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    ?AT,
    [site1] = lists:usort(
                lists:map(fun(_) -> 
                                  {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState2, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                  Result
                          end,
                          lists:seq(1, 20))),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    %% Kill the last remaining site
    ?AT,
    Site1RgName = ?BDB_RG:get_proc_name(test_group, site1),
    exit(whereis(Site1RgName), kill),
    ?AT,
    wait_until_all_sites_are_down(test_group, [site1], 30000, ?LINE),

    ?AT,
    %%user_default:dbgon(?BDB_RG_H),

    ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState3, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),

    %% RgCallerState should have changed (and the write function will send it to us)
    ?AT,
    RgCallerState4 = receive {new_rg_caller_state, R4} -> R4 after 0 -> throw({no_caller_state_4, ?AT}) end,
    ?AT,
    [] = lists:sort(RgCallerState4#?CS.up_site_locators),
    ?AT,
    [{{test_group, site1, _}, _}, {{test_group, site2, _}, _}, {{test_group, site3, _}, _}] = lists:sort(RgCallerState4#?CS.down_sites),
    ?AT,
    lists:foreach(fun(_) ->
                          ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState4, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000))
                  end,
                  lists:seq(1, 20)),

    %% Again, using old caller_states should not write a new version
    %% (it will try sites that are down, but won't write a new version as it will find a newer version has already been persisted)
    ?AT,
    lists:foreach(fun(_) ->
                          ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState1, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),
                          ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState2, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),
                          ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState3, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000))
                          %%?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState4, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000))
                  end,
                  lists:seq(1, 4)),
    ?AT,
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 30000, ?LINE),

    %% Resurrect the sites, and verify that we find them quickly
    io:format("************** Resurrecting sites again~n"),    
    {_GoodSitesPids2, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            RgAllSites,
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_ELECTION, 1}}, 
                                                              {ack_policy,		      bdb_DB_REPMGR_ACKS_ALL_PEERS}, 
                                                              {limit_transfer_per_call_bytes, 100000}
                                                             ]}]},
             %% disable respawing
             {spawn_other_sites, disabled}]),

    io:format("************** Waiting for a master to be elected~n"),    
    #master_location_hint{site_locator = {test_group, _MasterSiteName2, _}} = MasterHint2
        = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 20000),

    io:format("************** Got master #2 ~p~n", [_MasterSiteName2]),    
    {ok, MasterHint2} = ?BDB_RG:do_txn_on_master(MasterHint2,
                                                {?MODULE, txn_fun_put, {"test.db", "key2", "value2"}}),

    wait_until_all_sites_are_in_sync(MasterHint2, [site1, site2, site3], 30000, ?LINE),

    %% Wait for IgnoreDownSitesForMilliseconds to pass, so all sites should be resurrected on next use
    io:format("Sleeping for ~w seconds to allow IgnoreDownSitesForMilliseconds to expire~n", [IgnoreDownSitesForMilliseconds / 1000]),
    timer:sleep(IgnoreDownSitesForMilliseconds),
    ?AT,
    [site1, site2, site3] = lists:usort(
                              lists:map(fun(_) -> 
                                                {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState4, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                                Result
                                        end,
                                        lists:seq(1, 20))),
    ?AT,
    %% We should have a new rg_caller_state with all sites resurrected
    RgCallerState5 = receive {new_rg_caller_state, R5} -> R5 after 0 -> throw({no_caller_state_5, ?AT}) end,
    ?AT,
    [{test_group, site1, _}, {test_group, site2, _}, {test_group, site3, _}] = lists:sort(RgCallerState5#?CS.up_site_locators),
    ?AT,
    [] = lists:sort(RgCallerState5#?CS.down_sites),

    ?AT,
    [site1, site2, site3] = lists:usort(
                              lists:map(fun(_) -> 
                                                {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState5, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                                Result
                                        end,
                                        lists:seq(1, 20))),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    %% Test probing of believed-to-be-down sites, and refreshing of caller_state from real replication_group status

    %% Kill all the sites again
    exit(whereis(Site1RgName), kill),
    exit(whereis(Site2RgName), kill),
    exit(whereis(Site3RgName), kill),
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 30000, ?LINE),

    %% Get a NEW RgCallerState that has a very high IgnoreDownSitesForMilliseconds, to prevent timer-based resurrection
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,
    RgCallerState6 = ?BDB_RG_H:initialize_caller_state(
                       test_group, 
                       RgAllSites, 
                       500000, % IgnoreDownSitesForMilliseconds,
                       fun() -> 
                               [V] = ets:lookup(RgCallerStateTableId, test_group),
                               V
                       end,
                       fun(RgCallerState) -> 
                               io:format("Second RgCallerState being written to persistent store: ~p~n", [RgCallerState]),
                               TestDriver ! {new_rg_caller_state, RgCallerState},
                               ets:insert(RgCallerStateTableId, RgCallerState) 
                       end),

    %% The first write should send itself
    receive {new_rg_caller_state, RgCallerState6} -> ok end,
    %% It assumes that all sites are up, even though none are
    ?AT,
    [{test_group, site1, _}, {test_group, site2, _}, {test_group, site3, _}] = lists:sort(RgCallerState6#?CS.up_site_locators),
    ?AT,
    [] = lists:sort(RgCallerState6#?CS.down_sites),

    %% This should find all sites are down. It will modify caller_state 3 times (once for each site)
    ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState6, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),
    RgCallerState7 = receive {new_rg_caller_state,  R7}  -> R7 after 0 -> throw({no_caller_state_7, ?AT}) end,
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState7, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),
    RgCallerState8 = receive {new_rg_caller_state,  R8}  -> R8 after 0 -> throw({no_caller_state_7, ?AT}) end,
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState8, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000)),
    RgCallerState9 = receive {new_rg_caller_state,  R9}  -> R9 after 0 -> throw({no_caller_state_7, ?AT}) end,
    ?AT,
    [] = lists:sort(RgCallerState9#?CS.up_site_locators),
    ?AT,
    [{{test_group, site1, _}, _}, {{test_group, site2, _}, _}, {{test_group, site3, _}, _}] = lists:sort(RgCallerState9#?CS.down_sites),
    ?AT,
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    lists:foreach(fun(_) ->
                          ?EXPECT_RG_NO_SITES_ERROR(?BDB_RG_H:do_txn_on_any_site(RgCallerState9, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000))
                  end,
                  lists:seq(1, 20)),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,

    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 30000, ?LINE),

    %% Now restart the sites and ensure that we find them all immediately with RgCallerState7
    io:format("************** Resurrecting sites again~n"),    
    {_GoodSitesPids3, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            RgAllSites,
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_ELECTION, 1}}, 
                                                              {ack_policy,		      bdb_DB_REPMGR_ACKS_ALL_PEERS}, 
                                                              {limit_transfer_per_call_bytes, 100000}
                                                             ]}]},
             %% disable respawing
             {spawn_other_sites, disabled}]),

    io:format("************** Waiting for 3rd master to be elected~n"),    
    #master_location_hint{site_locator = {test_group, _MasterSiteName3, _}} = MasterHint3
        = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 20000),

    io:format("************** Got 3rd master ~p~n", [_MasterSiteName3]),    
    {ok, MasterHint3} = ?BDB_RG:do_txn_on_master(MasterHint3,
                                                {?MODULE, txn_fun_put, {"test.db", "key3", "value3"}}),

    wait_until_all_sites_are_in_sync(MasterHint3, [site1, site2, site3], 30000, ?LINE),

    ?AT,
    %% This will ask for real group status and will force a persistent write via update_caller_state_from_group_status()
    ?BDB_RG_H:do_txn_on_any_site(RgCallerState9, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
    ?AT,
    RgCallerState10 = receive {new_rg_caller_state,  R10}  -> R10 after 0 -> throw({no_caller_state_7, ?AT}) end,
    ?AT,
    [{test_group, site1, _}, {test_group, site2, _}, {test_group, site3, _}] = lists:sort(RgCallerState10#?CS.up_site_locators),
    ?AT,
    [] = lists:sort(RgCallerState10#?CS.down_sites),

    ?AT,
    [site1, site2, site3] = lists:usort(
                              lists:map(fun(_) -> 
                                                {Result, _} = ?BDB_RG_H:do_txn_on_any_site(RgCallerState10, {?MODULE, txn_fun_report_which_site_executes, {}}, 3, 1000),
                                                Result
                                        end,
                                        lists:seq(1, 30))),
    receive {new_rg_caller_state, _} -> throw({unexpected_persist_caller_state, ?AT}) after 0 -> ok end,


    io:format("********** stopping group~n"),    
    ok = ?BDB_RG:stop_entire_group({test_group, site1, node()}),

    %% As stop_entire_group is currently asynchronous, we need to wait for some time for all processes to exit (or next test will fail)
    timer:sleep( case StartFunc of 
                     start      -> 2000; 
                     start_pipe -> 4000 
                 end ),

    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 30000, ?LINE),

    ets:delete(RgCallerStateTableId),

    timer:sleep(200),
%%     {messages, Msgs} = process_info(self(), messages),
%%     lists:foreach(fun(M) -> io:format("pending message = ~p~n", [M]), timer:sleep(500) end, Msgs),
    {messages, []} = process_info(self(), messages),
    ok.


txn_fun_put(BdbPort, {DbId, Key, Value}) ->
    %%DB = ?BDB_PC:get_db_handle_with_db_config(BdbPort, bdb_DB_BTREE, DbId, [{page_size_bytes, 65536}]),
    DB = ?BDB_PC:get_db_handle(BdbPort, bdb_DB_BTREE, DbId),
    ok = ?BDB:db_put(BdbPort, DB, void, Key, Value, []).


txn_fun_report_which_site_executes(_BdbPort, _) ->
    {_ReplicationGroupName, SiteName} = get(txn_fun_replication_context),
    SiteName.


test_replication_perm_failed_events_limit(StartFunc) ->

    io:format("\nPerforming test_replication_perm_failed_events_limit\n"),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    MaxCount = 3, 
    WindowMs = 10000,

    {_GoodSitesPids, []} 
        = ?BDB_RG:start_sites(
            test_group, 
            [{{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 3}]}]}, % first master
             {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 2}]}]}, % failover master
             {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
            ], 
            %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
            %% (those are all set internally)
            [{log_removal_period_seconds, disabled},
             {perm_failed_events_limits,  {{waiting_for_durability, {MaxCount, WindowMs}},
                                           {available,              {MaxCount, WindowMs}}}},
             {common_port_coordinator_config, [{port_start_function, StartFunc},
                                               {checkpoint_seconds_and_min_kb,                disabled},  % we will do manual checkpoints
                                               {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_ELECTION, 2}}, 
                                                              {ack_policy,		       bdb_DB_REPMGR_ACKS_QUORUM}
                                                             ]}
                                              ]}
            ]),

    MLH = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, infinity),

    wait_until_all_sites_are_in_sync(MLH, [site1, site2, site3], 10000, ?LINE),

    %% Do more than MaxCount writes in WindowMs (should be fine as we have enough replicas)
    lists:foreach(fun(N) ->
                          ?BDB_RG:do_txn_on_master(MLH,
                                                   {?MODULE, test_txn_fun_put_1, {"keyA_" ++ integer_to_list(N), "value", ok}})
                  end,
                  lists:seq(1, MaxCount * 2)),

    %% Kill 1 replica and Do more than MaxCount writes in WindowMs
    ?AT,
    %%user_default:dbgon(berkeley_db_test),
    [{StopReplica1_SiteName, _}, {StopReplica2_SiteName, _}] = list_replicas({test_group, site1, node()}),

    ?AT,
    exit(whereis(?BDB_RG:get_proc_name(test_group, StopReplica1_SiteName)), shut_down1),

    ?AT,
    wait_until_all_sites_are_down(test_group, [StopReplica1_SiteName], 10000, ?LINE),

    %% TODO: sometimes the next call to do_txn_on_master gets NOTPERM
    %% -- perhaps due to the remaining replica doing a checkpoint?

    lists:foreach(fun(N) ->
                          ?AT,
                          ?BDB_RG:do_txn_on_master(MLH,
                                                   {?MODULE, test_txn_fun_put_1, {"keyA_" ++ integer_to_list(N), "value", ok}})
                  end,
                  lists:seq(1, MaxCount * 2)),
    ?AT,

    %% Kill the remaining replica then wait for all events to pass out of WindowMs
    %% Then do exactly MaxCount writes in WindowMs
    exit(whereis(?BDB_RG:get_proc_name(test_group, StopReplica2_SiteName)), shut_down2),

    ?AT,
    wait_until_all_sites_are_down(test_group, [StopReplica1_SiteName, StopReplica2_SiteName], 10000, ?LINE),

    ?AT,
    timer:sleep(WindowMs + 500),  % wait for a fresh window
    lists:foreach(fun(N) ->
                          ?AT,
                          ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_REP_NOTPERM},
                                                ?BDB_RG:do_txn_on_master(MLH,
                                                                         {?MODULE, test_txn_fun_put_1, {"keyA_" ++ integer_to_list(N), "value", ok}})),
                          timer:sleep(WindowMs div (MaxCount + 1))
                  end,
                  lists:seq(1, MaxCount)),
    ?AT,

    %% Wait for one of the events to pass out of WindowMs and do another
    timer:sleep((WindowMs div MaxCount) + 500),
    ?AT,
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_REP_NOTPERM},
                          ?BDB_RG:do_txn_on_master(MLH,
                                                   {?MODULE, test_txn_fun_put_1, {"keyD", "value", ok}})),
    ?AT,

    %% Now immediately do another (break the limit) and observe the master shutdown
    MasterPid = MLH#master_location_hint.pid,
    MonitorRef = erlang:monitor(process, MasterPid),

    receive 
        {'DOWN', MonitorRef, process, MasterPid, _} -> 
            throw({master_exited_unpectedly, ?AT})
    after 
        200 -> ok
    end,

    ?AT,
    ?EXPECT_RG_ERROR({site_is_down, _MasterSiteLocator, {{perm_failed_events_rate_too_high, _}, _}},
                     ?BDB_RG:do_txn_on_master(MLH,
                                              {?MODULE, test_txn_fun_put_1, {"keyE", "value", ok}})),
    ?AT,

    receive 
        {'DOWN', MonitorRef, process, MasterPid, {perm_failed_events_rate_too_high, _}} -> 
            ok
    after
        1000 ->
            throw({master_should_have_been_shutdown, ?AT})
    end,

    %% We don't need to call stop_entire_group as all sites are down

    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 15000, ?LINE),

    {messages, []} = process_info(self(), messages),

    ok.


test_replication_waiting_for_durability(StartFunc) ->

    io:format("\nPerforming test_replication_waiting_for_durability\n"),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    ?BDB_RG:start_sites(
       test_group, 
       [{{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 2}]}]}, 
        {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 1}]}]},
        {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]}
       ],
       %% Common port_coord config should not include any of 'this_hostname_port', 'peer_hostnames_ports', 'event_handlers'  
       %% (those are all set internally)
       [{log_removal_period_seconds,     disabled},
        {common_port_coordinator_config, [{port_start_function, StartFunc},
                                          {default_receive_reply_timeout_milliseconds, 4 * ?BDB_DEFAULT_RECEIVE_REPLY_TIMEOUT},  % verbose replication logging slows things down
                                          %%{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                          {replication, [{initial_role_and_threadcount, {bdb_DB_REP_ELECTION, 2}}, 
                                                         {ack_policy,		        bdb_DB_REPMGR_ACKS_QUORUM},
                                                         {config_flags,                 [bdb_DB_REP_CONF_DELAYCLIENT]}, % we want control over when sync happens
                                                         %% the test needs this:
                                                         {allow_not_durably_replicated_db_open, enabled}
                                                        ]}]}
       ]),

    %% The sites should elect a master at the BDB-HA level,
    %% but due to  bdb_DB_REP_CONF_DELAYCLIENT   that master should be stuck in waiting_for_durability

    Node = node(),
    ExpectedGroupStatus = lists:sort([{{site1, Node}, {master, waiting_for_durability}},
                                      {{site2, Node}, {replica, not_in_sync, {{site1,Node}, waiting_for_durability}}},
                                      {{site3, Node}, {replica, not_in_sync, {{site1,Node}, waiting_for_durability}}}]),
    wait_for_group_status(ExpectedGroupStatus, [site1, site2, site3], ?LINE),

    %% As waiting_for_durability only applies to get_master_location_hint and do_txn_on_master,
    %% we can dodge the check using ?BDB_RG:do_txn_on_site.  
    %% We get NOTPERM of course, but the write does hit BDB
    ?EXPECT_TXN_FUN_ERROR({bdb_error,bdb_DB_REP_NOTPERM},
                          ?BDB_RG:do_txn_on_site({test_group, site1, node()}, 
                                                 {?MODULE, test_txn_fun_put_1, {"key1", "committed1", ok}}, 
                                                 0)),

    %% Whereas do_txn_on_master doesn't even attempt the write
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:do_txn_on_master({test_group, site1, node()}, 
                                              {?MODULE, test_txn_fun_put_1, {"key2", "committed2", ok}}, 
                                              0)),

    %% polling for master fails to find one
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:get_master_location_hint({test_group, site1, node()}, 2000)),
    ?AT,
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:get_master_location_hint({test_group, site2, node()}, 2000)),
    ?AT,
    ?EXPECT_RG_ERROR(no_master, 
                     ?BDB_RG:get_master_location_hint({test_group, site3, node()}, 2000)),
    ?AT,

    %% Check that the group status hasn't changed
    wait_for_group_status(ExpectedGroupStatus, [site1, site2, site3], ?LINE),

    %% Now allow replicas to sync
    allow_sites_to_sync([{test_group, site2, node()},
                         {test_group, site3, node()}]),

    %% Wait for the master to become available
    MasterHint1 = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, infinity),

    %% Verify the group state (this also waits for the replicas to sync)
    ExpectedGroupStatus2 = lists:sort([{{site1, Node}, {master, available}},
                                       {{site2, Node}, {replica, in_sync, {{site1,Node}, available}}},
                                       {{site3, Node}, {replica, in_sync, {{site1,Node}, available}}}]),
    wait_for_group_status(ExpectedGroupStatus2, [site1, site2, site3], ?LINE),

    %% retry do_txn_on_master
    ok1 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, 
                                 {?MODULE, test_txn_fun_put_1, {"key3", "committed3", ok1}}, 
                                 0),

    {ok2, MasterHint1} = ?BDB_RG:do_txn_on_master(MasterHint1, 
                                                  {?MODULE, test_txn_fun_put_1, {"key4", "committed4", ok2}}, 
                                                  0),

    %% Verify that the initial do_txn_on_site worked, and that the initial do_txn_on_master did not
    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_site({test_group, site1, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),
    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_site({test_group, site2, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),
    {<<"key1">>, <<"committed1">>} = ?BDB_RG:do_txn_on_site({test_group, site3, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key1">>}}),

    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND}, ?BDB_RG:do_txn_on_site({test_group, site1, node()},
                                                                               {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}})),
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND}, ?BDB_RG:do_txn_on_site({test_group, site2, node()},
                                                                               {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}})),
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND}, ?BDB_RG:do_txn_on_site({test_group, site3, node()},
                                                                               {?MODULE, test_txn_fun_get, {"test.db", <<"key2">>}})),

    {<<"key3">>, <<"committed3">>} = ?BDB_RG:do_txn_on_site({test_group, site1, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key3">>}}),
    {<<"key3">>, <<"committed3">>} = ?BDB_RG:do_txn_on_site({test_group, site2, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key3">>}}),
    {<<"key3">>, <<"committed3">>} = ?BDB_RG:do_txn_on_site({test_group, site3, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key3">>}}),

    {<<"key4">>, <<"committed4">>} = ?BDB_RG:do_txn_on_site({test_group, site1, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key4">>}}),
    {<<"key4">>, <<"committed4">>} = ?BDB_RG:do_txn_on_site({test_group, site2, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key4">>}}),
    {<<"key4">>, <<"committed4">>} = ?BDB_RG:do_txn_on_site({test_group, site3, node()},
                                                            {?MODULE, test_txn_fun_get, {"test.db", <<"key4">>}}),

    %% Stop everything
    ok = ?BDB_RG:stop_entire_group({test_group, site1, node()}),

    %% Currently stop_entire_group is asynchronous, so wait for a bit for all processes to exit, or next test will fail
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 15000, ?LINE),

    {messages, []} = process_info(self(), messages),
    ok.

wait_for_group_status(ExpectedGroupStatus, SiteNames, Line) ->
    repeat_until_true(fun() ->
                              timer:sleep(1000),
                              try 
                                  lists:foreach(fun(SiteName) ->
                                                   ExpectedGroupStatus = lists:sort(?BDB_RG:get_group_status({test_group, SiteName, node()}))
                                                end,
                                                SiteNames),
                                  true
                              catch
                                  Class:Reason ->
                                      io:format("line ~w Failed with ~w,~w, retryint~n", [Line, Class, Reason]),
                                      false
                              end
                      end).

repeat_until_true(Fun) -> 
    case Fun() of 
        true  -> true;
        false -> repeat_until_true(Fun)
    end.


allow_sites_to_sync(SiteLocators) ->

    %% rep_sync only works after the site has received NEWMASTER.
    %%
    %% TODO: Currently there is know way to detect that without registering event handlers (outside the scope of this function).  
    %% The repeat_until_true code below is not good enough as the site's status may change to recognize a new master when we receive 
    %% an other_site_status message from the master, which can (and does) happen before the site receives NEWMASTER.
    %% This is part of our philiosphy -- we don't really use NEWMASTER events other than a sanity check.
    %% We queue off MASTER and STARTUPDONE, and propagate our own master-state changes.
    %% As a lot of thought went into that and it is very well test now, I really don't want to change it.
    %% So, as a gross, temporary (er, perhaps permanent) HACK we simply sleep for 'enough' time to wait for NEWMASTER.

    ?AT,
    io:format("Sleeping to ensure that NEWMASTER is received before calling rep_sync~n"),
    timer:sleep(10000),

    lists:foreach(
      fun({_Rg, _ThisSiteName, _ThisSiteNode} = SiteLocator) -> 

%%               repeat_until_true(
%%                 fun() ->
%%                         GroupStatus = ?BDB_RG:get_group_status(SiteLocator),
%%                         io:format("GroupStatus = ~p~n", [GroupStatus]),
%%                         RetVal = lists:foldl(fun({{SiteName, SiteNode}, {replica, _, {_Master, _State}} = SiteStatus}, _Acc) 
%%                                                 when SiteName =:= ThisSiteName, SiteNode =:= ThisSiteNode -> 
%%                                                      io:format("site ~w has status ~w~n", [SiteName, SiteStatus]),
%%                                                      true;
%%                                                 (_, Acc) -> 
%%                                                      Acc
%%                                              end,
%%                                              false,
%%                                              GroupStatus),
%%                         ?AT,
%%                         timer:sleep(500),
%%                         RetVal
%%                 end),

              ?AT,
              ?BDB_RG:do_txn_on_site(SiteLocator, 
                                     {?BDB_H, txn_fun_do_fun, fun(P) -> 
                                                                      ?BDB:rep_sync(P, []) 
                                                              end}, 
                                     0)
      end,
      SiteLocators).


test_write_errcall_coredump(StartFunc) ->
    io:format("\nPerforming test_write_errcall_coredump\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),
    P0 = ?BDB:StartFunc(),
    P = P0#bdb_port_handle{receive_reply_timeout = 30000},   % core dumps take a few seconds -- don't want a pointless timeout
    DB = ?BDB:db_create(P, []),

    %% This returns {{Year, Month, Day}, {Hour, Minute, Second}}
    TestStartTime = erlang:localtime(),
    [] = list_coredump_files_created_since(TestStartTime),

    IntervalSeconds = 5,

    %% Test disabled by default
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB, void, "key", [])),
    [] = list_coredump_files_created_since(TestStartTime),  

    %% Test inclusive
    ok = ?BDB:set_write_coredump_on_errcall(P, IntervalSeconds, regex_is_inclusive, "", "(environment specified)|(^DB->del: method not permitted before handle's open method$)"),
    %% Dumps core due to match on "DB->del: method not permitted before handle's open method"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB, void, "key", [])),
    [_] = List1 = list_coredump_files_created_since(TestStartTime),    
    timer:sleep((IntervalSeconds + 1) * 1000),

    %% Filtered out as first part of string differs ("DB->fd: method not permitted before handle's open method"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_fd(P, DB)),
    List1 = list_coredump_files_created_since(TestStartTime),  

    %% Dumps core due to "DB->set_cachesize: method not permitted when environment specified"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_set_cachesize(P, DB, 0, 256*1024, 0)),
    List2 = list_coredump_files_created_since(TestStartTime),  

    %% Test suppression within interval-secs (note no sleep after last dump)
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB, void, "key", [])),
    List2 = list_coredump_files_created_since(TestStartTime),  
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB, void, "key", [])),
    List2 = list_coredump_files_created_since(TestStartTime),  
    %% Now do the sleep, and verify that we do start getting dumps
    timer:sleep((IntervalSeconds + 1) * 1000),
    %% Dumps core
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_del(P, DB, void, "key", [])),
    List3 = list_coredump_files_created_since(TestStartTime),  
    3 = length(List3),
    timer:sleep((IntervalSeconds + 1) * 1000),

    %% Test disabling with interval-secs = 0
    ok = ?BDB:set_write_coredump_on_errcall(P, 0, regex_is_exclusive, "./", ".*"),
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_fd(P, DB)),
    List3 = list_coredump_files_created_since(TestStartTime),  
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_get_byteswapped(P, DB)),
    List3 = list_coredump_files_created_since(TestStartTime),  
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_get_type(P, DB)),
    List3 = list_coredump_files_created_since(TestStartTime),  

    %% Test exclusive
    ok = ?BDB:set_write_coredump_on_errcall(P, IntervalSeconds, regex_is_exclusive, "./", "(method not permitted before handle's open method)|(environment specified)"),

    %% Filtered out due to "DB->set_cachesize: method not permitted when environment specified"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_set_cachesize(P, DB, 0, 256*1024, 0)),
    List3 = list_coredump_files_created_since(TestStartTime),  

    DB2 = ?BDB:db_create(P, []),
    %% Dumps core with "environment not yet opened"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_open(P, DB2, void, "", "", bdb_DB_BTREE, [bdb_DB_CREATE], 0)),
    List4 = list_coredump_files_created_since(TestStartTime),  
    4 = length(List4),
    timer:sleep((IntervalSeconds + 1) * 1000),

    %% Is filtered out by "method not permitted before handle's open method"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_get_byteswapped(P, DB)),
    List4 = list_coredump_files_created_since(TestStartTime),  

    ok = env_open_with_retry(P, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK], 8#644, ?LINE),
    ok = ?BDB:db_open(P, DB2, void, "", "", bdb_DB_BTREE, [bdb_DB_CREATE], 0),
    C1 = ?BDB:db_cursor(P, DB2, void, []),
    %% Dumps core due to "Cursor position must be set before performing this operation"
    ?EXPECT_BDB_ERROR(einval, ?BDB:c_get(P, C1, <<>>, <<>>, [bdb_DB_CURRENT])),
    List5 = list_coredump_files_created_since(TestStartTime),  
    5 = length(List5),
    timer:sleep((IntervalSeconds + 1) * 1000),

    %% Filtered out due to end part of "DB->set_cachesize: method not permitted when environment specified"
    ?EXPECT_BDB_ERROR(einval, ?BDB:db_set_cachesize(P, DB, 0, 256*1024, 0)),
    List5 = list_coredump_files_created_since(TestStartTime),  

    normal = ?BDB:shutdown(P),
    
    timer:sleep(1000),
    List5 = list_coredump_files_created_since(TestStartTime),  

    %% Delete all of the coredumps
    lists:foreach(fun(File) -> ok = file:delete(File) end, List5),

    {messages, []} = process_info(self(), messages),
    ok.


test_write_coredump(StartFunc) ->

    io:format("\nPerforming test_write_coredump\n"),
    io:format("NOTE: this test often fails with 'echild' from the Google coredumper library; just retry it if it does\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc},
                                               %% coredumps take a few seconds -- don't want deadlock_detector or checkpointer/cache_trickler to hit a pointless timeout
                                               {default_receive_reply_timeout_milliseconds,    30000},
                                               {deadlock_detector_receive_reply_timeout_milliseconds, 30000},  % also used for memp_trickle
                                               {checkpoint_receive_reply_timeout_milliseconds, 30000}  % also used for memp_trickle
                                              ]),
    
    %% This returns {{Year, Month, Day}, {Hour, Minute, Second}}
    TestStartTime = erlang:localtime(),
    ?AT,
    [] = list_coredump_files_created_since(TestStartTime),

    Filename = "./core.test_coredump_" ++ atom_to_list(StartFunc),
    ok = ?BDB_H:do_txn({?MODULE, txn_fun_write_coredump, Filename}, 0),
    ?AT,
    [Filename] = list_coredump_files_created_since(TestStartTime),  

    %% An empty filename will choose a defaylt
    ok = ?BDB_H:do_txn({?MODULE, txn_fun_write_coredump, ""}, 0),
    ?AT,
    [Filename, _NewFilename] = List = list_coredump_files_created_since(TestStartTime),  

    %% Delete all of the coredumps
    lists:foreach(fun(File) -> ok = file:delete(File) end, List),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next  test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok
    after
        2000 -> throw({timeout_waiting_for_server_exit, ?AT})
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_write_coredump(P, Filename) ->
    ?BDB:write_coredump(P, Filename).


test_local_do_txn_error_handling(StartFunc) ->
                                
    io:format("\nPerforming test_local_do_txn_error_handling\n"),

    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    %% Test using a Pid that has been killed
    {ok, ServerPid1} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),

    %% Verify the success case first
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND},
                          ?BDB_H:do_txn(ServerPid1, {?MODULE, test_txn_fun_get, {"test.db", <<"key">>}}, 0)),

    unlink(ServerPid1),
    exit(ServerPid1, kill),
    timer:sleep(500),   % without this we get 'killed' instead 'of noproc'
    ?EXPECT_BDB_HELPER_ERROR({acquire_bdb_port_handle, {{exit, {noproc, {gen_server, call, [ServerPid1, acquire_bdb_port_handle]}}}, _StackTrace1}},
                             ?BDB_H:do_txn(ServerPid1, {?MODULE, test_txn_fun_get, {"test.db", <<"key">>}}, 0)),

    wait_until_process_is_unregistered(berkeley_db_port_coordinator),   

    %% Now test if the port-coordinator dies after the txn fun returns successfully
    %% -- e.g the txn fun kills the port-coordinator
    {ok, ServerPid2} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),
 
    ?EXPECT_TXN_FUN_ERROR({bdb_error, bdb_DB_NOTFOUND},
                          ?BDB_H:do_txn(ServerPid2, {?MODULE, test_txn_fun_get, {"test.db", <<"key">>}}, 0)),

    ?AT,
    case StartFunc of
        start ->
            ?EXPECT_BDB_HELPER_ERROR({release_bdb_port_handle, {{exit, {killed, {gen_server, call, [ServerPid2, release_bdb_port_handle]}}}, _StackTrace2}},
                                     ?BDB_H:do_txn(ServerPid2, {?MODULE, txn_fun_kill_port_coord, []}, 0));
        start_pipe ->
            ErlangPort2 = ?BDB_H:do_txn(ServerPid2, {?MODULE, txn_fun_get_erlang_port, []}, 0),
            ?EXPECT_BDB_HELPER_ERROR({release_bdb_port_handle, {{exit, {bdb_receive_reply_timeout, {#bdb_port_handle{erlang_port = ErlangPort2}, <<_/binary>>}}}, _StackTrace2}},
                                     ?BDB_H:do_txn(ServerPid2, {?MODULE, txn_fun_kill_port_coord, []}, 0))
    end,

    wait_until_process_is_unregistered(berkeley_db_port_coordinator),   

    %% Now test if release_bdb_port_handle() fails due to it's call to group_valmap_cleanup() finding that the port has been closed.
    ?AT,
    {ok, ServerPid3} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),
    ErlangPort3 = ?BDB_H:do_txn(ServerPid3, {?MODULE, txn_fun_get_erlang_port, []}, 0),
    ?AT,
    ?EXPECT_BDB_HELPER_ERROR({release_bdb_port_handle, {{port_is_unreachable, ErlangPort3}, _StackTrace3}},
                             ?BDB_H:do_txn(ServerPid3, {?MODULE, txn_fun_kill_port_coord_and_pause, []}, 0)),

    %% Now test error handling around do_txn's internal call to group_valmap_cleanup when handling a deadlock
    %% NOTE that we get {release_bdb_port_handle,{port_is_unreachable, _}} 
    %% and NOT a {bdb_helper_error, {group_valmap_cleanup, {error, badarg, StackTrace}}}}.
    %% That's because currently, release_bdb_port_handle is always executed after the retry-on-deadlock 
    %% group_valmap_cleanup (even if the latter fails).
    %% And release_db_port_handle always tries to use the port (for its own group_valmap_cleanup).
    ?AT,

    %% In SMP mode we get an already_started error here, as the previous server 
    wait_until_process_is_unregistered(berkeley_db_port_coordinator),

    ?AT,
    {ok, ServerPid4} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),
    ErlangPort4 = ?BDB_H:do_txn(ServerPid4, {?MODULE, txn_fun_get_erlang_port, []}, 0),
    ?EXPECT_BDB_HELPER_ERROR({release_bdb_port_handle, {{port_is_unreachable, ErlangPort4}, _StackTrace4}},
                             ?BDB_H:do_txn(ServerPid4, {?MODULE, txn_fun_close_port_and_throw_deadlock, []}, 1)),  % NumDeadlocRetries must be > 0

    %% In SMP mode we get an already_started error here, as the previous server 
    wait_until_process_is_unregistered(berkeley_db_port_coordinator),

    %% Now test if TxnFun fails due to finding that port has been closed
    %% NOTE that we get {release_bdb_port_handle,{port_is_unreachable, _}} 
    %% and NOT a txn_fun_error.  
    %% That's because release_bdb_port_handle always executes after TxnFun,
    %% and it always tries to use the port (for group_valmap_cleanup).
    %% And that kind of failure must mask TxnFun result.
    %% In theory we _could_ expose TxnFun result if it was a failure 
    ?AT,
    {ok, ServerPid5} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),
    ErlangPort5 = ?BDB_H:do_txn(ServerPid5, {?MODULE, txn_fun_get_erlang_port, []}, 0),
    ?EXPECT_BDB_HELPER_ERROR({release_bdb_port_handle, {{port_is_unreachable, ErlangPort5}, _StackTrace5}},
                             ?BDB_H:do_txn(ServerPid5, {?MODULE, txn_fun_close_port_and_attempt_to_use_it, []}, 0)),

    wait_until_process_is_unregistered(berkeley_db_port_coordinator),   
    ok.

txn_fun_get_erlang_port(#bdb_port_handle{erlang_port = ErlangPort}, _) ->
    ErlangPort.

txn_fun_kill_port_coord(#bdb_port_handle{port_coordinator_pid = Pid}, _) ->
    unlink(Pid),
    exit(Pid, kill),
    my_txn_fun_result.

txn_fun_kill_port_coord_and_pause(#bdb_port_handle{port_coordinator_pid = Pid}, _) ->
    unlink(Pid),
    exit(Pid, kill),
    timer:sleep(500),   % without this we get 'killed' instead 'of noproc'
    my_txn_fun_result.

wait_until_process_is_unregistered(ProcName) ->
    case whereis(ProcName) of 
        undefined -> ok;
        _Pid ->
            timer:sleep(100),
            wait_until_process_is_unregistered(ProcName)
    end.

txn_fun_close_port_and_throw_deadlock(#bdb_port_handle{port_coordinator_pid = Pid,
                                                       erlang_port          = ErlangPort}, _) ->
    unlink(Pid),
    true = erlang:port_close(ErlangPort),
    exit(Pid, kill),  % to avoid noise in logging (port sends exit signal to port-coord, which then crashes in terminate as port is closed)
    throw({bdb_error, bdb_DB_LOCK_DEADLOCK}).

txn_fun_close_port_and_attempt_to_use_it(#bdb_port_handle{port_coordinator_pid = Pid,
                                                          erlang_port          = ErlangPort} = BdbPort, _) ->
    unlink(Pid),
    true = erlang:port_close(ErlangPort),
    exit(Pid, kill),  % to avoid noise in logging (port sends exit signal to port-coord, which then crashes in terminate as port is closed)
    _Txn1 = ?BDB:txn_begin(BdbPort, void, []).



test_clone_bdb_port_handle(StartFunc) ->    
    io:format("\nPerforming test_clone_bdb_port_handle\n"),
 
    Dir = "txn_tests_env",
    ensure_empty_dir(Dir),

    {ok, ServerPid} = ?BDB_PC:start_link(Dir, [{port_start_function, StartFunc}]),

    ParentP   = ?BDB_PC:acquire_bdb_port_handle(),
    DB        = ?BDB_PC:get_db_handle(ParentP, bdb_DB_BTREE, "test_coord.db"),

    ParentTxn = ?BDB:txn_begin(ParentP, void, []),

    Key1 = <<"-Nestedkey1">>,
    Val1 = <<"-Nestedval1">>,

    TestDriver = self(),
    SpawnedPid = spawn_link(
                   fun() -> 
                           ChildP1 = ?BDB_PC:clone_bdb_port_handle(ParentP),

                           ChildTxn1 = ?BDB:txn_begin(ChildP1, ParentTxn, []),
                           ok = ?BDB:db_put(ChildP1, DB, ChildTxn1, Key1, Val1, []),

                           %% It doesn't matter for this test whether we commit explicitly, so we don't bother

                           TestDriver ! {self(), ready_to_commit}

                           %% Currently we MUST NOT pass ChildP1 to release_bdb_port_handle
                   end),

    %% wait for the OK to proceed
    receive {SpawnedPid, ready_to_commit} -> ok end,

    %% committing the parent also commits any unresolved children
    ok = ?BDB:txn_commit(ParentP, ParentTxn, []),

    %% verify
    {Key1, Val1} = ?BDB:db_get(ParentP, DB, void, Key1, <<>>, []),

    %% and the 'uncommitted' child txn should not need to be aborted
    ?AT,
    0 = ?BDB:group_valmap_cleanup(ParentP, ParentP#bdb_port_handle.group_valmap_id),
    ?AT,

    ok = ?BDB_PC:release_bdb_port_handle(ParentP),

    %% Can't release it twice
    ?EXPECT_PORT_COORD_ERROR(caller_does_not_have_a_bdb_port_handle, ?BDB_PC:release_bdb_port_handle(ParentP)),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next  test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, normal} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        2000 -> 
            throw(timeout_waiting_for_server_exit)
    end,
    undefined = whereis(?BDB_PC),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),

    ok.


test_replication_remove_txn_logs(StartFunc) ->

    io:format("\nPerforming test_replication_remove_txn_logs\n"),

    EnvDir1 = "receive_queue_env",
    ensure_empty_dir(EnvDir1),

    EnvDir2 = "txn_tests_env",
    ensure_empty_dir(EnvDir2),

    EnvDir3 = "send_queue_env",
    ensure_empty_dir(EnvDir3),

    MaxNumTrailingLogFilesIfAnyErrors = 7, % at least 7 or site 3 restart will fail to join environment
    LogRemovalPeriodSeconds = 4,
    SleepTimeMs = LogRemovalPeriodSeconds * 1010,

    RgConfig = [{find_other_sites_period_milliseconds,        800}, % fast re-discovery of down site
                {log_removal_period_seconds,                  LogRemovalPeriodSeconds},
                {log_removal_delete_file_period_milliseconds, 0},
                {log_removal_max_num_trailing_log_files,      MaxNumTrailingLogFilesIfAnyErrors},
                {common_port_coordinator_config, [{port_start_function, StartFunc},
                                                  {log_file_size_bytes, 51200},    % small log files (50 KB), to create lots of logs quickly
                                                  {env_set_flags,[bdb_DB_AUTO_COMMIT
                                                                  %%bdb_DB_TXN_NOSYNC, % don't use this or logs might not be flushed
                                                                 ]},
                                                  {checkpoint_seconds_and_min_kb, disabled}, % we do them manually to control when logs can be removed
                                                  {env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                                                  {port_log_directory, "."},
                                                  {replication, [{initial_role_and_threadcount,  {bdb_DB_REP_FULL_ELECTION, 1}}, 
                                                                 {ack_policy,		      bdb_DB_REPMGR_ACKS_QUORUM}, 
                                                                 {limit_transfer_per_call_bytes, 100000},
                                                                 {config_flags, [bdb_DB_REP_CONF_NOAUTOINIT]} % we want control over when sync happens
%%% with these settings we get infinite elections after restarting site 3
%%% in both start and start_pipe modes
%%% (that was when we were doing the test that causes PANIC, which leaves repmgr running
%%%  so that might have been messing with the next test. we disabled that test, so check this again)
%%                                                                  ,{{timeout, election_milliseconds},             5000}
%%                                                                  ,{{timeout, election_retry_milliseconds},       5000}
%% and these settings do the same in start mode.
%% 								 ,{{timeout, election_milliseconds},		10000}
%% 								 ,{{timeout, election_retry_milliseconds},	11000}
                                                                ]}]}],

    Site1 = {{site1, node()}, EnvDir1, {"localhost", 4000}, [{replication, [{this_site_priority, 2}]}]},
    Site2 = {{site2, node()}, EnvDir2, {"localhost", 4010}, [{replication, [{this_site_priority, 1}]}]},
    Site3 = {{site3, node()}, EnvDir3, {"localhost", 4020}, [{replication, [{this_site_priority, 1}]}]},

    {_GoodSitesPids, []} = ?BDB_RG:start_sites(test_group, [Site1, Site2, Site3], RgConfig),

    io:format("write some txns, verify logs are removed~n"),
    MLH = ?BDB_RG:get_master_location_hint({test_group, site1, node()}, infinity),
    wait_until_all_sites_are_in_sync(MLH, [site1, site2, site3], 20000, ?LINE),

    ?BDB_RG:do_txn_on_master(MLH,
                             {?MODULE, txn_fun_write_N_records_of_size_M, {10, 10000, "first-batch"}}, 
                             0,
                             10000), % timeout
    LogsBeforeRemoval = ["log.0000000001","log.0000000002","log.0000000003"],
    LogsBeforeRemoval = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    LogsBeforeRemoval = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    LogsBeforeRemoval = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),

    %% Due to the forced master-sleep of 30s delay added in 20070228.01, we have to use a really long timeout
    CheckpointTimeout = 35000,
    ?BDB_RG:do_txn_on_master(MLH, {?MODULE, txn_fun_checkpoint, CheckpointTimeout}, 0, CheckpointTimeout),

    io:format("Done checkpoint. Sleeping for ~w milliseconds", [SleepTimeMs]),
    timer:sleep(SleepTimeMs),
    ?AT,
    LogsAfterRemoval = ["log.0000000003"],
    ?AT,
    LogsAfterRemoval = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,

    io:format("kill 1 site, write more txns, verify logs are removed but a longer trail of logs is kept~n"),
    Site3PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site3),
    %%exit(whereis(Site3PortCoordName), kill),
    ?BDB_PC:stop(Site3PortCoordName),
    wait_until_all_sites_are_down(test_group, [site3], 15000, ?LINE),    

    ?BDB_RG:do_txn_on_master(MLH,
                             {?MODULE, txn_fun_write_N_records_of_size_M, {10, 20000, "first-batch"}}, 
                             0,
                             10000), % timeout
    ?AT,
    LogsBeforeRemoval2 = ["log.0000000003","log.0000000004","log.0000000005","log.0000000006","log.0000000007","log.0000000008","log.0000000009","log.0000000010"],
    LogsBeforeRemoval2 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    LogsBeforeRemoval2 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ThisNode = node(),
    ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site3, ThisNode}, ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0)),

    ?BDB_RG:do_txn_on_master(MLH, {?MODULE, txn_fun_checkpoint, CheckpointTimeout}, 0, CheckpointTimeout),

    io:format("Done checkpoint. Sleeping for ~w milliseconds", [SleepTimeMs]),
    timer:sleep(SleepTimeMs),

    ?AT,
    RevLogsBeforeRemoval = lists:sort(fun(A, B) -> A > B end, LogsBeforeRemoval2),
    SafeToRemoveLogs = tl(RevLogsBeforeRemoval),
    WillRemoveLogs = case length(SafeToRemoveLogs) - (MaxNumTrailingLogFilesIfAnyErrors + 1) of 
                         N when N > 0 -> lists:nthtail(N, SafeToRemoveLogs);
                         _            -> []
                     end,
    LogsAfterRemoval2 = lists:sort([hd(RevLogsBeforeRemoval)] ++ (SafeToRemoveLogs -- WillRemoveLogs)),
    ?AT,
    LogsBeforeRemoval2 = LogsAfterRemoval2,
    ?AT,
    LogsAfterRemoval2 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval2 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site3, ThisNode}, ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0)),
    

    io:format("restart site, let it sync up, verify that longer trail of logs is removed (even without adding more txns/checkpoints)~n"),
    ?AT,
    Site3Restart = {{site3, node()}, EnvDir3, {"localhost", 4020}, 
                    [{env_verbose_types, [bdb_DB_VERB_RECOVERY, bdb_DB_VERB_REGISTER, bdb_DB_VERB_REPLICATION]},
                     {port_log_directory, "."},
                     {replication, [{this_site_priority, 1},
                                    {config_flags, [bdb_DB_REP_CONF_NOAUTOINIT,
                                                    bdb_DB_REP_CONF_DELAYCLIENT]} % we want control over when sync happens
                                   ]}
%%% We only need this part if we test the failure-to-join case 
%%% later, but that is disabled for now as the associated PANIC event
%%% causes problems
%%                      ,{coredump_on_errcall_interval_secs_and_regex,
%%                       {300,                 % at most 1 core dump every 5 minutes (may be more if set_write_coredump_on_errcall() is called repeatedly as it resets the timer -- and port_coord init calls it)
%%                        regex_is_exclusive,  % we *don't* want core-dumps for the following...
%%                        "./",                % path prefix
%%                        %% these are just the defaults...
%%                        "(connecting to site)"                                       % repmgr noise
%%                        "|(EOF on connection)"                                       % repmgr noise
%%                        "|(socket writing failure)"                                  % repmgr noise
%%                        "|(can't read from site)"                                    % repmgr noise
%%                        "|(writing data)"                                            % repmgr noise
%%                        "|(redundant incoming connection will be ignored)"           % repmgr noise
%%                        "|(can't bind socket to listening address)"                  % repmgr noise
%%                        "|(DB_UNKNOWN type specified with DB_CREATE)"                % get_db_handle noise
%%                        "|(multiple databases specified but not supported by file)"  % get_db_handle_with_config noise
%%                        "|(Prepare disallowed on child transactions)"                % txn_prepare noise
%%                        "|(Operation locked out.  Waiting for replication lockout to complete)"
%%                        %% Additions just for this test:
%%                        %% We are going to force a join-failure, 
%%                        %% so suppress these
%%                        "|(DB_REP_JOIN_FAILURE: Unable to join replication group)"
%%                        "|(PANIC: fatal region error detected; run recovery)"
%%                       }}
                    ]},
    _Pid = ?BDB_RG:start(test_group, Site3Restart, [Site1, Site2], RgConfig),

    timer:sleep(1000), 

    ?AT,
    %% (site3 is using DELAYCIENT so has not started syncing yet)
    %% site3 initially only has it's last log from before it was killed
    Site3Logs = ["log.0000000003"],
    Site3Logs = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),

    %% Now we can add more txns and all sites will keep all logs back to
    %% the oldest (site3's) in-use log.  
    %% i.e. The log_removal_max_num_trailing_logs_files does NOT limit this 
    %% -- site3 *is* responding to log_archvie requests, it is just lagging a long way behind.
    %% log_removal_max_num_trailing_logs_files is only used when we cannot even connect 
    %% to a site (or we get some other kind of error), so we have no knowledge of what
    %% logs are on that machine (TODO: if remove_txn_logs_on_all_sites_process has seen a valid 
    %% response before from a process that is now down then it _could_ choose to remember/use the 
    %% last response for that site.  We don't do that because we have to call it a day somehow 
    %% (write off that site if it does not come back), so we might as well doing it by limiting 
    %% the number of trailing logs, as that is proportinate to disk-space used, which is what 
    %% we care about (instead of some timeout).

    ?BDB_RG:do_txn_on_master(MLH,
                             {?MODULE, txn_fun_write_N_records_of_size_M, {11, 20000, "second-batch"}}, 
                             0,
                             10000), % timeout
    ?AT,
    LogsBeforeRemoval3 = ["log.0000000003","log.0000000004","log.0000000005","log.0000000006","log.0000000007","log.0000000008","log.0000000009",
                          "log.0000000010","log.0000000011", "log.0000000012","log.0000000013","log.0000000014"],
    ?AT,
    LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),

    ?AT,    
    Site3Logs = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    %% TODO: In pipe mode, site3 has all the logs here, for some reason, but linked-in mode does not.
    %% (even if we sleep first in both modes)
%%     ?AT,    
%%     timer:sleep(1000),
%%     case StartFunc of 
%%         start ->
%%             ?AT,    
%%             Site3Logs = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0);
%%         start_pipe ->
%%             ?AT,    
%%             LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0)
%%         end,

    ?BDB_RG:do_txn_on_master(MLH, {?MODULE, txn_fun_checkpoint, CheckpointTimeout}, 0, CheckpointTimeout),

    io:format("Done checkpoint. Sleeping for ~w milliseconds", [SleepTimeMs]),
    timer:sleep(SleepTimeMs),

    %% cannot remove any of the logs until after site3 has synced up
    ?AT,
    LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    Site3Logs = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),

    %% now allow site3 to start syncing
    ?AT,
    allow_sites_to_sync([{test_group, site3, node()}]),
    ?AT,
    wait_until_all_sites_are_in_sync(MLH, [site1, site2, site3], 30000, ?LINE), 
    ?AT,
    %% Now Site3 should have all of the logs
    %% HOWEVER, this is a race condition; if starting site3 & doing sync-up takes longer than SleepTimeMs then some may have been purged
    %% LogsBeforeRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    %% ?AT,

    %% now the trailing removable logs will be removed from the other 2 sites
    %% HOWEVER, if the master performs a sync-up of another site, it suppresses log removal for 30 seconds,
    %% so the master will keep whatever logs it had when site3 restarted
    io:format("Sleeping for ~w milliseconds", [SleepTimeMs]),
    timer:sleep(SleepTimeMs),
    LogsAfterRemoval3 = ["log.0000000014"],
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    
    io:format("test no sites have any logs to remove~n"),
    io:format("Sleeping for ~w milliseconds", [SleepTimeMs]),
    timer:sleep(SleepTimeMs),
    ?AT,
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,
    LogsAfterRemoval3 = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
    ?AT,


    %% Now kill site3 again and this time wait for LONGER than log_removal_max_num_trailing_logs_files 
    %% before restarting it.
    %% When we let it sync, it should fail to join the environment, thereby proving
    %% that we do correctly remove logs past the threshold of log_removal_max_num_trailing_logs_files

    %% IMPORTANT
    %% TODO: currently this test is disabled (even though it passes) as in v4.5 it leaves a repmgr connection listening on port 4020
    %% -- apparently if a site gets a PANIC event then calling DbEnv:close won't shutdown repmgr.
    %% (need to confirm that with Sleepycat).   I know that v4.6 fixes a related issue -- BDB will close
    %% any open files that it has open, so we need to make sure they do this to.
    %% Until then our best bet is to call erlang:halt() if we get EVENT_PANIC, and let process-manager restart us.

%%     io:format("TODO: kill 1 site, write a LOT more txns, verify logs are removed but a longer trail of logs is kept~n"),
%%     Site3PortCoordName = ?BDB_RG:get_port_coordinator_proc_name(test_group, site3),
%%     %%exit(whereis(Site3PortCoordName), kill),
%%     ?BDB_PC:stop(Site3PortCoordName),

%%     ?BDB_RG:do_txn_on_master(MLH,
%%                              {?MODULE, txn_fun_write_N_records_of_size_M, {20, 20000, "third-batch"}}, 
%%                              0,
%%                              10000), % timeout
%%     ?AT,
%%     LogsBeforeRemoval4 = ["log.0000000014","log.0000000015","log.0000000016","log.0000000017","log.0000000018",
%%                           "log.0000000019","log.0000000020","log.0000000021","log.0000000022","log.0000000023"],
%%     ?AT,
%%     LogsBeforeRemoval4 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
%%     ?AT,
%%     LogsBeforeRemoval4 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
%%     ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site3, ThisNode}, ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0)),

%%     ?BDB_RG:do_txn_on_master(MLH, {?MODULE, txn_fun_checkpoint, CheckpointTimeout}, 0, CheckpointTimeout),

%%     io:format("Done checkpoint. Sleeping for ~w milliseconds", [SleepTimeMs]),
%%     timer:sleep(SleepTimeMs),

%%     ?AT,
%%     RevLogsBeforeRemoval4 = lists:sort(fun(A, B) -> A > B end, LogsBeforeRemoval4),
%%     SafeToRemoveLogs4 = tl(RevLogsBeforeRemoval4),
%%     WillRemoveLogs4 = case length(SafeToRemoveLogs4) - (MaxNumTrailingLogFilesIfAnyErrors + 1) of 
%%                          N2 when N2 > 0 -> lists:nthtail(N2, SafeToRemoveLogs4);
%%                          _            -> []
%%                      end,
%%     LogsAfterRemoval4 = lists:sort([hd(RevLogsBeforeRemoval4)] ++ (SafeToRemoveLogs4 -- WillRemoveLogs4)),
%%     ?AT,
%%     %% as #14 is missing, site3 (which is still on #14) won't be able to sync
%%     ?AT,
%%     ["log.0000000022","log.0000000023"] = LogsAfterRemoval4,
%%     LogsAfterRemoval4 = ?BDB_RG:do_txn_on_site({test_group, site1, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
%%     ?AT,
%%     LogsAfterRemoval4 = ?BDB_RG:do_txn_on_site({test_group, site2, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),
%%     ?AT,
%%     ?EXPECT_RG_SITE_DOWN_ERROR({test_group, site3, ThisNode}, ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0)),

%%     io:format("restart site3~n"),
%%     ?AT,
%%     Site3Pid3 = ?BDB_RG:start(test_group, Site3Restart, [Site1, Site2], RgConfig),
%%     MonitorRef = erlang:monitor(process, Site3Pid3),

%%     %% (site3 is using DELAYCIENT so has not started syncing yet)
%%     %% site3 initially only has it's last log from before it was killed
%%     Site3Logs2 = ["log.0000000014"],
%%     ?AT,
%%     Site3Logs2 = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?MODULE, txn_fun_list_all_logs, []}, 0),

%%     %% now allow site3 to start syncing -- it should fail
%%     ok = ?BDB_RG:do_txn_on_site({test_group, site3, node()}, {?BDB_H, txn_fun_do_fun, fun(P) -> ?BDB:rep_sync(P, []) end}, 0),
%%     ?AT,
%%     receive
%%         {'DOWN', MonitorRef, process, Site3Pid3, {this_site_port_coordinator_exited,{_Site3PortCoordPid3, {sleepycat_event, bdb_DB_EVENT_PANIC}}}} -> ok;
%%         Other -> throw({unexpected_msg, Other})
%%     after
%%         2000 -> throw({timeout_waiting_for_server_exit, ?AT})
%%     end,
%%     undefined = whereis(Site3PortCoordName),
%%     ?AT,

    io:format("****** stopping group~n"),    
    ok = ?BDB_RG:stop_entire_group({test_group, site2, node()}),
    wait_until_all_sites_are_down(test_group, [site1, site2, site3], 30000, ?LINE),
    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_write_N_records_of_size_M(P, {NumRecords, RecordSize, KeyPrefix}) ->

    DB = ?BDB_PC:get_db_handle_with_db_config(P, bdb_DB_BTREE, "test.db", []),
    Record = <<0:RecordSize/unit:8>>,
    lists:foreach(fun(N) -> 
                          io:format("Doing write ~w~n", [N]),
                          ?BDB:db_put(P, DB, void, KeyPrefix ++ integer_to_list(N), Record, [])
                          %%timer:sleep(2)
                  end,
                  lists:seq(1, NumRecords)).

txn_fun_list_all_logs(P, _) -> 
    ?BDB:log_archive(P, [bdb_DB_ARCH_LOG]).

txn_fun_checkpoint(P, CheckpointTimeout) ->
    ?BDB:txn_checkpoint(P#bdb_port_handle{receive_reply_timeout = CheckpointTimeout}, 0, 0, []).

txn_fun_list_removable_logs(P, _) ->
    ?BDB:log_archive(P, []).


test_panic_port_coordinator(StartFunc) ->
    io:format("\nPerforming test_panic_port_coordinator\n"),

    EnvDir = "txn_tests_env",
    ensure_empty_dir(EnvDir),

    {ok, ServerPid} = ?BDB_PC:start_link(
                         EnvDir, 
                         [{port_start_function, StartFunc},
                          %%{port_log_directory, "."},
                          {coredump_on_errcall_interval_secs_and_regex,
                           {300,                 % at most 1 core dump every 5 minutes (may be more if set_write_coredump_on_errcall() is called repeatedly as it resets the timer -- and port_coord init calls it)
                            regex_is_exclusive,  % we *don't* want core-dumps for the following...
                            "./",                % path prefix
                            "."}}
                         ]),
    unlink(ServerPid),
    ServerRef = erlang:monitor(process, ServerPid),

    case StartFunc of
        start ->
            %% This will call erlang:halt() due to the default value of run_mode_specific_funs_on_BDB_PANIC
            ?BDB_H:do_txn({?MODULE, txn_fun_panic, []}, 1);

        start_pipe ->
            %% The port coordinator will kill all clients and then exit, but the VM will not stop
            {ClientPid, ClientRef} = erlang:spawn_monitor(fun() -> ?BDB_H:do_txn({?MODULE, txn_fun_panic, []}, 1) end),
            ?AT,
            receive {'DOWN', ClientRef, process, ClientPid, killed} -> ok end,
            ?AT,
            receive {'DOWN', ServerRef, process, ServerPid, {sleepycat_event,bdb_DB_EVENT_PANIC}} -> ok end,
            ?AT,
            timer:sleep(2000)  % just to wait for error logging to finish
    end,
    {messages, []} = process_info(self(), messages),
    ok.

txn_fun_panic(P, _) -> ?BDB:env_set_flags(P, [bdb_DB_PANIC_ENVIRONMENT], 1).


test_env_interrupt_os_sleep(StartFunc) ->
    io:format("\nPerforming test_env_interrupt_os_sleep\n"),

    EnvDir = "txn_tests_env",
    ensure_empty_dir(EnvDir),

    {ok, ServerPid} = ?BDB_PC:start_link(
                         EnvDir, 
                         [{port_start_function, StartFunc},
                          {cache_trickle_milliseconds_and_percent,                  disabled},
                          {checkpoint_seconds_and_min_kb,                           {1,  0}},
                          {checkpoint_receive_reply_timeout_milliseconds,           100 * 1000},
                          {{thread_count_stacksize_qlenlimit, housekeeping},        {1, 32*1024, 10}}, % 1 thread
                          {replication, [{this_hostname_port,            {"localhost", 4000}},
                                         {this_site_priority,            1},
                                         {peer_hostnames_ports,          []},
                                         {initial_role_and_threadcount,  {bdb_DB_REP_MASTER, 2}},
                                         {{internal_delay, master_checkpoint_seconds}, 60}       %% **** long internal os_sleep
                                        ]}
                          ]),

    %% Generate a little bit of write-traffic (something to checkpoint)
    ok_result = ?BDB_H:do_txn(ServerPid, {?MODULE, test_txn_fun_put_1, {"key1", "committed1", ok_result}}, 0),

    %% Wait for checkpoint to start
    repeat_until_true(
      fun() ->
              case ?BDB_H:do_txn(ServerPid, 
                                 {?BDB_H, txn_fun_do_fun, fun(P) -> ?BDB:get_threadpool_info(P, ?BDB_THREADPOOL_ID_HOUSEKEEPING) end},
                                 0)
              of
                  [{edtk_threadpool_id,         ?BDB_THREADPOOL_ID_HOUSEKEEPING},
                   {curr_queue_len,             0},
                   {curr_idle_threads,          0},  %% these two show that 
                   {num_threads,                1},  %% a checkpoint is running (trickle is disabled)
                   {curr_enqueued_poison_pills, 0},
                   {queue_len_limit,            10}] -> 
                      true;

                  R ->
                      io:format("waiting for checkpoint to start ~w~n", [R]),
                      timer:sleep(1500),
                      false
              end
      end),

    %% Now tell server to shutdown (which will call env_interrupt_os_sleep) and time how long it takes to exit.
    %% It should be a lot faster than master_checkpoint_seconds.

    BeginShutdownTime = now(),

    %% As we are not trapping exits we DON'T receive 'EXIT' messages with Reason 'normal'.
    MonitorRef = erlang:monitor(process, ServerPid),
    ok = ?BDB_PC:stop(),
    %% Using hard-kill instead would exercise the same behaviour in bdb_stop() [berkeley_db.xml inserts a call 
    %% to env_interrupt_os_sleep() at the top of bdb_stop).  I checked this manually.
    %%exit(ServerPid, kill),

    %% It takes a few milliseconds to shutdown the server, so we need to wait 
    %% or the next  test might get 'already_started' when it tries to start the server.
    receive
        {'DOWN', MonitorRef, process, ServerPid, _} -> ok;
        Other -> throw({unexpected_msg, Other})
    after
        100 * 1000 -> 
            throw({?AT, timeout_waiting_for_server_exit})
    end,
    undefined = whereis(?BDB_PC),

    Elapsed = ?BDB_U:now_diff_milliseconds(now(), BeginShutdownTime),
    if Elapsed > 1000 -> throw({?AT, elapsed_time_too_large, Elapsed});
       true -> ok
    end,

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.



test_driver_stop_while_running_valmap_stop_commands(StartFunc) ->

    io:format("\nPerforming test_driver_stop_while_running_valmap_stop_commands\n"),

    %% This was a bug found during failover testing
    %% - open a lot of db handles
    %% - start closing them (enqueue db_close commands; in the test they were stuck behind slow-to-replicate txn_commits)
    %% - do bdb_stop()
    %%
    %% The bug was that 
    %% - bdb_stop sets shutdown_has_begun
    %% - bdb_stop sets CLEANUP_AS_SOON_AS_NOLONGER_IN_USE on all in-progress db handles
    %% - all invoke_bdb_db_close commands skipb the library call because of bdb_stop
    %% - but invoke_bdb_db_close then processes CLEANUP_AS_SOON_AS_NOLONGER_IN_USE, and calls cleanup_valmap_db() 
    %%   **with do_cleanup_func = 0**, because it thinks that the library call was performed.
    %%   This clears the valmap slot but leaks the db handle.
    %%
    %% A similar bug occures if the valmap_stop threadpool has a short queue-length-limit.
    %% The enqueue code fails to enqueue (hits limit) and sets c->suppress_library_call and executes the 
    %% command anyway. The library function is not called, but the tail cleanup knows that it is a valmap 'stop'
    %% function so clears the valmap stop.

    %% The fix is to _always_ do the library call (ignore shutdown_has_begun and c->suppress_library_call)
    %% if this is a 'valmap stop' function (for _any_ of its arguments).
    %% c->suppress_library_call was renamed c->suppress_library_call_if_not_valmap_stop for clarity.

    EnvDir = "txn_tests_env",
    ensure_empty_dir(EnvDir),

    P = ?BDB:StartFunc(),
    %%?BDB:adjust_port_logging(P, enable, [all]),
    ok = env_open_with_retry(P, EnvDir, [bdb_DB_REGISTER, bdb_DB_RECOVER, bdb_DB_CREATE, bdb_DB_INIT_MPOOL, bdb_DB_INIT_LOCK, bdb_DB_INIT_TXN], 8#644, ?LINE),

    ok = ?BDB:set_write_coredump_on_errcall(P, 500, regex_is_inclusive, "./", "Database handles still open at environment close"),

    NumDBHandles = 100,
    ok = ?BDB:set_num_valmap_slots(P, bdb_db, NumDBHandles),
    DBHandles = lists:map(fun(N) ->
                                  DB = ?BDB:db_create(P, []),
                                  ok = ?BDB:db_open(P, DB, void, "test.db", integer_to_list(N), bdb_DB_BTREE, [bdb_DB_CREATE, bdb_DB_AUTO_COMMIT], 8#644),
                                  DB
                          end,
                          lists:seq(1,NumDBHandles)),

    TestStartTime = erlang:localtime(),

    ReplyTags = lists:map(fun(DB) ->
                                  ?BDB:db_close_non_block(P, DB, [])
                          end,
                          DBHandles),    
    
    %% If the bug occurs, this writes a coredump
    normal = ?BDB:shutdown(P),	

    %% We might have got some replies, so 
    %% receive them with timeout 0 
    %% and catch the exits for those that we did not receive
    lists:foreach(fun(Tag) ->
                              (catch ?BDB:receive_db_close_reply(P#bdb_port_handle{receive_reply_timeout = 0}, Tag))
                  end,
                  ReplyTags),

    [] = list_coredump_files_created_since(TestStartTime),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.



test_c_get_multiple(StartFunc) ->

    io:format("\nPerforming test_c_get_multiple\n"),

    %% adapted from cursor_tests(), so doesn't use port-coordinator (but it could)

    %% Open an in-memory environment
    P = ?BDB:StartFunc(),
    ok = ?BDB:env_open(P, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644),

    %% a database that does not support duplicates
    DB = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB, void, "", "", bdb_DB_BTREE, [bdb_DB_CREATE], 0),

    %% Put records into the table so that it isn't completely empty.
    Key1 = <<"01">>,
    Val1 = <<"vAl01sdflkjsdflkjfsdlk">>,

    Key2 = <<"A2dfjfkdlkdfslkj">>,
    Val2 = <<"Val02">>,

    Key3 = <<"s3-j">>,
    Val3 = <<"val03">>,

    Key4 = <<"s4--large-value">>,
    Val4 = <<0:16150/integer-unit:8>>,

    Key5 = <<"t3-j">>,
    Val5 = <<"vlksjdflskdjfjfkfkal03">>,

    Key6 = <<"t4--!!!!!!e">>,
    Val6 = <<"jskfdkfkdfjlkjfdslkjlkljlkjlkjlkjklkjffufu">>,

    ok = ?BDB:db_put(P, DB, void, Key1, Val1, []),
    ok = ?BDB:db_put(P, DB, void, Key2, Val2, []),
    ok = ?BDB:db_put(P, DB, void, Key3, Val3, []),
    ok = ?BDB:db_put(P, DB, void, Key4, Val4, []),
    ok = ?BDB:db_put(P, DB, void, Key5, Val5, []),
    ok = ?BDB:db_put(P, DB, void, Key6, Val6, []),

    %% Test retrieving keys & records from btree/hash with no duplicates (DB_MULTIPLE_KEY)
    C1 = ?BDB:db_cursor(P, DB, void, []),

    [{Key1, Val1}, {Key2, Val2}, {Key3, Val3}] = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 16384, [bdb_DB_FIRST]),
    %% .. this is same as default

    [{Key1, Val1}, {Key2, Val2}, {Key3, Val3}] = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 16384, [bdb_DB_FIRST, bdb_DB_MULTIPLE_KEY]),


    %% Attempt to read next block with buffer that is too small even for one key/record
    ?EXPECT_BDB_ERROR(bdb_DB_BUFFER_SMALL, ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 8192, [bdb_DB_NEXT, bdb_DB_MULTIPLE_KEY])),
    %% read next key
    [{Key4, Val4}] = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 16384, [bdb_DB_NEXT, bdb_DB_MULTIPLE_KEY]),
    %% ... and next
    [{Key5, Val5}, {Key6, Val6}] = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 16384, [bdb_DB_NEXT, bdb_DB_MULTIPLE_KEY]),

    %% DB_MULTIPLE only gets (duplicate) records for the *current* key, not for subsequent keys
    {Key_d1, [Val1]} = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 16384, [bdb_DB_FIRST, bdb_DB_MULTIPLE]),


    %% Test illegal buffer size (is smaller than DB page)
    %% .. causes error "DB_MULTIPLE/DB_MULTIPLE_KEY buffers must be aligned, at least page size and multiples of 1KB"
    ?EXPECT_BDB_ERROR(einval, ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 1024, [bdb_DB_FIRST])),
    %% However, the requirement to be a multiple of 1024 is taken care of in outputv (berkeley_db.xml)
    [{Key1, Val1}, {Key2, Val2}, {Key3, Val3}] = ?BDB:c_get_multiple(P, C1, <<>>, <<>>, 8192 + 1, [bdb_DB_FIRST]),

    %% Test retrieving duplicate records for a single key from btree/hash with duplicates enabled (DB_MULTIPLE)
    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_flags(P, DB2, [bdb_DB_DUP, bdb_DB_DUPSORT]),
    ok = ?BDB:db_open(P, DB2, void, "", "", bdb_DB_BTREE, [bdb_DB_CREATE], 0),

    Key_d1 = <<"01">>,
    Val_d1 = <<"Val02">>,
    Val_d2 = <<"vAl01sdflkjsdflkjfsdlk">>,
    Val_d3 = <<"val03">>,

    ok = ?BDB:db_put(P, DB2, void, Key_d1, Val_d1, []),
    ok = ?BDB:db_put(P, DB2, void, Key_d1, Val_d2, []),
    ok = ?BDB:db_put(P, DB2, void, Key_d1, Val_d3, []),

    C2 = ?BDB:db_cursor(P, DB2, void, []),
    {Key_d1, [Val_d1, Val_d2, Val_d3]} = ?BDB:c_get_multiple(P, C2, <<>>, <<>>, 16384, [bdb_DB_FIRST, bdb_DB_MULTIPLE]),


    %% Test retrieving a subset of duplicate records for a single key from btree/hash with duplicates enabled (DB_GET_BOTH with DB_MULTIPLE)
    {Key_d1, [Val_d2, Val_d3]} = ?BDB:c_get_multiple(P, C2, Key_d1, Val_d2, 16384, [bdb_DB_GET_BOTH, bdb_DB_MULTIPLE]),
    %% ... prefix comparison on data
    {Key_d1, [Val_d2, Val_d3]} = ?BDB:c_get_multiple(P, C2, Key_d1, <<"Z">>, 16384, [bdb_DB_GET_BOTH_RANGE, bdb_DB_MULTIPLE]),


    %% Test retreiving keys (record numbers) and records from queue/recno (DB_MULTIPLE_KEY, must be decoded with DB_MULTIPLE_RECNO_NEXT)
    DB3 = ?BDB:db_create(P, []),
    ok = ?BDB:db_set_re_len(P, DB3, 4),             %% arg is bytes
    ok = ?BDB:db_set_re_pad(P, DB3, 32),           %% arg is byte value (default is 32, i.e. space character)
    ok = ?BDB:db_open(P, DB3, void, "", "", bdb_DB_QUEUE, [bdb_DB_CREATE], 0),

    QVal1 = <<"H   ">>,
    QVal2 = <<"T1  ">>,
    QVal3 = <<"T2  ">>,

    %% TODO: db_put_append should convert recno keys to integers, like post_process_c_get_multiple_result does
    %% (hmmm -- we'd have to do it everywhere that recno was accepted; maybe best to just have them as binary everywhere)

    <<1:32/native>> = ?BDB:db_put_append(P, DB3, void, QVal1, []),
    <<2:32/native>> = ?BDB:db_put_append(P, DB3, void, QVal2, []),
    <<3:32/native>> = ?BDB:db_put_append(P, DB3, void, QVal3, []),

    C3 = ?BDB:db_cursor(P, DB3, void, []),
    [{<<1:32/native>>, QVal1}, {<<2:32/native>>, QVal2}, {<<3:32/native>>, QVal3}] = ?BDB:c_get_multiple(P, C3, <<>>, <<>>, 16384, [bdb_DB_FIRST, bdb_DB_MULTIPLE_KEY]),


    %% Test retreiving records (and first record number) from queue/recno (DB_MULTIPLE)
    {<<1:32/native>>, [QVal1, QVal2, QVal3]} = ?BDB:c_get_multiple(P, C3, <<>>, <<>>, 16384, [bdb_DB_FIRST, bdb_DB_MULTIPLE]),


    %% TODO: 
    %% - configure for duplicates and try DB_GET_BOTH and/or DB_GET_BOTH_RANGE with buffer size smaller than input value, and buffer size larger than input value


    normal = ?BDB:shutdown(P),	

    ok.



%%
%% Helpers
%%

wait_until_all_sites_are_in_sync(#master_location_hint{site_locator = {ReplicationGroup, MasterSiteName, MasterSiteNode}}, 
                                 SiteNames, 
                                 Timeout,
                                 Line) ->
    MasterSiteId = {MasterSiteName, MasterSiteNode},
    wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, Timeout, Line).

%% SiteNames does NOT have to be the whole group
wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, Timeout, Line) when Timeout > 0 ->

    StartTime = now(),    %% Ask the first site
    {InSyncAllSites, NotInSyncAllSites} = list_in_sync_sites_not_in_sync_sites(ReplicationGroup, MasterSiteId, hd(SiteNames)),
    %% Wait until they ALL agree
    AgreeResult = {lists:sort(InSyncAllSites), lists:sort(NotInSyncAllSites)},
    try lists:foreach(fun(SiteName) ->
                              {InSyncAllSites2, NotInSyncAllSites2} = list_in_sync_sites_not_in_sync_sites(ReplicationGroup, MasterSiteId, SiteName),
                              case {lists:sort(InSyncAllSites2), lists:sort(NotInSyncAllSites2)} of
                                  AgreeResult    -> ok;
                                  DisagreeResult -> throw({disagree, SiteName, DisagreeResult})
                              end
                      end,
                      tl(SiteNames))
    catch
        {disagree, DisagreeingSiteName, DisagreeResult} ->
            io:format("FAILED wait_until_all_sites_are_in_sync Line:~p Master:~w SiteNames:~p  Site ~p says AgreeResult:~p but Site ~p disagrees with ~p~n", [Line, MasterSiteId, SiteNames, hd(SiteNames), AgreeResult, DisagreeingSiteName, DisagreeResult]),
            timer:sleep(500),
            wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, reduced_timeout(Timeout, StartTime), Line)
    end,
    %% Now see if all sites in SiteNames are in-sync.
    %% SiteNames does NOT have to be the whole group, so (Not)InSyncAllSites might contain other sites that we are not interested in.
    %% We want to know if any all members of SiteNames are in InSyncSites.
    case SiteNames -- InSyncAllSites of 
        [] -> 
            ok;
        NotInSyncRequestedSites ->
            io:format("FAILED wait_until_all_sites_are_in_sync Line:~p Master:~w SiteNames:~p NotInSyncRequestedSites:~p InSyncAllSites:~p NotInSyncAllSites:~p~n", [Line, MasterSiteId, SiteNames, NotInSyncRequestedSites, InSyncAllSites, NotInSyncAllSites]),
            timer:sleep(500),
            wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, reduced_timeout(Timeout, StartTime), Line)
    end;

%% Timeout =< 0
wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, Timeout, Line) ->
    lists:foreach(fun(SiteName) ->
                          {InSyncAllSites, NotInSyncAllSites} = list_in_sync_sites_not_in_sync_sites(ReplicationGroup, MasterSiteId, SiteName),
                          io:format("FAILED (GIVEN UP): wait_until_all_sites_are_in_sync Line:~p Master:~w SiteName:~p says InSyncAllSites:~p NotInSyncAllSites:~p~n", [Line, MasterSiteId, SiteName, InSyncAllSites, NotInSyncAllSites])
                  end,
                  SiteNames),
    %% endless loop (stuck in *this* clause) so we can attach a debugger
    timer:sleep(5000),
    wait_until_all_sites_are_in_sync(ReplicationGroup, MasterSiteId, SiteNames, Timeout, Line).


list_in_sync_sites_not_in_sync_sites(ReplicationGroup, SpecifiedMasterSiteId, SiteNameToAsk) ->
    GroupStatus = ?BDB_RG:get_group_status({ReplicationGroup, SiteNameToAsk, node()}),
    lists:foldl(
      fun({{SiteName, _Node}, {replica, in_sync, {MasterSiteId, _}}}, {InSyncSites, NotInSyncSites}) 
         when MasterSiteId == SpecifiedMasterSiteId -> 
              {[SiteName | InSyncSites], NotInSyncSites};
         ({{SiteName, _Node}, {master, _}}, {InSyncSites, NotInSyncSites}) -> 
              {[SiteName | InSyncSites], NotInSyncSites};
         ({{SiteName, _Node}, _}, {InSyncSites, NotInSyncSites}) -> 
              {InSyncSites, [SiteName | NotInSyncSites]}
      end,
      {[], []},
      GroupStatus).

list_replicas(SiteLocatorToAsk) ->
    lists:foldl(
      fun({SiteNameNode, {replica, _, _}}, Acc) -> [SiteNameNode|Acc];
         (_, Acc)                               -> Acc
      end,
      [],
      ?BDB_RG:get_group_status(SiteLocatorToAsk)).


%% SiteNames does NOT have to be the whole group
wait_until_all_sites_are_down(ReplicationGroup, SiteNames, Timeout, Line) when Timeout > 0 ->
    StartTime = now(),
    {UpAllSites, DownAllSites} = list_up_sites_down_sites(ReplicationGroup, SiteNames),
    %% SiteNames does NOT have to be the whole group, so UpAllSites might contain other sites that we are not interested in.
    %% We want to know if all members of SiteNames are in DownAllSites
    case SiteNames -- DownAllSites of 
        [] -> 
            ok;
        _ ->
            io:format("wait_until_all_sites_are_down Line:~p (SiteNames:~p) UpAllSites:~p DownAllSites:~p~n", [Line, SiteNames, UpAllSites, DownAllSites]),
            timer:sleep(500),
            wait_until_all_sites_are_down(ReplicationGroup, SiteNames, reduced_timeout(Timeout, StartTime), Line)
    end;
wait_until_all_sites_are_down(ReplicationGroup, SiteNames, Timeout, Line) ->
    {UpAllSites, DownAllSites} = list_up_sites_down_sites(ReplicationGroup, SiteNames),
    io:format("FAILED wait_until_all_sites_are_down Line:~p (SiteNames:~p) UpAllSites:~p DownAllSites:~p~n", [Line, SiteNames, UpAllSites, DownAllSites]),
    %% endless loop (stuck in *this* clause) so we can attach a debugger
    timer:sleep(5000),
    wait_until_all_sites_are_down(ReplicationGroup, SiteNames, Timeout, Line).

list_up_sites_down_sites(ReplicationGroup, SiteNames) ->
    lists:foldl(
      fun(SiteName, {UpSites, DownSites}) ->
              SiteRGName = ?BDB_RG:get_proc_name(ReplicationGroup, SiteName),
              SitePortCoordName = ?BDB_RG:get_port_coordinator_proc_name(ReplicationGroup, SiteName),
              case {whereis(SiteRGName), whereis(SitePortCoordName)} of
                  {undefined, undefined} -> {UpSites,              [SiteName | DownSites]};
                  _                      -> {[SiteName | UpSites], DownSites}
              end
      end,
      {[], []},
      SiteNames).

        
reduced_timeout(CurrTimeout, StartTime) ->
    case CurrTimeout - ?BDB_U:now_diff_milliseconds(now(), StartTime) of
        T when T >= 0 -> T;
        _             -> 0
    end.




%%
%% Scott's leak_test 
%%
%% [cnewcom: I have never tried this. I used the stress test & performance test to find leaks.]
%%

leak_test() ->
    Port1 = ?BDB:start(),
    ok = env_open_with_retry(Port1, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644, ?LINE),
    ok = leak_test(Port1),
    normal = ?BDB:shutdown(Port1),
    io:format("\nleak_test: linked-in driver test finished\n"),

    Port2 = ?BDB:start_pipe(),
    ok = env_open_with_retry(Port2, "", [bdb_DB_PRIVATE, bdb_DB_CREATE, bdb_DB_INIT_MPOOL], 8#644, ?LINE),
    ok = leak_test(Port2),
    normal = ?BDB:shutdown(Port2),
    io:format("\nleak_test: pipe driver test finished\n"),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

leak_test(P) ->
    DB2 = ?BDB:db_create(P, []),
    ok = ?BDB:db_open(P, DB2, void, "", "", bdb_DB_HASH, bdb_DB_CREATE, []),
    TheKey = list_to_binary("key"),
    TheValue = list_to_binary(lists:seq(0, 255)),
    Put = fun () ->
		  ok = ?BDB:db_put(P, DB2, void, TheKey, TheValue, [])
	  end,
    ok = do_loop0(99000, Put),
    Get = fun () ->
		  {TheKey, _TheVal} = ?BDB:db_get(P, DB2, void, TheKey, <<>>, [])
	  end,
    ok = do_loop0(99000, Get),

    timer:sleep(100),
    {messages, []} = process_info(self(), messages),
    ok.

do_loop0(0, _Fun) ->
    ok;
do_loop0(N, Fun) ->
    Fun(),
    do_loop0(N - 1, Fun).



%%
%% Helper functions
%%

get_all_keys_and_values(P, DB) ->
    Cursor = ?BDB:db_cursor(P, DB, void, []),
    get_all_keys_and_values_from_cursor(P, Cursor, []).

get_all_keys_and_values_from_cursor(P, Cursor, Acc) ->
    try ?BDB:c_get(P, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
	Pair -> 
            get_all_keys_and_values_from_cursor(P, Cursor, [Pair, Acc])
    catch
        {bdb_error, bdb_DB_NOTFOUND} -> 
            ok = ?BDB:c_close(P, Cursor),
            lists:reverse(Acc)
    end.


print_all_keys_and_values(P, DB) ->
    Cursor = ?BDB:db_cursor(P, DB, void, []),
    print_all_keys_and_values_from_cursor(P, Cursor).

print_all_keys_and_values_from_cursor(P, Cursor) ->
    try ?BDB:c_get(P, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
	{Key, Value} -> 
            io:format("Key = '~w', Value = '~w'\n", [Key, Value]),
            print_all_keys_and_values_from_cursor(P, Cursor)
    catch
	{bdb_error, bdb_DB_NOTFOUND} -> 
            ok = ?BDB:c_close(P, Cursor)
    end.


verify_all_keys(P, DB, ExpectedKeys) ->
    Cursor = ?BDB:db_cursor(P, DB, void, []),
    verify_all_keys_from_cursor(P, Cursor, ExpectedKeys).

verify_all_keys_from_cursor(P, Cursor, []) ->
    try ?BDB:c_get(P, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
	{ok, {ExtraKey, _}} -> 
            ok = ?BDB:c_close(P, Cursor),
            throw({too_many_keys_in_database, ExtraKey})
    catch
	{bdb_error, bdb_DB_NOTFOUND} ->
            ok = ?BDB:c_close(P, Cursor)
    end;

verify_all_keys_from_cursor(P, Cursor, [ExpectedKey | NextKeys]) ->
    try ?BDB:c_get(P, Cursor, <<>>, <<>>, [bdb_DB_NEXT]) of
	{ExpectedKey, _} -> 
            verify_all_keys_from_cursor(P, Cursor, NextKeys);

	{MismatchedKey, _} -> 
            ok = ?BDB:c_close(P, Cursor),
            throw({mismatched_key, ExpectedKey, MismatchedKey})
    catch
	{bdb_error, bdb_DB_NOTFOUND} -> 
            ok = ?BDB:c_close(P, Cursor),
            throw({not_enough_keys_in_database, ExpectedKey})
    end.


receive_tagged_port_replies(_Port, []) ->
    ok;

receive_tagged_port_replies(Port, [{Tag, ReceiveReplyFuncName}|T]) ->
    %%io:format("Waiting for tag (len ~p) ~w (~p more to go after this one)\n", [size(Tag), Tag, length(T)]),
    try ?BDB:ReceiveReplyFuncName(Port, Tag) of 
        _ ->     receive_tagged_port_replies(Port, T)
    catch
        Error -> throw({"Unexpected port reply", Error})
    end.



ensure_empty_dir(Dir) ->
    ok = filelib:ensure_dir(Dir), % this only ensures that *parent* directories exist!
    case file:make_dir(Dir) of
        ok ->
            ok;
        {error, eexist} ->
            ok;  % directory already exists
        {error, Reason} ->
            throw({'Failed to create regression test directory', Dir, Reason})
    end,
    lists:foreach(fun(File) -> 
                          %% io:format("Deleting old file ~p\n", [File]), 
                          file:delete(File) 
                  end, 
                  filelib:wildcard(Dir ++ "/*")).


clear_msg_queue() ->
    receive
	Msg -> io:format("clear_msg_queue: discarding msg ~w\n", [Msg]),
	       clear_msg_queue()
    after
	0 ->   %%io:format("clear_msg_queue: no more msgs\n"),
	       ok
    end.


%% This is used at the end of tests that set trap_exit = true, to verify that no spawned processes exited unexpectedly.
%% Important: this must be called BEFORE restoring trap_exit to false, as doing so seems to purge any EXIT messages 
%% from the message queue.
throw_if_any_exit_msgs() ->
    receive
        {'EXIT', Pid, Reason} -> 
            io:format("throw_if_any_exit_msgs: found EXIT pid=~p reason: ~p\n", [Pid, Reason]),
            throw({found_exit, Pid, Reason})
    after
        0 ->   %%io:format("throw_if_any_exit_msgs: no more msgs\n"),
               ok
    end.


wait_for_exit_messages([]) -> 
    ok;
wait_for_exit_messages([Pid|T]) -> 
    receive
        {'EXIT',Pid,killed} -> wait_for_exit_messages(T)
    end.






%% From http://schemecookbook.org/Erlang/NumberBiasedRands
%%
%% For this recipe, we will consider the normal (Gaussian)
%% distribution. If you need other distributions see either the CVS or
%% consult a numerical analyst.
%%
%% The function dis_var_new returns a stochastic variable (a thunk)
%% with mean mu and standard deviance sigma (norma;/Gaussian distribution).
%% 
%% An example of usage:
%% 1> X=dist_var_new(0, 1).      
%% #Fun<cookbook.4.38595032>
%% 2> X1().
%% -0.873932
%% 3> X1().
%% 4.91005e-2
%% 4> X1().
%% 1.55993e-2
%% 5> X1().
%% 0.181456 
%% If you are unsatisfied with the fact that you get the same numbers as above, 
%% then randomize the source of the random numbers: 
%% 6> random:seed() 
%%
%% The algorithm used is the polar Box Muller method. The algorithm
%% takes two independent uniformly distributed random numbers between
%% 0 and 1 (present in the code as random:uniform()) and generates two
%% numbers with a mean of mu and standard deviation sigma. Note that
%% the method produces two numbers at a time. Since we only need one,
%% the second is saved for later in the variable next.
%%
%% See also http://www.unix.org.ua/orelly/perl/cookbook/ch02_11.htm

%% -record( dist_state, {state, mu, sigma} ).

%% dist_var_new(Mu, Sigma) ->
%%     dist_var_new(false, Mu, Sigma).

%% dist_var_new(State, Mu, Sigma) ->
%%     This = #dist_state {state = State,
%%                         mu = 1.0 * Mu,
%%                         sigma = 1.0 * Sigma},
%%     IntPid = spawn(cookbook, dispatch, [This]),
%%     fun () ->
%%             IntPid ! {self(), value},
%%             receive
%%                 {retval, Any} -> Any
%%             end
%%     end.

%% dispatch(This) ->
%%     receive
%%         {Pid, value} ->
%%             {NewThis, Value} = value(This),
%%             Pid!{retval, Value},
%%             dispatch(This)    %% todo: should this be NewThis???
%%     end.

%% value(This) ->
%%     case This#dist_state.state of
%%         true -> Val = This#dist_state.mu
%%                     + (This#dist_state.sigma * This#dist_state.state),
%%                 {This#dist_state{state = false}, Val};
%%         _    -> sigma_loop(This)
%%     end.

%% sigma_loop(This) ->
%%     V1 = 2.0 * random:uniform() - 1.0,
%%     V2 = 2.0 * random:uniform() - 1.0,
%%     S  = (V1 * V1) + (V2 * V2),
%%     if
%%         S >= 1.0 ->
%%             sigma_loop(This);
%%         true     ->
%%             Scale = math:sqrt( (-2.0 * math:log(S)) / S),
%%             Val = This#dist_state.mu
%%                 + (This#dist_state.sigma * Scale * V1),
%%             {This#dist_state{state = Scale * V2}, Val}
%%     end. 
